/*
 *      calculations.c
 *      
 *      Copyright 2009 - 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/* Refrerences:
 * 1. Antenna pointing: Various scripts (Jens T. Satre), http://www.satellite-calculations.com
 * 2. Sun interferention: Sun Outage Calculator (Jens T. Satre and others), 
 * http://www.satellite-calculations.com/SUNcalc/SUNcalc.htm
 * 3. Sky tracker: Computing planetary positions - a tutorial with worked examples
 * (Paul Schlyter, Stockholm, Sweden), http://stjarnhimlen.se/
*/

#include <gtk/gtk.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include "calculations.h"

#define PI   3.14159265358979323846
#define EFF  1/298.257
#define RSAT 42164.570
#define REQ  6378.140
#define RG   42164.57
#define A0   0.58804392
#define A1   -0.17941557
#define A2   0.29906946E-1
#define A3   -0.25187400E-2
#define A4   0.82622101E-4

enum {
	SEASON_SPRING = 0,
	SEASON_FALL
};

enum {
	BAND_C = 0,
	BAND_KU
};

static gfloat degtorad(gfloat degrees);
static gfloat radtodeg(gfloat radians);
static gdouble rev(gdouble x);
static gdouble fmodv(gdouble value, gdouble modv);
static gdouble interferention_begin_angle(gdouble angle, gdouble longitude, gdouble latitude, gdouble satlongitude, gdouble *yangle);
static gdouble interferention_end_angle(gdouble angle, gdouble longitude, gdouble latitude, gdouble satlongitude);


static gfloat degtorad(gfloat degrees)
{
	return PI * degrees / 180.0;
}

static gfloat radtodeg(gfloat radians)
{
	return radians * (180.0 / PI);
}

void calculations_multifeed(gfloat longitude, gfloat latitude, gfloat sat1longitude, gfloat sat2longitude, gfloat altitude, gfloat focuslen, gfloat *xoffset, gfloat *yoffset)
{
	gfloat az1, el1, az2, el2;
	
	if ((xoffset == NULL) || (yoffset == NULL)) return;
	
	az1 = calculations_azimuth(longitude, latitude, sat1longitude, altitude);
	el1 = calculations_elevation(longitude, latitude, sat1longitude, altitude);
	az2 = calculations_azimuth(longitude, latitude, sat2longitude, altitude);
	el2 = calculations_elevation(longitude, latitude, sat2longitude, altitude);
	
	*xoffset = focuslen * sin(fabs(degtorad(az1) - degtorad(az2)));
	*yoffset = focuslen * sin(fabs(degtorad(el1) - degtorad(el2)));
}

gfloat calculations_azimuth(gfloat longitude, gfloat latitude, gfloat satlongitude, gfloat altitude)
{
	gfloat rst, ra, rz, alfar, alfarx, alfary, alfarz, alfarnorth, alfarzenith, res;
	
	rst = REQ / (sqrt(1 - EFF * (2 - EFF) * sin(degtorad(latitude)) * sin(degtorad(latitude))));
	ra = (rst + altitude) * cos(degtorad(latitude));
	rz = rst * (1-EFF) * (1-EFF) * sin(degtorad(latitude));
	alfar = RSAT - rst;
	alfarx = RSAT * cos(degtorad(satlongitude - longitude)) - ra;
	alfary = RSAT * sin(degtorad(satlongitude - longitude));
	alfarz = -rz;
	alfarnorth = -alfarx * sin(degtorad(latitude)) + alfarz * cos(degtorad(latitude));
	alfarzenith = alfarx * cos(degtorad(latitude)) + alfarz * sin(degtorad(latitude));
	if (alfarnorth <= 0) {
		return (180.0 + radtodeg(atan2((alfary / alfarnorth),1)));
	} else {
		res = 360.0 + radtodeg(atan2((alfary / alfarnorth),1));
		return (res - floor(res/360.0) * 360.0);
	}
}

gfloat calculations_elevation(gfloat longitude, gfloat latitude, gfloat satlongitude, gfloat altitude)
{
	gfloat rst, ra, rz, alfar, alfarx, alfary, alfarz, alfarnorth, alfarzenith, elgeom, refraction, x, elobs;
	
	rst = REQ / (sqrt(1 - EFF * (2 - EFF) * pow(sin(degtorad(latitude)), 2)));
	ra = (rst + altitude) * cos(degtorad(latitude));
	rz = rst * pow((1 - EFF), 2) * sin(degtorad(latitude));
	alfar = RSAT - rst;
	alfarx = RSAT * cos(degtorad(satlongitude - longitude)) - ra;
	alfary = RSAT * sin(degtorad(satlongitude - longitude));
	alfarz = -rz;
	alfarnorth = -alfarx * sin(degtorad(latitude)) + alfarz * cos(degtorad(latitude));
	alfarzenith = alfarx * cos(degtorad(latitude)) + alfarz * sin(degtorad(latitude));
	elgeom = radtodeg(atan2(alfarzenith, sqrt(pow(alfarnorth, 2) + pow(alfary, 2))));
	x = fabs(elgeom + 0.589);
	refraction = fabs(A0 + A1 * pow(x, 1) + A2 * pow(x, 2) + A3 * pow(x, 3) + A4 * pow(x, 4));
	if (elgeom > 10.2) {
		elobs = elgeom + 0.01617 * (cos(degtorad(fabs(elgeom))) / sin(degtorad(fabs(elgeom))));
	} else {
		elobs = elgeom + refraction;
	}
	if (alfarzenith < -3000.0) elobs = -99.0;
	
	return elobs;
}

gfloat calculations_declination(gfloat latitude, gfloat azimuth, gfloat elevation)
{
	
	return radtodeg(asin(sin(degtorad(elevation))*sin(degtorad(latitude))+cos(degtorad(elevation))*cos(degtorad(latitude))+cos(degtorad(azimuth))));
}

gfloat calculations_hourangle(gfloat latitude, gfloat azimuth, gfloat elevation)
{
	gdouble a, b, res;
	
	a = -cos(degtorad(elevation))*sin(degtorad(azimuth));
	b = sin(degtorad(elevation))*cos(degtorad(latitude))-cos(degtorad(elevation))*sin(degtorad(latitude))*cos(degtorad(azimuth));
	res = 180 + radtodeg(atan(a/b));
	
	if (azimuth > 270) {
		res = ((res-180)+360);
		if (res > 360) {
			res = 360-(res-360);
		}
	}
	
	if (azimuth < 90) {
		res = (180-res);
	}
	
	return res;
}

gfloat calculations_polangle(gfloat longitude, gfloat latitude, gfloat satlongitude)
{
	return radtodeg(atan2((sin(degtorad(longitude - satlongitude)) / tan(degtorad(latitude))), 1));
}

gfloat calculations_satheight(gfloat longitude, gfloat latitude, gfloat satlongitude)
{
	gfloat h, f, f1, f2, f3;
	
	h = RG - REQ;
	f1 = fabs(satlongitude - longitude - 360);
	f2 = fabs(longitude - satlongitude);
	f3 = fabs(longitude - satlongitude - 360);
	if (f1<f2) f = f1; else f = f2;
	if (f>f3) f = f3; 
	
	return sqrt(pow(REQ, 2) + pow((REQ + h), 2) - 2 * REQ * (REQ + h) * cos(degtorad(latitude)) * cos(degtorad(f)));
}

static gdouble rev(gdouble x)
{
	return  x - floor(x / 360.0) * 360.0;
}

sunmoonpos_t *calculations_sunmoon_tracker(gdouble longitude, gdouble latitude, time_t timestamp, sunmoonpos_t *sunmoonpos)
{
	gdouble utm, d, w, a, e, Msun, oblecl, L, E, x, y, r, v, lon, z, xequat, yequat, zequat, RA,
	decl, GMST0, SIDTIME, HA, xhor, yhor, zhor, azimuth, altitude, N, i, M, E0, E1, it, xeclip,
	yeclip, zeclip, mlong, mlat, Ls, Ms, Lm, Mm, D, F, plong, plat, pdist, MRA, Mdecl, refraction,
	raltitude, satlongitude, Z;
	struct tm *tms;
	
	
	if (sunmoonpos == NULL) return NULL;
	
	tms = gmtime(&timestamp);
	
	utm =  tms->tm_hour + (tms->tm_min / 60.0) + (tms->tm_sec / (60.0 * 60.0));
	d = (367 * (tms->tm_year + 1900)) - ((7 * ((tms->tm_year + 1900) + (((tms->tm_mon + 1) + 9) / 12))) / 4) + ((275 * (tms->tm_mon + 1)) / 9) + tms->tm_mday - 730530 + (utm / 24.0);	
	
	//SUN
	
	w = 282.9404 + 4.70935E-5 * d;            //longitude of perihelion
    a = 1.000000;                             //mean distance, a.u.
    e = 0.016709 - 1.151E-9 * d;              //eccentricity
    Msun = rev(356.0470 + 0.9856002585 * d);  //mean anomaly
	oblecl = 23.4393 - 3.563E-7 * d; 
	
	L = w + Msun;
	L = rev(L);
	
	E = Msun + (180.0 / PI) * e * sin(degtorad(Msun)) * (1.0 + e * cos(degtorad(Msun)));
	E = rev(E);
	
	x = a * cos(degtorad(E)) - e;
    y = a * sin(degtorad(E)) * sqrt(1 - e * e);
	r = sqrt(x*x + y*y);
    v = radtodeg(atan2(y, x));
	
	lon = rev(v + w);
	
	x = r * cos(degtorad(lon));
    y = r * sin(degtorad(lon));
    z = 0.0;
	
	xequat = x;
    yequat = y * cos(degtorad(oblecl)) - z * sin(degtorad(oblecl));
    zequat = y * sin(degtorad(oblecl)) + z * cos(degtorad(oblecl));
	
	RA = rev(radtodeg(atan2(yequat, xequat)));
	decl = radtodeg(atan2(zequat, sqrt(xequat*xequat + yequat*yequat)));
	
	GMST0 = (L / 15.0) + 12;
	SIDTIME = GMST0 + utm + longitude / 15.0;
	if (SIDTIME < 0) {
		SIDTIME += 24.0;
	} else if (SIDTIME > 24) {
		SIDTIME -= 24.0;
	}
	HA = rev(15.0 * (SIDTIME - RA / 15.0));
	
	x = cos(degtorad(HA)) * cos(degtorad(decl));
	y = sin(degtorad(HA)) * cos(degtorad(decl));
	z = sin(degtorad(decl));
	
	xhor = x * sin(degtorad(latitude)) - z * cos(degtorad(latitude));
    yhor = y;
    zhor = x * cos(degtorad(latitude)) + z * sin(degtorad(latitude));
	
	raltitude = atan2(zhor, sqrt(xhor*xhor + yhor*yhor));
		
	/*x = fabs(raltitude + 0.589);
	refraction = fabs(A0 + A1*x + A2*x*x + A3*x*x*x + A4*x*x*x*x);
	if (raltitude > 10.2) {
		raltitude = raltitude + 0.01617 * (cos(fabs(raltitude)) / sin(fabs(raltitude)));
	} else {
		raltitude = raltitude + refraction;
	}*/
	
	azimuth  = radtodeg(atan2(yhor, xhor)) + 180.0;
	altitude = radtodeg(raltitude);
	
	sunmoonpos->sunazimuth = azimuth;
	sunmoonpos->sunaltitude = altitude;
	
	/*printf("Sun azimuth, altitude: %f, %f\n", azimuth, altitude);*/
	
	satlongitude = 0;
	
	Z = degtorad(longitude) - atan(tan(degtorad(azimuth)) * sin(degtorad(latitude)));
	Z = rev(radtodeg(Z));
	
	if (((latitude > 0) && (azimuth > 90) && (azimuth < 270)) || ((latitude < 0) && (((azimuth >= 0) && (azimuth <= 90)) || ((azimuth >= 270) && (azimuth <= 360))))) {
		if (Z < 0) {
			Z = 360 - Z;
			satlongitude = -Z;
		}
		
		if (Z > 0) {
			if (Z > 180) {
				Z = 360 - Z;
				satlongitude = -Z;
			} else {
				satlongitude = Z;
			}
		}
		
		sunmoonpos->sunsat = satlongitude;
		
		/*printf("Satellite longitude: %f\n", satlongitude);*/
	} else {
		sunmoonpos->sunsat = -1;
	}
		
	//MOON
	
	N = rev(125.1228 - 0.0529538083 * d);   //Long asc. node
    i = 5.1454;                             //Inclination
    w = rev(318.0634 + 0.1643573223 * d);   //Arg. of perigee
    a = 60.2666;                            //Mean distance
    e = 0.054900;                           //Eccentricity
    M = rev(115.3654 + 13.0649929509 * d);  //Mean anomaly
	
	it = 0;
	E0 = rev(M + (180 / PI) * e * sin(degtorad(M)) * (1 + e * cos(degtorad(M))));
	
	while (1) {
		E1 = rev(E0 - (E0 - (180 / PI) * e * sin(degtorad(E0)) - M) / (1 - e * cos(degtorad(E0))));
		it++;
		if ((fabs(E0 - E1) <= 0.005) || (it >= 500)) break;
		E0 = E1;
	}
	
	E = E1;
	
	x = a * (cos(degtorad(E)) - e);
    y = a * sqrt(1 - e*e) * sin(degtorad(E));
	r = (sqrt(x*x + y*y));
    v = rev(radtodeg(atan2(y, x)));	
	
	xeclip = r * (cos(degtorad(N)) * cos(degtorad(v+w)) - sin(degtorad(N)) * sin(degtorad(v+w)) * cos(degtorad(i)));
    yeclip = r * (sin(degtorad(N)) * cos(degtorad(v+w)) + cos(degtorad(N)) * sin(degtorad(v+w)) * cos(degtorad(i)));
    zeclip = r * sin(degtorad(v+w)) * sin(degtorad(i));
	
	mlong = rev(radtodeg(atan2(yeclip, xeclip)));
	mlat = radtodeg(atan2(zeclip, sqrt(xeclip*xeclip +yeclip*yeclip)));
	
	Ls = L;
	Ms = Msun;
	Lm = rev(N + w + M);
    Mm = rev(M);
    D = rev(Lm - Ls);
    F = rev(Lm - N);
	
	plong = (-1.274 * sin(degtorad(Mm - 2*D))) + (0.658 * sin(degtorad(2*D))) + (-0.186 * sin(degtorad(Ms))) +
	(-0.059 * sin(degtorad(2*Mm - 2*D))) + (-0.057 * sin(degtorad(Mm - 2*D + Ms))) + (0.053 * sin(degtorad(Mm + 2*D))) +
	(0.046 * sin(degtorad(2*D - Ms))) + (0.041 * sin(degtorad(Mm - Ms))) + (-0.035 * sin(degtorad(D))) +
	(-0.031 * sin(degtorad(Mm + Ms))) + (-0.015 * sin(degtorad(2*F - 2*D))) + (0.011 * sin(degtorad(Mm - 4*D)));
	
	plat = (-0.173 * sin(degtorad(F - 2*D))) + (-0.055 * sin(degtorad(Mm - F - 2*D))) + (-0.046 * sin(degtorad(Mm + F - 2*D))) +
	(0.033 * sin(degtorad(F + 2*D))) + (0.017 * sin(degtorad(2*Mm + F)));
	
	pdist = (-0.58 * cos(degtorad(Mm - 2*D))) + (-0.46 * cos(degtorad(2*D)));
	
	mlong += plong;
	mlat += plat;
	r += pdist;
	
	xhor = r * cos(degtorad(mlong)) * cos(degtorad(mlat));
	yhor = r * sin(degtorad(mlong)) * cos(degtorad(mlat));
	zhor = r * sin(degtorad(mlat));
	
	xequat = xhor;
	yequat = yhor * cos(degtorad(oblecl)) - zhor * sin(degtorad(oblecl));
	zequat = yhor * sin(degtorad(oblecl)) + zhor * cos(degtorad(oblecl));
	
	MRA = rev(radtodeg(atan2(yequat, xequat)));
	Mdecl = radtodeg(atan2(zequat, sqrt(xequat*xequat + yequat*yequat)));
	
	GMST0 = (Ls / 15.0) + 12;
	SIDTIME = GMST0 + utm + longitude / 15.0;
	if (SIDTIME < 0) {
		SIDTIME += 24.0;
	} else if (SIDTIME > 24) {
		SIDTIME -= 24.0;
	}
	HA = rev(15.0 * (SIDTIME - MRA / 15.0));
	
	x = cos(degtorad(HA)) * cos(degtorad(Mdecl));
	y = sin(degtorad(HA)) * cos(degtorad(Mdecl));
	z = sin(degtorad(Mdecl));
	
	xhor = x * sin(degtorad(latitude)) - z * cos(degtorad(latitude));
    yhor = y;
    zhor = x * cos(degtorad(latitude)) + z * sin(degtorad(latitude));
	
	raltitude = asin(zhor);
	raltitude -= asin(1 / r * cos(raltitude));
		
	/*x = fabs(raltitude + 0.589);
	refraction = fabs(A0 + A1*x + A2*x*x + A3*x*x*x + A4*x*x*x*x);
	if (raltitude > 10.2) {
		raltitude = raltitude + 0.01617 * (cos(fabs(raltitude)) / sin(fabs(raltitude)));
	} else {
		raltitude = raltitude + refraction;
	}*/
    
    azimuth  = radtodeg(atan2(yhor, xhor)) + 180;
    altitude = radtodeg(raltitude);
    
    sunmoonpos->moonazimuth = azimuth;
	sunmoonpos->moonaltitude = altitude;
    
	/*printf("Moon azimuth, altitude: %f, %f\n", azimuth, altitude);*/
	
	satlongitude = 0;
	
	Z = degtorad(longitude) - atan(tan(degtorad(azimuth)) * sin(degtorad(latitude)));
	Z = rev(radtodeg(Z));
	
	if (((latitude > 0) && (azimuth > 90) && (azimuth < 270)) || ((latitude < 0) && (((azimuth >= 0) && (azimuth <= 90)) || ((azimuth >= 270) && (azimuth <= 360))))) {
		if (Z < 0) {
			Z = 360 - Z;
			satlongitude = -Z;
		}
		
		if (Z > 0) {
			if (Z > 180) {
				Z = 360 - Z;
				satlongitude = -Z;
			} else {
				satlongitude = Z;
			}
		}
		
		sunmoonpos->moonsat = satlongitude;
		
		/*printf("Satellite longitude: %f\n", satlongitude);*/
	} else {
		sunmoonpos->moonsat = -1;
	}
	
	return sunmoonpos;
}
	
static gdouble fmodv(gdouble value, gdouble modv)
{
	gdouble num;
	
	num = floor(value / modv);
	
	return (value - (modv * num));
}

static gdouble interferention_begin_angle(gdouble angle, gdouble longitude, gdouble latitude, gdouble satlongitude, gdouble *yangle)
{
	gdouble A26, D26, G26, H26, I26, L26, J26, K26, O26, M26b, M26c, M26a, N26,
	S26, T26, R26, X26, W26, Y26, V26, AA26, Sun_Dec, SC_Dec, E26, F26, M26, Q26, P26, U26, Z26;
		
	A26 = angle;
	D26 = A26 + 1460.5;
	E26 = (D26)/36525;
	F26 = fmodv((99.81355 + 0.9856474 * (A26 - 27759) + 360 * ((A26 - 27759) - floor(A26 - 27759))), 360.0);
	G26 = 0.016751 - (0.000042 * E26);
	H26 = 358.4759 + (35999.0498 * E26);
	I26 = 279.6964 + (36000.769 * E26);
	L26 = 23.4523 - (0.013 * E26);
	J26 = (2 * G26 * sin(H26 * PI / 180) + 1.25 * G26 * G26 * sin(H26 * PI / 90)) * (180 / PI);
	K26 = I26 + J26;
	O26 = 149504200 * (1 - G26 * cos(H26 * PI/180) - G26 * G26 / 2 * (cos(H26 * PI/90) - 1));
	M26b = cos(K26 * (PI / 180));
	M26c = (cos(L26 * PI / 180) * sin(K26 * PI / 180));
	M26a = atan2(M26c, M26b);
	M26 = fmodv((M26a  * 180 / PI + 360), 360.0);
	N26 = atan(tan(L26 * PI / 180) * sin(M26 * PI / 180)) * 180 / PI;
	Q26 = fmodv((F26 - longitude + 360), 360.0);
	P26 = fmodv((F26 - satlongitude + 360), 360.0);
	S26 = sin(M26 * (PI / 180)) * cos(N26 * PI / 180) * O26 - sin(Q26 * PI / 180) * cos(latitude * PI / 180) * 6380;
	T26 = sin(N26 * (PI / 180)) * O26 - sin(latitude * PI / 180) * 6380;
	R26 = cos(M26 * (PI / 180)) * cos(N26 * PI / 180) * O26 - cos(Q26 * PI / 180) * cos(latitude * PI / 180) * 6380;
	X26 = (sin(P26 * (PI / 180)) * 42164) - (sin(Q26 * (PI / 180)) * cos(latitude * PI / 180) * 6380);
	W26 = cos(P26 * PI / 180) * 42165 - cos(Q26 * PI / 180) * cos(latitude * PI / 180) * 6380;
	Y26 = -sin(latitude * PI / 180) * 6380;
	U26 = fmodv((atan2(S26, R26) * 180 / PI + 360), 360.0);
	Z26 = fmodv((atan2(X26, W26) * 180 / PI + 360), 360.0);
	V26 = atan(T26 / sqrt(R26 * R26 + S26 * S26)) * 180 / PI;
	AA26 = atan(Y26 / sqrt(W26 * W26 + X26 * X26)) * (180 / PI);
	Sun_Dec = V26;
	SC_Dec = AA26;
	*yangle = fabs(SC_Dec - Sun_Dec);
	
	return acos(sin(AA26 * PI / 180) * sin(V26 * PI / 180) + cos(AA26 * PI / 180) * cos(V26 * PI / 180) * cos((Z26 - U26) * PI / 180)) * 180 / PI;
}

static gdouble interferention_end_angle(gdouble angle, gdouble longitude, gdouble latitude, gdouble satlongitude)
{
	gdouble A27, D27, E27, F27, G27, H27, I27, L27, J27, K27, O27, M27b, M27c, M27a, M27, N27,
	Q27, P27, S27, T27, R27, X27, W27, Y27, U27, Z27, V27, AA27;
	
	A27 = angle;
	D27 = A27 + 1460.5;
	E27 = (D27)/36525;
	F27 = fmodv((99.81355 + 0.9856474 * (A27 - 27759) + 360 * ((A27 - 27759) - floor(A27 - 27759))), 360.0);
	G27 = 0.016751 - (0.000042 * E27);
	H27 = 358.4759 + (35999.0498 * E27);
	I27 = 279.6964 + (36000.769 * E27);
	L27 = 23.4523 - (0.013 * E27);
	J27 = (2 * G27 * sin(H27 * PI / 180) + 1.25 * G27 * G27 * sin(H27 * PI / 90)) * (180 / PI);
	K27 = I27 + J27;
	O27 = 149504200 * (1 - G27 * cos(H27 * PI/180) - G27 * G27 / 2 * (cos(H27 * PI/90) - 1));
	M27b = cos(K27 * (PI/180));
	M27c = (cos(L27 * PI/180) * sin(K27 * PI/180));
	M27a = atan2(M27c, M27b);
	M27 = fmodv((M27a  * 180/PI + 360), 360.0);
	N27 = atan(tan(L27 * PI / 180) * sin(M27 * PI / 180)) * 180 / PI;
	Q27 = fmodv((F27 - longitude + 360), 360.0);
	P27 = fmodv((F27 - satlongitude + 360), 360.0);
	S27 = sin(M27 * (PI / 180)) * cos(N27 * PI / 180) * O27 - sin(Q27 * PI / 180) * cos(latitude * PI / 180) * 6380;
	T27 = sin(N27 * (PI / 180)) * O27 - sin(latitude * PI / 180) * 6380;
	R27 = cos(M27 * (PI / 180)) * cos(N27 * PI / 180) * O27 - cos(Q27 * PI / 180) * cos(latitude * PI / 180) * 6380;
	X27 = (sin(P27 * (PI / 180)) * 42164) - (sin(Q27 * (PI / 180)) * cos(latitude * PI / 180) * 6380);
	W27 = cos(P27 * PI / 180) * 42165 - cos(Q27 * PI / 180) * cos(latitude * PI / 180) * 6380;
	Y27 = -sin(latitude * PI / 180) * 6380;
	U27 = fmodv((atan2(S27, R27) * 180 / PI + 360), 360.0);
	Z27 = fmodv((atan2(X27, W27) * 180 / PI + 360), 360.0);
	V27 = atan(T27 / sqrt(R27 * R27 + S27 * S27)) * 180 / PI;
	AA27 = atan(Y27 / sqrt(W27 * W27 + X27 * X27)) * 180 / PI;
	
	return acos(sin(AA27 * PI / 180) * sin (V27 * PI / 180) + cos(AA27 * PI / 180) * cos(V27 * PI / 180) * cos((Z27 - U27) * PI / 180)) * 180 / PI;
}

GSList *calculations_interferention_main(gdouble longitude, gdouble latitude, gdouble satlongitude, gshort year, gchar season, gchar band, gdouble dsize)
{
	gdouble epoch, cepoch, outageangle, boutage, refepoch, bangle, yangle, eoutage, eangle;
	gchar found = 0;
	unsigned int outnr = 0;
	
	struct tm timestamp;
	gchar outstr[200];
	time_t inttime;
	time_t starttime, endtime;
	intperiod *period;
	GSList *reslist = NULL;
	
	memset(&timestamp, 0, sizeof(struct tm));
	timestamp.tm_mday = 1;
	timestamp.tm_mon = 0;
	timestamp.tm_year = 4;
	timestamp.tm_hour = 2;
	timestamp.tm_min = 30;
	timestamp.tm_sec = 0;
	
	inttime = mktime(&timestamp);
	
	epoch = (year - 1904) * 365.25;
	if (season == SEASON_SPRING) {
		epoch += 45.0;
	} else {
		epoch += 229.0;
	}
	epoch = floor(epoch);
	cepoch = floor(epoch);
	
	if (band == BAND_C) {
		outageangle = 11 / 3.95 / dsize + 0.25;
	} else {
		outageangle = 11 / 11.95 / dsize + 0.25;
	}
	
	found = 0;
	
	longitude = -longitude;
	satlongitude = -satlongitude;
	
	while ((outnr == 0) || (found == 1)) {
		boutage = epoch;
		bangle = interferention_begin_angle(boutage, longitude, latitude, satlongitude, &yangle);
		refepoch = epoch;
		found = 0;
		while ((bangle > outageangle) && ((boutage - refepoch) < 1)) {
			if ((yangle) > (1 + outageangle / 2)) {
				boutage += (yangle * 0.5);
				bangle = interferention_begin_angle(boutage, longitude, latitude, satlongitude, &yangle);
			}
			
			if (bangle > (outageangle * 4)) {
				boutage += (bangle / 540);
				bangle = interferention_begin_angle(boutage, longitude, latitude, satlongitude, &yangle);
			}
			
			boutage += 0.00005787;
			bangle = interferention_begin_angle(boutage, longitude, latitude, satlongitude, &yangle);
		}
		epoch = boutage;
		
		if (bangle < outageangle) {
			outnr++;
			found = 1;
			eoutage = boutage;
			eangle = interferention_end_angle(eoutage, longitude, latitude, satlongitude);
			while (eangle < outageangle) {
				eoutage += 0.00005787;
				eangle = interferention_end_angle(eoutage, longitude, latitude, satlongitude);
			}
			eoutage -= 0.00005787;
			eangle = interferention_end_angle(eoutage, longitude, latitude, satlongitude);
			
			starttime = inttime + (unsigned int)trunc(boutage * 60.0 * 60.0 * 24.0);
			endtime = inttime + (unsigned int)trunc(eoutage * 60.0 * 60.0 * 24.0);
			
			period = (intperiod *)g_malloc0(sizeof(intperiod));
			period->start = starttime;
			period->end = endtime;
			
			reslist = g_slist_prepend(reslist, period);
			
			epoch = eoutage + 0.9;
			bangle = interferention_begin_angle(boutage, longitude, latitude, satlongitude, &yangle);
		}
		
	}
	
	reslist = g_slist_reverse(reslist);
	
	return reslist;   
}


