/*
 *      calcwindow.c
 *      
 *      Copyright 2011-2012 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <gtk/gtk.h>
#include <glib/gi18n.h>
#include <string.h>
#include "settings.h"
#include "calculations.h"
#include "calcwindow.h"

enum {
	CALC_INTLIST_PERIOD = 0,
	CALC_INTLIST_COLUMNS
};

enum {
	CALC_ENTRIES_COL_ICON = 0,
	CALC_ENTRIES_COL_ICON_SIZE,
	CALC_ENTRIES_COL_NAME,
	CALC_ENTRIES_COL_ID,
	CALC_ENTRIES_COLUMNS
};

enum {
	CALC_CONTROL_SATLONG = 0,
	CALC_CONTROL_SITELONG,
	CALC_CONTROL_SITELAT,
	CALC_CONTROL_SEASON,
	CALC_CONTROL_YEAR,
	CALC_CONTROL_DISHDIAMETER,
	CALC_CONTROL_DISHFOCUSLEN,
	CALC_CONTROL_DISHBAND
};

struct scalcwindow {
	gboolean visible;
	GtkWidget *win;
	gchar *title;
	GtkWidget *caption;
	GtkWidget *close;
	GtkWidget *list;
	GtkTreeSelection *selection;
	GtkWidget *panelcaption;
	GtkWidget *panel;
	gint pageid;
	GtkWidget *page;
	gulong changepagesignal;
};
struct scalcwindow calcwindow;

gint calcwindow_close(GtkWidget *widget, GdkEvent *event, gpointer data);
void calcwindow_init_list(GtkWidget *list);

GtkWidget **calcwindow_place_settings_create(void);
GtkWidget **calcwindow_antenna_settings_create(void);
GtkWidget **calcwindow_alignments_settings_create(void);
GtkWidget **calcwindow_multifeed_settings_create(void);
GtkWidget **calcwindow_trace_settings_create(void);
GtkWidget **calcwindow_outage_settings_create(void);

void calcwindow_place_settings_configure(gpointer page);
void calcwindow_antenna_settings_configure(gpointer page);
void calcwindow_alignments_settings_configure(gpointer page);
void calcwindow_multifeed_settings_configure(gpointer page);
void calcwindow_trace_settings_configure(gpointer page);

typedef GtkWidget **(*calc_page_init_func)(void);
typedef void (*calc_page_configure_func)(gpointer page);
typedef void (*calc_page_apply_func)(gpointer page);

typedef struct {
	GtkWidget **widgets;
	gboolean expired;
	const guint preferenceid;
	const gchar *title;
	const gchar *description;
	const gchar *stock;
	calc_page_init_func init_func;
	calc_page_configure_func configure_func;
	calc_page_apply_func apply_func;
} calcpage;

static calcpage calc_pages[6] = {
	{ NULL, FALSE, CALC_PREFERENCE_ID_PLACE,      N_("Place"),       N_("Coordinates"),       GTK_STOCK_HOME,            calcwindow_place_settings_create,      calcwindow_place_settings_configure,      NULL},
	{ NULL, FALSE, CALC_PREFERENCE_ID_ANTENNA,    N_("Antenna"),     N_("Dish parameters"),   GTK_STOCK_FILE,            calcwindow_antenna_settings_create,    calcwindow_antenna_settings_configure,    NULL},
	{ NULL, FALSE, CALC_PREFERENCE_ID_ALIGNMENTS, N_("Alignments"),  N_("Tuning directions"), GTK_STOCK_GO_FORWARD,      calcwindow_alignments_settings_create, calcwindow_alignments_settings_configure, NULL},
	{ NULL, FALSE, CALC_PREFERENCE_ID_MULTIFEED,  N_("Multifeed"),   N_("LNB positions"),     GTK_STOCK_PAGE_SETUP,      calcwindow_multifeed_settings_create,  calcwindow_multifeed_settings_configure,  NULL},
	{ NULL, FALSE, CALC_PREFERENCE_ID_TRACE,      N_("Trace"),       N_("Sun and moon"),      GTK_STOCK_SORT_DESCENDING, calcwindow_trace_settings_create,      calcwindow_trace_settings_configure,      NULL},
	{ NULL, FALSE, CALC_PREFERENCE_ID_OUTAGE,     N_("Sun outages"), N_("Outage periods"),    GTK_STOCK_INDEX,           calcwindow_outage_settings_create,     NULL,                                     NULL},
};

static gint timer;
static struct _calcdata *calcdata;


gint calcwindow_close(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	gint i;
	
	calcwindow.visible = FALSE;
	
	if (g_signal_handler_is_connected(G_OBJECT(calcwindow.list), calcwindow.changepagesignal)) {
		g_signal_handler_disconnect(G_OBJECT(calcwindow.list), calcwindow.changepagesignal);
	}
	
	for (i=0; i<(sizeof(calc_pages)/sizeof(calcpage)); i++) {
		if (calc_pages[i].widgets != NULL) {
			gtk_widget_destroy(calc_pages[i].widgets[0]);
			g_free(calc_pages[i].widgets);
			calc_pages[i].widgets = NULL;
		}
	}
	
	if (timer > 0) {
		//Stop timer
		g_source_remove(timer);
		timer = 0;
		//Write settings once
		settings_append_float("calc_site_lat", calcdata->sitelat);
		settings_append_float("calc_site_long", calcdata->sitelong);
		settings_append_float("calc_sat_long", calcdata->satlong);
		settings_append_int("calc_int_season", calcdata->season);
		settings_append_int("calc_int_year", calcdata->year);
		settings_append_int("calc_int_band", calcdata->dishband);
		settings_append_float("calc_int_antsize", calcdata->dishdiameter);
		settings_append_float("calc_int_antfocuslen", calcdata->dishfocuslen);
		settings_append_float("calc_mf_sat_long", calcdata->sat2long);
		settings_flush();
		//Free structure
		g_free(calcdata);
		calcdata = NULL;
	}
	
	gtk_widget_destroy(calcwindow.win);
	
	return FALSE;
}

static void calcwindow_action_response(GtkDialog *dialog, gint response_id, gpointer data)
{
	if (response_id == GTK_RESPONSE_CLOSE) {
		g_signal_emit_by_name(G_OBJECT(calcwindow.win), "destroy", NULL);
	} else if (response_id == GTK_RESPONSE_HELP) {
		
	}
}

void calcwindow_init_list(GtkWidget *list)
{
	GtkCellRenderer    *renderer;
	GtkTreeViewColumn  *column1, *column2;
	GtkListStore       *store;
		
	
	renderer = gtk_cell_renderer_pixbuf_new();
	column1 = gtk_tree_view_column_new();
	gtk_tree_view_column_set_title(column1, NULL);
	gtk_tree_view_column_pack_start(column1, renderer, FALSE);
	gtk_tree_view_column_set_attributes(column1, renderer, "stock-id", CALC_ENTRIES_COL_ICON, "stock-size", CALC_ENTRIES_COL_ICON_SIZE, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(list), column1);
	
	renderer = gtk_cell_renderer_text_new();
	g_object_set(renderer, "ellipsize", PANGO_ELLIPSIZE_END, "ellipsize-set", TRUE, NULL);
	column2 = gtk_tree_view_column_new_with_attributes(_("Device"), renderer, "markup", CALC_ENTRIES_COL_NAME, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(list), column2);
	
	store = gtk_list_store_new(CALC_ENTRIES_COLUMNS, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_INT);
	
	gtk_tree_view_set_model(GTK_TREE_VIEW(list), GTK_TREE_MODEL(store));
	g_object_unref(store);
}

gboolean calcwindow_fill_entries(void)
{
	GtkTreeModel *model;
	GtkTreeIter iter;
	gchar caption[256];
	gint len;
	gint i;
	
	if (calcwindow.visible) {
		if ((model = gtk_tree_view_get_model(GTK_TREE_VIEW(calcwindow.list))) != NULL) {
			g_object_ref(model);
			gtk_tree_view_set_model(GTK_TREE_VIEW(calcwindow.list), NULL);
			gtk_list_store_clear(GTK_LIST_STORE(model));
			for (i=0; i<(sizeof(calc_pages)/sizeof(calcpage)); i++) {
				if (calc_pages[i].widgets != NULL) continue;
				len = snprintf((gchar *)&caption, sizeof(caption), "<small><b>%s</b>\n%s</small>", gettext(calc_pages[i].title), gettext(calc_pages[i].description));
				caption[len] = '\0';
				gtk_list_store_append(GTK_LIST_STORE(model), &iter);
				gtk_list_store_set(GTK_LIST_STORE(model), &iter, CALC_ENTRIES_COL_ICON, calc_pages[i].stock,
																CALC_ENTRIES_COL_ICON_SIZE, GTK_ICON_SIZE_BUTTON,
																CALC_ENTRIES_COL_NAME, caption,
																CALC_ENTRIES_COL_ID, i,
																-1);
			}
			gtk_tree_view_set_model(GTK_TREE_VIEW(calcwindow.list), model);
			g_object_unref(model);
		}
	}
	
	return FALSE;
}


void calcwindow_entry_selected(GtkTreeView *tree_view, gpointer data)
{
	GtkTreeSelection *selection;
	GtkTreeModel *model;
	GtkTreeIter iter;
	gint id;
	gchar caption[256];
	gint len;
		
	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree_view));
	if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
		gtk_tree_model_get(model, &iter, CALC_ENTRIES_COL_ID, &id, -1);
		
		if (((calcwindow.pageid == -1) || (calcwindow.pageid != id)) && (id < (sizeof(calc_pages)/sizeof(calcpage)))) {
			if ((calc_pages[id].widgets == NULL) && (calc_pages[id].init_func != NULL)) {
				//Create new page and configure it
				calc_pages[id].widgets = calc_pages[id].init_func();
				if ((calc_pages[id].widgets != NULL) && (calc_pages[id].configure_func != NULL)) {
					calc_pages[id].configure_func(&calc_pages[id]);
				}
			} else if (calc_pages[id].expired) {
				//Reconfigure page
				if ((calc_pages[id].widgets != NULL) && (calc_pages[id].configure_func != NULL)) {
					calc_pages[id].configure_func(&calc_pages[id]);
				}
			} else if (calc_pages[id].init_func == NULL) {
				//No constructor
				g_warning("No constructor for page %i (%s)", id, calc_pages[id].title);
			}
				
			if (calc_pages[id].widgets != NULL) {
				if (calcwindow.page != NULL) {
					if (calcwindow.pageid == -1) {
						//After start page selected
						gtk_widget_set_sensitive(calcwindow.panel, TRUE);
					} else {
						//Increase refcount to avoid page re-creation
						g_object_ref(calcwindow.page);
					}
					//Remove old page
					gtk_container_remove(GTK_CONTAINER(calcwindow.panel), calcwindow.page);
				}
				//Add new page
				calcwindow.page = calc_pages[id].widgets[0];
				gtk_box_pack_end(GTK_BOX(calcwindow.panel), calcwindow.page, TRUE, TRUE, 10);
				calcwindow.pageid = id;
				len = snprintf((gchar *)&caption, sizeof(caption), "<b>%s</b>", gettext(calc_pages[id].title));
				caption[len] = '\0';
				gtk_label_set_markup(GTK_LABEL(calcwindow.panelcaption), caption);
				gtk_widget_show_all(calcwindow.panel);
			}
			
		}
		
	}
}	

void calcwindow_change_numeric(GtkSpinButton *spinbutton, gpointer data)
{
	gint control;
	
	control = GPOINTER_TO_INT(data);
	
	switch (control) {
		case CALC_CONTROL_SATLONG:
			calcdata->satlong = gtk_spin_button_get_value(GTK_SPIN_BUTTON(spinbutton));
			calcwindow_set_expired_page(CALC_PREFERENCE_ID_ALIGNMENTS);
			calcwindow_set_expired_page(CALC_PREFERENCE_ID_MULTIFEED);
			break;
		case CALC_CONTROL_SITELAT:
			calcdata->sitelat = gtk_spin_button_get_value(GTK_SPIN_BUTTON(spinbutton));
			calcwindow_set_expired_page(CALC_PREFERENCE_ID_ALIGNMENTS);
			calcwindow_set_expired_page(CALC_PREFERENCE_ID_MULTIFEED);
			break;
		case CALC_CONTROL_SITELONG:
			calcdata->sitelong = gtk_spin_button_get_value(GTK_SPIN_BUTTON(spinbutton));
			calcwindow_set_expired_page(CALC_PREFERENCE_ID_ALIGNMENTS);
			calcwindow_set_expired_page(CALC_PREFERENCE_ID_MULTIFEED);
			break;
		case CALC_CONTROL_YEAR:
			calcdata->year = gtk_spin_button_get_value(GTK_SPIN_BUTTON(spinbutton));
			break;
		case CALC_CONTROL_DISHDIAMETER:
			calcdata->dishdiameter = gtk_spin_button_get_value(GTK_SPIN_BUTTON(spinbutton));
			break;
		case CALC_CONTROL_DISHFOCUSLEN:
			calcdata->dishfocuslen = gtk_spin_button_get_value(GTK_SPIN_BUTTON(spinbutton));
			calcwindow_set_expired_page(CALC_PREFERENCE_ID_MULTIFEED);
			break;
		default:
			break;
	}
}

void calcwindow_change_preset(GtkComboBox *widget, gpointer data)
{
	gint control;
	
	control = GPOINTER_TO_INT(data);
	
	switch (control) {
		case CALC_CONTROL_SEASON:
			calcdata->season = gtk_combo_box_get_active(GTK_COMBO_BOX(widget));
			break;
		case CALC_CONTROL_DISHBAND:
			calcdata->dishband = gtk_combo_box_get_active(GTK_COMBO_BOX(widget));
			break;
		default:
			break;
	}
}

gboolean calcwindow_update_tracker(gpointer data)
{
	time_t timestamp;
	/*struct tm *timespec;*/
	
	timestamp = time(NULL);
	/*timespec = localtime(&timestamp);
	timestamp = mktime(timespec);*/
	if (calculations_sunmoon_tracker(calcdata->sitelong, calcdata->sitelat, timestamp, &calcdata->sunmoonpos) != NULL) {
		calcwindow_update_page(CALC_PREFERENCE_ID_TRACE);
	}
	
	return TRUE;
}

GtkWidget **calcwindow_place_settings_create(void)
{
	GtkWidget *table;
	GtkWidget *align;
	GtkWidget *satlon, *satlon_l;
	GtkWidget *sitelon, *sitelon_l;
	GtkWidget *sitelat, *sitelat_l;
	GtkWidget *season, *season_l;
	GtkWidget *year;
	GtkWidget *restart;
	GtkWidget **widgets;
	
	table = gtk_table_new(3, 4, FALSE);
	
	satlon_l = gtk_label_new(_("Satellite longitude"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), satlon_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 0, 1, GTK_FILL, GTK_SHRINK, 5, 5);
	
	satlon = gtk_spin_button_new_with_range(-180.0, 180.0, 1.0);
	gtk_spin_button_set_digits(GTK_SPIN_BUTTON(satlon), 2);
	g_signal_connect(G_OBJECT(satlon), "value-changed", G_CALLBACK(calcwindow_change_numeric), GINT_TO_POINTER(CALC_CONTROL_SATLONG));
	gtk_table_attach(GTK_TABLE(table), satlon, 1, 3, 0, 1, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 5);
		
	sitelon_l = gtk_label_new(_("Site longitude"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), sitelon_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 1, 2, GTK_FILL, GTK_SHRINK, 5, 5);
	
	sitelon = gtk_spin_button_new_with_range(-180.0, 180.0, 1.0);
	gtk_spin_button_set_digits(GTK_SPIN_BUTTON(sitelon), 2);
	g_signal_connect(G_OBJECT(sitelon), "value-changed", G_CALLBACK(calcwindow_change_numeric), GINT_TO_POINTER(CALC_CONTROL_SITELONG));
	gtk_table_attach(GTK_TABLE(table), sitelon, 1, 3, 1, 2, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 5);
	
	sitelat_l = gtk_label_new(_("Site latitude"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), sitelat_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 2, 3, GTK_FILL, GTK_SHRINK, 5, 5);
	
	sitelat = gtk_spin_button_new_with_range(-90.0, 90.0, 1.0);
	gtk_spin_button_set_digits(GTK_SPIN_BUTTON(sitelat), 2);
	g_signal_connect(G_OBJECT(sitelat), "value-changed", G_CALLBACK(calcwindow_change_numeric), GINT_TO_POINTER(CALC_CONTROL_SITELAT));
	gtk_table_attach(GTK_TABLE(table), sitelat, 1, 3, 2, 3, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 5);
	
	season_l = gtk_label_new(_("Season/year"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), season_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 3, 4, GTK_FILL, GTK_SHRINK, 5, 5);
	
	#if GTK_CHECK_VERSION(2, 24, 0)
		season = gtk_combo_box_text_new();
		gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(season), _("Spring"));
		gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(season), _("Fall"));
	#else
		season = gtk_combo_box_new_text();
		gtk_combo_box_append_text(GTK_COMBO_BOX(season), _("Spring"));
		gtk_combo_box_append_text(GTK_COMBO_BOX(season), _("Fall"));
	#endif
	gtk_combo_box_set_active(GTK_COMBO_BOX(season), 0);
	g_signal_connect(G_OBJECT(season), "changed", G_CALLBACK(calcwindow_change_preset), GINT_TO_POINTER(CALC_CONTROL_SEASON));
	gtk_table_attach(GTK_TABLE(table), season, 1, 2, 3, 4, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 5);
	
	year = gtk_spin_button_new_with_range(1900.0, 2100.0, 1.0);
	gtk_spin_button_set_digits(GTK_SPIN_BUTTON(year), 0);
	g_signal_connect(G_OBJECT(year), "value-changed", G_CALLBACK(calcwindow_change_numeric), GINT_TO_POINTER(CALC_CONTROL_YEAR));
	gtk_table_attach(GTK_TABLE(table), year, 2, 3, 3, 4, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 5);
	
	align = gtk_alignment_new(0.5, 0.5, 1.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), table);
	
	widgets = (GtkWidget **)g_malloc0(6*sizeof(GtkWidget *));
	widgets[0] = align;
	widgets[1] = satlon;
	widgets[2] = sitelon;
	widgets[3] = sitelat;
	widgets[4] = season;
	widgets[5] = year;
	
	return widgets;
}

void calcwindow_place_settings_configure(gpointer page)
{
	calcpage *pageptr;
	
	pageptr = (calcpage *)page;
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(pageptr->widgets[1]), calcdata->satlong);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(pageptr->widgets[2]), calcdata->sitelong);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(pageptr->widgets[3]), calcdata->sitelat);
	gtk_combo_box_set_active(GTK_COMBO_BOX(pageptr->widgets[4]), calcdata->season);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(pageptr->widgets[5]), calcdata->year);
}

GtkWidget **calcwindow_antenna_settings_create(void)
{
	GtkWidget *table;
	GtkWidget *align;
	GtkWidget *diameter, *diameter_l;
	GtkWidget *focuslen, *focuslen_l;
	GtkWidget *band, *band_l;
	GtkWidget **widgets;
	
	table = gtk_table_new(2, 3, FALSE);
	
	diameter_l = gtk_label_new(_("Diameter (m)"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), diameter_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 0, 1, GTK_FILL, GTK_SHRINK, 5, 5);
	
	diameter = gtk_spin_button_new_with_range(0.45, 5.0, 0.1);
	gtk_spin_button_set_digits(GTK_SPIN_BUTTON(diameter), 2);
	g_signal_connect(G_OBJECT(diameter), "value-changed", G_CALLBACK(calcwindow_change_numeric), GINT_TO_POINTER(CALC_CONTROL_DISHDIAMETER));
	gtk_table_attach(GTK_TABLE(table), diameter, 1, 2, 0, 1, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 5);
		
	focuslen_l = gtk_label_new(_("Focus length (mm)"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), focuslen_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 1, 2, GTK_FILL, GTK_SHRINK, 5, 5);
	
	focuslen = gtk_spin_button_new_with_range(150.0, 1000.0, 1.0);
	gtk_spin_button_set_digits(GTK_SPIN_BUTTON(focuslen), 2);
	g_signal_connect(G_OBJECT(focuslen), "value-changed", G_CALLBACK(calcwindow_change_numeric), GINT_TO_POINTER(CALC_CONTROL_DISHFOCUSLEN));
	gtk_table_attach(GTK_TABLE(table), focuslen, 1, 2, 1, 2, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 5);
	
	band_l = gtk_label_new(_("Band (GHz)"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), band_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 2, 3, GTK_FILL, GTK_SHRINK, 5, 5);
	
	#if GTK_CHECK_VERSION(2, 24, 0)
		band = gtk_combo_box_text_new();
		gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(band), "C (3.4-4.2)");
		gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(band), "Ku (10.7-12.75)");
	#else
		band = gtk_combo_box_new_text();
		gtk_combo_box_append_text(GTK_COMBO_BOX(band), "C (3.4-4.2)");
		gtk_combo_box_append_text(GTK_COMBO_BOX(band), "Ku (10.7-12.75)");
	#endif
	gtk_combo_box_set_active(GTK_COMBO_BOX(band), 0);
	g_signal_connect(G_OBJECT(band), "changed", G_CALLBACK(calcwindow_change_preset), GINT_TO_POINTER(CALC_CONTROL_DISHBAND));
	gtk_table_attach(GTK_TABLE(table), band, 1, 2, 2, 3, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 5);
	
	align = gtk_alignment_new(0.5, 0.5, 1.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), table);
	
	widgets = (GtkWidget **)g_malloc0(4*sizeof(GtkWidget *));
	widgets[0] = align;
	widgets[1] = diameter;
	widgets[2] = focuslen;
	widgets[3] = band;
		
	return widgets;
}

void calcwindow_antenna_settings_configure(gpointer page)
{
	calcpage *pageptr;
		
	pageptr = (calcpage *)page;
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(pageptr->widgets[1]), calcdata->dishdiameter);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(pageptr->widgets[2]), calcdata->dishfocuslen);
	gtk_combo_box_set_active(GTK_COMBO_BOX(pageptr->widgets[3]), calcdata->dishband);
}

GtkWidget **calcwindow_alignments_settings_create(void)
{
	GtkWidget *table;
	GtkWidget *align;
	GtkWidget *azimuth, *azimuth_l;
	GtkWidget *elevation, *elevation_l;
	GtkWidget *polangle, *polangle_l;
	GtkWidget *distance, *distance_l;
	GtkWidget **widgets;
	
	table = gtk_table_new(2, 4, FALSE);
	
	azimuth_l = gtk_label_new(_("Azimuth"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), azimuth_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 0, 1, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
	
	azimuth = gtk_label_new(NULL);
	gtk_label_set_selectable(GTK_LABEL(azimuth), TRUE);
	align = gtk_alignment_new(0.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), azimuth);
	gtk_table_attach(GTK_TABLE(table), align, 1, 2, 0, 1, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
		
	elevation_l = gtk_label_new(_("Elevation"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), elevation_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 1, 2, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
	
	elevation = gtk_label_new(NULL);
	gtk_label_set_selectable(GTK_LABEL(elevation), TRUE);
	align = gtk_alignment_new(0.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), elevation);
	gtk_table_attach(GTK_TABLE(table), align, 1, 2, 1, 2, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
	
	polangle_l = gtk_label_new(_("Polarisation angle"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), polangle_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 2, 3, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
	
	polangle = gtk_label_new(NULL);
	gtk_label_set_selectable(GTK_LABEL(polangle), TRUE);
	align = gtk_alignment_new(0.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), polangle);
	gtk_table_attach(GTK_TABLE(table), align, 1, 2, 2, 3, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
	
	distance_l = gtk_label_new(_("Satellite distance"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), distance_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 3, 4, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
	
	distance = gtk_label_new(NULL);
	gtk_label_set_selectable(GTK_LABEL(distance), TRUE);
	align = gtk_alignment_new(0.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), distance);
	gtk_table_attach(GTK_TABLE(table), align, 1, 2, 3, 4, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
	
	align = gtk_alignment_new(0.5, 0.5, 1.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), table);
	
	widgets = (GtkWidget **)g_malloc0(5*sizeof(GtkWidget *));
	widgets[0] = align;
	widgets[1] = azimuth;
	widgets[2] = elevation;
	widgets[3] = polangle;
	widgets[4] = distance;
		
	return widgets;
}

void calcwindow_alignments_settings_configure(gpointer page)
{
	calcpage *pageptr;
	float az, el, pol, sh;
	gchar strbuf[256];
	
	pageptr = (calcpage *)page;
	az = calculations_azimuth(calcdata->sitelong, calcdata->sitelat, calcdata->satlong, 200.0);
	el = calculations_elevation(calcdata->sitelong, calcdata->sitelat, calcdata->satlong, 200.0);
	pol = calculations_polangle(calcdata->sitelong, calcdata->sitelat, calcdata->satlong); 
	sh = calculations_satheight(calcdata->sitelong, calcdata->sitelat, calcdata->satlong);
	
	if (pageptr->widgets != NULL) {
		g_sprintf(strbuf, "%'.3f°", az);
		gtk_label_set_label(GTK_LABEL(pageptr->widgets[1]), strbuf);
		g_sprintf(strbuf, "%'.3f°", el);
		gtk_label_set_label(GTK_LABEL(pageptr->widgets[2]), strbuf);
		g_sprintf(strbuf, "%'.3f°", pol);
		gtk_label_set_label(GTK_LABEL(pageptr->widgets[3]), strbuf);
		g_sprintf(strbuf, "%'.3f km", sh);
		gtk_label_set_label(GTK_LABEL(pageptr->widgets[4]), strbuf);
	}
}

void calcwindow_change_multifeed_satlon(GtkSpinButton *spinbutton, gpointer data)
{
	calcdata->sat2long = gtk_spin_button_get_value(GTK_SPIN_BUTTON(spinbutton));
	
	calcwindow_update_page(CALC_PREFERENCE_ID_MULTIFEED);
}

GtkWidget **calcwindow_multifeed_settings_create(void)
{
	GtkWidget *table;
	GtkWidget *align;
	GtkWidget *satlon, *satlon_l;
	GtkWidget *centersat, *centersat_l;
	GtkWidget *xoffset, *xoffset_l;
	GtkWidget *yoffset, *yoffset_l;
	GtkWidget **widgets;
	
	table = gtk_table_new(2, 4, FALSE);
	
	satlon_l = gtk_label_new(_("New satellite longitude"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), satlon_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 0, 1, GTK_FILL, GTK_SHRINK, 5, 5);
	
	satlon = gtk_spin_button_new_with_range(-180.0, 180.0, 1.0);
	gtk_spin_button_set_digits(GTK_SPIN_BUTTON(satlon), 2);
	g_signal_connect(G_OBJECT(satlon), "value-changed", G_CALLBACK(calcwindow_change_multifeed_satlon), NULL);
	gtk_table_attach(GTK_TABLE(table), satlon, 1, 2, 0, 1, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 5);
	
	centersat_l = gtk_label_new(_("Central satellite longitude"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), centersat_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 1, 2, GTK_FILL, GTK_SHRINK, 5, 5);
	
	centersat = gtk_label_new(NULL);
	gtk_table_attach(GTK_TABLE(table), centersat, 1, 2, 1, 2, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 5);
		
	xoffset_l = gtk_label_new(_("Horizontal offset (mm)"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), xoffset_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 2, 3, GTK_FILL, GTK_SHRINK, 5, 5);
	
	xoffset = gtk_label_new(NULL);
	gtk_table_attach(GTK_TABLE(table), xoffset, 1, 3, 2, 3, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 5);
	
	yoffset_l = gtk_label_new(_("Vertical offset (mm)"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), yoffset_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 3, 4, GTK_FILL, GTK_SHRINK, 5, 5);
	
	yoffset = gtk_label_new(NULL);
	gtk_table_attach(GTK_TABLE(table), yoffset, 1, 3, 3, 4, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 5);
	
	align = gtk_alignment_new(0.5, 0.5, 1.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), table);
	
	widgets = (GtkWidget **)g_malloc0(5*sizeof(GtkWidget *));
	widgets[0] = align;
	widgets[1] = satlon;
	widgets[2] = centersat;
	widgets[3] = xoffset;
	widgets[4] = yoffset;
		
	return widgets;
}

void calcwindow_multifeed_settings_configure(gpointer page)
{
	calcpage *pageptr;
	gchar strbuf[256];
	gfloat xoffset, yoffset;
	
	pageptr = (calcpage *)page;
	
	if (pageptr->widgets != NULL) {
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(pageptr->widgets[1]), calcdata->sat2long);
		
		calculations_multifeed(calcdata->sitelong, calcdata->sitelat, calcdata->satlong,
								calcdata->sat2long, 200.0, calcdata->dishfocuslen, &xoffset, &yoffset);
		
		g_sprintf(strbuf, "%'.3f°", calcdata->satlong);
		gtk_label_set_text(GTK_LABEL(pageptr->widgets[2]), strbuf);
		
		if (xoffset <= 250.0) {
			g_sprintf(strbuf, "%'.3f", xoffset);
		} else {
			g_sprintf(strbuf, "--");
		}
		gtk_label_set_text(GTK_LABEL(pageptr->widgets[3]), strbuf);
		
		if (yoffset <= 250.0) {
			g_sprintf(strbuf, "%'.3f", yoffset);
		} else {
			g_sprintf(strbuf, "--");
		}
		gtk_label_set_text(GTK_LABEL(pageptr->widgets[4]), strbuf);
	}
}

GtkWidget **calcwindow_trace_settings_create(void)
{
	GtkWidget *table;
	GtkWidget *align;
	GtkWidget *sunazimuth, *sunazimuth_l;
	GtkWidget *sunelevation, *sunelevation_l;
	GtkWidget *sunsat, *sunsat_l;
	GtkWidget *moonazimuth, *moonazimuth_l;
	GtkWidget *moonelevation, *moonelevation_l;
	GtkWidget *moonsat, *moonsat_l;
	GtkWidget **widgets;
	
	table = gtk_table_new(2, 6, FALSE);
	
	sunazimuth_l = gtk_label_new(_("Sun azimuth"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), sunazimuth_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 0, 1, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
	
	sunazimuth = gtk_label_new(NULL);
	gtk_label_set_selectable(GTK_LABEL(sunazimuth), TRUE);
	align = gtk_alignment_new(0.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), sunazimuth);
	gtk_table_attach(GTK_TABLE(table), align, 1, 2, 0, 1, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
		
	sunelevation_l = gtk_label_new(_("Sun elevation"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), sunelevation_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 1, 2, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
	
	sunelevation = gtk_label_new(NULL);
	gtk_label_set_selectable(GTK_LABEL(sunelevation), TRUE);
	align = gtk_alignment_new(0.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), sunelevation);
	gtk_table_attach(GTK_TABLE(table), align, 1, 2, 1, 2, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
	
	sunsat_l = gtk_label_new(_("Satellite behind sun"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), sunsat_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 2, 3, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
	
	sunsat = gtk_label_new(NULL);
	gtk_label_set_selectable(GTK_LABEL(sunsat), TRUE);
	align = gtk_alignment_new(0.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), sunsat);
	gtk_table_attach(GTK_TABLE(table), align, 1, 2, 2, 3, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
	
	moonazimuth_l = gtk_label_new(_("Moon azimuth"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), moonazimuth_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 3, 4, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
	
	moonazimuth = gtk_label_new(NULL);
	gtk_label_set_selectable(GTK_LABEL(moonazimuth), TRUE);
	align = gtk_alignment_new(0.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), moonazimuth);
	gtk_table_attach(GTK_TABLE(table), align, 1, 2, 3, 4, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
	
	moonelevation_l = gtk_label_new(_("Moon elevation"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), moonelevation_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 4, 5, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
	
	moonelevation = gtk_label_new(NULL);
	gtk_label_set_selectable(GTK_LABEL(moonelevation), TRUE);
	align = gtk_alignment_new(0.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), moonelevation);
	gtk_table_attach(GTK_TABLE(table), align, 1, 2, 4, 5, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
	
	moonsat_l = gtk_label_new(_("Satellite behind moon"));
	align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), moonsat_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 5, 6, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
	
	moonsat = gtk_label_new(NULL);
	gtk_label_set_selectable(GTK_LABEL(moonsat), TRUE);
	align = gtk_alignment_new(0.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), moonsat);
	gtk_table_attach(GTK_TABLE(table), align, 1, 2, 5, 6, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 3);
	
	align = gtk_alignment_new(0.5, 0.5, 1.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), table);
	
	widgets = (GtkWidget **)g_malloc0(7*sizeof(GtkWidget *));
	widgets[0] = align;
	widgets[1] = sunazimuth;
	widgets[2] = sunelevation;
	widgets[3] = sunsat;
	widgets[4] = moonazimuth;
	widgets[5] = moonelevation;
	widgets[6] = moonsat;
	
	return widgets;
}

void calcwindow_trace_settings_configure(gpointer page)
{
	calcpage *pageptr;
	gchar strbuf[256];
	
	pageptr = (calcpage *)page;
	
	if (pageptr->widgets != NULL) {
		g_sprintf(strbuf, "%'.3f°", calcdata->sunmoonpos.sunazimuth);
		gtk_label_set_label(GTK_LABEL(pageptr->widgets[1]), strbuf);
		g_sprintf(strbuf, "%'.3f°", calcdata->sunmoonpos.sunaltitude);
		gtk_label_set_label(GTK_LABEL(pageptr->widgets[2]), strbuf);
		if (calcdata->sunmoonpos.sunsat == -1) {
			g_sprintf(strbuf, "--");
		} else {
			g_sprintf(strbuf, "%'.3f°", calcdata->sunmoonpos.sunsat);
		}
		gtk_label_set_label(GTK_LABEL(pageptr->widgets[3]), strbuf);
		
		g_sprintf(strbuf, "%'.3f°", calcdata->sunmoonpos.moonazimuth);
		gtk_label_set_label(GTK_LABEL(pageptr->widgets[4]), strbuf);
		g_sprintf(strbuf, "%'.3f°", calcdata->sunmoonpos.moonaltitude);
		gtk_label_set_label(GTK_LABEL(pageptr->widgets[5]), strbuf);
		if (calcdata->sunmoonpos.moonsat == -1) {
			g_sprintf(strbuf, "--");
		} else {
			g_sprintf(strbuf, "%'.3f°", calcdata->sunmoonpos.moonsat);
		}
		gtk_label_set_label(GTK_LABEL(pageptr->widgets[6]), strbuf);
	}
}

void calcwindow_outage_settings_refresh_foreach(intperiod *data, gpointer user_data)
{
	GtkListStore *store;
	GtkTreeIter iter;
	gchar strbuf[256], strbuf1[16], strbuf2[16], strbuf3[16];
	
	store = (GtkListStore *)user_data;
	
	if (store == NULL) return;
	
	if (strftime(strbuf1, sizeof(strbuf1), "%D", localtime(&data->start)) == 0) {
		g_sprintf(strbuf1, "unknown");
	}
	
	if (strftime(strbuf2, sizeof(strbuf2), "%T %Z", localtime(&data->start)) == 0) {
		g_sprintf(strbuf2, "unknown");
	}
	
	if (strftime(strbuf3, sizeof(strbuf3), "%T %Z", localtime(&data->end)) == 0) {
		g_sprintf(strbuf3, "unknown");
	}
	
	g_snprintf(strbuf, sizeof(strbuf), "%s - %s - %s", strbuf1, strbuf2, strbuf3);
	
	gtk_list_store_append(store, &iter);
	gtk_list_store_set(store, &iter, CALC_INTLIST_PERIOD, strbuf, -1);
	
}

void calcwindow_outage_settings_refresh(GtkButton *button, gpointer data)
{
	GtkListStore *store;
	GSList *reslist = NULL;
	
	store = (GtkListStore *)data;
	
	if (store == NULL) return;
	
	reslist = calculations_interferention_main(calcdata->sitelong, calcdata->sitelat, calcdata->sitelong, calcdata->year,
												calcdata->season, calcdata->dishband, calcdata->dishdiameter);
	if (reslist != NULL) {
		gtk_list_store_clear(store);
		g_slist_foreach(reslist, (GFunc)calcwindow_outage_settings_refresh_foreach, (gpointer)store);
		g_slist_free(reslist);
	}
}

void calcwindow_outage_settings_clear(GtkButton *button, gpointer data)
{
	GtkListStore *store;
	
	store = (GtkListStore *)data;
	
	if (store == NULL) return;
	
	gtk_list_store_clear(store);
}

GtkWidget **calcwindow_outage_settings_create(void)
{
	GtkWidget *table;
	GtkWidget *align;
	GtkWidget *interferention_l;
	GtkWidget *listscroll;
	GtkWidget *list;
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
	GtkListStore *store;
	GtkTreeIter iter;
	GtkTreeSelection *selection;
	GtkWidget *refresh, *clear;
	GtkWidget **widgets;
	
	
	table = gtk_table_new(2, 3, FALSE);
	
	interferention_l = gtk_label_new(_("Sun interferention periods"));
	align = gtk_alignment_new(0.0, 0.5, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), interferention_l);
	gtk_table_attach(GTK_TABLE(table), align, 0, 2, 0, 1, GTK_FILL, GTK_SHRINK, 5, 0);
	
	listscroll = gtk_scrolled_window_new(NULL, NULL);
	
	list = gtk_tree_view_new();
	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(list), FALSE);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(listscroll), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(listscroll), GTK_SHADOW_ETCHED_IN);
	gtk_container_add(GTK_CONTAINER(listscroll), list);
		
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(_("Period"), renderer, "markup", CALC_INTLIST_PERIOD, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(list), column);
	
	store = gtk_list_store_new(CALC_INTLIST_COLUMNS, G_TYPE_STRING);
	gtk_tree_view_set_model(GTK_TREE_VIEW(list), GTK_TREE_MODEL(store));
	g_object_unref(store);
	
	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(list));
	gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);
	
	gtk_table_attach(GTK_TABLE(table), listscroll, 0, 2, 1, 2, GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 5, 3);
	
	refresh = gtk_button_new_from_stock(GTK_STOCK_REFRESH);
	align = gtk_alignment_new(0.0, 0.0, 1.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), refresh);
	g_signal_connect(G_OBJECT(refresh), "clicked", G_CALLBACK(calcwindow_outage_settings_refresh), (gpointer)store);
	gtk_table_attach(GTK_TABLE(table), align, 0, 1, 2, 3, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 0);
	
	clear = gtk_button_new_from_stock(GTK_STOCK_CLEAR);
	align = gtk_alignment_new(0.0, 0.0, 1.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), clear);
	g_signal_connect(G_OBJECT(clear), "clicked", G_CALLBACK(calcwindow_outage_settings_clear), (gpointer)store);
	gtk_table_attach(GTK_TABLE(table), align, 1, 2, 2, 3, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 0);
	
	align = gtk_alignment_new(0.5, 0.5, 1.0, 1.0);
	gtk_container_add(GTK_CONTAINER(align), table);
	
	widgets = (GtkWidget **)g_malloc0(4*sizeof(GtkWidget *));
	widgets[0] = align;
	widgets[1] = list;
	widgets[2] = refresh;
	widgets[3] = clear;
	
	return widgets;
}

void calcwindow_set_expired_page(gint preference)
{
	gint i;
	
	if (calcwindow.visible) {
		for (i=0; i<(sizeof(calc_pages)/sizeof(calcpage)); i++) {
			if (calc_pages[i].preferenceid == preference) {
				if ((calc_pages[i].configure_func != NULL) && (calc_pages[i].widgets != NULL)) {
					calc_pages[i].expired = TRUE;
				}
			}
		}
	}
}

void calcwindow_update_page(gint preference)
{
	gint i;
	
	if (calcwindow.visible) {
		if (preference == CALC_PREFERENCE_ID_MENU) {
			calcwindow_fill_entries();
		} else {
			for (i=0; i<(sizeof(calc_pages)/sizeof(calcpage)); i++) {
				if (calc_pages[i].preferenceid == preference) {
					if ((calc_pages[i].configure_func != NULL) && (calc_pages[i].widgets != NULL)) {
						calc_pages[i].configure_func(&calc_pages[i]);
					} 
				}
			}
		}
	}
}

gboolean calcwindow_create(GtkWindow *parentwin)
{
	GtkWidget *vbox, *hbox, *hpaned;
	GtkWidget *align;
	GtkWidget *listscroll, *panelscroll;
	time_t timestamp;
	struct tm *localtimespec;
	
	if (!calcwindow.visible) {
		calcwindow.visible = TRUE;
		
		/*calcwindow.win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
		calcwindow.title = _("Calculations");
		gtk_window_set_position(GTK_WINDOW(calcwindow.win), GTK_WIN_POS_CENTER);
		gtk_window_set_title(GTK_WINDOW(calcwindow.win), calcwindow.title);
		gtk_widget_set_size_request(calcwindow.win, 500, 300);
		gtk_window_set_default_size(GTK_WINDOW(calcwindow.win), 500, 300);
		gtk_window_set_resizable(GTK_WINDOW(calcwindow.win), FALSE);
		gtk_window_set_transient_for(GTK_WINDOW(calcwindow.win), parentwin);*/
		
		calcwindow.win = gtk_dialog_new_with_buttons(_("Calculations"), 
													GTK_WINDOW(parentwin),
													GTK_DIALOG_DESTROY_WITH_PARENT,
													/*GTK_STOCK_HELP, GTK_RESPONSE_HELP,*/
													GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
													NULL);
		
		gtk_window_set_modal(GTK_WINDOW(calcwindow.win), TRUE);
		gtk_widget_set_size_request(GTK_WIDGET(calcwindow.win), 500, 300);
		gtk_window_set_default_size(GTK_WINDOW(calcwindow.win), 500, 300);
		gtk_window_set_resizable(GTK_WINDOW(calcwindow.win), FALSE);
		
		#if GTK_CHECK_VERSION(3, 0, 0)
			vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
			hpaned = gtk_paned_new(GTK_ORIENTATION_HORIZONTAL);
		#else
			vbox = gtk_vbox_new(FALSE, 0);
			hpaned = gtk_hpaned_new();
		#endif
		gtk_container_set_border_width(GTK_CONTAINER(hpaned), 5);
		
		listscroll = gtk_scrolled_window_new(NULL, NULL);
		
		calcwindow.list = gtk_tree_view_new();
		gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(calcwindow.list), FALSE);
		gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(listscroll), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
		gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(listscroll), GTK_SHADOW_ETCHED_IN);
		
		gtk_container_add(GTK_CONTAINER (listscroll), calcwindow.list);
		calcwindow.selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(calcwindow.list));
		gtk_tree_selection_set_mode(calcwindow.selection, GTK_SELECTION_BROWSE);
		
		calcwindow_init_list(calcwindow.list);
		
		gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(calcwindow.list), TRUE);
		
		gtk_paned_pack1(GTK_PANED(hpaned), listscroll, TRUE, FALSE);
		
		panelscroll = gtk_scrolled_window_new(NULL, NULL);
		gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(panelscroll), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
		gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(panelscroll), GTK_SHADOW_NONE);
		
		gtk_widget_set_size_request(listscroll, 150, -1);
		gtk_widget_set_size_request(panelscroll, 300, -1);
		
		#if GTK_CHECK_VERSION(3, 0, 0)
			calcwindow.panel = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
		#else
			calcwindow.panel = gtk_vbox_new(FALSE, 0);
		#endif
		
		calcwindow.panelcaption = gtk_label_new(NULL);
		gtk_label_set_markup(GTK_LABEL(calcwindow.panelcaption), _("<b>Calculations</b>"));
		gtk_label_set_max_width_chars(GTK_LABEL(calcwindow.panelcaption), 30);
		gtk_label_set_ellipsize(GTK_LABEL(calcwindow.panelcaption), PANGO_ELLIPSIZE_END);
		gtk_box_pack_start(GTK_BOX(calcwindow.panel), calcwindow.panelcaption, FALSE, FALSE, 10);
		
		calcwindow.page = gtk_image_new_from_stock(GTK_STOCK_CONVERT, GTK_ICON_SIZE_DIALOG);
		gtk_box_pack_end(GTK_BOX(calcwindow.panel), calcwindow.page, TRUE, FALSE, 10);
		
		gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(panelscroll), calcwindow.panel);
		
		align = gtk_alignment_new(0.5, 0.5, 1.0, 1.0);
		gtk_container_add(GTK_CONTAINER(align), panelscroll);
		
		gtk_paned_pack2(GTK_PANED(hpaned), align, TRUE, FALSE);
		
		gtk_box_pack_start(GTK_BOX(vbox), hpaned, TRUE, TRUE, 0);
		
		/*calcwindow.close = gtk_button_new_from_stock(GTK_STOCK_CLOSE);
		
		gtk_widget_set_sensitive(calcwindow.panel, FALSE);
		
		hbox = gtk_hbox_new(FALSE, 0);
		gtk_box_pack_end(GTK_BOX(hbox), calcwindow.close, FALSE, FALSE, 5);
		
		gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 5);*/
		
		/*g_signal_connect(G_OBJECT(calcwindow.win), "destroy", G_CALLBACK(calcwindow_close), NULL);
		g_signal_connect(G_OBJECT(calcwindow.close), "clicked", G_CALLBACK(calcwindow_close), NULL);
		g_signal_connect(G_OBJECT(calcwindow.list), "cursor-changed", G_CALLBACK(calcwindow_entry_selected), NULL);*/
		
		timestamp = time(NULL);
		localtimespec = localtime((const time_t *)&timestamp);
		
		calcdata = (struct _calcdata *)g_malloc0(sizeof(struct _calcdata));
		
		calcdata->satlong = settings_read_float("calc_sat_long", 60.00);
		calcdata->sitelong = settings_read_float("calc_site_long", 37.70);
		calcdata->sitelat = settings_read_float("calc_site_lat", 55.75);
		if (localtimespec != NULL) {
			if (localtimespec->tm_mon > 5) {
				calcdata->season = settings_read_int("calc_int_season", 1);
			} else {
				calcdata->season = settings_read_int("calc_int_season", 0);
			}
			calcdata->year = settings_read_int("calc_int_year", localtimespec->tm_year + 1900);
		} else {
			calcdata->season = settings_read_int("calc_int_season", 0);
			calcdata->year = settings_read_int("calc_int_year", 2010);
		}
		calcdata->dishdiameter = settings_read_float("calc_int_antsize", 1.2);
		calcdata->dishfocuslen = settings_read_float("calc_int_antfocuslen", 450.00);
		calcdata->dishband = settings_read_int("calc_int_band", 0);
		calcdata->sat2long = settings_read_float("calc_mf_sat_long", 80.00);
		
		calcwindow.pageid = -1;
		calcwindow_fill_entries();
		
		timer = g_timeout_add(1000, calcwindow_update_tracker, NULL);
		
		/*gtk_container_add(GTK_CONTAINER(calcwindow.win),vbox);
		gtk_widget_show_all(calcwindow.win);
		
		gtk_window_set_focus(GTK_WINDOW(calcwindow.win), calcwindow.list);*/
		
		GtkWidget *dialogarea = gtk_dialog_get_content_area(GTK_DIALOG(calcwindow.win));
		
		gtk_box_pack_start(GTK_BOX(dialogarea), vbox, TRUE, TRUE, 0);
		gtk_widget_show_all(calcwindow.win);
		
		g_signal_connect(G_OBJECT(calcwindow.win), "destroy", G_CALLBACK(calcwindow_close), NULL);
		g_signal_connect(G_OBJECT(calcwindow.win), "response", G_CALLBACK(calcwindow_action_response), NULL);
		calcwindow.changepagesignal = g_signal_connect(G_OBJECT(calcwindow.list), "cursor-changed", G_CALLBACK(calcwindow_entry_selected), NULL);
		
	} else {
		gtk_window_present(GTK_WINDOW(calcwindow.win));
	}
	
	return TRUE;
}
