/*
 *      log.c
 *      
 *      Copyright 2009 - 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <string.h>
#include <limits.h>

#include "log.h"
#include "sysutils.h"


static char *log_cut_linebreak(char *str, size_t *slen);
static char *log_escape_string(char *str, size_t *slen);
static char *log_unescape_string(char *str, size_t *slen, char rsymbol);


static char *log_cut_linebreak(char *str, size_t *slen)
{
	int len;
	char *res;
	
	if (str == NULL) {
		if (slen != NULL) *slen = 0;
		return NULL;
	}
	
	len = strlen(str);
	res = strdup(str);
	
	if (res[len-1] == '\n') {
		if ((len - 1) <= 0) {
			free(res);
			if (slen != NULL) *slen = 0;
			return NULL;
		} else {
			res[len-1] = '\0';
			if (slen != NULL) *slen = len - 1;
			return res;
		}
	}
		
	if (slen != NULL) *slen = len;
	
	return res;
}

static char *log_escape_string(char *str, size_t *slen)
{
	char symbols[6] = {'\\', '\a', '\b', '\t', '\r', '\n'};
	char sequences[6][2] = {"\\\\", "\\a", "\\b", "\\t", "\\r", "\\n"};
	int len, i, n, optr, symfound;
	char *res;
	
	if (str == NULL) {
		if (slen != NULL) *slen = 0;
		return NULL;
	}
	
	len = strlen(str);
	res = (char *)calloc(len*2, 1);
	
	if (res == NULL) {
		if (res != NULL) *slen = 0;
		return NULL;
	}
	
	optr = 0;
	
	for (i=0; i<=len; i++) {
		symfound = 0;
		
		if (i == len) {
			res[optr] = '\n';
			res[optr+1] = '\0';
			optr = optr + 2;
			break;
		}
		
		for (n=0; n<sizeof(symbols); n++) {
			if (str[i] == symbols[n]) {
				res[optr] = sequences[n][0];
				res[optr+1] = sequences[n][1];
				optr = optr + 2;
				symfound = 1;
				break;
			}
		}
		
		if (!symfound) {
			res[optr] = str[i];
			optr++;
		}
	
	}
	
	if (slen != NULL) *slen = optr - 1;
	
	return res;
}

static char *log_unescape_string(char *str, size_t *slen, char rsymbol)
{
	char symbols[6] = {'\\', '\a', '\b', '\t', '\r', '\n'};
	char sequences[6][2] = {"\\\\", "\\a", "\\b", "\\t", "\\r", "\\n"};
	int len, i, n, optr, symfound;
	char *res;
	
	if (str == NULL) {
		if (slen != NULL) *slen = 0;
		return NULL;
	}
	
	len = strlen(str);
	res = (char *)calloc(len, 1);
	
	if (res == NULL) {
		if (res != NULL) *slen = 0;
		return NULL;
	}
	
	optr = 0;
	
	for (i=0; i<len; i++) {
		symfound = 0;
		
		if (str[i] == '\\') {
			for (n=0; n<sizeof(sequences)/2; n++) {
				if (str[i+1] == sequences[n][1]) {
					if (rsymbol != '\0') {
						res[optr] = rsymbol;
					} else {
						res[optr] = symbols[n];
					}
					symfound = 1;
					optr++;
					break;
				}
			}
		}
		
		if (i < len-1) {
			if (!symfound) {
				res[optr] = str[i];
				optr++;
			} else {
				i++;
			}
		} else {
			res[optr] = '\0';
			optr++;
		}
		
	}
	
	if (slen != NULL) *slen = optr - 1;
	
	return res;
}

int log_add_message_static(struct log_t *log, int flags, char *message)
{
	char *rmsg;
	char ecode;
	time_t tmspec;
	struct tm *tms;
	char timebuf[30];
	int buflen;
	char *entry, *rentry, *cbentry;
	size_t rmsglen, entrylen, rentrylen, cbentrylen;
	
	
	if ((log == NULL) || (message == NULL)) return 0;
	
	if ((rmsg = log_cut_linebreak(message, &rmsglen)) == NULL) return 0;
	
	tmspec = time(NULL);
	
	if (flags & LOG_EVENT_MESSAGE) {
		ecode = '+';
		flags |= LOG_OUTPUT_STDOUT;
	} else if (flags & LOG_EVENT_WARNING) {
		ecode = '!';
		flags |= LOG_OUTPUT_STDOUT;
	} else if (flags & LOG_EVENT_ERROR) {
		ecode = '-';
		flags |= LOG_OUTPUT_STDERR;
	} else if (flags & LOG_EVENT_PRIVELEGED) {
		ecode = '^';
		flags |= LOG_OUTPUT_STDOUT;
	} else if (flags & LOG_EVENT_UNDEFINED) {
		ecode = '?';
		flags |= LOG_OUTPUT_STDOUT;
	} else {
		ecode = '?';
		flags |= LOG_OUTPUT_STDOUT;
	}
	
	if ((log->filefd != -1) && ((rentry = log_escape_string(rmsg, &rentrylen)) != NULL)) {
		entrylen = 40 + rentrylen;
	} else {
		entrylen = 40 + rmsglen;
	}
	
	if ((entry = (char *)calloc(entrylen, 1)) == NULL) {
		free(rmsg);
		if (rentry != NULL) free(rentry);
		return 0;
	}
	
	if ((log->filefd != -1) && (rentry != NULL)) {
		buflen = sprintf(entry, "[%c] %u ", ecode, tmspec);
		entry = strcat(entry, rentry);
		
		write(log->filefd, entry, buflen+rentrylen);
		
		if (log->addfunc != NULL) {
			cbentry = log_unescape_string(entry, &cbentrylen, ' ');
			log->addfunc(flags, tmspec, cbentry);
			free(cbentry);
		}
		
		free(rentry);
	}
	
	if (!log->silent) {
		if ((tms = localtime(&tmspec)) != NULL) {
			strftime(timebuf, sizeof(timebuf), "%T", tms);
		} else {
			sprintf(timebuf, "00:00:00");
		}
		
		buflen = sprintf(entry, "[%c] %s ", ecode, timebuf);
		entry = strcat(entry, rmsg);
		
		if (flags & LOG_OUTPUT_STDOUT) {
			fprintf(stdout, "%s\n", entry);
		} else if (flags & LOG_OUTPUT_STDERR) {
			fprintf(stderr, "%s\n", entry);
		}
	}
	
	free(rmsg);
	free(entry);
	
	return 1;
}

int log_add_message(struct log_t *log, int flags, char *format, ...)
{
	va_list arg;
	char *rformat;
	char ecode;
	time_t tmspec;
	struct tm *tms;
	char timebuf[30];
	int buflen, size, fentrylen;
	char *fentry, *entry, *rentry, *cbentry;
	size_t entrylen, rentrylen, cbentrylen;
	
	
	if ((log == NULL) || (format == NULL)) return 0;
	
	if ((rformat = log_cut_linebreak(format, NULL)) == NULL) return 0;
	
	tmspec = time(NULL);
	
	if (flags & LOG_EVENT_MESSAGE) {
		ecode = '+';
		flags |= LOG_OUTPUT_STDOUT;
	} else if (flags & LOG_EVENT_WARNING) {
		ecode = '!';
		flags |= LOG_OUTPUT_STDOUT;
	} else if (flags & LOG_EVENT_ERROR) {
		ecode = '-';
		flags |= LOG_OUTPUT_STDERR;
	} else if (flags & LOG_EVENT_PRIVELEGED) {
		ecode = '^';
		flags |= LOG_OUTPUT_STDOUT;
	} else if (flags & LOG_EVENT_UNDEFINED) {
		ecode = '?';
		flags |= LOG_OUTPUT_STDOUT;
	} else {
		ecode = '?';
		flags |= LOG_OUTPUT_STDOUT;
	}
	
	size = 128;
	
	if ((fentry = calloc(size, 1)) == NULL) {
		free(rformat);
		return 0;
	}
		
	while (1) {
		va_start(arg, format);
		fentrylen = vsnprintf(fentry, size, rformat, arg);
		va_end(arg);
		
		if ((fentrylen > -1) && (fentrylen < size)) break;
		
		if (fentrylen > -1) {
			size = fentrylen + 1;
		} else {
			size *= 2;
		}
		
		if ((fentry = realloc(fentry, size)) == NULL) {
			free(rformat);
			return 0;
		}
	}
	
	if ((log->filefd != -1) && ((rentry = log_escape_string(fentry, &rentrylen)) != NULL)) {
		entrylen = 40 + rentrylen;
	} else {
		entrylen = 40 + fentrylen;
	}
	
	if ((entry = (char *)calloc(entrylen, 1)) == NULL) {
		free(rformat);
		if (rentry != NULL) free(rentry);
		return 0;
	}
	
	if ((log->filefd != -1) && (rentry != NULL)) {
		buflen = sprintf(entry, "[%c] %u ", ecode, tmspec);
		entry = strcat(entry, rentry);
		
		write(log->filefd, entry, buflen+rentrylen);
		
		if (log->addfunc != NULL) {
			cbentry = log_unescape_string(entry, &cbentrylen, ' ');
			log->addfunc(flags, tmspec, cbentry);
			free(cbentry);
		}
		
		free(rentry);
	}
	
	if (!log->silent) {
		if ((tms = localtime(&tmspec)) != NULL) {
			strftime(timebuf, sizeof(timebuf), "%T", tms);
		} else {
			sprintf(timebuf, "00:00:00");
		}
		
		buflen = sprintf(entry, "[%c] %s ", ecode, timebuf);
		entry = strcat(entry, fentry);
		
		if (flags & LOG_OUTPUT_STDOUT) {
			fprintf(stdout, "%s\n", entry);
		} else if (flags & LOG_OUTPUT_STDERR) {
			fprintf(stderr, "%s\n", entry);
		}
	}
		
	free(fentry);
	free(entry);
	
	return 1;
}

int log_read_messages(struct log_t *log)
{
	struct stat statbuf;
	FILE *stream;
	char strbuf[4096];
	char *rentry;
	size_t rentrylen;
	char *space1, *space2;
	char timestr[30];
	time_t timestamp;
	int flags;
	char *entry;
	int res, linesnum;
	
	
	if (log == NULL) return 0;
	
	if ((stat(log->filename, &statbuf) == -1) || (log->readfunc == NULL)) return 0;
	
	if (statbuf.st_mtime == log->readtime) return 1;
	
	stream = fopen(log->filename, "r");
	
	linesnum = 0;
	
	while (fgets((char *)&strbuf, sizeof(strbuf), stream) != NULL) {
		
		rentry = log_unescape_string((char *)&strbuf, &rentrylen, ' ');
		
		linesnum++;
		
		if ((rentry == NULL) || (log->readlines > linesnum)) continue;
		
		space1 = strchr(rentry, ' ');
		space2 = strchr(space1 + 1, ' ');
		
		if ((space1 == NULL) || (space2 == NULL)) continue;
		
		if (rentry[1] == '+') {
			flags = LOG_EVENT_MESSAGE;
		} else if (rentry[1] == '!') {
			flags = LOG_EVENT_WARNING;
		} else if (rentry[1] == '-') {
			flags = LOG_EVENT_ERROR;
		} else if (rentry[1] == '^') {
			flags = LOG_EVENT_PRIVELEGED;
		} else if (rentry[1] == '?') {
			flags = LOG_EVENT_UNDEFINED;
		} else {
			flags = LOG_EVENT_UNDEFINED;
		}
		
		strncpy((char *)&timestr, space1 + 1, space2 - space1 - 1);
		
		timestamp = strtoul((char *)&timestr, (char **)NULL, 10);
		
		entry = space2 + 1;
		
		res = log->readfunc(flags, timestamp, entry);
		
		free(rentry);
		
		if (res == 0) break;
	}
	
	log->readlines = linesnum;
	log->readtime = statbuf.st_mtime;
		
	fclose(stream);
	
	return 1;
}

struct log_t *log_init_instance(char *program, char *filename, int readonly, int silent)
{
	const char *homedir;
	char fname[PATH_MAX];
	char *filepath;
	struct log_t *log;
	int filefd, flags;
	struct flock lock;
		
	if ((program == NULL) || (filename == NULL)) return NULL;
	
	if ((homedir = sysutils_gethomedir()) == NULL) return NULL;
	
	snprintf((char *)&fname, sizeof(fname), "%s/.config/%s/log/%s", homedir, program, filename);
	
	if (access((char *)&fname, readonly ? R_OK : W_OK) != 0) {
		filepath = sysutils_getfilepath((char *)&fname);
		if (!sysutils_mkdirs(filepath, S_IRWXU)) {
			free(filepath);
			return NULL;
		}
		free(filepath);
	}
		
	if (!readonly) {
		filefd = open(fname, O_CREAT | O_WRONLY | O_TRUNC, S_IRWXU);
		if (filefd != -1) {
			lock.l_type = F_WRLCK;
			lock.l_whence = SEEK_SET;
			lock.l_start = 0;
			lock.l_len = 0;
			
			if (fcntl(filefd, F_SETLK, &lock) == -1) {
				close(filefd);
				filefd = -1;
			}
		}
	} else {
		filefd = -1;
	}
	
	if ((log = (struct log_t *)calloc(sizeof(struct log_t), 1)) != NULL) {
		log->filefd = filefd;
		log->filename = strdup(fname);
		log->silent = silent;
		log->readlines = 0;
		log->readtime = 0;
		log->readfunc = NULL;
		log->addfunc = NULL;
	}
	
	return log;
}

void log_set_callbacks(struct log_t *log, log_read_func readfunc, log_add_func addfunc)
{
	if (log == NULL) return;
	
	log->readfunc = readfunc;
	if (log->filefd != -1) {
		log->addfunc = addfunc;
	} else {
		log->addfunc = NULL;
	}
}
	

void log_free_instance(struct log_t *log)
{
	if (log == NULL) return;
	
	free(log->filename);
	close(log->filefd);
	
	free(log);
}
