/*
 *      serialize.c
 *      
 *      Copyright 2009 - 2010 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <arpa/inet.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#define SERIALIZE_FORMAT_LENGTH 256

#define ntohll(x) ( ( (unsigned long long)(ntohl( (unsigned int)((x << 32) >> 32) )) << 32) | ntohl( ((unsigned int)(x >> 32)) ) )
#define htonll(x) ntohll(x)

enum {
	SERIALIZE_TYPE_BYTE = 0,
	SERIALIZE_TYPE_USHORT,
	SERIALIZE_TYPE_UINT,
	SERIALIZE_TYPE_UINT64,
	SERIALIZE_TYPE_STRING
};

unsigned int serialize_get_size(char *format)
{
	unsigned int count;
	char formatstr[SERIALIZE_FORMAT_LENGTH];
	unsigned int bytes;
	int i, p;
	
	count = 0;
	i = 0;
	while (format[i]) {
		if (format[i] == '%') count++;
		i++;
	}
	
	if (count == 0) return 0;
	
	p = 0;
	bytes = 0;
	while ((*format) && (p <= count)) {
		for (i = 0; format[i] && format[i] != '%'; i++)
			formatstr[i] = format[i];
		if ((i) && (p <= count)) {
			formatstr[i] = '\0';
			format += i;
			bytes += (size_t)atoi(formatstr);
		} else if (p < count){
			for (i = 0; !isalpha(format[i]); i++) {
				formatstr[i] = format[i];
				if (i && format[i] == '%')
					break;
			}
			formatstr[i] = format[i];
			formatstr[i + 1] = '\0';
			format += i + 1;
			
			switch (formatstr[1]) {
				case 'b': 
					bytes += sizeof(char);
					p++;
					break;
				case 'h': 
					bytes += sizeof(short);
					p++;
					break;
				case 'i': 
					bytes += sizeof(int);
					p++;
					break;
				case 'l': 
					bytes += sizeof(long long);
					p++;
					break;
				case 's': 
					p++;
					break;
				default:
					break;
			}	
	
		}
	}
	
	return bytes;
}

static unsigned int serialize_get_parameters(char *format, unsigned int **parametersv, unsigned int *countv)
{
	unsigned int count;
	unsigned int *parameters;
	char formatstr[SERIALIZE_FORMAT_LENGTH];
	unsigned int bytes;
	int i, p;
	
	count = 0;
	i = 0;
	while (format[i]) {
		if (format[i] == '%') count++;
		i++;
	}
	
	if ((count == 0) || ((parameters = (unsigned int *)malloc(sizeof(int)*count*2)) == NULL)) return 0;
	
	if (countv != NULL) *countv = count;
	
	if (parametersv != NULL) {
		p = 0;
		bytes = 0;
		while ((*format) && (p <= count)) {
			for (i = 0; format[i] && format[i] != '%'; i++)
				formatstr[i] = format[i];
			if ((i) && (p <= count)) {
				formatstr[i] = '\0';
				format += i;
				if (parameters[(p-1)*2] == SERIALIZE_TYPE_STRING) {
					parameters[(p-1)*2+1] = (size_t)atoi(formatstr);
					bytes += parameters[(p-1)*2+1];
				}
			} else if (p < count){
				for (i = 0; !isalpha(format[i]); i++) {
					formatstr[i] = format[i];
					if (i && format[i] == '%')
						break;
				}
				formatstr[i] = format[i];
				formatstr[i + 1] = '\0';
				format += i + 1;
				
				switch (formatstr[1]) {
					case 'b': 
						parameters[p*2]   = SERIALIZE_TYPE_BYTE;
						parameters[p*2+1] = sizeof(char);
						bytes += sizeof(char);
						p++;
						break;
					case 'h': 
						parameters[p*2]   = SERIALIZE_TYPE_USHORT;
						parameters[p*2+1] = sizeof(short);
						bytes += sizeof(short);
						p++;
						break;
					case 'i': 
						parameters[p*2]   = SERIALIZE_TYPE_UINT;
						parameters[p*2+1] = sizeof(int);
						bytes += sizeof(int);
						p++;
						break;
					case 'l': 
						parameters[p*2]   = SERIALIZE_TYPE_UINT64;
						parameters[p*2+1] = sizeof(long long);
						bytes += sizeof(long long);
						p++;
						break;
					case 's': 
						parameters[p*2]   = SERIALIZE_TYPE_STRING;
						parameters[p*2+1] = 0;
						p++;
						break;
					default:
						break;
				}	
		
			}
		}
		
		*parametersv = parameters;
	}
	
	return bytes;
}

static void serialize_add_byte(char value, char **buffer, unsigned int *position)
{
	memcpy((void *)*buffer+*position, (const void *)&value, sizeof(char));
	*position += sizeof(char);
}

static void serialize_add_ushort(unsigned short value, int endianess, char **buffer, unsigned int *position)
{
	if (endianess) value = htons(value);
	memcpy((void *)*buffer+*position, (const void *)&value, sizeof(unsigned short));
	*position += sizeof(unsigned short);
}

static void serialize_add_uint(unsigned int value, int endianess, char **buffer, unsigned int *position)
{
	if (endianess) value = htonl(value);
	memcpy((void *)*buffer+*position, (const void *)&value, sizeof(unsigned int));
	*position += sizeof(unsigned int);
}

static void serialize_add_uint64(unsigned long long value, int endianess, char **buffer, unsigned int *position)
{
	if (endianess) value = htonll(value);
	memcpy((void *)*buffer+*position, (const void *)&value, sizeof(unsigned long long));
	*position += sizeof(unsigned long long);
}

static void serialize_add_string(char *value, size_t length, char **buffer, unsigned int *position)
{
	if (value != NULL) {
		strncpy((char *)*buffer+*position, (const char *)value, length);
	} else {
		memset((char *)*buffer+*position, '\0', length);
	}
	*position += length;
}


char *serialize_form_message(char *format, int endianess, size_t *size, ...)
{
	va_list args;
	unsigned int bytes;
	char *buffer;
	int i;
	char bytevalue;
	unsigned short ushortvalue;
	unsigned int uintvalue;
	unsigned long long uint64value;
	char *stringvalue;
	unsigned int position;
	unsigned int *params;
	unsigned int count; 
	
	bytes = serialize_get_parameters(format, &params, &count);
	
	if (size != NULL) *size = (size_t)bytes;	
	
	if ((bytes == 0) || ((buffer = (char *)malloc(bytes)) == NULL)) {
		if (params != NULL) free(params);
		return NULL;
	}
		
	va_start(args, size);
	
	position = 0;
	
	for (i=0; i<count*2; i+=2) {
		switch (params[i]) {
			case SERIALIZE_TYPE_BYTE: 
				bytevalue = (char)va_arg(args, int);
				serialize_add_byte(bytevalue, &buffer, &position);
				break;
			case SERIALIZE_TYPE_USHORT: 
				ushortvalue = (unsigned short)va_arg(args, int);
				serialize_add_ushort(ushortvalue, endianess, &buffer, &position);
				break;
			case SERIALIZE_TYPE_UINT: 
				uintvalue = (unsigned int)va_arg(args, int);
				serialize_add_uint(uintvalue, endianess, &buffer, &position);
				break;
			case SERIALIZE_TYPE_UINT64: 
				uint64value = (unsigned long long)va_arg(args, long long);
				serialize_add_uint64(uint64value, endianess, &buffer, &position);
				break;
			case SERIALIZE_TYPE_STRING: 
				stringvalue = va_arg(args, char *);
				serialize_add_string(stringvalue, params[i+1], &buffer, &position);
				break;
			default:
				break;
		}
				
	}
	
	va_end(args);
	
	if (params != NULL) free(params);
	
	return buffer;
}

static char serialize_get_byte(char **buffer, unsigned int *position)
{
	char value;
	
	memcpy((void *)&value, (const void *)*buffer+*position, sizeof(char));
	*position += sizeof(char);
	return value;
}

static unsigned short serialize_get_ushort(int endianess, char **buffer, unsigned int *position)
{
	unsigned short value;
	
	memcpy((void *)&value, (const void *)*buffer+*position, sizeof(unsigned short));
	*position += sizeof(unsigned short);
	if (endianess) value = ntohs(value);
	return value;
}

static unsigned int serialize_get_uint(int endianess, char **buffer, unsigned int *position)
{
	unsigned int value;
	
	memcpy((void *)&value, (const void *)*buffer+*position, sizeof(unsigned int));
	*position += sizeof(unsigned int);
	if (endianess) value = ntohl(value);
	return value;
}

static unsigned long long serialize_get_uint64(int endianess, char **buffer, unsigned int *position)
{
	unsigned long long value;
	
	memcpy((void *)&value, (const void *)*buffer+*position, sizeof(unsigned long long));
	*position += sizeof(unsigned long long);
	if (endianess) value = ntohll(value);
	return value;
}

static char *serialize_get_string(size_t length, char **buffer, unsigned int *position)
{
	char *value;
	
	if ((length > 0) && ((value = malloc(length+1)) != NULL)) {
		strncpy((char *)value, (const char *)*buffer+*position, length);
		*position += length;
		value[length] = '\0';
		return value;
	} else {
		return NULL;
	}
}

static void serialize_get_string_static(char *value, size_t length, char **buffer, unsigned int *position)
{
	if (length > 0) {
		strncpy((char *)value, (const char *)*buffer+*position, length);
		*position += length;
		value[length] = '\0';
	}
}

int serialize_parse_message(char *format, char *message, int endianess, ...)
{
	va_list args;
	unsigned int bytes;
	int i;
	unsigned int position;
	unsigned int *params;
	unsigned int count; 
	
	
	bytes = serialize_get_parameters(format, &params, &count);
	
	if ((bytes == 0) || (message == NULL)) {
		if (params != NULL) free(params);
		return 0;
	}
		
	va_start(args, endianess);
	
	position = 0;
	
	for (i=0; i<count*2; i+=2) {
		switch (params[i]) {
			case SERIALIZE_TYPE_BYTE: 
				*(char *)va_arg(args, void *) = serialize_get_byte(&message, &position);
				break;
			case SERIALIZE_TYPE_USHORT: 
				*(unsigned short *)va_arg(args, void *) = serialize_get_ushort(endianess, &message, &position);
				break;
			case SERIALIZE_TYPE_UINT: 
				*(unsigned int *)va_arg(args, void *) = serialize_get_uint(endianess, &message, &position);
				break;
			case SERIALIZE_TYPE_UINT64: 
				*(unsigned long long *)va_arg(args, void *) = serialize_get_uint64(endianess, &message, &position);
				break;
			case SERIALIZE_TYPE_STRING: 
				*(char **)va_arg(args, void **) = serialize_get_string(params[i+1], &message, &position);
				break;
			default:
				break;
		}
				
	}
	
	va_end(args);
	
	if (params != NULL) free(params);
	
	return bytes;
}

int serialize_parse_message_static(char *format, char *message, int endianess, ...)
{
	va_list args;
	unsigned int bytes;
	int i;
	unsigned int position;
	unsigned int *params;
	unsigned int count; 
	
	
	bytes = serialize_get_parameters(format, &params, &count);
	
	if ((bytes == 0) || (message == NULL)) {
		if (params != NULL) free(params);
		return 0;
	}
		
	va_start(args, endianess);
	
	position = 0;
	
	for (i=0; i<count*2; i+=2) {
		switch (params[i]) {
			case SERIALIZE_TYPE_BYTE: 
				*(char *)va_arg(args, void *) = serialize_get_byte(&message, &position);
				break;
			case SERIALIZE_TYPE_USHORT: 
				*(unsigned short *)va_arg(args, void *) = serialize_get_ushort(endianess, &message, &position);
				break;
			case SERIALIZE_TYPE_UINT: 
				*(unsigned int *)va_arg(args, void *) = serialize_get_uint(endianess, &message, &position);
				break;
			case SERIALIZE_TYPE_UINT64: 
				*(unsigned long long *)va_arg(args, void *) = serialize_get_uint64(endianess, &message, &position);
				break;
			case SERIALIZE_TYPE_STRING: 
				serialize_get_string_static((char *)va_arg(args, void *), params[i+1], &message, &position);
				break;
			default:
				break;
		}
				
	}
	
	va_end(args);
	
	if (params != NULL) free(params);
	
	return bytes;
}

char *serialize_form_array_message(char *format, char cmd, unsigned int rows, size_t *size, void (callback)(unsigned int *params, unsigned int count, int seq, int size, char **buffer, void *data), void *data)
{
	va_list args;
	unsigned int bytes;
	char *buffer;
	int i;
	unsigned int *params;
	unsigned int count;
	unsigned int position; 
	
	if ((callback == NULL) || (rows <= 0)) return NULL;
	
	bytes = serialize_get_parameters(format, &params, &count);
	
	if (size != NULL) *size = (size_t)(bytes*rows)+sizeof(unsigned int)+sizeof(char);
	
	if ((bytes == 0) || ((buffer = (char *)malloc((bytes*rows)+sizeof(unsigned int)+sizeof(char))) == NULL)) {
		if (params != NULL) free(params);
		return NULL;
	}
	
	position = 0;
	
	serialize_add_byte(cmd, &buffer, &position);
	serialize_add_uint(rows, 0, &buffer, &position);
	
	for (i=0; i<rows; i++) {
		callback(params, count, i, bytes, &buffer, data);
	}
	
	if (params != NULL) free(params);
	
	return buffer;
}

void serialize_form_message_buffer(unsigned int *params, unsigned int count, char **buf, unsigned int offset, ...)
{
	va_list args;
	char *buffer;
	int i;
	char bytevalue;
	unsigned short ushortvalue;
	unsigned int uintvalue;
	unsigned long long uint64value;
	char *stringvalue;
	unsigned int position;
	
	buffer = (char *)*buf+offset;
	
	va_start(args, offset);
	
	position = 0;
	
	for (i=0; i<count*2; i+=2) {
		switch (params[i]) {
			case SERIALIZE_TYPE_BYTE: 
				bytevalue = (char)va_arg(args, int);
				serialize_add_byte(bytevalue, &buffer, &position);
				break;
			case SERIALIZE_TYPE_USHORT: 
				ushortvalue = (unsigned short)va_arg(args, int);
				serialize_add_ushort(ushortvalue, 0, &buffer, &position);
				break;
			case SERIALIZE_TYPE_UINT: 
				uintvalue = (unsigned int)va_arg(args, int);
				serialize_add_uint(uintvalue, 0, &buffer, &position);
				break;
			case SERIALIZE_TYPE_UINT64: 
				uint64value = (unsigned long long)va_arg(args, long long);
				serialize_add_uint64(uint64value, 0, &buffer, &position);
				break;
			case SERIALIZE_TYPE_STRING: 
				stringvalue = va_arg(args, char *);
				serialize_add_string(stringvalue, params[i+1], &buffer, &position);
				break;
			default:
				break;
		}
				
	}
	
	va_end(args);
}


int serialize_parse_array_message(char *format, unsigned int *arows, char *message, char *acmd, void (callback)(unsigned int *params, unsigned int count, int seq, int size, char **message, void *data), void *data)
{
	//va_list args;
	unsigned int bytes;
	int i;
	unsigned int *params;
	unsigned int count; 
	unsigned int position;
	unsigned int rows;
	char cmd;
	
	bytes = serialize_get_parameters(format, &params, &count);
	
	if ((bytes == 0) || (message == NULL)) {
		if (params != NULL) free(params);
		return 0;
	}
	
	position = 0;
	
	cmd = serialize_get_byte(&message, &position);
	
	if (acmd != NULL) *acmd = cmd;
	
	rows = serialize_get_uint(0, &message, &position);
	
	if (arows != NULL) *arows = rows;
		
	if (rows == 0) {
		if (params != NULL) free(params);
		return 0;
	}
	
	for (i=0; i<rows; i++) {
		callback(params, count, i, bytes, &message, data);
	}
		
	if (params != NULL) free(params);
	
	return bytes*rows;
}


void serialize_parse_message_buffer(unsigned int *params, unsigned int count, char **msg, unsigned int offset, ...)
{
	va_list args;
	char *message;
	int i;
	unsigned int position;
	
		
	message = (char *)*msg+offset;
	
	va_start(args, offset);
	
	position = 0;
	
	for (i=0; i<count*2; i+=2) {
		switch (params[i]) {
			case SERIALIZE_TYPE_BYTE: 
				*(char *)va_arg(args, void *) = serialize_get_byte(&message, &position);
				break;
			case SERIALIZE_TYPE_USHORT: 
				*(unsigned short *)va_arg(args, void *) = serialize_get_ushort(0, &message, &position);
				break;
			case SERIALIZE_TYPE_UINT: 
				*(unsigned int *)va_arg(args, void *) = serialize_get_uint(0, &message, &position);
				break;
			case SERIALIZE_TYPE_UINT64: 
				*(unsigned long long *)va_arg(args, void *) = serialize_get_uint64(0, &message, &position);
				break;
			case SERIALIZE_TYPE_STRING: 
				//*(char **)va_arg(args, void **) = serialize_get_string(params[i+1], &message, &position);
				serialize_get_string_static((char *)va_arg(args, void *), params[i+1], &message, &position);
				break;
			default:
				break;
		}
				
	}
	
	va_end(args);
}

void serialize_fill_static_string(char *svalue, char *value, size_t length)
{
	size_t vlength;
	
	if (value == NULL) return;
		
	if (value[0] == '\0') {
		free(value);
		return;
	}  
		
	vlength = strlen(value);
	if (vlength >= length) {
		strncpy(svalue, value, length);
		svalue[length] = '\0';
	} else {
		memset(svalue, '\0', length);
		strncpy(svalue, value, vlength);
	}
	free(value);
}

