/*
 *      sysutils.c
 *      
 *      Copyright 2009 - 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pwd.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <sys/stat.h>


char *sysutils_strdup_printf(const char *format, ...)
{
	int n, size;
	char *str;
	va_list params;
	
	if ((str = calloc(size, 1)) == NULL) return NULL;
	
	size = 128;
	
	while (1) {
		va_start(params, format);
		
		n = vsnprintf(str, size, format, params);
		
		va_end (params);
		
		if ((n > -1) && (n < size)) break;
		
		if (n > -1) {
			size = n + 1;
		} else {
			size *= 2;
		}
		
		if ((str = realloc(str, size)) == NULL) {
			return NULL;
		}
	}
	
	return str;
}

const char *sysutils_gethomedir(void)
{
	const char *homedir;
	struct passwd *passwd;
		
	homedir = getenv("HOME");
	
	if (homedir != NULL) {
		if (access(homedir, F_OK) == 0) {
			return homedir;
		}
	}
	
	passwd = getpwuid(getuid());
	
	if (passwd != NULL) {
		if (access(homedir, F_OK) == 0) {
			return passwd->pw_dir;
		}
	}
	
	return NULL;
}

const char *sysutils_getusername(void)
{
	const char *username;
	struct passwd *passwd;
		
	username = getenv("USERNAME");
	
	if (username != NULL) {
		return username;
	}
		
	passwd = getpwuid(getuid());
	
	if (passwd != NULL) {
		return passwd->pw_name;
	}
	
	return NULL;
}

char *sysutils_getfilepath(char *path)
{
	char *fullpath;
	size_t len;
	int i;
	
	if (path == NULL) return NULL;
	
	if (strchr(path, '/') == NULL) return NULL;
	
	fullpath = strdup(path);
	len = strlen(path);
	
	if (fullpath == NULL) return NULL;
	
	for (i=len-1; i>=0; i--) {
		if (fullpath[i] == '/') {
			fullpath[i] = '\0';
			break;
		}
	}
	
	return fullpath;
}

int sysutils_mkdirs(char *path, mode_t mode)
{
	char *fullpath;
	char *rootpath;
	int offset, i, optr, lastslash;
	size_t len;
	int status;
	
	if (path == NULL) return 0;
	
	if ((path[0] != '/') || (strchr(path, '/') == NULL)) return 0;
	
	len = strlen(path);
	fullpath = calloc(len, 1);
	
	if (fullpath == NULL) return 0;
	
	lastslash = 0;
	optr = 0;
	
	for (i=0; i<len; i++) {
		if ((path[i] == '/') && (lastslash)) {
			continue;
		} else if ((path[i] == '/') && (!lastslash)) {
			lastslash = 1;
			fullpath[optr] = path[i];
			optr++;
		} else if (path[i] != '/') {
			if (lastslash) lastslash = 0;
			fullpath[optr] = path[i];
			optr++;
		}
	}
	
	if (fullpath[optr-1] == '/') {
		fullpath[optr-1] = '\0';
	} else {
		fullpath[optr] = '\0';
	}
	
	offset = 1;
	
	do {
		rootpath = strchr(fullpath+offset, '/');
		
		if (rootpath != NULL) {
			fullpath[rootpath - fullpath] = '\0';
		}
		
		status = mkdir(fullpath, mode);
		if ((status == -1) && (errno != EEXIST)) {
			perror("mkdirs()");
			free(fullpath);
			return 0;
		}
		
		if (rootpath != NULL) {
			fullpath[rootpath - fullpath] = '/';
			offset = 1 + (rootpath - fullpath);
		}
	} while (rootpath != NULL);
	
	free(fullpath);
	
	return 1;
}

int sysutils_process_exists(pid_t pid)
{
	if (pid == 0) return 0;
	
	
	if (kill(pid, 0) == 0) {
		return 1;
	} else {
		return 0;
	}
}
