/*
 *      dbstorage.c
 *      
 *      Copyright 2009 - 2012 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib.h>
#include <glib/gstdio.h>
#include <gdbm.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <utime.h>
#include "dbstorage.h"
#include "gsfcore.h"

#define SERVICE_XML "<service>\n\t<name>%s</name>\n\t<pic>%s</pic>\n\t<type>%u</type>\n\t<pos>%i</pos>\n\t<freq>%u</freq>\n\t<sr>%u</sr>\n\t<pol>%u</pol>\n\t<fec>%u</fec>\n\t<sys>%u</sys>\n\t<mod>%u</mod>\n\t<rof>%u</rof>\n\t<switch>%u:%u:%u</switch>\n\t<lnb>%u:%u:%u:%u:%u</lnb>\n\t<positioner>%u</positioner>\n\t<possatnum>%u</possatnum>\n\t<npid>%u</npid>\n\t<nip>%s</nip>\n\t<nmask>%s</nmask>\n\t<nmac>%s</nmac>\n\t<nprog>%s</nprog>\n\t<caid>%u</caid>\n\t<sid>%u</sid>\n\t<key>%s</key>\n\t<pids>%s</pids>\n</service>\n\n"

#define SVC_XML_LIST_HEADER "<?xml version=\'1.0\' encoding='UTF-8'?>\n\n<servicelist>\n"
#define SVC_XML_LIST_ENTRY "\t<service id=\"%u\">%s</service>\n"
#define SVC_XML_LIST_END "</servicelist>\n\n"

enum {
	SVC_XML_PARAM_NAME = 0,
	SVC_XML_PARAM_PIC,
	SVC_XML_PARAM_TYPE,
	SVC_XML_PARAM_POS,
	SVC_XML_PARAM_FREQ,
	SVC_XML_PARAM_SR,
	SVC_XML_PARAM_POL,
	SVC_XML_PARAM_FEC,
	SVC_XML_PARAM_SYS,
	SVC_XML_PARAM_MOD,
	SVC_XML_PARAM_ROF,
	SVC_XML_PARAM_SWITCH,
	SVC_XML_PARAM_LNB,
	SVC_XML_PARAM_POSITIONER,
	SVC_XML_PARAM_POSSATNUM,
	SVC_XML_PARAM_NPID,
	SVC_XML_PARAM_NIP,
	SVC_XML_PARAM_NMASK,
	SVC_XML_PARAM_NMAC,
	SVC_XML_PARAM_NPROG,
	SVC_XML_PARAM_CAID,
	SVC_XML_PARAM_SID,
	SVC_XML_PARAM_KEY,
	SVC_XML_PARAM_PIDS,
	SVC_XML_PARAM_NULL
};

static gint xml_parameter = SVC_XML_PARAM_NULL;

static guint32 *svcindex = NULL;
static GHashTable *svchash = NULL;
static GSList *netsvclist = NULL;

static GString *netxmllist = NULL;

static time_t svcmodtime;

static gchar *svcstoragepath = NULL;
static gchar *imgstoragepath = NULL;

struct _multiplesvc {
	GDBM_FILE dbf;
	GHashTable *hash;
	guint startid;
	guint32 *index;
};

struct _netsvcsearch {
	guint id;
	netservice *netsvc;
};

gboolean service_image_add(guint32 oldid, guint32 newid);
gboolean service_image_remove(guint32 id);
void service_index_refresh(guint32 *indexvec, gsize indexlen);

gchar *service_xml_create(service *ch);
service *service_xml_parse(gchar *xml);
static void service_xml_get_element(GMarkupParseContext *context, const gchar *element, const gchar **attr_names, const gchar **attr_values, gpointer data, GError **error);
static void service_xml_get_value(GMarkupParseContext *context, const gchar *text, gsize size, gpointer data, GError **error);
static void service_xml_end_element(GMarkupParseContext *context, const gchar *element, gpointer data, GError **error);


gboolean service_xml_list_rebuild(void)
{
	service *svc;
	guint i;
		
	if ((svcindex == NULL) || (svchash == NULL)) return FALSE;
	
	if (netxmllist != NULL) {
		g_string_free(netxmllist, TRUE);
		netxmllist = NULL;
	}
	
	netxmllist = g_string_new(SVC_XML_LIST_HEADER);
	i = 0;
	while (svcindex[i] != 0) {
		svc = g_hash_table_lookup(svchash, &svcindex[i]);
		if (svc != NULL) 
			g_string_append_printf(netxmllist, SVC_XML_LIST_ENTRY, svc->id, svc->name);
		i++;
	}
	g_string_append(netxmllist, SVC_XML_LIST_END);
	
	return TRUE;
}

guint service_xml_list_get_num_fragments(gsize fragmentsize)
{
	guint fragments = 0;
	
	if ((netxmllist == NULL) || (netxmllist->len == 0) || (fragmentsize == 0)) return 0;
	
	if (netxmllist->len % fragmentsize > 0) {
		fragments = (guint)((netxmllist->len / fragmentsize) + 1);
	} else {
		fragments = (guint)(netxmllist->len / fragmentsize);
	}
	
	return fragments;
}

gchar *service_xml_list_get_fragment(gsize fragmentsize, guint number)
{
	guint fragments;
	gchar *fragment;
	
	if ((netxmllist == NULL) || (fragmentsize == 0)) return NULL;
	
	fragments = service_xml_list_get_num_fragments(fragmentsize);
	
	if (number > (fragments - 1)) return NULL;
	
	fragment = (gchar *)g_malloc0(fragmentsize+1);
	if (number == (fragments - 1)) {
		strncpy(fragment, &netxmllist->str[number*fragmentsize], netxmllist->len - (fragmentsize*number));
		fragment[netxmllist->len - (fragmentsize*number)] = '\0';
	} else {
		strncpy(fragment, &netxmllist->str[number*fragmentsize], fragmentsize);
		fragment[fragmentsize] = '\0';
	}
	
	return fragment;
}

guint service_xml_list_get_length(void)
{
	if (netxmllist == NULL) {
		return 0;
	} else {
		return netxmllist->len;
	}
}

gchar *service_xml_list_get_string(void)
{
	if (netxmllist == NULL) {
		return NULL;
	} else {
		return netxmllist->str;
	}
}

gboolean service_net_add(guint32 id, gchar *name)
{
	netservice *nsvc;
	
	if ((id != 0) && (name != NULL)) {
		nsvc = g_malloc0(sizeof(netservice));
		nsvc->id = id;
		memset(nsvc->name, 0, sizeof(nsvc->name));
		if (strlen(name) >= 256) {
			g_memmove(nsvc->name, name, 255);
		} else {
			g_memmove(nsvc->name, name, strlen(name));
		}
		
		netsvclist = g_slist_append(netsvclist, nsvc);
		
		return TRUE;
	}
	
	return FALSE;
}

gboolean service_net_clear(void)
{
	if (netsvclist != NULL) {
		g_slist_free(netsvclist);
		netsvclist = NULL;
		
		return TRUE;
	}
	
	return FALSE;
}

guint service_net_get_number(void)
{
	if (netsvclist == NULL) return 0;
	
	return g_slist_length(netsvclist);
}

gboolean service_net_read_foreach(gboolean (*add_netservice_cb)(netservice *nsvc))
{
	guint i;
	gint netsvclength = 0;
	
	if ((netsvclist != NULL) && (add_netservice_cb != NULL)) {
		netsvclength = g_slist_length(netsvclist);
		for (i=0; i<netsvclength; i++) {
			add_netservice_cb(g_slist_nth_data(netsvclist, i));
		}
	}	
	
	return TRUE;
}

static void service_net_read_id_foreach(gpointer data, gpointer user_data)
{
	netservice *netsvc;
	struct _netsvcsearch *netsearch;
	
	netsvc = (netservice *)data;
	netsearch = (struct _netsvcsearch *)user_data;
	
	if (netsvc->id == netsearch->id) {
		netsearch->netsvc = netsvc;
	}
}

netservice *service_net_read_id(guint id)
{
	struct _netsvcsearch netsearch;
	
	if (netsvclist != NULL) {
		netsearch.id = id;
		netsearch.netsvc = NULL;
		
		g_slist_foreach(netsvclist, service_net_read_id_foreach, &netsearch);
		
		return netsearch.netsvc;
	}
	
	return NULL;
}

tuneparams_t service_search_position_params(gint position)
{
	service *svc;
	tuneparams_t tuneparams;
	guint i;
	
	i = 0;
	tuneparams = NULL;
	
	while (svcindex[i] != 0) {
		svc = g_hash_table_lookup(svchash, &svcindex[i]);
		if (svc != NULL) {
			if (svc->pos == position) {
				tuneparams = (tuneparams_t)g_malloc0(sizeof(struct _tuneparams));
				tuneparams->pos = position;
				tuneparams->freq = svc->freq;
				tuneparams->sr = svc->sr;
				tuneparams->realfreq = svc->freq;
				tuneparams->realsr = svc->sr;
				tuneparams->tuned = 0;
				tuneparams->pol = svc->pol;
				tuneparams->fec = svc->fec;
				tuneparams->sys = svc->sys;
				tuneparams->mod = svc->mod;
				tuneparams->rof = svc->rof;
				tuneparams->port10 = svc->port10;
				tuneparams->port11 = svc->port11;
				tuneparams->rep11 = svc->rep11;
				tuneparams->lnbtype = svc->lnbtype;
				tuneparams->lnb22khz = svc->lnb22khz;
				tuneparams->positioner = svc->positioner;
				tuneparams->possatnum = svc->possatnum;
				
				break;
			}
		}
		i++;
	}
	
	return tuneparams;
}

gboolean service_storage_init(void)
{
	const gchar *homedir;
	gchar *dbfilename;
	//GDBM_FILE dbf;
	gchar *imgpath;
	
	if ((homedir = g_getenv("HOME")) == NULL) homedir = g_get_home_dir();
	
	dbfilename = g_build_filename(homedir, ".config", "gsf", "services.gdbm", NULL);
	svcstoragepath = dbfilename;
	g_message("Database file: %s", svcstoragepath);
	
	imgpath = g_build_path(G_DIR_SEPARATOR_S, homedir, ".config", "gsf", "images", NULL);
	if (!g_file_test(imgpath, G_FILE_TEST_EXISTS)) {
		if (g_mkdir_with_parents(imgpath, 0755) == -1) {
			g_warning("Failed to create images path");
		}
	}
	imgstoragepath = imgpath;
	g_message("Images path: %s", imgpath);
	
	service_read_foreach(NULL, NULL, NULL);
	service_xml_list_rebuild();
		
	return TRUE;	
}

gboolean service_image_add(guint32 oldid, guint32 newid)
{
	gchar strid[16];
	gchar *oldfn, *newfn;
	gint oldfd, newfd;
	guchar cpbuf[4096];
	gsize cpbytes;
	
	memset(strid, 0, sizeof(strid));
	g_sprintf(strid, "%u.png", oldid);
	oldfn = g_build_filename(g_get_tmp_dir(), strid, NULL);
	if (!g_file_test(oldfn, G_FILE_TEST_EXISTS)) {
		g_free(oldfn);
		g_warning("Image not found");
		return FALSE;
	}
	oldfd = g_open(oldfn, O_RDONLY, S_IRWXU);
	if (oldfd == -1) {
		g_free(oldfn);
		g_warning("Error while opening temporary file");
		return FALSE;
	}
	
	memset(strid, 0, sizeof(strid));
	g_sprintf(strid, "%u.png", newid);
	newfn = g_build_filename(imgstoragepath, strid, NULL);
	newfd = g_open(newfn, O_WRONLY | O_CREAT, S_IRWXU);
	g_free(newfn);
	if (newfd == -1) {
		g_free(oldfn);
		close(oldfd);
		g_warning("Error while opening destination file");
		return FALSE;
	}
	
	do {
		cpbytes = read(oldfd, cpbuf, sizeof(cpbuf));
		if (cpbytes) write(newfd, cpbuf, cpbytes);
	} while (cpbytes);
	
	close(oldfd);
	close(newfd);
	unlink(oldfn);
	g_free(oldfn);
	
	return TRUE;
}

gboolean service_image_exists(guint32 id)
{
	gchar strid[16];
	gchar *imgfn;
	
	memset(strid, 0, sizeof(strid));
	g_sprintf(strid, "%u.png", id);
	imgfn = g_build_filename(imgstoragepath, strid, NULL);
	if (!g_file_test(imgfn, G_FILE_TEST_EXISTS)) {
		g_free(imgfn);
		return FALSE;
	}
	g_free(imgfn);
	
	return TRUE;
}

gboolean service_image_remove(guint32 id)
{
	gchar strid[16];
	gchar *imgfn;
	
	//TODO: Unfinished function
	memset(strid, 0, sizeof(strid));
	g_sprintf(strid, "%u.png", id);
	imgfn = g_build_filename(imgstoragepath, strid, NULL);
	if (!g_file_test(imgfn, G_FILE_TEST_EXISTS)) {
		g_free(imgfn);
		g_warning("Image not found");
		return FALSE;
	}
	unlink(imgfn);
	g_free(imgfn);
	
	return TRUE;
}

void service_index_refresh(guint32 *indexvec, gsize indexlen)
{
	if ((indexvec == NULL) || ((indexvec != NULL) && (indexvec[0] == 0))) return;
	
	if (svcindex != NULL) {
		g_free(svcindex);
		svcindex = NULL;
	}
	
	svcindex = (guint32 *)g_malloc0(sizeof(guint32)*indexlen);
	memcpy(svcindex, indexvec, sizeof(guint32)*indexlen);
}

guint32 *service_index_rebuild(GDBM_FILE dbf, gsize *indexlen)
{
	datum key, data;
	GSList *keys = NULL;
	guint len = 0, i;
	guint32 *svcindex;
	
	key = gdbm_firstkey(dbf);
	if (key.dptr == NULL) {
		if (indexlen != NULL) *indexlen = 0;
		return NULL;
	}
		
	do {
		if (*(guint32 *)key.dptr != 0) {
			keys = g_slist_prepend(keys, (gpointer)key.dptr);
			len += 1;
		}
		key = gdbm_nextkey(dbf, key);
	} while (key.dptr != NULL);
	
	if (len == 0) {
		if (indexlen != NULL) *indexlen = 0;
		return NULL;
	}
		
	keys = g_slist_reverse(keys);
		
	svcindex = (guint32 *)g_malloc0(sizeof(guint32)*(len+1));
	
	for (i=0; i<len; i++) {
		svcindex[i] = *(guint32 *)g_slist_nth_data(keys, i);
		free(g_slist_nth_data(keys, i));
	}
	g_slist_free(keys);
	svcindex[len] = 0;
	len += 1;
	
	if (indexlen != NULL) *indexlen = len;
	return svcindex;
}

//Get service by ID
service *service_read(guint id)
{
	if ((svchash == NULL) || (id > ULONG_MAX) || (id < 0)) return NULL;
	
	return (service *)g_hash_table_lookup(svchash, (gpointer)&id);
}

//Write new index to database
gboolean service_write_index(guint32 *indexvec, gsize indexlen)
{
	GDBM_FILE dbf;
	datum key, data;
	gint ret;
	guint32 indexid = 0;
	struct stat statbuf;
	struct utimbuf utimebuf;
	gboolean setmtime = FALSE;
	
	if ((indexvec == NULL) || ((indexvec != NULL) && (indexvec[0] == 0))) return FALSE;
	
	if (g_stat(svcstoragepath, &statbuf) != -1) {
		setmtime = TRUE;
		svcmodtime = statbuf.st_mtime;
	} else {
		g_warning("Failed to set database modification time");
	}
	
	if ((dbf = gdbm_open(svcstoragepath, 0, GDBM_WRCREAT|GDBM_SYNC, 0755, 0)) == NULL) {
		g_warning("Failed to open database");
		return FALSE;
	}
	
	key.dptr = (gchar *)&indexid;
	key.dsize = sizeof(guint32);
	if (!gdbm_exists(dbf, key)) {
		data.dptr = (gchar *)indexvec;
		data.dsize = sizeof(guint32)*indexlen;
	} else {
		data = gdbm_fetch(dbf, key);
		free(data.dptr);
		if (data.dsize != (sizeof(guint32)*indexlen)) {
			g_warning("Index size mismatch %u - %u", data.dsize, sizeof(guint32)*indexlen);
			gdbm_close(dbf);
			return FALSE;
		} else {
			data.dptr = (gchar *)indexvec;
			data.dsize = sizeof(guint32)*indexlen;
		}
	}
	ret = gdbm_store(dbf, key, data, GDBM_REPLACE);
	if (ret == -1) {
		gdbm_close(dbf);
		g_warning("Unable to write database index");
		return FALSE;
	}
	
	if (setmtime) {
		utimebuf.actime = statbuf.st_atime;
		utimebuf.modtime = statbuf.st_mtime;
		if (g_utime(svcstoragepath, &utimebuf) == -1) {
			g_warning("Unable to restore database modification time");
		}
	}
	
	service_index_refresh(indexvec, indexlen);
	
	gdbm_close(dbf);
	
	return TRUE;
}

//Fill list with services
gboolean service_read_foreach(gboolean (*add_service_cb)(service *svc), gboolean (*add_netservice_cb)(netservice *nsvc), gboolean (*add_svcxml_cb)(service *svc, gboolean initialize))
{
	GDBM_FILE dbf;
	datum key, data;
	service *svc;
	guint32 indexid = 0;
	guint32 *indexvec;
	gsize indexsize;
	guint len, i;
	struct stat statbuf;
	gint netsvclength = 0;
	
	//Get database modification time
	if (g_stat(svcstoragepath, &statbuf) != -1) {
		if (svcmodtime == statbuf.st_mtime) {
			i = 0;
			if (add_svcxml_cb != NULL) add_svcxml_cb(NULL, TRUE);
			while (svcindex[i] != 0) {
				svc = g_hash_table_lookup(svchash, &svcindex[i]);
				if (svc != NULL) {
					if (add_service_cb != NULL) add_service_cb(svc);
					if (add_svcxml_cb != NULL) add_svcxml_cb(svc, FALSE);
				}
				i++;
			}
			if (add_svcxml_cb != NULL) add_svcxml_cb(NULL, FALSE);
			goto netservices;
		} else {
			svcmodtime = statbuf.st_mtime;
		}
	} else {
		g_warning("Failed to get database modification time");
	}
	
	//Open database and read from it
	if ((dbf = gdbm_open(svcstoragepath, 0, GDBM_READER, 0755, 0)) == NULL) {
		//Make new hash table containing services
		if (svchash != NULL) {
			g_hash_table_destroy(svchash);
			svchash = NULL;
		}
		svchash = g_hash_table_new_full(g_int_hash, g_int_equal, NULL, (GDestroyNotify)service_free);
		g_warning("Failed to open database - new one will be created automatically");
		goto netservices;
	}
	
	key.dptr = (gchar *)&indexid;
	key.dsize = sizeof(guint32);
	
	if (!gdbm_exists(dbf, key)) {
		indexvec = service_index_rebuild(dbf, &indexsize);
		if (indexsize > 0) {
			service_write_index(indexvec, indexsize);
			g_warning("No index key - index rebuilt");
		} else {
			g_warning("No keys in database - unable to read");
			goto netservices;
		}
	} else {
		data = gdbm_fetch(dbf, key);
		indexvec = (guint32 *)data.dptr;
		indexsize = (gsize)data.dsize;
		service_index_refresh(indexvec, indexsize);
	}
	
	if ((len = indexsize/sizeof(guint32)) > 0) {
		//Make new hash table containing services
		if (svchash != NULL) {
			g_hash_table_destroy(svchash);
			svchash = NULL;
		}
		svchash = g_hash_table_new_full(g_int_hash, g_int_equal, NULL, (GDestroyNotify)service_free);
		//Make new service xml template for network control
		if (add_svcxml_cb != NULL) add_svcxml_cb(NULL, TRUE);
		//Read services from database and fill hash table and xml template
		for (i=0; i<len; i++) {
			if (indexvec[i] == 0) break;
			key.dptr = (gchar *)&indexvec[i];
			key.dsize = sizeof(guint32);
			data = gdbm_fetch(dbf, key);
			if (data.dptr != NULL) {
				svc = service_xml_parse(data.dptr);
				svc->id = indexvec[i];
				g_hash_table_insert(svchash, (gpointer)&(svc->id), (gpointer)svc);
				if (add_svcxml_cb != NULL) add_svcxml_cb(svc, FALSE);
				if (add_service_cb != NULL) add_service_cb(svc);
				free(data.dptr);
			}
		}
		//End element of GString with service list XML
		if (add_svcxml_cb != NULL) add_svcxml_cb(NULL, FALSE);
		
	}
	
	gdbm_close(dbf);
	
	free(indexvec);
	
	netservices:;
	
	if ((netsvclist != NULL) && (add_netservice_cb != NULL)) {
		netsvclength = g_slist_length(netsvclist);
		for (i=0; i<netsvclength; i++) {
			add_netservice_cb(g_slist_nth_data(netsvclist, i));
		}
	}	
	
	return TRUE;
}

//Add new service into service list and database
gboolean service_add(service *svc, gboolean replace, gboolean (*add_service_cb)(service *svc))
{
	GDBM_FILE dbf;
	datum key, data;
	gint ret;
	gboolean chexists;
	guint32 indexid = 0;
	guint32 *indexvec;
	guint len, i;
	guint32 *newvec;
	guint32 oldid;
		
	if ((dbf = gdbm_open(svcstoragepath, 0, GDBM_WRCREAT|GDBM_SYNC, 0755, 0)) == NULL) {
		g_warning("Failed to open database");
		return FALSE;
	}
	
	oldid = svc->id;
	
	if (svc->id != 0) {
		key.dptr = (gchar *)&(svc->id);
		key.dsize = sizeof(guint32);
		chexists = (gboolean)gdbm_exists(dbf, key);
	} else {
		do {
			svc->id = g_random_int();
			key.dptr = (gchar *)&(svc->id);
			key.dsize = sizeof(guint32);
			chexists = (gboolean)gdbm_exists(dbf, key);
		} while ((chexists) || (svc->id == 0));
	}
	
	if ((svc->picture != NULL) && (strlen(svc->picture) > 0)) service_image_add(oldid, svc->id);
	
	if (!chexists) {
		key.dptr = (gchar *)&indexid;
		key.dsize = sizeof(guint32);
		if (!gdbm_exists(dbf, key)) {
			newvec = (guint32 *)g_malloc(sizeof(guint32)*2);
			newvec[0] = svc->id;
			newvec[1] = 0;
			data.dptr = (gchar *)newvec;
			data.dsize = sizeof(guint32)*2;
		} else {
			data = gdbm_fetch(dbf, key);
			indexvec = (guint32 *)data.dptr;
			len = data.dsize/sizeof(guint32);
			newvec = (guint32 *)g_malloc0(data.dsize + sizeof(guint32));
			for (i=0; i<len-1; i++) {
				newvec[i] = indexvec[i];
			}
			newvec[len-1] = svc->id;
			newvec[len] = 0;
			data.dptr = (gchar *)newvec;
			data.dsize = data.dsize + sizeof(guint32);
			free(indexvec);
		}
		ret = gdbm_store(dbf, key, data, GDBM_REPLACE);
		if (ret == -1) {
			gdbm_close(dbf);
			g_warning("Unable to write database index");
			return FALSE;
		}
		service_index_refresh((guint32 *)data.dptr, (gsize)data.dsize);
		g_free(newvec);
	}
	
	if ((!chexists) || (chexists && replace)) {
		key.dptr = (gchar *)&(svc->id);
		key.dsize = sizeof(guint32);
		data.dptr = service_xml_create(svc);
		data.dsize = strlen(data.dptr)+1;
		ret = gdbm_store(dbf, key, data, GDBM_REPLACE);
		g_free(data.dptr);
		if (ret == -1) {
			gdbm_close(dbf);
			g_warning("Unable to add service to database");
			return FALSE;
		}
	}
	
	gdbm_close(dbf);
		
	g_hash_table_insert(svchash, (gpointer)&(svc->id), (gpointer)svc);
	
	service_xml_list_rebuild();
	
	if (add_service_cb != NULL) add_service_cb(svc);
	
	return TRUE;
}

//Add single service foreach
void service_add_multiple_write_foreach(gpointer data, gpointer user_data)
{
	struct _service *svc;
	struct _multiplesvc *mpsvc;
	datum dbkey, dbdata;
	gint ret;
	guint32 value;
	guint32 *hashval;
	
	svc = (struct _service *)data;
	mpsvc = (struct _multiplesvc *)user_data;
	
	do {
		value = g_random_int();
	} while ((g_hash_table_lookup(mpsvc->hash, &value) != NULL) || (g_hash_table_lookup(svchash, &value) != NULL));
	
	hashval = (guint32 *)g_malloc0(sizeof(guint32));
	*hashval = value;
	g_hash_table_insert(mpsvc->hash, hashval, hashval);
	
	svc->id = value;
	
	mpsvc->index[mpsvc->startid] = value;
	mpsvc->startid++;
	
	if (mpsvc->dbf != NULL) {
		dbkey.dptr = (gchar *)&(svc->id);
		dbkey.dsize = sizeof(guint32);
		dbdata.dptr = service_xml_create(svc);
		dbdata.dsize = strlen(dbdata.dptr) + 1;
		if (gdbm_store(mpsvc->dbf, dbkey, dbdata, GDBM_REPLACE) == -1) {
			g_warning("Unable to add service to database");
		}
		g_free(dbdata.dptr);
	}
	
	g_hash_table_insert(svchash, (gpointer)&(svc->id), (gpointer)svc);
	
}

//Destroy hash table
void service_add_multiple_free_hash(gpointer data)
{
	guint32 *hashval;
	
	hashval = (guint32 *)data;
	
	g_free(hashval);
}

//Add multiple services to databse in one call
gboolean service_add_multiple(GSList **servicelist)
{
	GDBM_FILE dbf = NULL;
	datum dbkey, dbdata;
	guint32 indexid = 0;
	struct _multiplesvc mpsvc;
	gsize number;
	guint indexnum;
		
	if ((number = g_slist_length(*servicelist)) == 0) return FALSE;
	
	if (svcindex != NULL) {
		indexnum = 0;
		while (svcindex[indexnum] != 0) indexnum++;
		mpsvc.index = (guint32 *)g_malloc0((indexnum + number + 1) * sizeof(guint32));
		mpsvc.startid = indexnum;
		memcpy(&mpsvc.index[0], &svcindex[0], indexnum * sizeof(guint32));
	} else {
		mpsvc.index = (guint32 *)g_malloc0((number + 1) * sizeof(guint32));
		mpsvc.startid = 0;
	}
	
	g_message("Adding services to database");
	
	if ((dbf = gdbm_open(svcstoragepath, 0, GDBM_WRCREAT|GDBM_SYNC, 0755, 0)) == NULL) {
		g_warning("Failed to open database");
	}
	
	mpsvc.hash = g_hash_table_new_full(g_int_hash, g_int_equal, NULL, (GDestroyNotify)service_add_multiple_free_hash);
	mpsvc.dbf = dbf;
	
	g_slist_foreach(*servicelist, service_add_multiple_write_foreach, &mpsvc);
	
	g_hash_table_destroy(mpsvc.hash);
	
	g_message("Rebuilding database index");
	
	indexnum = indexnum + number;
	
	if (dbf != NULL) {
		dbkey.dptr = (gchar *)&indexid;
		dbkey.dsize = sizeof(guint32);
		dbdata.dptr = (gchar *)mpsvc.index;
		dbdata.dsize = (indexnum + 1) * sizeof(guint32);
		if (gdbm_store(dbf, dbkey, dbdata, GDBM_REPLACE) == -1) {
			g_warning("Unable to write database index");
		}
		gdbm_close(dbf);
	}
	
	service_index_refresh(mpsvc.index, (indexnum + 1) * sizeof(guint32));
	
	service_xml_list_rebuild();
	
	return TRUE;
}

//Merge changes in service into service list and database
gboolean service_merge(service *svc, gboolean (*update_service_cb)(service *svc))
{
	GDBM_FILE dbf;
	datum key, data;
	gint ret;
	gboolean chexists;
	
	if ((dbf = gdbm_open(svcstoragepath, 0, GDBM_WRCREAT | GDBM_SYNC, 0755, 0)) == NULL) {
		g_warning("Failed to open database");
		return FALSE;
	}
	
	if (svc->id != 0) {
		key.dptr = (gchar *)&(svc->id);
		key.dsize = sizeof(guint32);
		chexists = (gboolean)gdbm_exists(dbf, key);
	} else {
		g_warning("Unable to merge changes without service ID");
		return FALSE;
	}
	
	if ((svc->picture != NULL) && (strlen(svc->picture) > 0)) service_image_add(svc->id, svc->id);
	
	if (chexists) {
		key.dptr = (gchar *)&(svc->id);
		key.dsize = sizeof(guint32);
		data.dptr = service_xml_create(svc);
		data.dsize = strlen(data.dptr)+1;
		ret = gdbm_store(dbf, key, data, GDBM_REPLACE);
		g_free(data.dptr);
		if (ret == -1) {
			gdbm_close(dbf);
			g_warning("Unable to delete service from database");
			return FALSE;
		}
	} else {
		g_warning("Unable to merge changes - no such service ID");
		return FALSE;
	}
	
	gdbm_close(dbf);
	
	g_hash_table_replace(svchash, (gpointer)&(svc->id), (gpointer)svc);
	
	if (update_service_cb != NULL) update_service_cb(svc);
	
	return TRUE;
}

//Remove service
gboolean service_remove(service *svc, gboolean (*remove_service_cb)(service *svc))
{
	GDBM_FILE dbf;
	datum key, data;
	gint ret;
	gboolean chexists;
	guint len, i, pos = -1;
	guint32 indexid = 0;
	guint32 *indexvec;
	guint32 *newvec;
	
	if ((dbf = gdbm_open(svcstoragepath, 0, GDBM_WRCREAT|GDBM_SYNC, 0755, 0)) == NULL) {
		g_warning("Failed to open database");
		return FALSE;
	}
	
	if (svc->id != 0) {
		key.dptr = (gchar *)&(svc->id);
		key.dsize = sizeof(guint32);
		chexists = (gboolean)gdbm_exists(dbf, key);
	} else {
		gdbm_close(dbf);
		g_warning("Unable to remove service without valid service ID");
		return FALSE;
	}
	
	if ((svc->picture != NULL) && (strlen(svc->picture) > 0)) service_image_remove(svc->id);
	
	if (chexists) {
		key.dptr = (gchar *)&indexid;
		key.dsize = sizeof(guint32);
		if (!gdbm_exists(dbf, key)) {
			gdbm_close(dbf);
			g_warning("Unable to read database index\n");
			return FALSE;
		} else {
			data = gdbm_fetch(dbf, key);
			if (data.dptr == NULL) {
				gdbm_close(dbf);
				g_warning("Unable to read database index");
				return FALSE;
			} else {
				indexvec = (guint32 *)data.dptr;
				len = data.dsize/sizeof(guint32);
				for (i=0; i<len; i++) {
					if (indexvec[i] == svc->id) {
						pos = i;
						break;
					} 
				}
				if (pos == -1) {
					gdbm_close(dbf);
					g_warning("No such service in index");
					return FALSE;
				} else {
					newvec = (guint32 *)g_malloc0(data.dsize-sizeof(guint32));
					for (i=0; i<len; i++) {
						if (i == pos) {
							continue;
						} else if (i < pos) {
							newvec[i] = indexvec[i];
						} else if (i > pos) {
							newvec[i-1] = indexvec[i];
						}
					}
					data.dptr = (gchar *)newvec;
					data.dsize = data.dsize-sizeof(guint32);
										
					ret = gdbm_store(dbf, key, data, GDBM_REPLACE);
					if (ret == -1) {
						gdbm_close(dbf);
						g_warning("Unable to write new database index");
						return FALSE;
					}
					service_index_refresh(newvec, data.dsize-sizeof(guint32));
					g_free(data.dptr);
					
					key.dptr = (gchar *)&(svc->id);
					key.dsize = sizeof(guint32);
					
					ret = gdbm_delete (dbf, key);
					if (ret == -1) {
						gdbm_close(dbf);
						g_warning("Unable to delete service from database");
						return FALSE;
					}
				}
			}
		}
		
	} else {
		gdbm_close(dbf);
		g_warning("Unable to remove service - no such service ID");
		return FALSE;
	}
	
	gdbm_close(dbf);
	
	
	//SVC Will be freed later - callback must be called first
	if (remove_service_cb != NULL) remove_service_cb(svc);
	
	if (!g_hash_table_remove(svchash, (gpointer)&(svc->id))) {
		g_warning("Unable to remove service from local hash table");
		return FALSE;
	} else {
		service_xml_list_rebuild();
	}
	
	return TRUE;
}

/*Make new service entry with default values*/
service *service_new(gchar *picture, gchar *name, guchar type)
{
	service *svc;
	
	svc = (service *)g_malloc0(sizeof(service));
	svc->id = 0;
	svc->picture = g_strdup(picture);
	svc->name = g_strdup(name);
	svc->node = NULL;
	svc->type = type;
	svc->pos = 0;
	svc->freq = 11044;
	svc->sr = 44950;
	svc->pol = 0;
	svc->fec = 3;
	svc->sys = 0;
	svc->mod = 0;
	svc->rof = 0;
	svc->port10 = 0;
	svc->port11 = 0;
	svc->rep11 = 1;
	svc->lnbtype = 0;
	svc->lnbhifreq = 10600;
	svc->lnblofreq = 9750;
	svc->lnbswfreq = 11700;
	svc->lnb22khz = 0;
	svc->npid = 0;
	svc->nip = g_strdup("192.168.0.0");
	svc->nmask = g_strdup("255.255.255.0");
	svc->nmac = g_strdup("00:01:02:03:04:05");
	svc->nprog = NULL;
	svc->caid = 0;
	svc->key = NULL;
	svc->pids = NULL;
	
	return svc;
}

/*We can make copy of service structure to do anything with it using service assistant*/
service *service_clone(service *svc)
{
	service *dsvc;
	guint i, len = 0; 
	pid *oldpid, *newpid;
	
	if (svc == NULL) {
		dsvc = service_new(NULL, "New service", 0);
		return dsvc;
	}
	
	dsvc = (service *)g_malloc0(sizeof(service));
	dsvc->id = svc->id;
	dsvc->picture = g_strdup(svc->picture);
	dsvc->name = g_strdup(svc->name);
	if (svc->node != NULL) dsvc->node = g_strdup(svc->node); else dsvc->node = NULL;
	if (svc->type == SERVICE_RADIO) dsvc->type = SERVICE_TV_RADIO; else dsvc->type = svc->type;
	//dsvc->type = svc->type;
	dsvc->pos = svc->pos;
	dsvc->freq = svc->freq;
	dsvc->sr = svc->sr;
	dsvc->pol = svc->pol;
	dsvc->fec = svc->fec;
	dsvc->sys = svc->sys;
	dsvc->mod = svc->mod;
	dsvc->rof = svc->rof;
	dsvc->port10 = svc->port10;
	dsvc->port11 = svc->port11;
	dsvc->rep11 = svc->rep11;
	dsvc->lnbtype = svc->lnbtype;
	dsvc->lnbhifreq = svc->lnbhifreq;
	dsvc->lnblofreq = svc->lnblofreq;
	dsvc->lnbswfreq = svc->lnbswfreq;
	dsvc->lnb22khz = svc->lnb22khz;
	dsvc->positioner = svc->positioner;
	dsvc->possatnum = svc->possatnum;
	dsvc->npid = svc->npid;
	dsvc->nip = g_strdup(svc->nip);
	dsvc->nmask = g_strdup(svc->nmask);
	dsvc->nmac = g_strdup(svc->nmac);
	dsvc->nprog = g_strdup(svc->nprog);
	dsvc->caid = svc->caid;
	dsvc->sid = svc->sid;
	dsvc->key = g_strdup(svc->key);
	dsvc->pids = NULL;
	
	if (svc->pids != NULL) {
		len = g_slist_length(svc->pids);
		for (i=0; i<len; i++) {
			oldpid = (pid *)g_slist_nth_data(svc->pids, i);
			newpid = (pid *)g_malloc0(sizeof(pid));
			newpid->enabled = oldpid->enabled;
			newpid->type = oldpid->type;
			newpid->value = oldpid->value;
			dsvc->pids = g_slist_prepend(dsvc->pids, (gpointer)newpid);			
		}
		dsvc->pids = g_slist_reverse(dsvc->pids);
	}
	
	return dsvc;
}

void service_pids_free_foreach(pid *curpid, gpointer data)
{
	g_free(curpid);
}

void service_pids_free(service *svc)
{
	if (svc->pids != NULL) {
		g_slist_foreach(svc->pids, (GFunc)service_pids_free_foreach, NULL);
		g_slist_free(svc->pids);
		svc->pids = NULL;
	}
}

void service_free(service *svc)
{
	if (svc != NULL) {
		service_pids_free(svc);
		if (svc->key != NULL) g_free(svc->key);
		if (svc->nprog != NULL) g_free(svc->nprog);
		if (svc->nmac != NULL) g_free(svc->nmac);
		if (svc->nmask != NULL) g_free(svc->nmask);
		if (svc->nip != NULL) g_free(svc->nip);
		if (svc->name != NULL) g_free(svc->name);
		if (svc->picture != NULL) g_free(svc->picture);
		if (svc->node != NULL) g_free(svc->node);
		g_free(svc);
	}
	
}

gchar *service_xml_create(service *svc)
{
	gchar **res;
	gchar *sres;
	guint len;
	gint i = 0;
	pid *curpid;
	gchar *resstr;
	
	len = g_slist_length(svc->pids);
	res = (gchar **)g_malloc0(sizeof(gchar *)*len + 1);
	for (i=0; i<len; i++) {
		curpid = g_slist_nth_data(svc->pids, i);
		res[i] = g_strdup_printf("%u:%u:%u", curpid->type, curpid->value, curpid->enabled);
	}
	res[len] = NULL;
	sres = g_strjoinv(";", res);
	
	resstr = g_strdup_printf(SERVICE_XML,
								svc->name == NULL ? "" : svc->name,
								svc->picture == NULL ? "" : svc->picture,
								svc->type,
								svc->pos,
								svc->freq,
								svc->sr,
								svc->pol,
								svc->fec,
								svc->sys,
								svc->mod,
								svc->rof,
								svc->port10,
								svc->port11,
								svc->rep11,
								svc->lnbtype,
								svc->lnb22khz,
								svc->lnbhifreq,
								svc->lnblofreq,
								svc->lnbswfreq,
								svc->positioner,
								svc->possatnum,
								svc->npid,
								svc->nip == NULL ? "" : svc->nip,
								svc->nmask == NULL ? "" : svc->nmask,
								svc->nmac == NULL ? "" : svc->nmac,
								svc->nprog == NULL ? "" : svc->nprog,
								svc->caid,
								svc->sid,
								svc->key == NULL ? "" : svc->key,
								sres == NULL ? "" : sres);
	g_strfreev(res);
	g_free(sres);
	
	return resstr;
}

service *service_xml_parse(gchar *xml)
{
	service *svc;
	GMarkupParser mp;
	GMarkupParseContext *mpc;
	GError *error = NULL;
	
	svc = service_new(NULL, NULL, 2);
	
	mp.start_element = service_xml_get_element;
	mp.end_element = service_xml_end_element;
	mp.text = service_xml_get_value;
	mp.passthrough = NULL;
	mp.error = NULL;
	
	mpc = g_markup_parse_context_new(&mp, 0, (gpointer)svc, NULL);
	g_markup_parse_context_parse(mpc, xml, -1, &error);
	if (error != NULL) {
		g_warning(error->message);
		g_error_free(error);
		g_markup_parse_context_free(mpc);
		return NULL;
	}
	g_markup_parse_context_free(mpc);
	
	return svc;
}

static void service_xml_get_element(GMarkupParseContext *context, const gchar *element, const gchar **attr_names, const gchar **attr_values, gpointer data, GError **error)
{
	if (g_str_equal(element, "name")) xml_parameter = SVC_XML_PARAM_NAME;
	else if (g_str_equal(element, "pic")) xml_parameter = SVC_XML_PARAM_PIC;
	else if (g_str_equal(element, "type")) xml_parameter = SVC_XML_PARAM_TYPE;
	else if (g_str_equal(element, "pos")) xml_parameter = SVC_XML_PARAM_POS;
	else if (g_str_equal(element, "freq")) xml_parameter = SVC_XML_PARAM_FREQ;
	else if (g_str_equal(element, "sr")) xml_parameter = SVC_XML_PARAM_SR;
	else if (g_str_equal(element, "pol")) xml_parameter = SVC_XML_PARAM_POL;
	else if (g_str_equal(element, "fec")) xml_parameter = SVC_XML_PARAM_FEC;
	else if (g_str_equal(element, "sys")) xml_parameter = SVC_XML_PARAM_SYS;
	else if (g_str_equal(element, "mod")) xml_parameter = SVC_XML_PARAM_MOD;
	else if (g_str_equal(element, "rof")) xml_parameter = SVC_XML_PARAM_ROF;
	else if (g_str_equal(element, "switch")) xml_parameter = SVC_XML_PARAM_SWITCH;
	else if (g_str_equal(element, "lnb")) xml_parameter = SVC_XML_PARAM_LNB;
	else if (g_str_equal(element, "positioner")) xml_parameter = SVC_XML_PARAM_POSITIONER;
	else if (g_str_equal(element, "possatnum")) xml_parameter = SVC_XML_PARAM_POSSATNUM;
	else if (g_str_equal(element, "npid")) xml_parameter = SVC_XML_PARAM_NPID;
	else if (g_str_equal(element, "nip")) xml_parameter = SVC_XML_PARAM_NIP;
	else if (g_str_equal(element, "nmask")) xml_parameter = SVC_XML_PARAM_NMASK;
	else if (g_str_equal(element, "nmac")) xml_parameter = SVC_XML_PARAM_NMAC;
	else if (g_str_equal(element, "nprog")) xml_parameter = SVC_XML_PARAM_NPROG;
	else if (g_str_equal(element, "caid")) xml_parameter = SVC_XML_PARAM_CAID;
	else if (g_str_equal(element, "sid")) xml_parameter = SVC_XML_PARAM_SID;
	else if (g_str_equal(element, "key")) xml_parameter = SVC_XML_PARAM_KEY;
	else if (g_str_equal(element, "pids")) xml_parameter = SVC_XML_PARAM_PIDS;
}

static void service_xml_get_value(GMarkupParseContext *context, const gchar *text, gsize size, gpointer data, GError **error)
{
	gchar **strv, **strv2;	
	gint i = 0;
	guint strvlen;
	pid *curpid;
	
	service *svc = (service *)data;
	
	if (xml_parameter == SVC_XML_PARAM_NULL) return;
	
	switch (xml_parameter) {
		case SVC_XML_PARAM_NAME:
			if (svc->name != NULL) g_free(svc->name);
			svc->name = g_markup_escape_text(text, -1);
			break;
		case SVC_XML_PARAM_PIC:
			if (svc->picture != NULL) g_free(svc->picture);
			svc->picture = g_markup_escape_text(text, -1);
			break;
		case SVC_XML_PARAM_TYPE:
			svc->type = atoi(text);
			break;
		case SVC_XML_PARAM_POS:
			svc->pos = atoi(text);
			break;
		case SVC_XML_PARAM_FREQ:
			svc->freq = atoi(text);
			break;
		case SVC_XML_PARAM_SR:
			svc->sr = atoi(text);
			break;
		case SVC_XML_PARAM_POL:
			svc->pol = atoi(text);
			break;
		case SVC_XML_PARAM_FEC:
			svc->fec = atoi(text);
			break;
		case SVC_XML_PARAM_SYS:
			svc->sys = atoi(text);
			break;
		case SVC_XML_PARAM_MOD:
			svc->mod = atoi(text);
			break;
		case SVC_XML_PARAM_ROF:
			svc->rof = atoi(text);
			break;
		case SVC_XML_PARAM_SWITCH:
			strv = g_strsplit(text, ":", 3);
			if (strv != NULL) {
				svc->port10 = atoi(strv[0]);
				svc->port11 = atoi(strv[1]);
				svc->rep11 = atoi(strv[2]);
				g_strfreev(strv);
			}
			break;
		case SVC_XML_PARAM_LNB:
			strv = g_strsplit(text, ":", 5);
			if (strv != NULL) {
				strvlen = g_strv_length(strv);
				if (strvlen >= 2) {
					svc->lnbtype = atoi(strv[0]);
					svc->lnb22khz = atoi(strv[1]);
					if (strvlen == 5) {
						svc->lnbhifreq = atoi(strv[2]);
						svc->lnblofreq = atoi(strv[3]);
						svc->lnbswfreq = atoi(strv[4]);
					}
				}
				g_strfreev(strv);
			}
			break;
		case SVC_XML_PARAM_POSITIONER:
			svc->positioner = atoi(text);
			if ((svc->positioner != POSITIONER_POSITION) && 
				(svc->positioner != POSITIONER_NUMBER)) {
				svc->positioner = POSITIONER_DIASABLED;
			}
			break;
		case SVC_XML_PARAM_POSSATNUM:
			svc->possatnum = atoi(text);
			if (svc->possatnum > 128) {
				svc->possatnum = 0;
			}
			break;
		case SVC_XML_PARAM_NPID:
			svc->npid = atoi(text);
			break;
		case SVC_XML_PARAM_NIP:
			if (svc->nip != NULL) g_free(svc->nip);
			svc->nip = g_markup_escape_text(text, -1);
			break;
		case SVC_XML_PARAM_NMASK:
			if (svc->nmask != NULL) g_free(svc->nmask);
			svc->nmask = g_strdup(text);
			break;
		case SVC_XML_PARAM_NMAC:
			if (svc->nmac != NULL) g_free(svc->nmac);
			svc->nmac = g_markup_escape_text(text, -1);
			break;
		case SVC_XML_PARAM_NPROG:
			if (svc->nprog != NULL) g_free(svc->nprog);
			svc->nprog = g_markup_escape_text(text, -1);
			break;
		case SVC_XML_PARAM_CAID:
			svc->caid = atoi(text);
			break;
		case SVC_XML_PARAM_SID:
			svc->sid = atoi(text);
			break;
		case SVC_XML_PARAM_KEY:
			if (svc->key != NULL) g_free(svc->key);
			svc->key = g_markup_escape_text(text, -1);
			break;
		case SVC_XML_PARAM_PIDS:
			strv = g_strsplit(text, ";", -1);
			i = 0;
			while (strv[i] != NULL) {
				strv2 = g_strsplit(strv[i], ":", 3);
				curpid = g_malloc0(sizeof(pid));
				curpid->type = atoi(strv2[0]);
				curpid->value = atoi(strv2[1]);
				if (strv2[2] != NULL) curpid->enabled = (gboolean)atoi(strv2[2]); else curpid->enabled = TRUE; 
				svc->pids = g_slist_prepend(svc->pids, curpid); 
				g_strfreev(strv2);
				i++;
			}
			svc->pids = g_slist_reverse(svc->pids);
			g_strfreev(strv);
			break;
		default: break;
	}
	
}

static void service_xml_end_element(GMarkupParseContext *context, const gchar *element, gpointer data, GError **error)
{
	if (!g_str_equal(element, "service")) {
		xml_parameter = SVC_XML_PARAM_NULL;
	}
	
	return;
}
