/*
 *      dvbscan.c
 *      
 *      Copyright 2009 - 2012 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


#include <glib.h>
#include <stdio.h>
#include <stdint.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <string.h>
#include <sys/ioctl.h>
#include <time.h>
#include <poll.h>
#include <sys/socket.h>
#include <linux/dvb/dmx.h>
#include "dvbscan.h"
#include "common/iso3166.h"

enum {
	DVBSCAN_COMMAND_NULL = 0,
	DVBSCAN_COMMAND_STOP
};

enum {
	DVBSCAN_ANSWER_SUCCESS = 0,
	DVBSCAN_ANSWER_FAILURE
};

static gint dvbscan_search_entry_compare(gconstpointer a, gconstpointer b);
static gint dvbscan_search_entry_sid_compare(gconstpointer a, gconstpointer b);
static void dvbscan_count_enabled_foreach(gpointer data, gpointer user_data);

static gboolean dvbscan_start_filter(dvbscan_t instance, guint16 pid, guint8 tid);
static gboolean dvbscan_stop_filter(dvbscan_t instance);
static gint dvbscan_read_filter(dvbscan_t instance, gchar *buffer, gint bufsize, guchar timeout);
static gint dvbscan_send_command(dvbscan_t instance, gint command);
static void dvbscan_free_list_foreach(gpointer data, gpointer user_data);
static void dvbscan_free_list(dvbscan_t instance);

static void dvbscan_add_caid(struct _scanentry *entry, gushort caid);
static gchar dvbscan_add_audio_pid(struct _scanentry *entry, gushort audiopid);
static gchar dvbscan_add_data_pid(struct _scanentry *entry, gushort datapid);
static void dvbscan_add_audio_language(struct _scanentry *entry, gchar *buffer, gchar langid);
static time_t dvbscan_decode_time(guint MJD_time, guint UTC_time);
static gdouble dvbscan_decode_duration(guint UTC_time);

static gpointer dvbscan_transponder_thread(gpointer data);

static void dvbscan_free_event_list_foreach(gpointer data, gpointer user_data);
static void dvbscan_free_event_list(dvbscan_t instance);
static gint dvbscan_sort_event_list(gconstpointer a, gconstpointer b);

static gpointer dvbscan_epg_thread(gpointer data);


static gint dvbscan_search_entry_compare(gconstpointer a, gconstpointer b)
{
	struct _scanentry *entry;
	guint id;
	
	entry = (struct _scanentry *)a;
	id = GPOINTER_TO_UINT(b);
	
	if (entry->id < id) {
		return 1;
	} else if (entry->id > id) {
		return -1;
	} else {
		return 0;
	}
}

gpointer dvbscan_search_entry(dvbscan_t instance, guint id)
{
	GSList *entry;
	
	if ((instance == NULL) || ((instance != NULL) && (instance->scanlist == NULL))) return NULL;
	
	entry = g_slist_find_custom(instance->scanlist, GUINT_TO_POINTER(id), dvbscan_search_entry_compare);
	
	if (entry == NULL) {
		return NULL;
	} else {
		return entry->data;
	}
}

static gint dvbscan_search_entry_sid_compare(gconstpointer a, gconstpointer b)
{
	struct _scanentry *entry;
	guint sid;
	
	entry = (struct _scanentry *)a;
	sid = GPOINTER_TO_UINT(b);
	
	if (entry->type == DVBSCAN_ENTRY_TRANSPONDER) return -1;
	
	if (entry->values.service.sid < sid) {
		return 1;
	} else if (entry->values.service.sid > sid) {
		return -1;
	} else {
		return 0;
	}
}

gpointer dvbscan_search_entry_sid(dvbscan_t instance, guint sid)
{
	GSList *entry;
	
	if ((instance == NULL) || ((instance != NULL) && (instance->scanlist == NULL))) return NULL;
	
	entry = g_slist_find_custom(instance->scanlist, GUINT_TO_POINTER(sid), dvbscan_search_entry_sid_compare);
	
	if (entry == NULL) {
		return NULL;
	} else {
		return entry->data;
	}
}

static void dvbscan_count_enabled_foreach(gpointer data, gpointer user_data)
{
	struct _scanentry *entry;
	guint *counter;
	
	entry = (struct _scanentry *)data;
	counter = (guint *)user_data;
	
	if (entry->sticky) (*counter)++;
}

guint dvbscan_count_enabled(dvbscan_t instance)
{
	guint counter;
	
	if ((instance == NULL) || ((instance != NULL) && (instance->scanlist == NULL))) return 0;
	
	counter = 0;
	
	g_slist_foreach(instance->scanlist, dvbscan_count_enabled_foreach, &counter);
	
	return counter;
}

GSList *dvbscan_get_list(dvbscan_t instance)
{
	if (instance == NULL) return NULL;
	
	return instance->scanlist;
}

gchar *dvbscan_get_network_name(dvbscan_t instance)
{
	if (instance == NULL) return NULL;
	
	return instance->networkname;
}

struct tm *dvbscan_get_tptime(dvbscan_t instance)
{
	if (instance == NULL) return NULL;
	
	return instance->timestamp;
}

static gboolean dvbscan_start_filter(dvbscan_t instance, guint16 pid, guint8 tid)
{
	struct dmx_sct_filter_params f;
	gint fd;
	gchar *demuxfn;
	gint flags;
	
	if (instance == NULL) return FALSE;
	
	demuxfn = g_strdup_printf("/dev/dvb/adapter%u/demux%u", instance->adapter, instance->demux, NULL);
	
	if ((instance->filterfd = open(demuxfn, O_RDWR | O_NONBLOCK)) < 0) {
		g_free(demuxfn);
		return FALSE;
	}
	
	g_free(demuxfn);
	
	memset(&f, 0, sizeof(f));

	f.pid = pid;
	f.filter.filter[0] = tid;
	f.filter.mask[0]   = 0xff;
	

	f.timeout = 0;
	f.flags = DMX_IMMEDIATE_START | DMX_CHECK_CRC;

	if (ioctl(instance->filterfd, DMX_SET_FILTER, &f) == -1) {
		printf("ioctl DMX_SET_FILTER failed\n");
		dvbscan_stop_filter(instance);
		return FALSE;
	}
	
	return TRUE;
}

static gboolean dvbscan_stop_filter(dvbscan_t instance)
{
	if (instance == NULL) return FALSE;
	
	ioctl(instance->filterfd, DMX_STOP);
	close(instance->filterfd);
	
	return TRUE;
}

static gint dvbscan_read_filter(dvbscan_t instance, gchar *buffer, gint bufsize, guchar timeout)
{
	struct pollfd pollfds[2];
	ssize_t readb;
	gint answer;
	
	if ((instance == NULL) || (buffer == NULL) || (bufsize == 0)) return -1;
	
	pollfds[0].fd = instance->filterfd;
	pollfds[0].events = POLLIN;
	pollfds[1].fd = instance->ctlfd[1];
	pollfds[1].events = POLLIN;
	
	instance->command = DVBSCAN_COMMAND_NULL;
	readb = -1;
	
	if (poll(&pollfds[0], 2, timeout*1000) > 0) {
		if (pollfds[1].revents & POLLIN) {
			readb = read(instance->ctlfd[1], &(instance->command), sizeof(instance->command));
			if (readb > 0) {
				if (instance->command == DVBSCAN_COMMAND_STOP) {
					answer = DVBSCAN_ANSWER_SUCCESS;
					write(instance->ctlfd[1], &answer, sizeof(answer));
					readb = 0;
					return readb;
				}
			}
		}
		
		if (pollfds[0].revents & POLLIN) {
			readb = read(instance->filterfd, buffer, bufsize);
		}
	}
	
	return readb;
}

static gint dvbscan_send_command(dvbscan_t instance, gint command)
{
	struct pollfd pollfds[1];
		
	if (instance == NULL) return DVBSCAN_ANSWER_FAILURE;
	
	if (write(instance->ctlfd[0], &command, sizeof(command)) > 0) {
		pollfds[0].fd = instance->ctlfd[0];
		pollfds[0].events = POLLIN;
		
		if (poll(&pollfds[0], 1, 5000) > 0) {
			if (pollfds[0].revents & POLLIN) {
				if (read(instance->ctlfd[0], &command, sizeof(command)) > 0) {
					return command;
				}
			}
		}
	}
	
	return DVBSCAN_ANSWER_FAILURE;
}

static void dvbscan_free_list_foreach(gpointer data, gpointer user_data)
{
	struct _scanentry *entry;
	
	entry = (struct _scanentry *)data;
	
	if (entry->values.service.name != NULL) {
		g_free(entry->values.service.name);
		entry->values.service.name = NULL;
	}
	
	if (entry->values.service.provider != NULL) {
		g_free(entry->values.service.provider);
		entry->values.service.provider = NULL;
	}
}

static void dvbscan_free_list(dvbscan_t instance)
{
	if (instance->scanlist != NULL) {
		g_slist_foreach(instance->scanlist, dvbscan_free_list_foreach, NULL);
		g_slist_free(instance->scanlist);
		instance->scanlist = NULL;
	}
	
	if (instance->networkname != NULL) {
		g_free(instance->networkname);
		instance->networkname = NULL;
	}
	
	if (instance->timestamp != NULL) {
		g_free(instance->timestamp);
		instance->timestamp = NULL;
	}
}

static gchar *dvbscan_convert_encoding(gchar *string)
{
	gchar *encoding;
	const gchar *utf = "UTF-8";
	gsize bread, bwritten;
	gboolean needoffset = TRUE;
	gchar *convstr, *escstr;
	
	switch (string[0]) {
		case 0x01: encoding = "ISO8859-5"; break;
		case 0x02: encoding = "ISO8859-6"; break;
		case 0x03: encoding = "ISO8859-7"; break;
		case 0x04: encoding = "ISO8859-8"; break;
		case 0x05: encoding = "ISO8859-9"; break;
		case 0x06: encoding = "ISO8859-10"; break;
		case 0x07: encoding = "ISO8859-11"; break;
		case 0x08: encoding = "ISO8859-12"; break;
		case 0x09: encoding = "ISO8859-13"; break;
		case 0x0A: encoding = "ISO8859-14"; break;
		case 0x0B: encoding = "ISO8859-15"; break;
		default:   encoding = "ISO_6937";  needoffset = FALSE; break;
	}
	
	convstr = g_convert(needoffset ? string+1 : string, -1, utf, encoding, &bread, &bwritten, NULL);
	
	escstr = g_markup_escape_text(convstr, -1);
	
	g_free(convstr);
	
	return escstr;
}

gchar *dvbscan_get_cas_name(guint caid) 
{
	gint i = 0;
	gboolean found = FALSE;
	gchar *res;
	
	typedef struct {
		guint caid;
		guint rangecaid;
		const gchar *str;
	} enc_table;
	
	enc_table enctable[] = {
		{0x0000, 0x0000, "Reserved"},
		{0x0001, 0x00FF, "DVB CAS"},
		{0x0100, 0x01FF, "Seca/MediaGuard"},
		{0x0200, 0x02FF, "CCETT"},
		{0x0300, 0x03FF, "Kabel Deutschland"},
		{0x0400, 0x04FF, "EuroDec"},
		{0x0500, 0x05FF, "Viaccess"},
		{0x0600, 0x06FF, "Irdeto"},
		{0x0700, 0x07FF, "Jerrold/GI/Motorola"},
		{0x0800, 0x08FF, "Matra Communication"},
		{0x0900, 0x09FF, "Videoguard"},
		{0x0A00, 0x0AFF, "Nokia"},
		{0x0B00, 0x0BFF, "Conax"},
		{0x0C00, 0x0CFF, "NTL"},
		{0x0D00, 0x0DFF, "Cryptoworks"},
		{0x0E00, 0x0EFF, "Power VU"},
		{0x0F00, 0x0FFF, "Sony"},
		{0x1000, 0x10FF, "RAS"},
		{0x1100, 0x11FF, "Thomson"},
		{0x1200, 0x12FF, "TV//Com"},
		{0x1300, 0x13FF, "HPT"},
		{0x1400, 0x14FF, "HRT"},
		{0x1500, 0x15FF, "IBM"},
		{0x1600, 0x16FF, "Nera"},
		{0x1700, 0x17FF, "Betacrypt"},
		{0x1800, 0x18FF, "Nagravision"},
		{0x1900, 0x19FF, "Titan Information Systems"},
		{0x2000, 0x20FF, "Telefonica Servicios Audiovisuales"},
		{0x2100, 0x21FF, "STENTOR"},
		{0x2200, 0x22FF, "Codicrypt"},
		{0x2300, 0x23FF, "BARCO AS"},
		{0x2400, 0x24FF, "StarGuide Digital Networks"},
		{0x2500, 0x25FF, "Mentor Data System, Inc."},
		{0x2600, 0x26FF, "BISS"},
		{0x2700, 0x27FF, "PolyCipher"},
		{0x4347, 0x4347, "CryptOn"},
		{0x4700, 0x47FF, "Motorola"},
		{0x4800, 0x48FF, "Accessgate"},
		{0x4900, 0x49FF, "China Crypt"},
		{0x4A00, 0x4A0F, "Novel-SuperTV"},
		{0x4A10, 0x4A1F, "EasyCas"},
		{0x4A20, 0x4A2F, "AlphaCrypt"},
		{0x4A30, 0x4A3F, "DVN Holdings"},
		{0x4A40, 0x4A4F, "ADT"},
		{0x4A50, 0x4A5F, "Shenzhen Kingsky"},
		{0x4A60, 0x4A6F, "@Sky"},
		{0x4A70, 0x4A7F, "Dreamcrypt"},
		{0x4A80, 0x4A8F, "THALESCrypt"},
		{0x4A90, 0x4A9F, "Runcom Technologies"},
		{0x4AA0, 0x4AAF, "KeyFly"},
		{0x4AB0, 0x4ABF, "CTI-CAS"},
		{0x4AC0, 0x4ACF, "Latens Systems Ltd "},
		{0x4AD0, 0x4AD1, "XCrypt"},
		{0x4AD2, 0x4AD3, "Beijing Digital Video Technology Co., Ltd. "},
		{0x4AD4, 0x4AD5, "OmniCrypt"},
		{0x4AD6, 0x4AD7, "SK Telecom Co., Ltd. "},
		{0x4AD8, 0x4AD9, "Enigma Systems"},
		{0x4ADA, 0x4ADA, "Wyplay SAS"},
		{0x4ADB, 0x4ADB, "Jinan Taixin Electronics, Co., Ltd."},
		{0x4ADC, 0x4ADC, "LogiWays"},
		{0x4ADD, 0x4ADD, "ATSC SRM"},
		{0x4ADE, 0x4ADE, "CerberCrypt"},
		{0x4ADF, 0x4ADF, "Caston Co., Ltd."},
		{0x4AE0, 0x4AE1, "DRE-Crypt"},
		{0x4AE2, 0x4AE3, "Microsoft Corp."},
		{0x4AE4, 0x4AE4, "Coretrust, Inc."},
		{0x4AE5, 0x4AE5, "PRO-Crypt"},
		{0x4AE6, 0x4AE6, "SypherMedia International"},
		{0x4AE7, 0x4AE7, "Guangzhou Ewider Technology Corporation Limited"},
		{0x4AE8, 0x4AE8, "FG Crypt"},
		{0x4AE9, 0x4AE9, "Dreamer-i Co., Ltd."},
		{0x4AEA, 0x4AEA, "Cryptoguard AB"},
		{0x4AEB, 0x4AEB, "Abel Quintic"},
		{0x4AEC, 0x4AEC, "FTS DVL SRL"},
		{0x5347, 0x5347, "GkWare e.K."},
		{0x5500, 0x5500, "DRE-Crypt"},
		{0x5501, 0x5501, "Griffin"},
		{0x5581, 0x5581, "Bulcrypt"},
		{0x5601, 0x5601, "Verimatrix"},
		{0xA101, 0xA101, "RosCrypt-M1"},
		{0,      0,      NULL}
	};
	
	while (enctable[i].str != NULL) {
		if ((caid >= enctable[i].caid) && (caid <= enctable[i].rangecaid)) {
			return g_strdup(enctable[i].str);
		}		
		i++;
	}
	
	return g_strdup_printf("Unknown 0x%x", caid); 
}


static void dvbscan_add_caid(struct _scanentry *entry, gushort caid)
{
	gint i, freeindex;
	gboolean idexists;
	
	freeindex = -1;
	idexists = FALSE;
	
	if ((caid == 0) || (entry == NULL)) return;
	
	if (entry->type == DVBSCAN_ENTRY_TRANSPONDER) return;
	
	for (i=0; i<16; i++) {
		if ((freeindex == -1) && (entry->values.service.caid[i] == 0)) freeindex = i;
		if (caid == entry->values.service.caid[i]) {
			idexists = TRUE;
			break;
		}
	}
	
	if ((!idexists) && (freeindex != -1)) {
		entry->values.service.caid[freeindex] = caid;
		printf(" CAID: 0x%x\n", caid);
	}
}

static gchar dvbscan_add_audio_pid(struct _scanentry *entry, gushort audiopid)
{
	gint i, freeindex;
	gboolean idexists;
	
	freeindex = -1;
	idexists = FALSE;
	
	if ((audiopid == 0) || (entry == NULL)) return -1;
	
	if (entry->type == DVBSCAN_ENTRY_TRANSPONDER) return -1;
	
	for (i=0; i<16; i++) {
		if ((freeindex == -1) && (entry->values.service.apid[i] == 0)) freeindex = i;
		if (audiopid == entry->values.service.apid[i]) {
			idexists = TRUE;
			break;
		}
	}
	
	if ((!idexists) && (freeindex != -1)) {
		entry->values.service.apid[freeindex] = audiopid;
		printf(" Audio PID: 0x%x\n", audiopid);
	}
	
	return (guchar)freeindex;
}

static gchar dvbscan_add_data_pid(struct _scanentry *entry, gushort datapid)
{
	gint i, freeindex;
	gboolean idexists;
	
	freeindex = -1;
	idexists = FALSE;
	
	if ((datapid == 0) || (entry == NULL)) return -1;
	
	if (entry->type == DVBSCAN_ENTRY_TRANSPONDER) return -1;
	
	for (i=0; i<16; i++) {
		if ((freeindex == -1) && (entry->values.service.dpid[i] == 0)) freeindex = i;
		if (datapid == entry->values.service.dpid[i]) {
			idexists = TRUE;
			break;
		}
	}
	
	if ((!idexists) && (freeindex != -1)) {
		entry->values.service.dpid[freeindex] = datapid;
		printf(" Data PID: 0x%x\n", datapid);
	}
	
	return (guchar)freeindex;
}

static void dvbscan_add_audio_language(struct _scanentry *entry, gchar *buffer, gchar langid)
{
	gchar langbuf[4];
	gushort langcode;
	
	if ((langid < 0) || (langid > DVBSCAN_SERVICE_LANG_ID) || (entry == NULL) || (buffer == NULL)) return;
	
	if (entry->type == DVBSCAN_ENTRY_TRANSPONDER) return;
	
	memset(langbuf, 0, sizeof(langbuf));
	memcpy(&langbuf, buffer, 3);
	
	if (langid == DVBSCAN_SERVICE_LANG_ID) {
		if ((langcode = iso3166_get_code_from_a3((const char *)buffer)) != 0) {
			entry->values.service.svclang = langcode;
			printf(" Service language: %u (%s)\n", langcode, langbuf);
		}
	} else {
		if ((langcode = iso3166_get_code_from_a3((const char *)buffer)) != 0) {
			entry->values.service.lang[langid] = langcode;
			printf(" Language: %u (%s)\n", langcode, langbuf);
		}
	}
}

static time_t dvbscan_decode_time(guint MJD_time, guint UTC_time)
{
	struct tm timestamp;
	guint  y, m, d, k;
	time_t ttime;
	
	y = (guint)((MJD_time - 15078.2) / 365.25);
	m = (guint)((MJD_time - 14956.1 - (guint)(y * 365.25)) / 30.6001);
	d = (guint)(MJD_time - 14956 - (guint)(y * 365.25) - (guint)(m * 30.6001));
	if ((m == 14) || (m == 15)) k = 1; else k = 0;
	y = y + k;
	m = m - 1 - (k * 12);
	
	timestamp.tm_mday = d;
	timestamp.tm_mon =  m-1;
	timestamp.tm_year = y;
	timestamp.tm_hour = (((UTC_time >> 20) & 0x0F) * 10) + ((UTC_time >> 16) & 0x0F);
	timestamp.tm_min =  (((UTC_time >> 12) & 0x0F) * 10) + ((UTC_time >> 8) & 0x0F);
	timestamp.tm_sec =  (((UTC_time >> 4) & 0x0F) * 10) + (UTC_time & 0x0F);
	
	ttime = mktime(&timestamp) - timezone;
	
	return ttime;
}

static gdouble dvbscan_decode_duration(guint UTC_time)
{
	gdouble duration;
	
	duration  = ((((UTC_time >> 20) & 0x0F) * 10) + ((UTC_time >> 16) & 0x0F))*60.0*60.0;
	duration += ((((UTC_time >> 12) & 0x0F) * 10) + ((UTC_time >> 8) & 0x0F))*60.0;
	duration +=  (((UTC_time >> 4) & 0x0F) * 10) + (UTC_time & 0x0F);
	
	return duration;
}

dvbscan_t dvbscan_new(guint datasrc[3], gint type, dvbscan_status_callback callback)
{
	dvbscan_t instance;
	
	if (callback == NULL) return NULL;
	
	instance = g_malloc0(sizeof(struct _dvbscan));
	
	if (instance != NULL) {
		instance->adapter = datasrc[0];
		instance->demux = datasrc[1];
		instance->sid = datasrc[2];
		instance->type = type;
		instance->scanlist = NULL;
		instance->eventlist = NULL;
		instance->networkname = NULL;
		instance->timestamp = NULL;
		instance->callback = callback;
		
		if (socketpair(AF_LOCAL, SOCK_STREAM, 0, instance->ctlfd) == -1) {
			g_free(instance);
			return NULL;
		}
		
		if (type == DVBSCAN_TYPE_TRANSPONDER) {
			instance->status = DVBSCAN_STATUS_PAT;
			instance->thread = g_thread_create(dvbscan_transponder_thread, instance, FALSE, NULL);
		} else if (type == DVBSCAN_TYPE_EPG) {
			instance->status = DVBSCAN_EPG_STATUS_CURRENT;
			instance->thread = g_thread_create(dvbscan_epg_thread, instance, FALSE, NULL);
		}
		
	}
	
	return instance;
}

gboolean dvbscan_free(dvbscan_t instance)
{
	if (instance == NULL) return FALSE;
	
	instance->callback = NULL;
	
	if (instance->type == DVBSCAN_TYPE_TRANSPONDER) {
		if (instance->status == DVBSCAN_STATUS_NULL) {
			dvbscan_free_list(instance);
		} else {
			dvbscan_send_command(instance, DVBSCAN_COMMAND_STOP);
			dvbscan_free_list(instance);
		}
	} else if (instance->type == DVBSCAN_TYPE_EPG) {
		if (instance->status == DVBSCAN_EPG_STATUS_NULL) {
			dvbscan_free_event_list(instance);
		} else {
			dvbscan_send_command(instance, DVBSCAN_COMMAND_STOP);
			dvbscan_free_event_list(instance);
		}
	}
	
	close(instance->ctlfd[0]);
	close(instance->ctlfd[1]);
	
	g_free(instance);
	instance = NULL;
	
	return TRUE;
}

static gpointer dvbscan_transponder_thread(gpointer data)
{
	const gint fec_tt[16] = {0, 1, 2, 3, 6, 8, 9, 4, 5, 10, 0, 0, 0, 0, 0, 0};
	const gint mod_tt[4]  = {0, 0, 1, 0};
	const gint sys_tt[2]  = {0, 1};
	const gint rol_tt[4]  = {1, 2, 3, 0};
	const gint pol_tt[4]  = {1, 0, 2, 3};
	
	struct _scanentry *entry;
	dvbscan_t instance;
	
	guint  listlen;
	guint  res;
	guchar buffer[4096];
	guchar table_id;
	guint  table_length;
	guint  transport_stream_id;
	guchar version_number;
	guchar section_number;
	guchar last_section_number;
	guint  program_info_length;
	guchar stream_type;
	gushort  caid;
	guint  elementary_PID;
	guint  ES_info_length;
	guchar descriptor_tag;
	guchar descriptor_length;
	
	guint  num_services;	
	gint   i,i2;
	gint   ptr = 0;
	gint   descriptorptr = 0;
	gchar  audioid;
	
	guint  original_network_id;
	gint   service_id;
	guchar service_status;
	guchar free_CA_mode;
	guint  descriptors_loop_length;
	guchar service_provider_name_length;
	guchar service_name_length;
	gchar  strbuf[256];
	
	guint  network_id;
	guint  network_descriptors_length;
	guchar service_type;
	guint  transport_stream_loop_length;
	guint  transport_descriptors_length;
		
	guint  MJD_time;
	guint  UTC_time;
	struct tm *tspec;
	time_t ttime;
		
	
	if (data == NULL) return NULL;
	
	instance = (dvbscan_t)data;
	
	//PAT SECTION
	instance->status = DVBSCAN_STATUS_PAT;
	if (instance->callback != NULL) (instance->callback)(DVBSCAN_STATUS_PAT);
		
	if (!dvbscan_start_filter(instance, 0x0000, 0x00)) {
		printf("[-] Failed to set filter for PAT section\n");
		instance->status = DVBSCAN_STATUS_NULL;
		if (instance->callback != NULL) (instance->callback)(DVBSCAN_STATUS_NULL);
		return NULL;
	}
	
	res = dvbscan_read_filter(instance, (char *)&buffer, sizeof(buffer), DVBSCAN_PAT_READ_TIMEOUT);
	
	dvbscan_stop_filter(instance);
	
	if ((res == -1) || ((res == 0) && (instance->command == DVBSCAN_COMMAND_STOP))) {
		if (res == -1) {
			printf("[-] Failed to read PAT section\n");
		} else if (res == 0) {
			printf("[-] PAT section read process stopped by user\n");
		}
		instance->status = DVBSCAN_STATUS_NULL;
		if (instance->callback != NULL) (instance->callback)(DVBSCAN_STATUS_NULL);
		return NULL;
	}
	
	printf("PAT section\n");
	
	table_id = buffer[0];
	table_length = (((buffer[1] & 0x0f) << 8) | buffer[2]);
	transport_stream_id = (((buffer[3] & 0xff) << 8) | buffer[4]);
	version_number = (buffer[5] & 0x3e);
	section_number = buffer[6];
	last_section_number = buffer[7];
	
	num_services = ((table_length + 3 - DVBSCAN_SECTION_HEADER_LENGTH - DVBSCAN_SECTION_CRC_LENGTH) / 4);	
		
	if (instance->scanlist != NULL) dvbscan_free_list(instance);
	
	listlen = 0;
	
	for (i=0; i<num_services; i++) {
		entry = g_malloc0(sizeof(struct _scanentry));
		entry->values.service.sid = (((buffer[8+(4*i)] & 0xff) << 8) | buffer[9+(4*i)]);
		entry->values.service.pmt = (((buffer[10+(4*i)] & 0x1f) << 8) | buffer[11+(4*i)]);
		entry->values.service.name = g_strdup_printf("Service %u", entry->values.service.sid);
		entry->values.service.provider = g_strdup_printf("Provider %u", entry->values.service.sid);
		if (entry->values.service.sid == 0) {
			g_free(entry);
		} else {
			entry->id = listlen;
			entry->type = DVBSCAN_ENTRY_SERVICE;
			entry->sticky = TRUE;
			instance->scanlist = g_slist_append(instance->scanlist, entry);
			printf(" %u. SID = %u, PID = 0x%x\n", i, entry->values.service.sid, entry->values.service.pmt);
			listlen++;
		}
	}
	
	if (listlen > 0) {
		printf("Read %u services from PAT\n", listlen);
	} else {
		g_warning("Services not found in PAT");
		if (instance->callback != NULL) (instance->callback)(DVBSCAN_STATUS_NULL);
		return NULL;
	}
	
	
	//PMT SECTION
	instance->status = DVBSCAN_STATUS_PMT;
	if (instance->callback != NULL) (instance->callback)(DVBSCAN_STATUS_PMT);
		
	for (i=0; i<listlen; i++) {
		entry = g_slist_nth_data(instance->scanlist, i);
		
		if (!dvbscan_start_filter(instance, entry->values.service.pmt, 0x02)) {
			printf("[-] Failed to set filter for PMT section\n");
			continue;
		}
		
		res = dvbscan_read_filter(instance, (char *)&buffer, sizeof(buffer), DVBSCAN_PMT_READ_TIMEOUT);
		
		dvbscan_stop_filter(instance);
		
		if ((res == 0) && (instance->command == DVBSCAN_COMMAND_STOP)) {
			printf("[-] PMT section read process stopped by user\n");
			instance->status = DVBSCAN_STATUS_NULL;
			if (instance->callback != NULL) (instance->callback)(DVBSCAN_STATUS_NULL);
			return NULL;
		} else if (res == -1) {
			g_warning("Failed to read PMT table for PID %i", entry->values.service.pmt);
			continue;
		}
		
		printf(" %i. PMT section PID 0x%x\n", i, entry->values.service.pmt);
		
		table_id = buffer[0];
		table_length = (((buffer[1] & 0x0f) << 8) | buffer[2]);
		transport_stream_id = (((buffer[3] & 0xff) << 8) | buffer[4]);
		version_number = (buffer[5] & 0x3e);
		section_number = buffer[6];
		last_section_number = buffer[7];
		entry->values.service.pcr = (((buffer[8] & 0x1f) << 8) | buffer[9]);
		program_info_length = (((buffer[10] & 0x0f) << 8) | buffer[11]);
		
		ptr = 12;
		descriptorptr = 0;
		
		while ((program_info_length > 0) && (descriptorptr < program_info_length-1)) {
			descriptor_tag = buffer[ptr];
			descriptor_length = buffer[ptr+1] + 2;
			if (descriptor_tag == 0x0a) {
				dvbscan_add_audio_language(entry, &buffer[ptr+2], DVBSCAN_SERVICE_LANG_ID);
			} else if (descriptor_tag == 0x09) {
				caid = (buffer[ptr+2] << 8) | buffer[ptr+3];
				dvbscan_add_caid(entry, caid);
			} /*else printf("1. Descriptorlength: %i, tag 0x%x, length %i\n", ES_info_length, descriptor_tag, descriptor_length);*/
			ptr += descriptor_length;
			descriptorptr += descriptor_length;
		}
		
		while (ptr < table_length-1) {
			stream_type = buffer[ptr];
			elementary_PID = (((buffer[ptr+1] & 0x1f) << 8) | buffer[ptr+2]);
			ES_info_length = (((buffer[ptr+3] & 0x0f) << 8) | buffer[ptr+4]);
			if ((stream_type == 0x01) || (stream_type == 0x02) || (stream_type == 0x1b) || (stream_type == 0x10) || (stream_type == 0x80) || (stream_type == 0xea)) {
				/*MPEG1Video (0x01), MPEG2Video (0x02), H264Video (0x1b), MPEG4Video (0x10), OpenCableVideo (0x80), VC1Video (0xea)*/
				entry->values.service.vpid = elementary_PID;
				printf(" Video PID: 0x%x\n", elementary_PID);
			} else if ((stream_type == 0x03) || (stream_type == 0x04) || (stream_type == 0x81) || (stream_type == 0x0f) || (stream_type == 0x11) || (stream_type == 0x8a)) {
				/*MPEG1Audio (0x03), MPEG2Audio (0x04), MPEG2AACAudio (0x0f), MPEG2AudioAmd1 (0x11), AC3Audio (0x81), DTSAudio (0x8a)*/
				audioid = dvbscan_add_audio_pid(entry, elementary_PID);
			} else if (stream_type == 0x06) {
				dvbscan_add_data_pid(entry, elementary_PID);
			} /*else printf("Other: 0x%x\n", stream_type);*/

			if (ES_info_length > 3) {
				descriptorptr = 0;
				while (descriptorptr < ES_info_length-1) {
					descriptor_tag = buffer[ptr+5];
					descriptor_length = buffer[ptr+6] + 2;
					if (descriptor_tag == 0x0a) {
						dvbscan_add_audio_language(entry, &buffer[ptr+7], audioid);
					} else if (descriptor_tag == 0x09) {
						caid = (buffer[ptr+7] << 8) | buffer[ptr+8];
						dvbscan_add_caid(entry, caid);
					} else if ((stream_type == 0x06) && (descriptor_tag == 0x56)) {
						entry->values.service.ttpid = elementary_PID;
						printf(" Teletext PID: 0x%x\n", elementary_PID);
					} else if ((stream_type == 0x06) && (descriptor_tag == 0x59)) {
						entry->values.service.stpid = elementary_PID;
						printf(" Subtitles PID: 0x%x\n", elementary_PID);
					} else if ((stream_type == 0x06) && (descriptor_tag == 0x6a)) {
						entry->values.service.ac3pid = elementary_PID;
						printf(" AC3 Audio PID: 0x%x\n", elementary_PID);
					} /*else printf("2. Descriptorlength: %i, type: %i, tag 0x%x, length %i\n", ES_info_length, stream_type, descriptor_tag, descriptor_length);*/
					descriptorptr += descriptor_length;
				}
			}
			
			ptr += ES_info_length + 5;
		}
		
		if (entry->values.service.caid[0] != 0) entry->sticky = FALSE;
	}
			
	//SDT SECTION
	instance->status = DVBSCAN_STATUS_SDT;
	if (instance->callback != NULL) (instance->callback)(DVBSCAN_STATUS_SDT);
	
	if (!dvbscan_start_filter(instance, 0x0011, 0x42)) {
		printf("[-] Failed to set filter for SDT section\n");
		instance->status = DVBSCAN_STATUS_NULL;
		if (instance->callback != NULL) (instance->callback)(DVBSCAN_STATUS_NULL);
		return NULL;
	}
	
	res = dvbscan_read_filter(instance, (char *)&buffer, sizeof(buffer), DVBSCAN_SDT_READ_TIMEOUT);
	
	dvbscan_stop_filter(instance);
	
	if ((res == -1) || ((res == 0) && (instance->command == DVBSCAN_COMMAND_STOP))) {
		if (res == -1) {
			printf("[-] Failed to read SDT section\n");
		} else if (res == 0) {
			printf("[-] SDT section read process stopped by user\n");
		}
		instance->status = DVBSCAN_STATUS_NULL;
		if (instance->callback != NULL) (instance->callback)(DVBSCAN_STATUS_NULL);
		return NULL;
	}
	
	printf("SDT section\n");
	
	table_id = buffer[0];
	table_length = (((buffer[1] & 0x0f) << 8) | buffer[2]);
	transport_stream_id = (((buffer[3] & 0xff) << 8) | buffer[4]);
	version_number = (buffer[5] & 0x3e);
	section_number = buffer[6];
	last_section_number = buffer[7];
	original_network_id = (((buffer[8] & 0xff) << 8) | buffer[9]);
	
	ptr = 11;
	i = 0;
	
	while (ptr < (table_length-1)) {
		service_id = (((buffer[ptr+0] & 0xff) << 8) | buffer[ptr+1]);
		service_status = ((buffer[ptr+2] & 0xe0) >> 5);
		free_CA_mode = ((buffer[ptr+2] & 0x10) >> 4);
		descriptors_loop_length = (((buffer[ptr+3] & 0x0f) << 8) | buffer[ptr+4]);
		descriptorptr = 0;
		entry = (struct _scanentry *)dvbscan_search_entry_sid(instance, service_id);
		if (entry != NULL) {
			entry->values.service.status = service_status;
			while (descriptorptr < (descriptors_loop_length - 1)) {
				descriptor_tag = buffer[ptr+descriptorptr+5];
				descriptor_length = buffer[ptr+descriptorptr+6]+2;
				if (descriptor_tag == 0x48) {
					entry->values.service.type = buffer[ptr+descriptorptr+7];
					service_provider_name_length = buffer[ptr+descriptorptr+8];
					
					memset(strbuf, 0, sizeof(strbuf));
					memcpy(strbuf, &buffer[ptr+descriptorptr+9], service_provider_name_length);
					strbuf[service_provider_name_length] = '\0';
					if (entry->values.service.provider != NULL) g_free(entry->values.service.provider);
					entry->values.service.provider = dvbscan_convert_encoding((gchar *)&strbuf);
					service_name_length = buffer[service_provider_name_length+ptr+descriptorptr+9];
					
					memset(strbuf, 0, sizeof(strbuf));
					memcpy(strbuf, &buffer[service_provider_name_length+ptr+descriptorptr+10], service_name_length);
					strbuf[service_name_length] = '\0';
					if (entry->values.service.name != NULL) g_free(entry->values.service.name);
					entry->values.service.name = dvbscan_convert_encoding((gchar *)&strbuf);
				} else if (descriptor_tag == 0x53) {
					for (i=0; i<((descriptor_length-2)/2); i++) {
						caid = (((buffer[ptr+descriptorptr+7] & 0xff) << 8) | buffer[ptr+descriptorptr+8]);
						dvbscan_add_caid(entry, caid);
					}
				} /*else printf("3. descriptor_length = %i (0x%x) loop %i\n", descriptor_length, descriptor_tag, descriptors_loop_length);*/
				descriptorptr += descriptor_length;
			}
			printf(" %u. %s - %s\n", i+1, entry->values.service.provider, entry->values.service.name/*scansvc->provider, scansvc->name*/);
			i++;
		} else {
			printf("SDT NOT FOUND\n");
		}
		ptr += 5 + descriptors_loop_length;
	}
		
	
	//NIT SECTION
	instance->status = DVBSCAN_STATUS_NIT;
	if (instance->callback != NULL) (instance->callback)(DVBSCAN_STATUS_NIT);
	
	if (!dvbscan_start_filter(instance, 0x10, 0x40)) {
		printf("[-] Failed to set filter for NIT section\n");
		instance->status = DVBSCAN_STATUS_NULL;
		if (instance->callback != NULL) (instance->callback)(DVBSCAN_STATUS_NULL);
		return NULL;
	}
	
	res = dvbscan_read_filter(instance, (char *)&buffer, sizeof(buffer), DVBSCAN_NIT_READ_TIMEOUT);
	
	dvbscan_stop_filter(instance);
	
	if ((res == -1) || ((res == 0) && (instance->command == DVBSCAN_COMMAND_STOP))) {
		if (res == -1) {
			printf("[-] Failed to read NIT section\n");
		} else if (res == 0) {
			printf("[-] NIT section read process stopped by user\n");
		}
		instance->status = DVBSCAN_STATUS_NULL;
		if (instance->callback != NULL) (instance->callback)(DVBSCAN_STATUS_NULL);
		return NULL;
	}
		
	printf("NIT section\n");
	
	table_id = buffer[0];
	table_length = (((buffer[1] & 0x0f) << 8) | buffer[2]);
	network_id = (((buffer[3] & 0xff) << 8) | buffer[4]);
	version_number = (buffer[5] & 0x3e);
	section_number = buffer[6];
	last_section_number = buffer[7];
	network_descriptors_length = (((buffer[8] & 0x0f) << 8) | buffer[9]);
	
	i = 0;
	ptr = 10;
	descriptorptr = 0;
	
	while ((network_descriptors_length > 0) && (descriptorptr < network_descriptors_length-1)) {
		descriptor_tag = buffer[ptr];
		descriptor_length = buffer[ptr+1] + 2;
		if (descriptor_tag == 0x40) {
			memset(strbuf, 0, sizeof(strbuf));
			memcpy(strbuf, &buffer[ptr+2], (descriptor_length-2)>255 ? 255 : (descriptor_length-2));
			instance->networkname = dvbscan_convert_encoding((gchar *)&strbuf);
			printf("Network name: %s\n", instance->networkname);
		} /*else printf("4. descriptor_tag = 0x%x, descriptor_length = %i\n", descriptor_tag, descriptor_length);*/
		/*else if (descriptor_tag == 0x09) {
			caid = (buffer[ptr+2] << 8) | buffer[ptr+3];
			printf("CAID: 0x%x\n", caid);
		}*/
		ptr += descriptor_length;
		descriptorptr += descriptor_length;
	}
	
	transport_stream_loop_length = (((buffer[ptr] & 0x1f) << 8) | buffer[ptr+1]);
	ptr = ptr + 2;
	
	while (ptr < table_length-1) {
		transport_stream_id = (((buffer[ptr] & 0x1f) << 8) | buffer[ptr+1]);
		original_network_id = (((buffer[ptr+2] & 0x1f) << 8) | buffer[ptr+3]);
		transport_descriptors_length = (((buffer[ptr+4] & 0x0f) << 8) | buffer[ptr+5]);
		
		if (transport_descriptors_length > 0) {
			descriptorptr = 0;
			while (descriptorptr < transport_descriptors_length-1) {
				descriptor_tag = buffer[ptr+descriptorptr+6];
				descriptor_length = buffer[ptr+descriptorptr+7] + 2;
				if (descriptor_tag == 0x41) {
					service_id = (((buffer[ptr+descriptorptr+8] & 0x1f) << 8) | buffer[ptr+descriptorptr+9]);
					service_type = buffer[ptr+descriptorptr+10];
					entry = (struct _scanentry *)dvbscan_search_entry_sid(instance, service_id);
					if (entry != NULL) {
						entry->values.service.type = service_type;
					}
					printf("SID: 0x%x, Service type: 0x%x\n", service_id, service_type);
				} else if (descriptor_tag == 0x43) {
					entry = g_malloc0(sizeof(struct _scanentry));
					
					entry->sticky = TRUE;
					entry->type = DVBSCAN_ENTRY_TRANSPONDER;
					
					entry->values.transponder.frequency = 10*(((buffer[ptr+descriptorptr+8]  >> 4) & 0x0f) * 10000000 + (buffer[ptr+descriptorptr+8] & 0x0f) * 1000000 +
																((buffer[ptr+descriptorptr+9]  >> 4) & 0x0f) * 100000   + (buffer[ptr+descriptorptr+9] & 0x0f) * 10000 +
																((buffer[ptr+descriptorptr+10] >> 4) & 0x0f) * 1000     + (buffer[ptr+descriptorptr+10] & 0x0f) * 100 +
																((buffer[ptr+descriptorptr+11] >> 4) & 0x0f) * 10       + (buffer[ptr+descriptorptr+11] & 0x0f));
					entry->values.transponder.position =    	(((buffer[ptr+descriptorptr+12] >> 4) & 0x0f) * 1000     + (buffer[ptr+descriptorptr+12] & 0x0f) * 100 +
																((buffer[ptr+descriptorptr+13] >> 4) & 0x0f) * 10       + (buffer[ptr+descriptorptr+13] & 0x0f));
					if (!(buffer[ptr+descriptorptr+14] >> 7)) {
						entry->values.transponder.position = -(entry->values.transponder.position);
					}
					entry->values.transponder.polarisation =  	(buffer[ptr+descriptorptr+14] >> 5) & 0x03;
					entry->values.transponder.rolloff =       	(buffer[ptr+descriptorptr+14] >> 3) & 0x03;
					entry->values.transponder.system =          (buffer[ptr+descriptorptr+14] >> 2) & 0x01;
					entry->values.transponder.modulation =      (buffer[ptr+descriptorptr+14]) & 0x03;
					entry->values.transponder.sr =        	10*(((buffer[ptr+descriptorptr+15] >> 4) & 0x0f) * 10000000 + (buffer[ptr+descriptorptr+15] & 0x0f) * 1000000 +
																((buffer[ptr+descriptorptr+16] >> 4) & 0x0f) * 100000   + (buffer[ptr+descriptorptr+16] & 0x0f) * 10000 +
																((buffer[ptr+descriptorptr+17] >> 4) & 0x0f) * 1000     + (buffer[ptr+descriptorptr+17] & 0x0f) * 100 +
																(((buffer[ptr+descriptorptr+18] & 0xf0) >> 4) & 0x0f) * 10 + ((buffer[ptr+descriptorptr+18] & 0xf0) & 0x0f));
					entry->values.transponder.fec =             buffer[ptr+descriptorptr+18] & 0x0f;
					
					if (entry->values.transponder.modulation == 0x02) entry->values.transponder.system = 0x01;
					
					entry->values.transponder.frequency = entry->values.transponder.frequency/1000;
					entry->values.transponder.sr = entry->values.transponder.sr/1000;
					entry->values.transponder.polarisation = pol_tt[entry->values.transponder.polarisation];
					entry->values.transponder.rolloff = rol_tt[entry->values.transponder.rolloff];
					entry->values.transponder.system = sys_tt[entry->values.transponder.system];
					entry->values.transponder.modulation = mod_tt[entry->values.transponder.modulation];
					entry->values.transponder.fec = fec_tt[entry->values.transponder.fec];
					
					if ((entry->values.transponder.frequency >= 3200) && (entry->values.transponder.frequency <= 12750) && (entry->values.transponder.sr > 0) && (abs(entry->values.transponder.position) <= 1800)) {
						entry->id = listlen;
						instance->scanlist = g_slist_append(instance->scanlist, entry);
						listlen++;
						i++;
						printf(" %i. fr: %i, pos: %i, sr: %i, pol: %i, mod: %i, fec: %i, sys: %i, rof: %i\n", entry->id, entry->values.transponder.frequency, entry->values.transponder.position, entry->values.transponder.sr,
																											entry->values.transponder.polarisation, entry->values.transponder.modulation, entry->values.transponder.fec, entry->values.transponder.system, entry->values.transponder.rolloff);
					} else {
						g_warning("Wrong parameters in discarded transponder entry");
						g_free(entry);
					}
				} else {
					printf("5. Descriptorlength: %i, tag 0x%x, length %i\n", ES_info_length, descriptor_tag, descriptor_length);
				}
				
				descriptorptr += descriptor_length;
			}
			
		}
		ptr += transport_descriptors_length + 6;	
	}
	
	//TDT SECTION
	instance->status = DVBSCAN_STATUS_TDT;
	if (instance->callback != NULL) (instance->callback)(DVBSCAN_STATUS_TDT);
	
	if (!dvbscan_start_filter(instance, 0x14, 0x70)) {
		printf("[-] Failed to set filter for TDT section\n");
		instance->status = DVBSCAN_STATUS_NULL;
		if (instance->callback != NULL) (instance->callback)(DVBSCAN_STATUS_NULL);
		return NULL;
	}
	
	res = dvbscan_read_filter(instance, (char *)&buffer, sizeof(buffer), DVBSCAN_TDT_READ_TIMEOUT);
	
	dvbscan_stop_filter(instance);
	
	if ((res == -1) || ((res == 0) && (instance->command == DVBSCAN_COMMAND_STOP))) {
		if (res == -1) {
			printf("[-] Failed to read TDT section\n");
		} else if (res == 0) {
			printf("[-] TDT section read process stopped by user\n");
		}
		instance->status = DVBSCAN_STATUS_NULL;
		if (instance->callback != NULL) (instance->callback)(DVBSCAN_STATUS_NULL);
		return NULL;
	}
			
	printf("TDT section\n");
		
	table_id = buffer[0];
	table_length = (((buffer[1] & 0x0f) << 8) | buffer[2]);
	MJD_time = (((buffer[3] & 0xff) << 8) | buffer[4]);
	UTC_time = (((buffer[5] & 0xff) << 16) | ((buffer[6] & 0xff) << 8) | buffer[7]);
	
	instance->timestamp = (struct tm *)g_malloc0(sizeof(struct tm));
	
	ttime = dvbscan_decode_time(MJD_time, UTC_time);
	
	tspec = gmtime(&ttime);
	
	memcpy(instance->timestamp, tspec, sizeof(struct tm));
		
	if (strftime(strbuf, sizeof(strbuf), "%c", instance->timestamp) == 0) {
		printf(" UTC time: broken\n");
	} else {
		printf(" UTC time: %s\n", strbuf);
	}
	
	instance->status = DVBSCAN_STATUS_NULL;
	if (instance->callback != NULL) (instance->callback)(DVBSCAN_STATUS_NULL);
	
	return NULL;
}

GSList *dvbscan_get_event_list(dvbscan_t instance)
{
	if (instance == NULL) return NULL;
	
	return instance->eventlist;
}

static void dvbscan_free_event_list_foreach(gpointer data, gpointer user_data)
{
	struct _scanevent *event;
	
	event = (struct _scanevent *)data;
	
	if (event->name != NULL) {
		g_free(event->name);
		event->name = NULL;
	}
	
	if (event->description != NULL) {
		g_free(event->description);
		event->description = NULL;
	}
}

static void dvbscan_free_event_list(dvbscan_t instance)
{
	if (instance->eventlist != NULL) {
		g_slist_foreach(instance->eventlist, dvbscan_free_event_list_foreach, NULL);
		g_slist_free(instance->eventlist);
		instance->eventlist = NULL;
	}
}

static gint dvbscan_sort_event_list(gconstpointer a, gconstpointer b)
{
	struct _scanevent *event1, *event2;
	
	event1 = (struct _scanevent *)a;
	event2 = (struct _scanevent *)b;
	
	if (event1->start == event2->start) {
		return 0;
	} else if (event1->start < event2->start) {
		return -1;
	} else {
		return 1;
	}
}

//gint dvbscan_epg_start(guint adapter, guint demux, guint sid, void (status_cb)(gint status))
static gpointer dvbscan_epg_thread(gpointer data)
{
	dvbscan_t instance;
	
	//gint    eitfilter;
	guchar  buffer[4096];
	guint8  table_id;
	guint16 section_length;
	guint16 service_id;
	guint8  version_number;
	guint8  section_number;
	guint8  last_section_number;
	guint16 transport_stream_id;
	guint16 original_network_id;
	guint8  segment_last_section_number;
	guint8  last_table_id;
	
	guint16 event_id;
	guint16 MJD_time;
	guint32 UTC_time;
	guint32 UTC_time_duration;
	guint8  running_status;
	guint8  free_CA_mode;
	guint16 descriptors_loop_length;
	
	guint8  descriptor_tag;
	guint8  descriptor_length;
	guint8  stream_content;
	guint8  component_type;
	guint8  component_tag;
	
	guint8  descriptor_number;
	guint8  last_descriptor_number;
	guint16 length_of_items;
	
	guint   di;
	guint8  content_nibble_level_1;
	guint8  content_nibble_level_2;
	guint8  user_nibble_1;
	guint8  user_nibble_2;
	
	guint8  raiting;
	
	time_t start;
	gdouble duration;
	
	guchar  language[4];
	guchar  string[256];
	guint8  event_name_length;
	guint8  text_length;
	
	guint   ptr, descriptorptr;
	gint    readb;
	guint   allevents, readevents, repevents, emptyevents;
	gint    attempt;
		
	struct _scanevent *event;
	GHashTable *eventhash;
	
	
	if (data == NULL) return NULL;
	
	instance = (dvbscan_t)data;
	
	if (instance->eventlist != NULL) dvbscan_free_event_list(instance);
	
	eventhash = g_hash_table_new(g_int_hash, g_int_equal);
	
	for (attempt=0; attempt<2; attempt++) {
		if (attempt == 0) {
			instance->status = DVBSCAN_EPG_STATUS_CURRENT;
			if (instance->callback != NULL) (instance->callback)(DVBSCAN_EPG_STATUS_CURRENT);
			if (!dvbscan_start_filter(instance, 0x12, 0x4E)) {
				instance->status = DVBSCAN_EPG_STATUS_NULL;
				if (instance->callback != NULL) (instance->callback)(DVBSCAN_EPG_STATUS_NULL);
				return NULL;
			}
		} else if (attempt == 1) {
			instance->status = DVBSCAN_EPG_STATUS_FULL;
			if (instance->callback != NULL) (instance->callback)(DVBSCAN_EPG_STATUS_FULL);
			if (!dvbscan_start_filter(instance, 0x12, 0x50)) {
				instance->status = DVBSCAN_EPG_STATUS_NULL;
				if (instance->callback != NULL) (instance->callback)(DVBSCAN_EPG_STATUS_NULL);
				return NULL;
			}
		}
		
		allevents = 0;
		readevents = 0;
		repevents = 0;
		emptyevents = 0;
		
		while (((attempt == 1) && (allevents <= 300) && (repevents < 25)) || ((attempt == 0) && (allevents <= 100) && (repevents < 5) && (emptyevents < 5))) {
			readb = dvbscan_read_filter(instance, (gchar *)&buffer, sizeof(buffer), DVBSCAN_EIT_READ_TIMEOUT);
			if (readb > 0) {
				table_id = buffer[0];
				section_length = (((buffer[1] & 0x0f) << 8) | buffer[2]);
				service_id = (((buffer[3] & 0xff) << 8) | buffer[4]);
				version_number = (buffer[5] & 0x3e);
				section_number = buffer[6];
				last_section_number = buffer[7];
				transport_stream_id = (((buffer[8] & 0xff) << 8) | buffer[9]);
				original_network_id = (((buffer[10] & 0xff) << 8) | buffer[11]);
				segment_last_section_number = buffer[12];
				last_table_id = buffer[13];
				ptr = 14;
				
				allevents++;
				
				//printf("PACKET SID, LEN: %i, %i\n", service_id, section_length);
			
				if ((service_id == instance->sid) && (section_length > 15)) {
					while (ptr < section_length - 4) {
						event_id = (((buffer[ptr] & 0xff) << 8) | buffer[ptr+1]);
						MJD_time = (((buffer[ptr+2] & 0xff) << 8) | buffer[ptr+3]);
						UTC_time = (((buffer[ptr+4] & 0xff) << 16) | ((buffer[ptr+5] & 0xff) << 8) | buffer[ptr+6]);
						UTC_time_duration  = (((buffer[ptr+7] & 0xff) << 16) | ((buffer[ptr+8] & 0xff) << 8) | buffer[ptr+9]);
						running_status = (buffer[ptr+10] & 0xe0);
						free_CA_mode = (buffer[ptr+10] & 0x10);
						descriptors_loop_length = (((buffer[ptr+10] & 0x0f) << 8) | buffer[ptr+11]);
						ptr += 12;
						
						if (g_hash_table_lookup(eventhash, &event_id) == NULL) {
							event = (struct _scanevent *)g_malloc0(sizeof(struct _scanevent));
							event->id = event_id;
							event->start = dvbscan_decode_time(MJD_time, UTC_time);
							event->duration = dvbscan_decode_duration(UTC_time_duration);
							event->name = g_strdup("Undefined event");
							event->description = NULL;
							if (attempt == 0) {
								event->current = TRUE;
							} else {
								event->current = FALSE;
							}
							
							g_hash_table_insert(eventhash, &event_id, event);
							
							instance->eventlist = g_slist_prepend(instance->eventlist, event);
							
							readevents++;
							
							printf("EVENT ID: %i\n", event->id);
							printf("TIME %s", ctime(&event->start));
							printf("DURATION %4.0f\n", event->duration);
							
							if (descriptors_loop_length > 0) {
								descriptorptr = 0;
								while (descriptorptr < descriptors_loop_length) {
									descriptor_tag = buffer[ptr+descriptorptr];
									descriptor_length = buffer[ptr+descriptorptr+1];
									if (descriptor_tag == 0x50) {
										printf("COMPONENT DESCRIPTOR\n");
										stream_content = (buffer[ptr+descriptorptr+2] & 0x0f);
										component_type = buffer[ptr+descriptorptr+3];
										component_tag = buffer[ptr+descriptorptr+4];
										memset(event->language, 0, sizeof(event->language));
										strncpy(event->language, &buffer[ptr+descriptorptr+5], 3);
										printf(" Content: 0x%x, type: 0x%x (%s)\n", stream_content, component_type, event->language);
										if (descriptor_length > 6) {
											memset(string, 0, sizeof(string));
											strncpy(string, &buffer[ptr+descriptorptr+8], descriptor_length - 6);
											printf(" Text description: %s\n", string);
										}
									} else if (descriptor_tag == 0x4D) {
										printf("SHORT DESCRIPTOR\n");
										memset(event->language, 0, sizeof(event->language));
										strncpy(event->language, &buffer[ptr+descriptorptr+2], 3);
										printf(" Languge: %s\n", event->language);
										event_name_length = buffer[ptr+descriptorptr+5];
										if (event_name_length > 0) {
											memset(string, 0, sizeof(string));
											strncpy(string, &buffer[ptr+descriptorptr+6], event_name_length);
											if (event->name != NULL) g_free(event->name);
											event->name = dvbscan_convert_encoding((gchar *)&string);
											printf(" Event: %s\n", event->name);
										}
										text_length = buffer[ptr+descriptorptr+event_name_length+6];
										if (text_length > 0) {
											memset(string, 0, sizeof(string));
											strncpy(string, &buffer[ptr+descriptorptr+event_name_length+7], text_length);
											if (event->description != NULL) g_free(event->description);
											event->description = dvbscan_convert_encoding((gchar *)&string);
											printf(" Description: %s\n", event->description);
										}
									} else if (descriptor_tag == 0x4E) {
										printf("EXTENDED DESCRIPTOR\n");
										descriptor_number = (buffer[ptr+descriptorptr+2] >> 4) & 0x0F;
										last_descriptor_number = (buffer[ptr+descriptorptr+2]) & 0x0F;
										memset(event->language, 0, sizeof(event->language));
										strncpy(event->language, &buffer[ptr+descriptorptr+3], 3);
										printf(" Languge: %s\n", event->language);
										length_of_items = buffer[ptr+descriptorptr+6];
										event_name_length = buffer[ptr+descriptorptr+6+length_of_items];
										if (event_name_length > 0) {
											memset(string, 0, sizeof(string));
											strncpy(string, &buffer[ptr+descriptorptr+7+length_of_items], event_name_length);
											if (event->name != NULL) g_free(event->name);
											event->name = dvbscan_convert_encoding((gchar *)&string);
											printf(" Event: %s\n", string);
										}
									} else if (descriptor_tag == 0x54) {
										printf("CONTENT DESCRIPTOR (%i)\n", descriptor_length/2);
										for (di=ptr+descriptorptr+2; di<ptr+descriptorptr+2+descriptor_length; di=di+2) {
											content_nibble_level_1 = (buffer[di] >> 4) & 0x0F;
											content_nibble_level_2 = (buffer[di]) & 0x0F;
											user_nibble_1 = (buffer[di+1] >> 4) & 0x0F;
											user_nibble_2 = (buffer[di+1]) & 0x0F;
											printf(" Content cn1: 0x%1x, cn2: 0x%1x, un1: 0x%1x, un2: 0x%1x\n", content_nibble_level_1, content_nibble_level_2, user_nibble_1, user_nibble_2);
										}
									} else if (descriptor_tag == 0x55) {
										printf("PARENTAL RAITING DESCRIPTOR (%i)\n", descriptor_length/4);
										for (di=ptr+descriptorptr+2; di<ptr+descriptorptr+2+descriptor_length; di=di+4) {
											memset(language, 0, sizeof(language));
											strncpy(language, &buffer[di], 3);
											raiting = buffer[di+3];
											printf(" Country code: %s, raiting: 0x%1x\n", language, raiting);
										}
									} else {
										printf("Unknown Descriptor: 0x%2x\n", descriptor_tag);
									}
									descriptorptr += descriptor_length + 2;
								}
							}
						} else {
							repevents++;
						}
						ptr += descriptors_loop_length;
					}
				} else if (section_length <= 15) {
					emptyevents++;
				}
			} else if ((readb == 0) && (instance->command == DVBSCAN_COMMAND_STOP)) {
				printf("[-] EPG read process stopped by user\n");
				instance->status = DVBSCAN_EPG_STATUS_NULL;
				if (instance->callback != NULL) (instance->callback)(DVBSCAN_EPG_STATUS_NULL);
				return NULL;
			} else {
				g_warning("Timeout while reading event info");
				break;
			}
		}
		dvbscan_stop_filter(instance);
	}
	
	g_hash_table_destroy(eventhash);
	
	if (instance->eventlist != NULL) instance->eventlist = g_slist_sort(instance->eventlist, dvbscan_sort_event_list);
	
	instance->status = DVBSCAN_EPG_STATUS_NULL;
	if (instance->callback != NULL) (instance->callback)(DVBSCAN_EPG_STATUS_NULL);
		
	return NULL;
}
