/*
 *      fswindow.c
 *      
 *      Copyright 2009 - 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <cairo.h>
#include <glib/gi18n.h>

#include "fswindow.h"
#include "dbstorage.h"
#include "settings.h"
#include "gsfcore.h"

typedef struct {
	guint id;
	gdouble r;
	gdouble g;
	gdouble b;
} color_table;

color_table colortable[] = {
	{FSWIN_COLOR_BACKGROUND, 0.2, 0.5, 0.8},
	{FSWIN_COLOR_TEXT,       0.1, 0.1, 0.1},
	{FSWIN_COLOR_HILEVEL,    0.0, 1.0, 0.0},
	{FSWIN_COLOR_MIDLEVEL,   1.0, 1.0, 0.0},
	{FSWIN_COLOR_LOWLEVEL,   1.0, 0.0, 0.5},
	{FSWIN_COLOR_NULL,       0.0, 0.0, 0.0}
};


struct _fswindow {
	gboolean visible;
	GtkWidget *win;
	gchar *title;
	GtkWidget *darea;
	//Cached values
	gdouble snrlevel;
	gdouble agclevel;
	gdouble berlevel;
	guint snr;
	guint agc;
	guint ber;
	gboolean lock;
	gboolean carrier;
	gchar svcname[256];
	gchar svcparams[256];
};

struct _fswindow fswindow;

#if GTK_CHECK_VERSION(3, 0, 0)
static gboolean fswindow_on_draw_event(GtkWidget *widget, cairo_t *cr);
#else
static gboolean fswindow_on_expose_event(GtkWidget *widget, GdkEventExpose *event, gpointer data);
#endif
static gdouble fswindow_calculate_fontsize(gint height, gint border, gdouble scale);
static gint fswindow_get_color_id(gdouble value);
gboolean fswindow_set_color(guint id, gdouble r, gdouble g, gdouble b);
gboolean fswindow_get_color(guint id, gdouble *r, gdouble *g, gdouble *b);


#if GTK_CHECK_VERSION(3, 0, 0)
static gboolean fswindow_on_draw_event(GtkWidget *widget, cairo_t *cr)
{
	gint winheight, winwidth;
	gdouble r, g, b;
	
	winwidth = gtk_widget_get_allocated_width(widget);
	winheight = gtk_widget_get_allocated_height(widget);
		
	fswindow_get_color(FSWIN_COLOR_BACKGROUND, &r, &g, &b);
	cairo_set_source_rgb(cr, r, g, b);
	cairo_rectangle(cr, 0, 0, winwidth, winheight);
	cairo_fill(cr);
	
	fswindow_get_color(FSWIN_COLOR_TEXT, &r, &g, &b);
	cairo_set_source_rgb(cr, r, g, b);
	
	cairo_rectangle(cr, winwidth*0.025, winheight*0.1, winwidth-winwidth*0.05, winheight*0.15);
	cairo_set_line_width(cr, 3);
	cairo_set_line_join(cr, CAIRO_LINE_JOIN_MITER); 
	cairo_stroke(cr);
	
	cairo_rectangle(cr, winwidth*0.025, winheight*0.35, winwidth-winwidth*0.05, winheight*0.15);
	cairo_set_line_width(cr, 3);
	cairo_set_line_join(cr, CAIRO_LINE_JOIN_MITER); 
	cairo_stroke(cr);
	
	cairo_rectangle(cr, winwidth*0.025, winheight*0.6, winwidth-winwidth*0.05, winheight*0.15);
	cairo_set_line_width(cr, 3);
	cairo_set_line_join(cr, CAIRO_LINE_JOIN_MITER); 
	cairo_stroke(cr);
	
	fswindow_display_signal_status(NULL);
	
	fswindow_display_infobar(NULL);
	
	return FALSE;
}
#else
static gboolean fswindow_on_expose_event(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
	GdkWindow *window;
	cairo_t *cr;
	gint winheight, winwidth;
	gdouble r, g, b;
	
	window = gtk_widget_get_window(widget);
	
	#if GTK_CHECK_VERSION(2, 24, 0)
		winwidth = gdk_window_get_width(window);
		winheight = gdk_window_get_height(window);
	#else
		gdk_drawable_get_size(GDK_DRAWABLE(window), &winwidth, &winheight);
	#endif
		
	cr = gdk_cairo_create(window);
	
	fswindow_get_color(FSWIN_COLOR_BACKGROUND, &r, &g, &b);
	cairo_set_source_rgb(cr, r, g, b);
	cairo_rectangle(cr, 0, 0, winwidth, winheight);
	cairo_fill(cr);
	
	fswindow_get_color(FSWIN_COLOR_TEXT, &r, &g, &b);
	cairo_set_source_rgb(cr, r, g, b);
	
	cairo_rectangle(cr, winwidth*0.025, winheight*0.1, winwidth-winwidth*0.05, winheight*0.15);
	cairo_set_line_width(cr, 3);
	cairo_set_line_join(cr, CAIRO_LINE_JOIN_MITER); 
	cairo_stroke(cr);
	
	cairo_rectangle(cr, winwidth*0.025, winheight*0.35, winwidth-winwidth*0.05, winheight*0.15);
	cairo_set_line_width(cr, 3);
	cairo_set_line_join(cr, CAIRO_LINE_JOIN_MITER); 
	cairo_stroke(cr);
	
	cairo_rectangle(cr, winwidth*0.025, winheight*0.6, winwidth-winwidth*0.05, winheight*0.15);
	cairo_set_line_width(cr, 3);
	cairo_set_line_join(cr, CAIRO_LINE_JOIN_MITER); 
	cairo_stroke(cr);
	
	cairo_destroy(cr);
	
	fswindow_display_signal_status(NULL);
	
	fswindow_display_infobar(NULL);
	
	return FALSE;
}
#endif

static gdouble fswindow_calculate_fontsize(gint height, gint border, gdouble scale)
{
	GdkScreen *screen;
	gdouble dpi;
	gdouble fontsize;
	
	screen = gdk_screen_get_default();
	
	dpi = gdk_screen_get_resolution(screen);
	
	fontsize = ((height*scale)-border)*72.0/dpi;
	
	return fontsize;
}

static gint fswindow_get_color_id(gdouble value)
{
	gint colorid;
	
	if (value <= 0.4) {
		colorid = FSWIN_COLOR_LOWLEVEL;
	} else if (value <= 0.6) {
		colorid = FSWIN_COLOR_MIDLEVEL;
	} else {
		colorid = FSWIN_COLOR_HILEVEL;
	}
	
	return colorid;
}

void fswindow_display_signal_status(struct _frontend *frontend)
{
	GdkWindow *window;
	cairo_t *cr;
	gint winheight, winwidth;
	gint colorid;
	gdouble r, g, b;
	gdouble fontsize;
	cairo_text_extents_t textextents;
	gchar strbuf[256];
	
	if (frontend != NULL) {
		if ((fswindow.snrlevel = (gdouble)(frontend->reset ? 0 : frontend->snr)/settings_read_float("scale_snr", 100.0)) > 1.0) fswindow.snrlevel = 1.0;
		if ((fswindow.agclevel = (gdouble)((frontend->signal)/settings_read_float("scale_agc", 1000.0))) > 1.0) fswindow.agclevel = 1.0;
		if ((fswindow.berlevel = (gdouble)((frontend->ber)/settings_read_float("scale_ber", 65535.0))) > 1.0) fswindow.berlevel = 1.0;
		fswindow.snr = frontend->snr;
		fswindow.agc = frontend->signal;
		fswindow.ber = frontend->ber;
		fswindow.lock = frontend->lock;
		fswindow.carrier = frontend->carrier;
	}
	
	if (!fswindow.visible) return;
	
	window = gtk_widget_get_window(fswindow.darea);
	
	#if GTK_CHECK_VERSION(2, 24, 0)
		winwidth = gdk_window_get_width(window);
		winheight = gdk_window_get_height(window);
	#else
		gdk_drawable_get_size(GDK_DRAWABLE(window), &winwidth, &winheight);
	#endif
	
	cr = gdk_cairo_create(window);
	
	fswindow_get_color(FSWIN_COLOR_BACKGROUND, &r, &g, &b);
	cairo_set_source_rgb(cr, r, g, b);
	cairo_rectangle(cr, (winwidth*0.025)+3, (winheight*0.10)+3, (winwidth-((winwidth*0.05)+3+3)), (winheight*0.15)-3-3);
	cairo_fill(cr);
	cairo_rectangle(cr, (winwidth*0.025)+3, (winheight*0.35)+3, (winwidth-((winwidth*0.05)+3+3)), (winheight*0.15)-3-3);
	cairo_fill(cr);
	cairo_rectangle(cr, (winwidth*0.025)+3, (winheight*0.60)+3, (winwidth-((winwidth*0.05)+3+3)), (winheight*0.15)-3-3);
	cairo_fill(cr);
	
	colorid = fswindow_get_color_id(fswindow.snrlevel);
	fswindow_get_color(colorid, &r, &g, &b);
	cairo_set_source_rgb(cr, r, g, b);
	cairo_rectangle(cr, (winwidth*0.025)+3, (winheight*0.10)+3, (winwidth-((winwidth*0.05)+3+3))*fswindow.snrlevel, (winheight*0.15)-3-3);
	cairo_fill(cr);
	
	colorid = fswindow_get_color_id(fswindow.agclevel);
	fswindow_get_color(colorid, &r, &g, &b);
	cairo_set_source_rgb(cr, r, g, b);
	cairo_rectangle(cr, (winwidth*0.025)+3, (winheight*0.35)+3, (winwidth-((winwidth*0.05)+3+3))*fswindow.agclevel, (winheight*0.15)-3-3);
	cairo_fill(cr);
	
	colorid = fswindow_get_color_id(fswindow.berlevel);
	fswindow_get_color(colorid, &r, &g, &b);
	cairo_set_source_rgb(cr, r, g, b);
	cairo_rectangle(cr, (winwidth*0.025)+3, (winheight*0.60)+3, (winwidth-((winwidth*0.05)+3+3))*fswindow.berlevel, (winheight*0.15)-3-3);
	cairo_fill(cr);
	
	fswindow_get_color(FSWIN_COLOR_TEXT, &r, &g, &b);
	cairo_set_source_rgb(cr, r, g, b);
	
	cairo_select_font_face(cr, "Liberation Sans", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_BOLD);
	
	fontsize = fswindow_calculate_fontsize(winheight*0.15, 3+3, 0.80);
	
	cairo_set_font_size(cr, fontsize);
	
	g_snprintf((gchar *)&strbuf, sizeof(strbuf), fswindow.lock ? "%u LOCK" : "%u", fswindow.snr);
	cairo_text_extents(cr, (gchar *)&strbuf, &textextents);
	cairo_move_to(cr, (winwidth+3-textextents.width)/2, ((winheight*0.25)-3)-((((winheight*0.15)-3-3)-textextents.height)/2));
	cairo_show_text(cr, (gchar *)&strbuf);
	
	g_snprintf((gchar *)&strbuf, sizeof(strbuf), fswindow.carrier ? "%u CARRIER" : "%u", fswindow.agc);
	cairo_text_extents(cr, (gchar *)&strbuf, &textextents);
	cairo_move_to(cr, (winwidth+3-textextents.width)/2, ((winheight*0.50)-3)-((((winheight*0.15)-3-3)-textextents.height)/2));
	cairo_show_text(cr, (gchar *)&strbuf);
		
	g_snprintf((gchar *)&strbuf, sizeof(strbuf), "%u", fswindow.ber);
	cairo_text_extents(cr, (gchar *)&strbuf, &textextents);
	cairo_move_to(cr, (winwidth+3-textextents.width)/2, ((winheight*0.75)-3)-((((winheight*0.15)-3-3)-textextents.height)/2));
	cairo_show_text(cr, (gchar *)&strbuf);
		
	cairo_destroy(cr);
}

void fswindow_display_infobar(struct _service *svc)
{
	GdkWindow *window;
	cairo_t *cr;
	gint winheight, winwidth;
	gdouble r, g, b;
	gdouble fontsize;
	cairo_text_extents_t textextents;
		
	if (svc != NULL) {
		g_snprintf((gchar *)&fswindow.svcname, sizeof(fswindow.svcname), "%s", svc->name);
		if (svc->type == SERVICE_NETSHARE) {
			g_snprintf((gchar *)&fswindow.svcparams, sizeof(fswindow.svcparams), "%s", svc->node);
		} else {
			g_snprintf((gchar *)&fswindow.svcparams, sizeof(fswindow.svcparams), "%u %s %u %s %s %s", svc->freq, gsfcore_string_definition(DEFINITION_POLARISATION, svc->pol), svc->sr,
						gsfcore_string_definition(DEFINITION_FEC, svc->fec), gsfcore_string_definition(DEFINITION_SYSTEM, svc->sys), gsfcore_string_definition(DEFINITION_MODULATION, svc->mod));
		}
	}
	
	if (!fswindow.visible) return;
	
	window = gtk_widget_get_window(fswindow.darea);
	
	#if GTK_CHECK_VERSION(2, 24, 0)
		winwidth = gdk_window_get_width(window);
		winheight = gdk_window_get_height(window);
	#else
		gdk_drawable_get_size(GDK_DRAWABLE(window), &winwidth, &winheight);
	#endif
	
	cr = gdk_cairo_create(window);
	
	fswindow_get_color(FSWIN_COLOR_BACKGROUND, &r, &g, &b);
	cairo_set_source_rgb(cr, r, g, b);
	cairo_rectangle(cr, (winwidth*0.025)+3, (winheight*0.75)+3, (winwidth-((winwidth*0.05)+3+3)), (winheight*0.10)-3-3);
	cairo_fill(cr);
	cairo_set_source_rgb(cr, r, g, b);
	cairo_rectangle(cr, (winwidth*0.025)+3, (winheight*0.85)+3, (winwidth-((winwidth*0.05)+3+3)), (winheight*0.10)-3-3);
	cairo_fill(cr);
	
	fswindow_get_color(FSWIN_COLOR_TEXT, &r, &g, &b);
	cairo_set_source_rgb(cr, r, g, b);
	
	cairo_select_font_face(cr, "Liberation Sans", CAIRO_FONT_SLANT_ITALIC, CAIRO_FONT_WEIGHT_NORMAL);
	
	fontsize = fswindow_calculate_fontsize(winheight*0.10, 3+3, 0.80);
		
	cairo_set_font_size(cr, fontsize);
	
	cairo_text_extents(cr, fswindow.svcname, &textextents);
	cairo_move_to(cr, (winwidth+3-textextents.width)/2, ((winheight*0.85)-3)-((((winheight*0.10)-3-3)-textextents.height)/2));
	cairo_show_text(cr, fswindow.svcname);
	
	cairo_text_extents(cr, fswindow.svcparams, &textextents);
	cairo_move_to(cr, (winwidth+3-textextents.width)/2, ((winheight*0.95)-3)-((((winheight*0.10)-3-3)-textextents.height)/2));
	cairo_show_text(cr, fswindow.svcparams);
	
	cairo_destroy(cr);
}

gboolean fswindow_key_event(GtkWidget *widget, GdkEventKey *event, gpointer data)
{
	if (event != NULL) {
		#if GTK_CHECK_VERSION(2, 22, 0)
			if ((event->keyval == GDK_KEY_F11) || (event->keyval == GDK_KEY_Escape)) {
		#else
			if ((event->keyval == GDK_F11) || (event->keyval == GDK_Escape)) {
		#endif
				g_signal_emit_by_name(widget, "destroy", NULL);
			}
	}
}

gboolean fswindow_set_color(guint id, gdouble r, gdouble g, gdouble b)
{
	guint i = 0;
	
	while (colortable[i].id != FSWIN_COLOR_NULL) {
		if (colortable[i].id == id) {
			colortable[i].r = r;
			colortable[i].g = g;
			colortable[i].b = b;
			
			if ((fswindow.darea != NULL) && (GTK_IS_WIDGET(fswindow.darea))) {
				g_signal_emit_by_name(fswindow.darea, "expose-event", NULL);
			}
			
			return TRUE;
		}
		i++;
	}
	
	return FALSE;
}

gboolean fswindow_get_color(guint id, gdouble *r, gdouble *g, gdouble *b)
{
	guint i = 0;
	
	while (colortable[i].id != FSWIN_COLOR_NULL) {
		if (colortable[i].id == id) {
			*r = colortable[i].r;
			*g = colortable[i].g;
			*b = colortable[i].b;
			
			return TRUE;
		}
		i++;
	}
	
	*r = 1.0;
	*g = 1.0;
	*b = 1.0;
	
	return FALSE;
}

gint fswindow_close(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	gtk_widget_destroy(fswindow.win);
	fswindow.darea = NULL;
	fswindow.visible = FALSE;
	
	return FALSE;
}

gboolean fswindow_init(void)
{
	GdkColor color;
	gchar *tmpstr;
	gdouble r, g, b;
	
	tmpstr = settings_read_string("fullscreen_low_color", "#ffff00007fff");
	gdk_color_parse(tmpstr, &color);
	r = (gdouble)color.red/65535.0;
	g = (gdouble)color.green/65535.0;
	b = (gdouble)color.blue/65535.0;
	fswindow_set_color(FSWIN_COLOR_LOWLEVEL, r, g, b);
	g_free(tmpstr);
	
	tmpstr = settings_read_string("fullscreen_middle_color", "#ffffffff0000");
	gdk_color_parse(tmpstr, &color);
	r = (gdouble)color.red/65535.0;
	g = (gdouble)color.green/65535.0;
	b = (gdouble)color.blue/65535.0;
	fswindow_set_color(FSWIN_COLOR_MIDLEVEL, r, g, b);
	g_free(tmpstr);
	
	tmpstr = settings_read_string("fullscreen_high_color", "#0000ffff0000");
	gdk_color_parse(tmpstr, &color);
	r = (gdouble)color.red/65535.0;
	g = (gdouble)color.green/65535.0;
	b = (gdouble)color.blue/65535.0;
	fswindow_set_color(FSWIN_COLOR_HILEVEL, r, g, b);
	g_free(tmpstr);
	
	tmpstr = settings_read_string("fullscreen_background_color", "#33337fffcccc");
	gdk_color_parse(tmpstr, &color);
	r = (gdouble)color.red/65535.0;
	g = (gdouble)color.green/65535.0;
	b = (gdouble)color.blue/65535.0;
	fswindow_set_color(FSWIN_COLOR_BACKGROUND, r, g, b);
	g_free(tmpstr);
	
	tmpstr = settings_read_string("fullscreen_text_color", "#199919991999");
	gdk_color_parse(tmpstr, &color);
	r = (gdouble)color.red/65535.0;
	g = (gdouble)color.green/65535.0;
	b = (gdouble)color.blue/65535.0;
	fswindow_set_color(FSWIN_COLOR_TEXT, r, g, b);
	g_free(tmpstr);
	
	return TRUE;
}

gboolean fswindow_create(void)
{
	if (!fswindow.visible) {
		fswindow.win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
		
		fswindow.darea = gtk_drawing_area_new();
		gtk_container_add(GTK_CONTAINER(fswindow.win), fswindow.darea);
		
		#if GTK_CHECK_VERSION(3, 0, 0)
			g_signal_connect(fswindow.darea, "draw", G_CALLBACK(fswindow_on_draw_event), NULL);
		#else
			g_signal_connect(fswindow.darea, "expose-event", G_CALLBACK(fswindow_on_expose_event), NULL);
		#endif
		
		g_signal_connect(fswindow.win, "key-press-event", G_CALLBACK(fswindow_key_event), NULL);
		g_signal_connect(fswindow.win, "destroy", G_CALLBACK(fswindow_close), NULL);
		
		gtk_window_set_position(GTK_WINDOW(fswindow.win), GTK_WIN_POS_CENTER);
		gtk_window_set_default_size(GTK_WINDOW(fswindow.win), 300, 280);
		
		gtk_window_fullscreen(GTK_WINDOW(fswindow.win));
		gtk_widget_show_all(fswindow.win);
			
		fswindow.visible = TRUE;
	} else {
		gtk_window_present(GTK_WINDOW(fswindow.win));
	}
	
	return TRUE;
}
