/*
 *      grstatus.c
 *      
 *      Copyright 2010-2012 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#include "grstatus.h"

#define GTK_GRSTATUS_HEIGHT 20
#define GTK_GRSTATUS_WIDTH  100

#define GTK_GRSTATUS_STYLE ".frame {\n   border-style: solid;\n   border-color: rgb(%i,%i,%i);\n   border-width: 1;\n   border-radius: 2;\n}\n"

struct _GtkGrstatusPrivate {
	gchar *values;
	gint valsize;
	gint ptr;
	GdkColor mincolor;
	GdkColor maxcolor;
	#if GTK_CHECK_VERSION(3, 0, 0)
		GtkStyleProvider *provider;
	#endif
};

static void gtk_grstatus_class_init(GtkGrstatusClass *klass);
static void gtk_grstatus_init(GtkGrstatus *grstatus);
#if GTK_CHECK_VERSION(3, 0, 0)
static void gtk_grstatus_get_preferred_width(GtkWidget *widget, gint *minimum_size, gint *natural_size);
static void gtk_grstatus_get_preferred_height(GtkWidget *widget, gint *minimum_size, gint *natural_size);
#else
static void gtk_grstatus_size_request(GtkWidget *widget, GtkRequisition *requisition);
#endif
#if !GTK_CHECK_VERSION(3, 0, 0)
static gboolean gtk_grstatus_expose(GtkWidget *widget, GdkEventExpose *event);
#endif
static gboolean gtk_grstatus_draw(GtkWidget *widget, cairo_t *cr);
#if GTK_CHECK_VERSION(3, 0, 0)
static void gtk_grstatus_destroy(GtkWidget *widget);
#else
static void gtk_grstatus_destroy(GtkObject *object);
#endif

static gint vsize;

G_DEFINE_TYPE (GtkGrstatus, gtk_grstatus, GTK_TYPE_WIDGET)

static void gtk_grstatus_class_init(GtkGrstatusClass *klass)
{
	GtkWidgetClass *widget_class;
	GObjectClass *object_class;
	#if !GTK_CHECK_VERSION(3, 0, 0)
		GtkObjectClass *gtk_object_class;
	#endif
	
	widget_class = GTK_WIDGET_CLASS(klass);
	#if GTK_CHECK_VERSION(3, 0, 0)
		widget_class->get_preferred_width = gtk_grstatus_get_preferred_width;
		widget_class->get_preferred_height = gtk_grstatus_get_preferred_height;
		widget_class->draw = gtk_grstatus_draw;
		widget_class->destroy = gtk_grstatus_destroy;
	#else
		widget_class->size_request = gtk_grstatus_size_request;
		widget_class->expose_event = gtk_grstatus_expose;
	#endif
	
	object_class = G_OBJECT_CLASS(klass);
	g_type_class_add_private(object_class, sizeof(GtkGrstatusPrivate));
	
	#if !GTK_CHECK_VERSION(3, 0, 0)
		gtk_object_class = GTK_OBJECT_CLASS(klass);
		gtk_object_class->destroy = gtk_grstatus_destroy;
	#endif
}

static void gtk_grstatus_init(GtkGrstatus *grstatus)
{
	GtkGrstatusPrivate *priv;
	#if GTK_CHECK_VERSION(3, 0, 0)
		GtkStyleContext *context;
		gchar *stylebuf;
		GdkRGBA color;
	#endif
		
	priv = G_TYPE_INSTANCE_GET_PRIVATE(grstatus, GTK_TYPE_GRSTATUS, GtkGrstatusPrivate);
	grstatus->priv = priv;
	
	priv->valsize = vsize;
	priv->values = (gchar *)g_malloc0(priv->valsize);
	priv->ptr = 0;
	gdk_color_parse("white", &priv->mincolor);
	gdk_color_parse("green3", &priv->maxcolor);
	
	#if GTK_CHECK_VERSION(3, 0, 0)
		context = gtk_widget_get_style_context(GTK_WIDGET(grstatus));
		gtk_style_context_get_border_color(context, GTK_STATE_FLAG_NORMAL, &color);
		stylebuf = g_strdup_printf(GTK_GRSTATUS_STYLE, color.red, color.green, color.blue);
		priv->provider = (GtkStyleProvider *)gtk_css_provider_new();
		gtk_css_provider_load_from_data(GTK_CSS_PROVIDER(priv->provider), stylebuf, -1, NULL);
		gtk_style_context_add_provider(context, priv->provider, GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);
		gtk_style_context_add_class(context, GTK_STYLE_CLASS_FRAME);
		g_free(stylebuf);
	#endif
	
	#if GTK_CHECK_VERSION(2, 18, 0)
		gtk_widget_set_has_window(GTK_WIDGET(grstatus), FALSE);
	#else
		GTK_WIDGET_SET_FLAGS(GTK_WIDGET(grstatus), GTK_NO_WINDOW);
	#endif
}

#if GTK_CHECK_VERSION(3, 0, 0)
static void gtk_grstatus_get_preferred_width(GtkWidget *widget, gint *minimum_size, gint *natural_size)
{
	GtkGrstatusPrivate *priv;
	
	priv = GTK_GRSTATUS(widget)->priv;
	
	if (minimum_size)
		*minimum_size = priv->valsize;
	
	if (natural_size)
		*natural_size = priv->valsize;
}

static void gtk_grstatus_get_preferred_height(GtkWidget *widget, gint *minimum_size, gint *natural_size)
{
	if (minimum_size)
		*minimum_size = GTK_GRSTATUS_HEIGHT;
	
	if (natural_size)
		*natural_size = GTK_GRSTATUS_HEIGHT;
}
#else
static void gtk_grstatus_size_request(GtkWidget *widget, GtkRequisition *requisition)
{
	GtkGrstatusPrivate *priv;
	
	priv = GTK_GRSTATUS(widget)->priv;
	
	requisition->width = priv->valsize;
	requisition->height = 20;
}
#endif

void gtk_grstatus_set_value(GtkGrstatus *grstatus, gchar value)
{
	GtkGrstatusPrivate *priv;
	
	priv = grstatus->priv;
	
	priv->values[priv->ptr] = value;
	
	if (priv->ptr+1 > priv->valsize-2) priv->ptr = 0; else priv->ptr++;
	
	gtk_widget_queue_draw(GTK_WIDGET(grstatus));
}

void gtk_grstatus_set_colors(GtkGrstatus *grstatus, GdkColor *mincolor, GdkColor *maxcolor)
{
	GtkGrstatusPrivate *priv;
	
	priv = grstatus->priv;
	
	priv->mincolor = *mincolor;
	priv->maxcolor = *maxcolor;
	
	gtk_widget_queue_draw(GTK_WIDGET(grstatus));
}

GtkWidget *gtk_grstatus_new(gint valsize)
{
	if (valsize <= 0) {
		vsize = GTK_GRSTATUS_WIDTH;
	} else {
		vsize = valsize;
	}
	
	return g_object_new(GTK_TYPE_GRSTATUS, NULL);
}

#if !GTK_CHECK_VERSION(3, 0, 0)
static gboolean gtk_grstatus_expose(GtkWidget *widget, GdkEventExpose *event)
{
	cairo_t *cr;
	
	cr = gdk_cairo_create(widget->window);
	
	gdk_cairo_rectangle(cr, &(event->area));
	
	cairo_clip(cr);
	
	cairo_translate(cr, event->area.x, event->area.y);
	
	gtk_grstatus_draw(widget, cr);
	
	gtk_paint_shadow(widget->style, widget->window, GTK_STATE_NORMAL, GTK_SHADOW_IN,
						&(event->area), widget, "", event->area.x, event->area.y,
						event->area.width, event->area.height);
	
	cairo_destroy(cr);
	
	return FALSE;
}
#endif

static gboolean gtk_grstatus_draw(GtkWidget *widget, cairo_t *cr)
{
	GtkGrstatusPrivate *priv;
	gint i, locptr;
	gdouble peakvalue;
	gdouble minr, ming, minb;
	gdouble slr, slg, slb;
	#if GTK_CHECK_VERSION(3, 0, 0)
		GtkStyleContext *context;
	#endif
		
	priv = GTK_GRSTATUS(widget)->priv;
	
	cairo_set_source_rgb(cr, priv->mincolor.red, priv->mincolor.green, priv->mincolor.blue);
	cairo_paint(cr);
	
	cairo_set_line_width(cr, 1);
		
	minr = priv->mincolor.red/65535.0;
	slr = (priv->maxcolor.red - priv->mincolor.red)/65535.0;
	ming = priv->mincolor.green/65535.0;
	slg = (priv->maxcolor.green - priv->mincolor.green)/65535.0;
	minb = priv->mincolor.blue/65535.0;
	slb = (priv->maxcolor.blue - priv->mincolor.blue)/65535.0;
	
	locptr = priv->ptr; 
	
	for (i=0; i<=priv->valsize-2; i++) {
		peakvalue = (priv->values[locptr])/(100.0/(gfloat)(GTK_GRSTATUS_HEIGHT-2));
		cairo_set_source_rgb(cr, minr+(slr*peakvalue/(gfloat)(GTK_GRSTATUS_HEIGHT-2)), 
								ming+(slg*peakvalue/(gfloat)(GTK_GRSTATUS_HEIGHT-2)),
								minb+(slb*peakvalue/(gfloat)(GTK_GRSTATUS_HEIGHT-2)));
		cairo_move_to(cr, (gdouble)(i+1), GTK_GRSTATUS_HEIGHT-1);
		cairo_line_to(cr, (gdouble)(i+1), GTK_GRSTATUS_HEIGHT-1-peakvalue);
		cairo_stroke(cr);
		if (locptr+1 > priv->valsize-2) locptr = 0; else locptr++;
	}
	
	#if GTK_CHECK_VERSION(3, 0, 0)
		context = gtk_widget_get_style_context(widget);
		gtk_render_frame(context, cr, 0, 0, priv->valsize, GTK_GRSTATUS_HEIGHT);
	#endif
	
	return FALSE;
}

#if GTK_CHECK_VERSION(3, 0, 0)
static void gtk_grstatus_destroy(GtkWidget *widget)
{
	GtkGrstatus *grstatus = GTK_GRSTATUS(widget);
	GtkGrstatusPrivate *priv = grstatus->priv;
	GtkStyleContext *context;
	
	if (priv->values != NULL) {
		g_free(priv->values);
		priv->values = NULL;
	}
	
	context = gtk_widget_get_style_context(widget);
	gtk_style_context_remove_provider(context, priv->provider);
	gtk_style_context_remove_class(context, GTK_STYLE_CLASS_FRAME);
	
	GTK_WIDGET_CLASS(gtk_grstatus_parent_class)->destroy (widget);
}
#else
static void gtk_grstatus_destroy(GtkObject *object)
{
	GtkGrstatus *grstatus;
	GtkGrstatusClass *klass;
	GtkGrstatusPrivate *priv;
	
	grstatus = GTK_GRSTATUS(object);
	priv = grstatus->priv;
	
	if (priv->values != NULL) {
		g_free(priv->values);
		priv->values = NULL;
	}
	
	klass = gtk_type_class(gtk_widget_get_type());
	
	if (GTK_OBJECT_CLASS(klass)->destroy) {
		(* GTK_OBJECT_CLASS(klass)->destroy) (object);
	}
}
#endif
