/*
 *      gsfcore.c
 *      
 *      Copyright 2012 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>

#include "settings.h"
#include "gsfcore.h"
#include "dbstorage.h"
#include "dvbstatus.h"
#include "dvbname.h"
#include "streamconfig.h"
#include "satscanwindow.h"

#define THREAD_TIMEOUT  1000000

enum {
	SOUND_THREAD_POLL = 0,
	SOUND_THREAD_PLAY
};

enum {
	NETSTATUS_THREAD_POLL = 0,
	NETSTATUS_THREAD_WORK
};

enum {
	SIGNAL_THREAD_POLL = 0,
	SIGNAL_THREAD_FRONTEND_READ,
	SIGNAL_THREAD_FRONTEND_TUNE,
	SIGNAL_THREAD_SATSCAN,
	SIGNAL_THREAD_NET_INIT,
	SIGNAL_THREAD_NET_READ,
	SIGNAL_THREAD_NET_TUNE,
	SIGNAL_THREAD_NET_REFRESH
};

struct _workthread {
	GThread *thread;
	gchar action;
	gpointer data;
	gboolean alive;
};

static GSList *devices = NULL;

static gsfcore_event_callback eventcallback = NULL;

static struct _workthread soundthread, netstatusthread, signalthread;

static struct _netgate netgate;
static struct _frontend frontend;
static struct _service *svc;
static struct _demuxer demuxer;
static struct _streamconfig streamconfig;
static struct _location location;

static gboolean networkdisabled, streamingdisabled;


gchar *gsfcore_string_definition(guint table, guint id)
{
	gchar *polarisation_tab[] = {"V","H","L","R"};
	gchar *fec_tab[] = {"AUTO","1/2","2/3","3/4","3/5","4/5","5/6","6/7","7/8","8/9","9/10","NONE"};
	gchar *system_tab[] = {"DVB-S","DVB-S2","Undefined"};
	gchar *modulation_tab[] = {"QPSK","8PSK"};
	gchar *rolloff_tab[] = {"Unknown","0.35","0.25","0.20"};
	gchar *port10_tab[] = {"0","A","B","C","D"};
	gchar *port11_tab[] = {"0","1","2","3","4","5","6","7","8"};
	
	switch (table) {
		case DEFINITION_POLARISATION:
			return polarisation_tab[id];
		case DEFINITION_FEC:
			return fec_tab[id];
		case DEFINITION_SYSTEM:
			return system_tab[id];
		case DEFINITION_MODULATION:
			return modulation_tab[id];
		case DEFINITION_ROLLOFF:
			return rolloff_tab[id];
		case DEFINITION_PORT10:
			return port10_tab[id];
		case DEFINITION_PORT11:
			return port11_tab[id];
		default: break;
	}
	
	return "Unknown";
}

static gpointer gsfcore_sound_thread(gpointer data)
{
	while (soundthread.alive) {
		switch(soundthread.action) {
			case SOUND_THREAD_PLAY: { 
				sound_play(&frontend, THREAD_TIMEOUT);
				break;
			}
			
			case SOUND_THREAD_POLL:
			default:
				g_usleep(THREAD_TIMEOUT);
				break;
		}
	}
	
    return NULL;
}

void gsfcore_sound_set_enabled(gboolean enabled)
{
	if (soundthread.alive) {
		if (enabled) {
			soundthread.action = SOUND_THREAD_PLAY;
		} else {
			soundthread.action = SOUND_THREAD_POLL;
		}
	}
}

gboolean gsfcore_sound_get_enabled(void)
{
	if (!soundthread.alive) return FALSE;
	
	if (soundthread.action == SOUND_THREAD_PLAY) {
		return TRUE;
	} else {
		return FALSE;
	}
}

static void gsfcore_sound_start(void)
{
	soundthread.action = SOUND_THREAD_POLL;
	soundthread.alive = TRUE;
	soundthread.data = NULL;
	soundthread.thread = g_thread_create(gsfcore_sound_thread, NULL, FALSE, NULL);
	if (soundthread.thread == NULL) soundthread.alive = FALSE;
}

static void gsfcore_sound_stop(void)
{
	if (soundthread.alive) {
		soundthread.action = SOUND_THREAD_POLL;
		soundthread.alive = FALSE;
	}
}


static gpointer gsfcore_netstatus_thread(gpointer data)
{
	communication_initialize();
	
	return NULL;
}

gchar *gsfcore_get_netstatus_client_address(void)
{
	return g_strdup(netgate.clientipstr);
}

netgate_t gsfcore_get_netgate(void)
{
	return &netgate;
}

static void gsfcore_netstatus_start(void)
{
	udpclient_init(&netgate);
	
	netstatusthread.action = NETSTATUS_THREAD_WORK;
	netstatusthread.alive = TRUE;
	netstatusthread.data = NULL;
	netstatusthread.thread = g_thread_create(gsfcore_netstatus_thread, NULL, FALSE, NULL);
	if (netstatusthread.thread == NULL) netstatusthread.alive = FALSE;
}

static void gsfcore_netstatus_stop(void)
{
	if (netstatusthread.alive) {
		netstatusthread.action = NETSTATUS_THREAD_POLL;
		netstatusthread.alive = FALSE;
	}
}

frontend_t gsfcore_get_frontend(void)
{
	return &frontend;
}

service_t gsfcore_get_service(void)
{
	return svc;
}

void gsfcore_network_servicelist_refresh(void)
{
	if (netgate.clientconnected)
		signalthread.action = SIGNAL_THREAD_NET_REFRESH;
}

tuneparams_t gsfcore_get_tuneparams(void)
{
	static struct _tuneparams tuneparams;
	
	if (svc == NULL) return NULL;
	
	tuneparams.pos = svc->pos;
	tuneparams.freq = svc->freq;
	tuneparams.sr = svc->sr;
	tuneparams.pol = svc->pol;
	tuneparams.fec = svc->fec;
	tuneparams.sys = svc->sys;
	tuneparams.mod = svc->mod;
	tuneparams.rof = svc->rof;
	tuneparams.port10 = svc->port10;
	tuneparams.port11 = svc->port11;
	tuneparams.rep11 = svc->rep11;
	tuneparams.lnbtype = svc->lnbtype;
	tuneparams.lnb22khz = svc->lnb22khz;
	tuneparams.positioner = svc->positioner;
	tuneparams.possatnum = svc->possatnum;
	
	if (frontend.tuned) {
		tuneparams.tuned = (guchar)frontend.tuned;
		tuneparams.realfreq = frontend.realfreq;
		tuneparams.realsr = frontend.realsr;
	} else {
		tuneparams.tuned = (guchar)frontend.tuned;
		tuneparams.realfreq = 0;
		tuneparams.realsr = 0;
	} 
	
	return &tuneparams;
}

void gsfcore_get_signal_status(struct _sigstatus *sigstatus)
{
	if (sigstatus == NULL) return;
	
	memset(sigstatus, 0, sizeof(sigstatus));
	
	if ((!frontend.tuned) || (svc == NULL)) {
		g_snprintf(sigstatus->svcname, sizeof(sigstatus->svcname), "Unknown service");
		sigstatus->snr = 0;
		sigstatus->agc = 0;
		sigstatus->ber = 0;
		sigstatus->lock = FALSE;
		sigstatus->carrier = FALSE;
	} else {
		g_snprintf(sigstatus->svcname, sizeof(sigstatus->svcname), "%s", svc->name);
		if (!frontend.reset) {
			sigstatus->snr = frontend.snr;
			sigstatus->agc = frontend.signal;
			sigstatus->ber = frontend.ber;
			sigstatus->lock = frontend.lock;
			sigstatus->carrier = frontend.carrier;
		}
	}
}

static gint gsfcore_read_event(gint status, uint32_t freq, uint32_t sr, gint fec)
{
	if (status == 0) {
		//correct if and sr in frontend struct
		frontend.realfreq = freq/1000;
		frontend.realsr = sr/1000;
		frontend.realfec = fec;
		//set snr scale to zero
		frontend.reset = TRUE;
		frontend.oldsnr = frontend.snr;
		//no need to restart demuxers if signal is lost
		if (!frontend.tuned) {
			dvb_demux_open(&frontend, &demuxer, svc);
			streamconfig_set_stream_data(&streamconfig, svc->name, demuxer.mediastreams, svc->caid, svc->key);
		}
		
		if (!frontend.tuned) {
			if (eventcallback != NULL) {
				(eventcallback)(GSF_EVENT_LOCK_TUNE_BUTTONS, GINT_TO_POINTER(FALSE));
				(eventcallback)(GSF_EVENT_LOCK_SCAN_BUTTONS, GINT_TO_POINTER(FALSE));
				(eventcallback)(GSF_EVENT_STATUS_MESSAGE, GINT_TO_POINTER(STATUS_MESSAGE_SVC));
				(eventcallback)(GSF_EVENT_TRANSPONDER_SELECTED, svc);
				(eventcallback)(GSF_EVENT_UPDATE_TUNE_PARAMS, svc);
			}
			frontend.tuned = TRUE;
		}
	} else if (status == 30) {
		if (eventcallback != NULL) (eventcallback)(GSF_EVENT_TRANSPONDER_TUNED, svc);
	}
	
	g_printf("[+] Handled event status: %u, frequency: %u, symbolrate: %u, fec: %u\n", status, freq/1000, sr/1000, fec);
	
	return 1;
}

static void gsfcore_zero_signal(frontend_t frontend)
{
	if (frontend != NULL) {
		if (frontend->snr != 0) frontend->snr = 0;
		if (frontend->signal != 0) frontend->signal = 0;
		if (frontend->ber != 0) frontend->ber = 0;
		if (frontend->lock) frontend->lock = FALSE;
		if (frontend->carrier) frontend->carrier = FALSE;
		
		if (eventcallback != NULL) (eventcallback)(GSF_EVENT_UPDATE_SIGNAL, &frontend);
	}
}

static gpointer gsfcore_signal_thread(gpointer data)
{
	transponders_t transponders;
	
    while (signalthread.alive) {
		switch(signalthread.action) {
			case SIGNAL_THREAD_FRONTEND_READ:
				if (dvb_status(&frontend, svc, (frontend.open_mode_readonly) ? NULL : gsfcore_read_event)) {
					if ((frontend.snr != frontend.oldsnr) && (frontend.reset)) frontend.reset = FALSE;
					if (eventcallback != NULL) (eventcallback)(GSF_EVENT_UPDATE_SIGNAL, &frontend);
				}
				break;
			
			case SIGNAL_THREAD_FRONTEND_TUNE: 
				if (dvb_tune(&frontend, svc)) {
					frontend.tuned = FALSE;
					signalthread.action = SIGNAL_THREAD_FRONTEND_READ;
				} else {
					gsfcore_zero_signal(&frontend);
					signalthread.action = SIGNAL_THREAD_POLL;
					if (eventcallback != NULL) (eventcallback)(GSF_EVENT_STATUS_MESSAGE, GINT_TO_POINTER(STATUS_MESSAGE_TP_TUNING_FAIL));
				}
				break;
			
			case SIGNAL_THREAD_SATSCAN:
				if ((transponders = satscan_window_get_transponders()) != NULL) {
					if (dvb_satscan_init(&frontend, transponders)) {
						satscan_window_scan(&frontend);
					}
				}			
				if (eventcallback != NULL) (eventcallback)(GSF_EVENT_SATSCAN_STATUS, NULL);
				signalthread.action = SIGNAL_THREAD_POLL;
				break;
			
			case SIGNAL_THREAD_NET_INIT: {
				frontend.tuned = FALSE;
				if (!udpclient_connect(&netgate)) {
					gsfcore_zero_signal(&frontend);
					signalthread.action = SIGNAL_THREAD_POLL;
					if ((netgate.error != NULL) && (eventcallback != NULL)) (eventcallback)(GSF_EVENT_ERROR_MESSAGE, netgate.error);
				} else {
					frontend.reset = TRUE;
					if (!dvb_frontend_opened(frontend.frontend_fd)) {
						if ((service_net_get_number() > 0) && (eventcallback != NULL)) (eventcallback)(GSF_EVENT_LOCK_TUNE_BUTTONS, GINT_TO_POINTER(FALSE));
					}
					signalthread.action = SIGNAL_THREAD_NET_READ;
				}
				break;
			}
						
			case SIGNAL_THREAD_NET_READ: {
				if (netgate.clientconnected) {
					udpclient_get_signal_status(&netgate, svc, &frontend);
					
					if (frontend.reset) {
						gsfcore_zero_signal(&frontend);
						if (eventcallback != NULL) (eventcallback)(GSF_EVENT_STATUS_MESSAGE, GINT_TO_POINTER(STATUS_MESSAGE_SVC_NETSARE));
						frontend.reset = FALSE;
					}
					if (eventcallback != NULL) (eventcallback)(GSF_EVENT_UPDATE_SIGNAL, &frontend);
				} else {
					gsfcore_zero_signal(&frontend);
					signalthread.action = SIGNAL_THREAD_POLL;
					if ((netgate.error != NULL) && (eventcallback != NULL)) (eventcallback)(GSF_EVENT_ERROR_MESSAGE, netgate.error);
				}
				break;
			}
			
			case SIGNAL_THREAD_NET_TUNE: {
				frontend.tuned = FALSE;
				if (!udpclient_tune_service(&netgate)) {
					if ((netgate.error != NULL) && (eventcallback != NULL)) (eventcallback)(GSF_EVENT_ERROR_MESSAGE, netgate.error);
				}
				signalthread.action = SIGNAL_THREAD_NET_READ;
				frontend.reset = TRUE;
				break;
			}
			
			case SIGNAL_THREAD_NET_REFRESH: {
				if (!udpclient_get_channels_list(&netgate)) {
					if ((netgate.error != NULL) && (eventcallback != NULL)) (eventcallback)(GSF_EVENT_ERROR_MESSAGE, netgate.error);
				}
				signalthread.action = SIGNAL_THREAD_NET_READ;
				break;
			}
			
			case SIGNAL_THREAD_POLL: break;
			
			default: break;
		}
		usleep(THREAD_TIMEOUT);
	}
	
    return NULL;
}

static void gsfcore_signalstatus_start(void)
{
	signalthread.action = SIGNAL_THREAD_POLL;
	signalthread.alive = TRUE;
	signalthread.data = NULL;
	signalthread.thread = g_thread_create(gsfcore_signal_thread, NULL, FALSE, NULL);
	if (signalthread.thread == NULL) signalthread.alive = FALSE;
}

static void gsfcore_signalstatus_stop(void)
{
	if (signalthread.alive) {
		signalthread.action = SIGNAL_THREAD_POLL;
		signalthread.alive = FALSE;
	}
}

gboolean gsfcore_tune_start(guint32 id, guchar type)
{
	struct _service *currentservice;
	gint calcres;
	
	if (type == SERVICE_NETSHARE) {
		netgate.netsvcid = id;
		g_printf("[+] Tuning to channel at server: ID:%u\n", netgate.netsvcid);
		signalthread.action = SIGNAL_THREAD_NET_TUNE;
		return TRUE;
	} else {
		currentservice = (struct _service *)service_read(id);
		if ((dvb_frontend_opened(frontend.frontend_fd)) || (currentservice == NULL)) {
			dvb_demux_close(&frontend, &demuxer, svc);
			calcres = dvb_calculate_values(&frontend, currentservice, TRUE);
			if (calcres == DVBSTATUS_VALUES_ALL_GOOD) {
				if (svc != NULL) service_free(svc);
				svc = service_clone(currentservice);
				signalthread.action = SIGNAL_THREAD_FRONTEND_TUNE;
				if (eventcallback != NULL) {
					(eventcallback)(GSF_EVENT_LOCK_TUNE_BUTTONS, GINT_TO_POINTER(TRUE));
					(eventcallback)(GSF_EVENT_LOCK_SCAN_BUTTONS, GINT_TO_POINTER(TRUE));
					(eventcallback)(GSF_EVENT_STATUS_MESSAGE, GINT_TO_POINTER(STATUS_MESSAGE_TP_TUNING));
				}
				return TRUE;
			} else {
				switch (calcres) {
					case DVBSTATUS_VALUES_IF_WRONG:
						g_printf("[-] Service frequency not supported by selected device tuner or driver\n");
						break;
					case DVBSTATUS_VALUES_SR_WRONG:
						g_printf("[-] Service symbol rate not supported by selected device demodulator or driver\n");
						break;
					default:
						g_printf("[-] Service parameters not supported by selected hardware or driver\n");
						break;
				}
				if (eventcallback != NULL) (eventcallback)(GSF_EVENT_TUNE_ERROR_MESSAGE, GINT_TO_POINTER(calcres));
				return FALSE;
			}
		} else {
			if (currentservice != NULL) service_free(currentservice);
			gsfcore_zero_signal(&frontend);
			signalthread.action = SIGNAL_THREAD_POLL;
			return FALSE;
		}
	}
}

gboolean gsfcore_network_start(gchar *address, gushort port, gchar *password)
{
	if ((address != NULL) && (password != NULL) && (port > 1024)) {
		if (udpclient_set_parameters(&netgate, address, port, password)) {
			if (svc != NULL) service_free(svc);
			svc = service_new(NULL, "Service", SERVICE_NETSHARE);
			signalthread.action = SIGNAL_THREAD_NET_INIT;
				
			return TRUE;
		}
	}
	
	return FALSE;
}

void gsfcore_satscan_start(gboolean resume)
{ 
	if (!resume) {
		if (dvb_frontend_opened(frontend.frontend_fd)) {
			dvb_demux_close(&frontend, &demuxer, svc);
		}
		frontend.frozen = TRUE;
		signalthread.action = SIGNAL_THREAD_SATSCAN;
	} else {
		if (frontend.frozen) {
			if ((frontend.tuned) && (svc != NULL)) {
				signalthread.action = SIGNAL_THREAD_FRONTEND_TUNE;
			} else {
				gsfcore_zero_signal(&frontend);
				signalthread.action = SIGNAL_THREAD_POLL;
			}
			frontend.frozen = FALSE;
		}
	}
}

gboolean gsfcore_service_has_pid_type(guint type, gushort *pidvalue)
{
	guint i, len;
	gboolean pidexists;
	struct _pid *currentpid;
	
	if (svc == NULL) return FALSE;
	
	if (svc->pids == NULL) return FALSE;
	
	len = 0;
	pidexists = FALSE;
	
	if (svc->pids != NULL) {
		len = g_slist_length(svc->pids);
		for (i=0; i<len; i++) {
			currentpid = (struct _pid *)g_slist_nth_data(svc->pids, i);
			if ((currentpid->type == type) && (currentpid->enabled)) {
				if (pidvalue != NULL) {
					*pidvalue = currentpid->value;
				}
				pidexists = TRUE;
				break;
			}
		}
	}
	
	return pidexists;
}

static gboolean gsfcore_device_close(gpointer data)
{
	if (dvb_frontend_opened(frontend.frontend_fd)) {
		signalthread.action = SIGNAL_THREAD_POLL;
		if (dvb_close(&frontend)) frontend.frontend_fd = -1;
		frontend.tuned = FALSE;
		g_free(frontend.frontend_fn);
		frontend.frontend_nr = 0;
		frontend.adapter_nr = 0;
		memset(frontend.driver, 0, 128);
		frontend.frequency_min = 0;
		frontend.frequency_max = 0;
		frontend.frequency_stepsize = 0;
		frontend.symbol_rate_min = 0;
		frontend.symbol_rate_max = 0;
		frontend.dvbs_support = FALSE;
		frontend.dvbs2_support = FALSE;
		frontend.open_mode_readonly = FALSE;
		frontend.api_version = 0;
		frontend.locked = FALSE;
		frontend.timestamp = 0;
		frontend.lockattempts = 0;
		frontend.frozen = FALSE;
		frontend.error = NULL;
		frontend.lastposition = 0;
		
		gsfcore_zero_signal(&frontend);
		if (eventcallback != NULL) (eventcallback)(GSF_EVENT_LOCK_TUNE_BUTTONS, GINT_TO_POINTER(TRUE));
		if (eventcallback != NULL) (eventcallback)(GSF_EVENT_LOCK_SCAN_BUTTONS, GINT_TO_POINTER(TRUE));
	}
	
	return FALSE;
}

static gboolean gsfcore_device_open(struct _device *device)
{
	if (device != NULL) {
		gsfcore_device_close(NULL); //if changing frontend
		if (eventcallback != NULL) (eventcallback)(GSF_EVENT_STATUS_MESSAGE, GINT_TO_POINTER(STATUS_MESSAGE_FR_OPENING));
		frontend.adapter_nr = device->adapter;
		frontend.frontend_nr = device->frontend;
		frontend.frontend_fn = dvbname_get_frontend_filename(device->adapter, device->frontend);
		frontend.lastposition = 0;
		
		signalthread.action = SIGNAL_THREAD_POLL;
		
		if ((frontend.frontend_fd = dvb_open(&frontend)) < 0 ) {
			//error while opening
			if (eventcallback != NULL) (eventcallback)(GSF_EVENT_LOCK_TUNE_BUTTONS, GINT_TO_POINTER(TRUE));
			if (eventcallback != NULL) (eventcallback)(GSF_EVENT_LOCK_SCAN_BUTTONS, GINT_TO_POINTER(TRUE));
			if (eventcallback != NULL) (eventcallback)(GSF_EVENT_STATUS_MESSAGE, GINT_TO_POINTER(STATUS_MESSAGE_FR_OPEN_FAIL));
			//Device configuration update
			if (eventcallback != NULL) (eventcallback)(GSF_EVENT_UPDATE_DEVICE_PARAMS, &frontend);
			signalthread.action = SIGNAL_THREAD_POLL;
		} else {
			if (!frontend.open_mode_readonly) {
				//open in full access mode
				if (eventcallback != NULL) (eventcallback)(GSF_EVENT_LOCK_TUNE_BUTTONS, GINT_TO_POINTER(FALSE));
				if (eventcallback != NULL) (eventcallback)(GSF_EVENT_STATUS_MESSAGE, GINT_TO_POINTER(STATUS_MESSAGE_TP_SELECTION));
				//Device configuration update
				if (eventcallback != NULL) (eventcallback)(GSF_EVENT_UPDATE_DEVICE_PARAMS, &frontend);
				signalthread.action = SIGNAL_THREAD_POLL;
			} else {
				//open in readonly mode
				if (eventcallback != NULL) (eventcallback)(GSF_EVENT_LOCK_TUNE_BUTTONS, GINT_TO_POINTER(TRUE));
				if (eventcallback != NULL) (eventcallback)(GSF_EVENT_LOCK_SCAN_BUTTONS, GINT_TO_POINTER(TRUE));
				//Need to clone empty service for reading parameters
				if (svc != NULL) service_free(svc);
				svc = service_new(NULL, "Readonly mode service", SERVICE_TRANSPONDER);
				dvb_parameters(&frontend, svc);
				frontend.realfec = svc->fec;
				frontend.realfreq = svc->freq;
				frontend.realsr = svc->sr;
				if (eventcallback != NULL) (eventcallback)(GSF_EVENT_STATUS_MESSAGE, GINT_TO_POINTER(STATUS_MESSAGE_FR_READONLY_MODE));
				//Device configuration update
				if (eventcallback != NULL) (eventcallback)(GSF_EVENT_UPDATE_DEVICE_PARAMS, &frontend);
				signalthread.action = SIGNAL_THREAD_FRONTEND_READ;
			}
			//save opened device idents
			settings_append_int("adapter", device->adapter);
			settings_append_int("frontend", device->frontend);
			settings_flush();
		}
	} else {
		if (eventcallback != NULL) (eventcallback)(GSF_EVENT_STATUS_MESSAGE, GINT_TO_POINTER(STATUS_MESSAGE_FR_NOT_FOUND));
		//Device configuration update
		if (eventcallback != NULL) (eventcallback)(GSF_EVENT_UPDATE_DEVICE_PARAMS, &frontend);
	}
	
	return FALSE;
}

GSList *gsfcore_get_devices(gboolean rescan)
{
	if (rescan) {
		dvbname_free_devices(&devices);
		devices = dvbname_scan_devices();
	}
	
	return devices;
}

gint gsfcore_device_search(guint adapterid, guint frontendid, gboolean rescan, gboolean restart)
{
	gsize length;
	gint i;
	struct _device *device;
	gboolean devfound;
	gboolean unique;
	
	if (rescan) gsfcore_get_devices(TRUE);
	
	devfound = FALSE;
	
	if ((length = g_slist_length(devices)) > 0) {
		for (i=0; i<length; i++) {
			device = (struct _device *)g_slist_nth_data(devices, i);
			if ((adapterid == device->adapter) && (frontendid == device->frontend)) {
				devfound = TRUE;
				unique = ((device->adapter != frontend.adapter_nr) || (device->frontend != frontend.frontend_nr));
				if ((!restart) || ((restart) && (!unique))) {
					gsfcore_device_open(device);
				}
				break;
			}
		}
		
		if (!devfound) {
			device = (struct _device *)g_slist_nth_data(devices, 0);
			unique = ((device->adapter != frontend.adapter_nr) || (device->frontend != frontend.frontend_nr));
			if ((!restart) || ((restart) && (!unique))) {
				gsfcore_device_open(device);
			}
			return -(length);
		}
	}
	
	return length;
}	

void gsfcore_get_location(gdouble *longitude, gdouble *latitude)
{
	if (longitude != NULL) {
		*longitude = location.longitude;
	}
	
	if (latitude != NULL) {
		*latitude = location.latitude;
	}
}

void gsfcore_set_location(gdouble longitude, gdouble latitude)
{
	location.longitude = longitude;
	location.latitude = latitude;
	
	settings_append_float("place_longitude", longitude);
	settings_append_float("place_latitude", latitude);
}

streamconfig_t gsfcore_get_streamconfig(void)
{
	return &streamconfig;
}

gboolean gsfcore_restart_streamserver(void)
{
	streamconfig.serverctl = streamconfig_service_restart(&streamconfig);
	
	return TRUE;
}

pid_t gsfcore_get_streamserver_pid(void)
{
	return streamconfig.processpid;
}

gboolean gsfcore_get_streamserver_control(void)
{
	return streamconfig.serverctl;
}

gboolean gsfcore_handle_streamserver_stop_signal(pid_t sender)
{
	if ((streamconfig.serverctl) && ((sender == streamconfig.processpid) || (sender == getpid()))) {
		streamconfig.serverctl = FALSE;
		return TRUE;
	} else {
		return FALSE;
	}
}

gboolean gsfcore_set_streamserver_record_status(gboolean status)
{
	return streamconfig_set_record_status(&streamconfig, status);
}

gboolean gsfcore_set_streamserver_udpbc_status(gboolean status)
{
	return streamconfig_set_udp_broadcast_status(&streamconfig, status);
}

gboolean gsfcore_set_streamserver_udpmc_status(gboolean status)
{
	return streamconfig_set_udp_multicast_status(&streamconfig, status);
}

gboolean gsfcore_set_streamserver_tcpuc_status(gboolean status)
{
	return streamconfig_set_tcp_unicast_status(&streamconfig, status);
}

gboolean gsfcore_get_streamserver_services(gboolean *record, gboolean *udpbc, gboolean *udpmc, gboolean *tcpuc)
{
	return streamconfig_get_service_flags(&streamconfig, record, udpbc, udpmc, tcpuc);
}

gboolean gsfcore_get_streamserver_available(void)
{
	if ((streamconfig_service_state(&streamconfig) != -1) && (streamconfig.serverctl)) {
		return TRUE;
	} else {
		return FALSE;
	}
}

gboolean gsfcore_get_streaming_disabled(void)
{
	return streamingdisabled;
}

gboolean gsfcore_get_network_disabled(void)
{
	return networkdisabled;
}


gboolean gsfcore_init(gboolean disablenetwork, gboolean disablestreaming, gsfcore_event_callback callback)
{
	if (callback == NULL) return FALSE;
	
	eventcallback = callback;
	networkdisabled = disablenetwork;
	streamingdisabled = disablestreaming;
	
	if (sound_init()) {
		gsfcore_sound_start();
	} else {
		if (eventcallback != NULL) (eventcallback)(GSF_EVENT_LOCK_SOUND_BUTTONS, GINT_TO_POINTER(TRUE));
	}
	
	gsfcore_signalstatus_start();
	
	if (!disablenetwork) {
		gsfcore_netstatus_start();
	}
	
	if (!disablestreaming) {
		streamconfig_init(&streamconfig);
	}
	
	frontend.adapter_nr = 0;
	frontend.frontend_nr = 0;
	frontend.frontend_fd = -1;
	demuxer.netif_nr = -1;
	
	location.longitude = settings_read_float("place_longitude", 37.70);
	location.latitude = settings_read_float("place_latitude", 55.75);
	
	if (gsfcore_device_search(settings_read_int("adapter", 0), settings_read_int("frontend", 0), TRUE, FALSE) == 0) { 
		if (eventcallback != NULL) (eventcallback)(GSF_EVENT_STATUS_MESSAGE, GINT_TO_POINTER(STATUS_MESSAGE_DEV_NOT_FOUND));
	}
	
	return TRUE;
}

gboolean gsfcore_stop(void)
{
	gsfcore_sound_stop();
	gsfcore_signalstatus_stop();
	gsfcore_netstatus_stop();
	
	streamconfig_close(&streamconfig);
	
	if (dvb_frontend_opened(frontend.frontend_fd)) {
		dvb_demux_close(&frontend, &demuxer, svc);
		gsfcore_device_close(NULL);
	}
	
	return TRUE;
}
