/*
 *      saferwlocks.c
 *      
 *      Copyright 2010 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <string.h>

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/mman.h>
#include <signal.h>

#include "saferwlocks.h"

#define SHMSIZE(x) ( sizeof(int)*3 + sizeof(pid_t)*1 + sizeof(pid_t)*(x) )
#define SEMID(x) ( ((int *)(x))[0] )
#define PROCESSES(x) ( ((int *)(x))[1] )
#define LATENCY(x) ( ((int *)(x))[2] )
#define WRITER(x) ( ((pid_t *)((char *)(x) + sizeof(int)*3))[0] )
#define READERS(x, n) ( ((pid_t *)((char *)(x) + sizeof(int)*3))[(n) + 1] )

union semun {
   int val;
   struct semid_ds *buf;
   ushort *array;
};


void rwlockshed_init(rwlockshed_t *rwlockshed, key_t svkey, int maxprocesses, int latency)
{
	union semun vsctl;
	unsigned short initvals[4];
	
	*rwlockshed = (rwlockshed_t)mmap(NULL, SHMSIZE(maxprocesses), PROT_READ|PROT_WRITE, MAP_SHARED|MAP_ANONYMOUS, 0, 0);
	memset(*rwlockshed, 0, SHMSIZE(maxprocesses));
	
	SEMID(*rwlockshed) = semget(svkey, 4, 0666 | IPC_CREAT);
	
	initvals[0] = maxprocesses;
	initvals[1] = 1;
	initvals[2] = 1;
	initvals[3] = 1;
	vsctl.array = (unsigned short *)&initvals;
	semctl(SEMID(*rwlockshed), 4, SETALL, vsctl); 
	
	PROCESSES(*rwlockshed) = maxprocesses;
	LATENCY(*rwlockshed) = latency;
	WRITER(*rwlockshed) = 0;
}

void rwlockshed_destroy(rwlockshed_t rwlockshed)
{
	int semid, processes;
	
	semid = SEMID(rwlockshed);
	processes = PROCESSES(rwlockshed);
	semctl(semid, 4, IPC_RMID);
	munmap(rwlockshed, (size_t)(SHMSIZE(processes)));
}

void rwlockshed_readlock(rwlockshed_t rwlockshed)
{
	struct sembuf vsbuf[4];
	int i;
	struct timespec timeout;
	
	timeout.tv_sec = 0;
	timeout.tv_nsec = LATENCY(rwlockshed)*1000;
	
	vsbuf[1].sem_num = 1;
	vsbuf[1].sem_op = -1;
	vsbuf[1].sem_flg = 0;
	semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[1], 1);
	
	do {
		vsbuf[0].sem_num = 0;
		vsbuf[0].sem_op = -1;
		vsbuf[0].sem_flg = 0;
		if (semtimedop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[0], 1, &timeout) == 0) {
			break;
		} else if (errno == EAGAIN) {
			vsbuf[2].sem_num = 2;
			vsbuf[2].sem_op = -1;
			vsbuf[2].sem_flg = 0;
			semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[2], 1);
			for (i=0; i<PROCESSES(rwlockshed); i++) {
				if ((READERS(rwlockshed, i)) && (kill(READERS(rwlockshed, i), 0) == -1) && (errno == ESRCH)) {
					printf("[RWLOCK SHEDULER] deadlock in process invoked read lock (%d) resolved\n", READERS(rwlockshed, i));
					READERS(rwlockshed, i) = 0;
					vsbuf[0].sem_num = 0;
					vsbuf[0].sem_op = 1;
					vsbuf[0].sem_flg = 0;
					semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[0], 1);
					break;
				}
			}
			vsbuf[2].sem_num = 2;
			vsbuf[2].sem_op = 1;
			vsbuf[2].sem_flg = 0;
			semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[2], 1);
		}
	} while (1);
	
	vsbuf[2].sem_num = 2;
	vsbuf[2].sem_op = -1;
	vsbuf[2].sem_flg = 0;
	semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[2], 1);
	for (i=0; i<PROCESSES(rwlockshed); i++) {
		if (READERS(rwlockshed, i) == 0) {
			READERS(rwlockshed, i) = getpid();
			break;
		}
	}
	vsbuf[2].sem_num = 2;
	vsbuf[2].sem_op = 1;
	vsbuf[2].sem_flg = 0;
	semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[2], 1);
	
	vsbuf[1].sem_num = 1;
	vsbuf[1].sem_op = 1;
	vsbuf[1].sem_flg = 0;
	semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[1], 1);
}

void rwlockshed_readunlock(rwlockshed_t rwlockshed)
{
	struct sembuf vsbuf[4];
	int i;
	pid_t cpid = getpid();
	
	vsbuf[2].sem_num = 2;
	vsbuf[2].sem_op = -1;
	vsbuf[2].sem_flg = 0;
	semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[2], 1);
	
	for (i=0; i<PROCESSES(rwlockshed); i++) {
		if (READERS(rwlockshed, i) == cpid) {
			READERS(rwlockshed, i) = 0;
			break;
		}
	}
	
	vsbuf[2].sem_num = 2;
	vsbuf[2].sem_op = 1;
	vsbuf[2].sem_flg = 0;
	semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[2], 1);
	
	vsbuf[0].sem_num = 0;
	vsbuf[0].sem_op = 1;
	vsbuf[0].sem_flg = 0;
	semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[0], 1);
}

void rwlockshed_writelock(rwlockshed_t rwlockshed)
{
	struct sembuf vsbuf[4];
	int semval, i;
	pid_t cpid = getpid();
	struct timespec timeout;
	time_t ts0, ts1;
	
	timeout.tv_sec = 0;
	timeout.tv_nsec = LATENCY(rwlockshed)*1000;
	
	do {
		vsbuf[1].sem_num = 1;
		vsbuf[1].sem_op = -1;
		vsbuf[1].sem_flg = 0;
		if (semtimedop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[1], 1, &timeout) == 0) {
			break;
		} else if (errno == EAGAIN) {
			vsbuf[3].sem_num = 3;
			vsbuf[3].sem_op = -1;
			vsbuf[3].sem_flg = 0;
			semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[3], 1);
			if ((WRITER(rwlockshed)) && (kill(WRITER(rwlockshed), 0) == -1) && (errno == ESRCH)) {
				printf("[RWLOCK SHEDULER] deadlock in process invoked write lock (%d) resolved\n", WRITER(rwlockshed));
				WRITER(rwlockshed) = 0;
				vsbuf[1].sem_num = 1;
				vsbuf[1].sem_op = 1;
				vsbuf[1].sem_flg = 0;
				semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[1], 1);
			}
			vsbuf[3].sem_num = 3;
			vsbuf[3].sem_op = 1;
			vsbuf[3].sem_flg = 0;
			semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[3], 1);
		}
	} while (1);
	
	vsbuf[3].sem_num = 3;
	vsbuf[3].sem_op = -1;
	vsbuf[3].sem_flg = 0;
	semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[3], 1);
	
	WRITER(rwlockshed) = cpid;
	
	vsbuf[3].sem_num = 3;
	vsbuf[3].sem_op = 1;
	vsbuf[3].sem_flg = 0;
	semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[3], 1);
	
	ts0 = time(NULL);
	do {
		ts1 = time(NULL);
		if ((ts1 - ts0) > 1) {
			vsbuf[2].sem_num = 2;
			vsbuf[2].sem_op = -1;
			vsbuf[2].sem_flg = 0;
			semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[2], 1);
			for (i=0; i<PROCESSES(rwlockshed); i++) {
				if ((READERS(rwlockshed, i)) && (kill(READERS(rwlockshed, i), 0) == -1) && (errno == ESRCH)) {
					printf("[RWLOCK SHEDULER] deadlock in process invoked read lock (%d) resolved\n", READERS(rwlockshed, i));
					READERS(rwlockshed, i) = 0;
					vsbuf[0].sem_num = 0;
					vsbuf[0].sem_op = 1;
					vsbuf[0].sem_flg = 0;
					semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[0], 1);
					break;
				}
			}
			vsbuf[2].sem_num = 2;
			vsbuf[2].sem_op = 1;
			vsbuf[2].sem_flg = 0;
			semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[2], 1);
			ts0 = ts1;
		}
		semval = semctl(SEMID(rwlockshed), 0, GETVAL);
	} while (semval != PROCESSES(rwlockshed));
}

void rwlockshed_writeunlock(rwlockshed_t rwlockshed)
{
	struct sembuf vsbuf[4];
	
	vsbuf[3].sem_num = 3;
	vsbuf[3].sem_op = -1;
	vsbuf[3].sem_flg = 0;
	semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[3], 1);
	WRITER(rwlockshed) = 0;
	vsbuf[3].sem_num = 3;
	vsbuf[3].sem_op = 1;
	vsbuf[3].sem_flg = 0;
	semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[3], 1);
	
	vsbuf[1].sem_num = 1;
	vsbuf[1].sem_op = 1;
	vsbuf[1].sem_flg = 0;
	semop(SEMID(rwlockshed), (struct sembuf *)&vsbuf[1], 1);
}
