/*
 *      signalshed.c
 *      
 *      Copyright 2010 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <string.h>

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/mman.h>
#include <signal.h>

#include "signalshed.h"

#define SHMSIZE(x) ( sizeof(int)*3 + sizeof(pid_t)*(x) )
#define SEMID(x) ( ((int *)(x))[0] )
#define MAXCLIENTS(x) ( ((int *)(x))[1] )
#define SIGDELAY(x) ( ((int *)(x))[2] )
#define CLIENTS(x, n) ( ((pid_t *)((char *)(x) + sizeof(int)*3))[(n)] )

union semun {
	int val;
	struct semid_ds *buf;
	ushort *array;
};

static volatile int contflag;
static unsigned long starttime, endtime;

static void signalshed_sighandler(int sig);
static int signalshed_sigdelay_detect(void);


static void signalshed_sighandler(int sig)
{
	struct timeval tval;
	
	if (sig == SIGUSR1) {
		gettimeofday(&tval, NULL);
		endtime = tval.tv_sec*1000*1000 + tval.tv_usec;
	} else if (sig == SIGALRM) {
		endtime = 0;
	}
	
	contflag = 1; 
}

static int signalshed_sigdelay_detect(void)
{
	struct timeval tval;
	unsigned long interval;
	
	contflag = 0;
	
	signal(SIGUSR1, signalshed_sighandler);
	signal(SIGALRM, signalshed_sighandler);
	
	alarm(1);
	
	gettimeofday(&tval, NULL);
	starttime = tval.tv_sec*1000*1000 + tval.tv_usec;
	
	raise(SIGUSR1);
	
	while (!contflag) { };
	
	signal(SIGUSR1, SIG_DFL);
	signal(SIGALRM, SIG_IGN);
	
	if (endtime != 0) {
		interval = endtime - starttime;
	} else {
		interval = 0;
	}
	
	return interval;
}

void signalshed_init(signalshed_t *signalshed, key_t svkey, int maxprocesses)
{
	unsigned short semvalues[1];
	union semun semarg;
	int sigdelay;
	
	*signalshed = (signalshed_t)mmap(NULL, SHMSIZE(maxprocesses), PROT_READ|PROT_WRITE, MAP_SHARED|MAP_ANONYMOUS, 0, 0);
	memset(*signalshed, 0, SHMSIZE(maxprocesses));
	
	SEMID(*signalshed) = semget(svkey, 1, 0666 | IPC_CREAT);
	MAXCLIENTS(*signalshed) = maxprocesses;
	
	sigdelay = signalshed_sigdelay_detect();
	if (sigdelay == 0) {
		sigdelay = 100;
	} else {
		sigdelay = sigdelay * 5;
	}
	SIGDELAY(*signalshed) = sigdelay;
	
	semvalues[0] = 0;
	
	semarg.array = (unsigned short *)&semvalues;
	if (semctl(SEMID(*signalshed), 0, SETALL, semarg) == -1) {
		perror("[SIGNAL SHEDULER] semctl call:");
	}
	
}

void signalshed_destroy(signalshed_t signalshed)
{
	int semid, clients;
	int i;
	
	/*for (i=MAXCLIENTS(signalshed); i>0; i--) {
		if ((CLIENTS(signalshed, i) > 0) && (kill(CLIENTS(signalshed, i), 0) == 0)) {
			kill(CLIENTS(signalshed, i), SIGTERM);
		}
	}*/
	
	semid = SEMID(signalshed);
	clients = MAXCLIENTS(signalshed);
	semctl(semid, 1, IPC_RMID);
	munmap(signalshed, (size_t)(SHMSIZE(clients)));
}

void signalshed_add_client(signalshed_t signalshed, pid_t pid, int seq)
{
	int i;
	
	if (seq < 0) {
		for (i=0; i<MAXCLIENTS(signalshed); i++) {
			if (CLIENTS(signalshed, i) == 0) {
				CLIENTS(signalshed, i) = pid;
			}
		}
	} else {
		CLIENTS(signalshed, seq) = pid;
	}
}

void signalshed_remove_client(signalshed_t signalshed, pid_t pid, int seq)
{
	int i;
	int res = 0;
	union semun semarg;
		
	if (seq < 0) {
		for (i=0; i<MAXCLIENTS(signalshed); i++) {
			if (CLIENTS(signalshed, i) == pid) {
				CLIENTS(signalshed, i) = 0;
				res = 1;
			}
		}
	} else {
		CLIENTS(signalshed, seq) = 0;
		res = 1;
	}
	
	if (res) {
		if (semctl(SEMID(signalshed), 0, GETZCNT, semarg) > 0) {
			semarg.val = 0;
			if (semctl(SEMID(signalshed), 0, SETVAL, semarg) == -1) {
				perror("[SIGNAL SHEDULER] semctl call:");
			}
		}
	}
}

void signalshed_wait_value(signalshed_t signalshed, int val)
{
	union semun semarg;
	
	if (val <= 0) return;
	
	semarg.val = val;
	if (semctl(SEMID(signalshed), 0, SETVAL, semarg) == -1) {
		perror("[SIGNAL SHEDULER] semctl call:");
	}
}

void signalshed_wait(signalshed_t signalshed)
{
	struct sembuf buf[1];
	struct timespec ts;
	
	ts.tv_sec = 0;
	ts.tv_nsec = 500000;
	
	buf[0].sem_num = 0;
	buf[0].sem_op = 0;
	buf[0].sem_flg = 0;
	
	/*#ifdef linux
		if (semtimedop(SEMID(signalshed), (struct sembuf*)&buf, 1, &ts) == -1) {
			perror("[SIGNAL SHEDULER] semop call:");
		}
	#else*/
		if (semop(SEMID(signalshed), (struct sembuf*)&buf, 1) == -1) {
			perror("[SIGNAL SHEDULER] semop call:");
		}
	/*#endif*/
}

void signalshed_post(signalshed_t signalshed)
{
	union semun semarg;
	
	semarg.val = 0;
	if (semctl(SEMID(signalshed), 0, SETVAL, semarg) == -1) {
		perror("[SIGNAL SHEDULER] semctl call:");
	}
}

void signalshed_continue(signalshed_t signalshed)
{
	int i;
	struct timespec sleeptime = {.tv_sec = 0, .tv_nsec = 50 * 1000 };	
	
	sleeptime.tv_sec = 0;
	sleeptime.tv_nsec = SIGDELAY(signalshed) * 1000;
		
	for (i=0; i<MAXCLIENTS(signalshed); i++) {
		if ((CLIENTS(signalshed, i) > 0) && (kill(CLIENTS(signalshed, i), 0) == 0)) {
			kill(CLIENTS(signalshed, i), SIGCONT);
		}
	}
	
	nanosleep(&sleeptime, NULL);
}

int signalshed_main(signalshed_t signalshed, int *index, unsigned int *errorcount, unsigned int *idlecount, int (*send_cb)(int *index, unsigned int *errorcount, unsigned int *idlecount))
{
	sigset_t sigmask;
	int signum;
	
	if ((send_cb == NULL) || (*index <= 0)) return 1;
	
	sigemptyset(&sigmask);
	sigaddset(&sigmask, SIGCONT);
	sigaddset(&sigmask, SIGTERM);
	sigaddset(&sigmask, SIGINT);
	
	if (sigwait(&sigmask, &signum) == 0) {
		if (signum == SIGCONT) {
			send_cb(index, errorcount, idlecount);
		} else if ((signum == SIGTERM) || (signum == SIGINT)) {
			return 1;
		}
	}
	
	return 0;
}
