/*
 *      streamserver.c
 *      
 *      Copyright 2009-2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


#include <unistd.h>
#include <signal.h>
#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/select.h>
#include <sys/poll.h>
#include <string.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/msg.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <time.h>
#include <limits.h>
#include <stdarg.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <sys/utsname.h>
#include <sys/statvfs.h>

#ifdef linux
	#include <sys/prctl.h>
	#include <linux/prctl.h>
#endif

#include "csadec.h"
#include "saferwlocks.h"
#include "signalshed.h"
#include "streamserver.h"
#include "svrsettings.h"
#include "transportstream.h"
#include "../common/serialize.h"
#include "../common/sysutils.h"
#include "../common/log.h"
#include "../resources.h"


#ifndef INFTIM
	#define INFTIM -1
#endif

union semun {
	int val;
	struct semid_ds *buf;
	unsigned short *array;
};

enum {
	FDTRANSPORT_START_RECORD = 0,
	FDTRANSPORT_STOP_RECORD,
	FDTRANSPORT_CHANGE_DEVICE,
	FDTRANSPORT_NEW_TCP_CLIENT,
	FDTRANSPORT_LOG_MESSAGE
	/*FDTRANSPORT_SERIVCE_ACTIVATION*/
};

#define SERVICE_MESSAGE_ACTIVATION  1

struct svcmsg_t {
	long mtype;
	pid_t svcpid;
	int svcrunning;
};

struct logmsg_t {
	int flags;
	char message[4000];
};

//Prototypes
static int add_client(struct sclient *client);
static int remove_client(pid_t pid);
static void streamserver_cleanup_client_slots(void);
static int transmission_process_send(int *index, unsigned int *errorcount, unsigned int *idlecount);
static void transmission_process_signal_handler(int sig, siginfo_t *si, void *unused);
static void transmission_process(struct sclient *client);
static int streamserver_find_pid(unsigned short pid);

static int streamserver_get_client(pid_t pid, struct sclient *client);

static int streamserver_fdtransport_send_message(int fd, int cmd, int ctlfd, void *data, size_t size);
static int streamserver_fdtransport_recv_message(int fd, int *cmd, int *ctlfd, void *data, size_t size);

static void streamserver_tcp_listen_process_signal_handler(int sig, siginfo_t *si, void *unused);

static int streamserver_recv_log_message(struct logmsg_t *msg);
static void streamserver_main_recv_message(int fd);

static int streamserver_lock_file(const char *homedir, int *runpid, int killdaemon, int daemonize);

static void main_signal_handler(int sig, siginfo_t *si, void *unused);
static void streamserver_start_daemon(void(*main_callback)(void));

static long int streamserver_stop_clients(void);


rwlockshed_t rwlockshed;
signalshed_t signalshed;

int shmid, clisemid;
struct memblock *procmblock;
int fdtransport[2];
struct log_t *mainlog;
volatile int readconfflag = 0, readsockflag = 0, cleanupflag = 0, exitflag = 0;
volatile int tmexitflag = 0;
volatile int tlpexitflag = 0;
volatile int rpexitflag = 0;


int streamserver_client_operation_lock(int clisemid)
{
	int res;
	struct sembuf semoperations[1];
	
	res = 1;
	
	semoperations[0].sem_num = 0;
	semoperations[0].sem_op = -1;
	semoperations[0].sem_flg = 0;
	
	if (semop(clisemid, semoperations, 1) == -1) {
		res = 0;
		log_add_message(mainlog, LOG_EVENT_ERROR, "Client operation lock semop call");
	}
	
	return res;
}

int streamserver_client_operation_unlock(int clisemid)
{
	int res;
	struct sembuf semoperations[1];
	
	res = 1;
	
	semoperations[0].sem_num = 0;
	semoperations[0].sem_op =  1;
	semoperations[0].sem_flg = 0;
	
	if (semop(clisemid, semoperations, 1) == -1) {
		res = 0;
		log_add_message(mainlog, LOG_EVENT_ERROR, "Client operation unlock semop call");
	}
	
	return res;
}

static int add_client(struct sclient *client)
{
	int i;
	int resindex;
	int lockres;
	char abuf[64];
	char processname[16];
	
	resindex = -1;
	memset(processname, 0, sizeof(processname));
	
	lockres = streamserver_client_operation_lock(clisemid);
	
	switch (client->type) {
		case CLIENT_TYPE_READ_HELPER:
			procmblock->clients[READ_PROCESS_INDEX].type = client->type;
			procmblock->clients[READ_PROCESS_INDEX].family = client->family;
			procmblock->clients[READ_PROCESS_INDEX].pid = client->pid;
			procmblock->clients[READ_PROCESS_INDEX].fdtransport = fdtransport[1];
			procmblock->clients[READ_PROCESS_INDEX].socketfd = client->socketfd;
			procmblock->clients[READ_PROCESS_INDEX].pidfilter = 0;
			procmblock->clients[READ_PROCESS_INDEX].transmittedbytes = 0;
			procmblock->clients[READ_PROCESS_INDEX].ready = 0;
			procmblock->clients[READ_PROCESS_INDEX].starttime = time(NULL);
			if (client->useragent[0] != '\0') strncpy((char *)&procmblock->clients[READ_PROCESS_INDEX].useragent, client->useragent, strlen(client->useragent)); else memset(&procmblock->clients[READ_PROCESS_INDEX].useragent, 0, sizeof(procmblock->clients[READ_PROCESS_INDEX].useragent));
			memset(procmblock->clients[READ_PROCESS_INDEX].ifname, 0, sizeof(procmblock->clients[LISTEN_PROCESS_INDEX].ifname));
			snprintf(processname, sizeof(processname), "gsfstrsvr/read");
			log_add_message(mainlog, LOG_EVENT_MESSAGE, "READ PROCESS started (ID: %i/%i)", client->pid, READ_PROCESS_INDEX);
			resindex = READ_PROCESS_INDEX;
			break;
		
		case CLIENT_TYPE_TCP_LISTEN_HELPER:
			procmblock->clients[LISTEN_PROCESS_INDEX].type = client->type;
			procmblock->clients[LISTEN_PROCESS_INDEX].family = client->family;
			procmblock->clients[LISTEN_PROCESS_INDEX].pid = client->pid;
			procmblock->clients[LISTEN_PROCESS_INDEX].fdtransport = fdtransport[1];
			procmblock->clients[LISTEN_PROCESS_INDEX].socketfd = client->socketfd;
			procmblock->clients[LISTEN_PROCESS_INDEX].pidfilter = 0;
			procmblock->clients[LISTEN_PROCESS_INDEX].transmittedbytes = 0;
			procmblock->clients[LISTEN_PROCESS_INDEX].ready = 0;
			procmblock->clients[LISTEN_PROCESS_INDEX].starttime = time(NULL);
			if (client->useragent[0] != '\0') strncpy((char *)&procmblock->clients[LISTEN_PROCESS_INDEX].useragent, client->useragent, strlen(client->useragent)); else memset(&procmblock->clients[LISTEN_PROCESS_INDEX].useragent, 0, sizeof(procmblock->clients[LISTEN_PROCESS_INDEX].useragent));
			memset(procmblock->clients[LISTEN_PROCESS_INDEX].ifname, 0, sizeof(procmblock->clients[LISTEN_PROCESS_INDEX].ifname));
			snprintf(processname, sizeof(processname), "gsfstrsvr/tcplp");
			log_add_message(mainlog, LOG_EVENT_MESSAGE, "TCP LISTEN PROCESS started at port %d (ID: %i/%i)", ntohs(client->address.sin_port), client->pid, LISTEN_PROCESS_INDEX);
			resindex = LISTEN_PROCESS_INDEX;
			break;
		
		
		case CLIENT_TYPE_UDP_BROADCAST:
			procmblock->clients[UDP_BC_PROCESS_INDEX].type = client->type;
			procmblock->clients[UDP_BC_PROCESS_INDEX].family = client->family;
			procmblock->clients[UDP_BC_PROCESS_INDEX].pid = client->pid;
			procmblock->clients[UDP_BC_PROCESS_INDEX].fdtransport = fdtransport[1];
			procmblock->clients[UDP_BC_PROCESS_INDEX].socketfd = client->socketfd;
			procmblock->clients[UDP_BC_PROCESS_INDEX].pidfilter = 0;
			procmblock->clients[UDP_BC_PROCESS_INDEX].transmittedbytes = 0;
			procmblock->clients[UDP_BC_PROCESS_INDEX].ready = 0;
			procmblock->clients[UDP_BC_PROCESS_INDEX].starttime = time(NULL);
			procmblock->clients[UDP_BC_PROCESS_INDEX].family = CLIENT_FAMILY_V4_TRANSPORT;
			if (client->useragent[0] != '\0') strncpy((char *)&procmblock->clients[UDP_BC_PROCESS_INDEX].useragent, client->useragent, strlen(client->useragent)); else memset(&procmblock->clients[UDP_BC_PROCESS_INDEX].useragent, 0, sizeof(procmblock->clients[UDP_BC_PROCESS_INDEX].useragent));
			memset(procmblock->clients[UDP_BC_PROCESS_INDEX].ifname, 0, sizeof(procmblock->clients[UDP_BC_PROCESS_INDEX].ifname));
			procmblock->numclients++;
			close(fdtransport[1]);
			snprintf(processname, sizeof(processname), "gsfstrsvr/udpbc");
			log_add_message(mainlog, LOG_EVENT_MESSAGE, "UDP BROADCAST started %s:%d (ID: %i/%i)", inet_ntop(AF_INET, (struct in_addr *)&(client->address.sin_addr), abuf, sizeof(abuf)), ntohs(client->address.sin_port), client->pid, UDP_BC_PROCESS_INDEX);
			resindex = UDP_BC_PROCESS_INDEX;
			break;
		
		
		case CLIENT_TYPE_UDP_MULTICAST:
			procmblock->clients[UDP_MC_PROCESS_INDEX].type = client->type;
			procmblock->clients[UDP_MC_PROCESS_INDEX].family = client->family;
			procmblock->clients[UDP_MC_PROCESS_INDEX].pid = client->pid;
			procmblock->clients[UDP_MC_PROCESS_INDEX].fdtransport = fdtransport[1];
			procmblock->clients[UDP_MC_PROCESS_INDEX].socketfd = client->socketfd;
			procmblock->clients[UDP_MC_PROCESS_INDEX].pidfilter = 0;
			procmblock->clients[UDP_MC_PROCESS_INDEX].transmittedbytes = 0;
			procmblock->clients[UDP_MC_PROCESS_INDEX].ready = 0;
			procmblock->clients[UDP_MC_PROCESS_INDEX].starttime = time(NULL);
			procmblock->clients[UDP_MC_PROCESS_INDEX].family = CLIENT_FAMILY_V4_TRANSPORT;
			if (client->useragent[0] != '\0') strncpy((char *)&procmblock->clients[UDP_MC_PROCESS_INDEX].useragent, client->useragent, strlen(client->useragent)); else memset(&procmblock->clients[UDP_MC_PROCESS_INDEX].useragent, 0, sizeof(procmblock->clients[UDP_MC_PROCESS_INDEX].useragent));
			memset(procmblock->clients[UDP_MC_PROCESS_INDEX].ifname, 0, sizeof(procmblock->clients[UDP_MC_PROCESS_INDEX].ifname));
			procmblock->numclients++;
			close(fdtransport[1]);
			snprintf(processname, sizeof(processname), "gsfstrsvr/udpmc");
			log_add_message(mainlog, LOG_EVENT_MESSAGE, "UDP MULTICAST started %s:%d (ID: %i/%i)", inet_ntop(AF_INET, (struct in_addr *)&(client->address.sin_addr), abuf, sizeof(abuf)), ntohs(client->address.sin_port), client->pid, UDP_MC_PROCESS_INDEX);
			resindex = UDP_MC_PROCESS_INDEX;
			break;
		
		
		case CLIENT_TYPE_TCP_UNICAST:
			for (i=TCP_FIRST_CLIENT; i<MAXCLIENTS; i++) {
				if (procmblock->clients[i].pid == 0) {
					procmblock->clients[i].type = client->type;
					procmblock->clients[i].family = client->family;
					procmblock->clients[i].pid = client->pid;
					procmblock->clients[i].fdtransport = fdtransport[1];
					procmblock->clients[i].socketfd = client->socketfd;
					procmblock->clients[i].ready = 0;
					procmblock->clients[i].starttime = time(NULL);
					procmblock->clients[i].family = CLIENT_FAMILY_V6_TRANSPORT;
					procmblock->clients[i].transmittedbytes = client->transmittedbytes;
					memcpy(&procmblock->clients[i].addressv6, &client->addressv6, sizeof(struct sockaddr_in6));
					if (client->useragent[0] != '\0') strncpy((char *)&procmblock->clients[i].useragent, client->useragent, strlen(client->useragent)); else memset(&procmblock->clients[i].useragent, 0, sizeof(procmblock->clients[i].useragent));
					if ((client->pidfilter >= 32) && (client->pidfilter <= 8191)) procmblock->clients[i].pidfilter = client->pidfilter; else procmblock->clients[i].pidfilter = 0;
					procmblock->numclients++;
					procmblock->numtcpclients++;
					close(fdtransport[1]);
					snprintf(processname, sizeof(processname), "gsfstrsvr/tcpuc");
					log_add_message(mainlog, LOG_EVENT_MESSAGE, "TCP UNICAST started with client %s:%d (ID: %i/%i)", inet_ntop(AF_INET6, (struct in_addr6 *)&(client->addressv6.sin6_addr), abuf, sizeof(abuf)), ntohs(client->addressv6.sin6_port), client->pid, i);
					resindex = i;
					break;
				}
			}
			break;
		
		default: 
			resindex = -1;
			break;
	}
	
	if (lockres) streamserver_client_operation_unlock(clisemid);
	
	#ifdef linux
		if (resindex != -1) prctl(PR_SET_NAME, (char *)&processname, 0, 0, 0);
	#endif
	
	if ((procmblock->sheduler == SHEDULER_SIGNAL) && (resindex != -1) && (resindex != LISTEN_PROCESS_INDEX)) 
		signalshed_add_client(signalshed, client->pid, resindex);
	
	return resindex;
}

static int remove_client(pid_t pid)
{
	int i;
	int lockres;
	char abuf[64];
	
	for (i=0; i<MAXCLIENTS; i++) {
		if (procmblock->clients[i].pid == pid) {
			
			lockres = streamserver_client_operation_lock(clisemid);
			
			switch (i) {
				case READ_PROCESS_INDEX: {
					log_add_message(mainlog, LOG_EVENT_MESSAGE, "READ PROCESS stopped (ID: %i/%i)", procmblock->clients[i].pid, i);
					break;
				}
				
				case LISTEN_PROCESS_INDEX: {
					log_add_message(mainlog, LOG_EVENT_MESSAGE, "TCP LISTEN PROCESS stopped at port %d (ID: %i/%i)", ntohs(procmblock->clients[i].address.sin_port), procmblock->clients[i].pid, i);
					break;
				}
				
				case UDP_BC_PROCESS_INDEX: {
					log_add_message(mainlog, LOG_EVENT_MESSAGE, "UDP BROADCAST stopped %s:%d (ID: %i/%i)", inet_ntop(AF_INET, (struct in_addr *)&(procmblock->clients[i].address.sin_addr), abuf, sizeof(abuf)), ntohs(procmblock->clients[i].address.sin_port), procmblock->clients[i].pid, i);
					break;
				}
				
				case UDP_MC_PROCESS_INDEX: {
					log_add_message(mainlog, LOG_EVENT_MESSAGE, "UDP MULTICAST stopped %s:%d (ID: %i/%i)", inet_ntop(AF_INET, (struct in_addr *)&(procmblock->clients[i].address.sin_addr), abuf, sizeof(abuf)), ntohs(procmblock->clients[i].address.sin_port), procmblock->clients[i].pid, i);
					break;				
				}
				
				default: {
					log_add_message(mainlog, LOG_EVENT_MESSAGE, "TCP UNICAST connection closed with client %s:%d (ID: %i/%i)", inet_ntop(AF_INET6, (struct in_addr6 *)&(procmblock->clients[i].addressv6.sin6_addr), abuf, sizeof(abuf)), ntohs(procmblock->clients[i].addressv6.sin6_port), procmblock->clients[i].pid, i);
					memset(&procmblock->clients[i].addressv6, 0, sizeof(struct sockaddr_in6));
					procmblock->numtcpclients--;
					break;
				}
			}
			
			//procmblock->clients[i].pid = 0;
			procmblock->clients[i].fdtransport = 0;
			procmblock->clients[i].family = CLIENT_FAMILY_STOPPED;
			procmblock->clients[i].pidfilter = 0;
			procmblock->clients[i].socketfd = 0;
			//procmblock->clients[i].starttime = 0;
			procmblock->clients[i].transmittedbytes = 0;
			memset(&procmblock->clients[i].useragent, 0, sizeof(procmblock->clients[i].useragent));
			
			if ((i > 0) && (i != LISTEN_PROCESS_INDEX)) {
				procmblock->numclients--;
				procmblock->readyclients--;
			}
			
			if (lockres) streamserver_client_operation_unlock(clisemid);
			
			if (procmblock->sheduler == SHEDULER_SIGNAL) signalshed_remove_client(signalshed, pid, i);
			
			return i;
		}
	}
	
	return -1;
}

static void streamserver_cleanup_client_slots(void)
{
	int i, status;
	
	for (i=0; i<MAXCLIENTS; i++) {
		if ((procmblock->clients[i].pid > 0) && ((procmblock->clients[i].family == CLIENT_FAMILY_STOPPED) || (!sysutils_process_exists(procmblock->clients[i].pid) /*kill(procmblock->clients[i].pid, 0) != 0*/))) {
			if (!sysutils_process_exists(procmblock->clients[i].pid)) {
				remove_client(procmblock->clients[i].pid);
			}
			if (waitpid(procmblock->clients[i].pid, &status, WNOHANG) != -1) {
				log_add_message(mainlog, LOG_EVENT_MESSAGE, "Freed resource slot of process (ID: %i/%i)", procmblock->clients[i].pid, i);
				procmblock->clients[i].pid = 0;
				procmblock->clients[i].starttime = 0;
			}
		}
	}
}

unsigned int streamserver_get_free_tcp_client_slots(void)
{
	int i;
	unsigned int count = 0;
	
	for (i=TCP_FIRST_CLIENT; i<MAXCLIENTS; i++) {
		if (procmblock->clients[i].pid == 0) {
			count++;
		}
	}
	
	return count;
}

static int transmission_process_send(int *index, unsigned int *errorcount, unsigned int *idlecount)
{
	int socketfd;
	int bufptr;
	unsigned char recvbuffer[128];
	ssize_t movedbytes;
	int errorcode;
	int lockres;
	unsigned char dmxbuffer[BUFSIZE];
	unsigned int dmxbytes;
	unsigned int bytes;
	
	unsigned char packetbuf[188];
	unsigned int bytesread, tsstart;
	uint16_t pid;
	uint8_t payload;
	uint8_t priority;
	uint8_t adfield;
	uint8_t counter;
	uint32_t adfield_length;
	unsigned int packstart, pesfound, pesstart, pesoffset, peslen, peshdr, j;
	
	
	
	if (procmblock->clients[*index].pid == getpid()) {
		errno = 0;
		if (recv(procmblock->clients[*index].socketfd, recvbuffer, sizeof(recvbuffer), 0) <= 0) {
			if (errno != EWOULDBLOCK) {
				tmexitflag = 1;
				goto finalizeclient;
			}
		}
		
		bufptr = 0;
		
		switch(*index) {
			case LISTEN_PROCESS_INDEX: {
				goto finalizeclient;
				break;
			}
			
			case UDP_BC_PROCESS_INDEX: {
				do {
					if ((movedbytes = sendto(procmblock->clients[*index].socketfd, &procmblock->buffer[bufptr], SENDSIZE, 0, (struct sockaddr *)&procmblock->clients[*index].address, sizeof(struct sockaddr))) <= 0) {
						*errorcount++;
						errorcode = errno;
						log_add_message(mainlog, LOG_EVENT_ERROR, "UDP BROADCAST error while sendto call: %s", strerror(errno));
						goto finalizeclient;
					}
					bufptr = bufptr + SENDSIZE;
				} while (bufptr != BUFSIZE);
				break;				
			}
			
			case UDP_MC_PROCESS_INDEX: {
				do {
					if ((movedbytes = sendto(procmblock->clients[*index].socketfd, &procmblock->buffer[bufptr], SENDSIZE, 0, (struct sockaddr *)&procmblock->clients[*index].address, sizeof(struct sockaddr))) <= 0) {
						*errorcount++;
						errorcode = errno;
						log_add_message(mainlog, LOG_EVENT_ERROR, "UDP MULTICAST error while sendto call: %s", strerror(errno));
						goto finalizeclient;
					}
					bufptr = bufptr + SENDSIZE;
				} while (bufptr != BUFSIZE);
				break;
			}
			
			default: {
				if (procmblock->clients[*index].pidfilter != 0) {
					//log_add_message(mainlog, LOG_EVENT_ERROR, "PID FILTER %i", procmblock->clients[*index].pidfilter);
					
					/*dmxbytes = transportstream_demux_pes((unsigned char *)&procmblock->buffer[0], sizeof(procmblock->buffer), 
																(unsigned char *)&dmxbuffer[0], sizeof(dmxbuffer), procmblock->clients[*index].pidfilter);*/
					
					dmxbytes = 0;
					bytesread = 0;
					
					while (bytesread < sizeof(procmblock->buffer)) {
						memcpy(&packetbuf[0], &procmblock->buffer[bytesread], PACKETSIZE);
						bytesread += PACKETSIZE;
						
						payload = (packetbuf[1] & 0x40) >> 6;
						priority = (packetbuf[1] & 0x20) >> 5;
						pid = ((packetbuf[1] & 0x1f) << 8) | packetbuf[2];
						adfield = (packetbuf[3] & 0x30) >> 4;
						counter = (packetbuf[3] & 0x0f);
						
						if (adfield == 0x3) {
							adfield_length = packetbuf[4] + 1;
						} else {
							adfield_length = 0;
						}
						
						tsstart = 4 + adfield_length;
						
						if (pid == procmblock->clients[*index].pidfilter) {
							if (!payload) {
								memcpy(&dmxbuffer[dmxbytes], &packetbuf[tsstart], PACKETSIZE-tsstart);
								dmxbytes += PACKETSIZE-tsstart;
							} else {
								peslen = 0;
								peshdr = 0;
								pesfound = 0;
								
								for (pesstart=tsstart; pesstart<PACKETSIZE-6; pesstart++) {
									if ((packetbuf[pesstart] == 0x00) && (packetbuf[pesstart+1] == 0x00) && (packetbuf[pesstart+2] == 0x01)) {
										peslen = (((packetbuf[pesstart+4] & 0xff) << 8) | packetbuf[pesstart+5]) & 0xffff;
										pesfound = 1;
										break;
									}
								}
								
								if (!pesfound) {
									memcpy(&dmxbuffer[dmxbytes], &packetbuf[tsstart], PACKETSIZE-tsstart);
									dmxbytes += PACKETSIZE-tsstart;
								} else {
									pesoffset = tsstart-pesstart;
									
									if ((packetbuf[pesstart + 6] & 0xC0) == 0x80) {
										peshdr = packetbuf[pesstart + 8] + 9;
									} else {
										for (j=pesstart+6; j<PACKETSIZE; j++) {
											if (packetbuf[j] != 0xFF) break;
										}
										if ((packetbuf[j] & 0xC0) == 0x40) {
											j += 2;
										}
										if ((packetbuf[j] & 0xF0) == 0x20) {
											j += 5;
										} else if ((packetbuf[j] & 0xF0) == 0x30) {
											j += 10;
										} else {
											j++;
										}
										if (j < PACKETSIZE) {
											peshdr = j;
										}
									}
									
									memcpy(&dmxbuffer[dmxbytes], &packetbuf[tsstart+pesoffset+peshdr], PACKETSIZE-tsstart-pesoffset-peshdr);
									dmxbytes += PACKETSIZE-tsstart-pesoffset-peshdr;
								}
							}
						}
					}
					
					
					if (dmxbytes > 0) {
						bufptr = 0;
						while (bufptr < dmxbytes) {
							if (dmxbytes - bufptr > SENDSIZE) bytes = SENDSIZE; else bytes = dmxbytes - bufptr;
							if ((movedbytes = send(procmblock->clients[*index].socketfd, &dmxbuffer[bufptr], bytes, 0)) <= 0) {
								if (errno != EWOULDBLOCK) {
									*errorcount++;
									errorcode = errno;
									log_add_message(mainlog, LOG_EVENT_ERROR, "TCP UNICAST error while send call: %s", strerror(errno));
									goto finalizeclient;
								}
							}
							bufptr = bufptr + bytes;
						}
						break;
					} else {
						log_add_message(mainlog, LOG_EVENT_ERROR, "DMX Buffer is empty");
						break;
					}
				} else {
					do {
						if (BUFSIZE - bufptr >= SENDSIZE) bytes = SENDSIZE; else bytes = BUFSIZE - bufptr;
						if ((movedbytes = send(procmblock->clients[*index].socketfd, &procmblock->buffer[bufptr], bytes/*SENDSIZE*/, 0)) <= 0) {
							if (errno != EWOULDBLOCK) {
								*errorcount++;
								errorcode = errno;
								log_add_message(mainlog, LOG_EVENT_ERROR, "TCP UNICAST error while send call: %s", strerror(errno));
								goto finalizeclient;
							}
						}
						bufptr = bufptr + bytes;
					} while (bufptr != BUFSIZE);
				}
			}
		}
		
		procmblock->clients[*index].transmittedbytes += movedbytes;
		procmblock->clients[*index].ready = 1;
	}
		
	finalizeclient:
	
	lockres = streamserver_client_operation_lock(clisemid);
	
	procmblock->readyclients++;
	
	if (lockres) streamserver_client_operation_unlock(clisemid);
		
	if (movedbytes > 0) {
		*idlecount = 0;
	} else if (errorcode != EWOULDBLOCK) {
		*idlecount++;
	}
	
	if ((procmblock->sheduler == SHEDULER_SIGNAL) && (procmblock->readyclients >= procmblock->numclients)) {
		signalshed_post(signalshed);
	}
}

static void transmission_process_signal_handler(int sig, siginfo_t *si, void *unused)
{
	if ((sig == SIGINT) || (sig == SIGTERM)) {
		tmexitflag = 1;
	}
}

static void transmission_process(struct sclient *client)
{
	pid_t pid;
	struct sigaction sa;
	int flagsval;
	int index;
	unsigned int errorcounter;
	unsigned int idlecounter;
	int dsignal = SIGTERM;
	
	
	if ((pid = fork()) < 0) {
		//error
		log_add_message(mainlog, LOG_EVENT_ERROR, "Transmission process in fork system call: %s", strerror(errno));
		close(client->socketfd);
	} else if (pid == 0) {
		switch (client->type) {
			case CLIENT_TYPE_UDP_BROADCAST:
				log_add_message(mainlog, LOG_EVENT_MESSAGE, "UDP BROADCAST transmission process started with PID: %i", getpid());
				break;
			
			case CLIENT_TYPE_UDP_MULTICAST:
				log_add_message(mainlog, LOG_EVENT_MESSAGE, "UDP MULTICAST transmission process started with PID: %i", getpid());
				break;
			
			case CLIENT_TYPE_TCP_UNICAST:
				log_add_message(mainlog, LOG_EVENT_MESSAGE, "TCP UNICAST transmission process started with PID: %i", getpid());
				break;
				
			default:
				log_add_message(mainlog, LOG_EVENT_MESSAGE, "Special transmission process started with PID: %i", getpid());
				break;
		}
		
		
		close(fdtransport[0]);
		
		client->pid = getpid();
		
		errorcounter = 0;
		idlecounter = 0;
		
		sa.sa_flags = 0;
		sigemptyset(&sa.sa_mask);
		sa.sa_sigaction = transmission_process_signal_handler;
		
		if (sigaction(SIGCONT, &sa, NULL) == -1) {
			log_add_message(mainlog, LOG_EVENT_ERROR, "Error calling SIGCONT handler procedure in client process");
		}
		
		if (sigaction(SIGINT, &sa, NULL) == -1) {
			log_add_message(mainlog, LOG_EVENT_ERROR, "Error calling SIGINT handler procedure in client process");
		}
		
		if (sigaction(SIGTERM, &sa, NULL) == -1) {
			log_add_message(mainlog, LOG_EVENT_ERROR, "Error calling SIGTERM handler procedure in client process");
		}
		
		#ifdef linux
			prctl(PR_SET_PDEATHSIG, dsignal, 0, 0, 0);
		#endif
		
		flagsval = fcntl(client->socketfd, F_GETFL, 0);
		fcntl(client->socketfd, F_SETFL, flagsval | O_NONBLOCK);
		
		if ((index = add_client(client)) == -1) {
			log_add_message(mainlog, LOG_EVENT_ERROR, "Failed to add client info into storage");
			close(client->socketfd);
			exit(EXIT_FAILURE);
		}
		
		while (!tmexitflag) {
			if (procmblock->sheduler == SHEDULER_SIGNAL) {
				tmexitflag |= signalshed_main(signalshed, &index, &errorcounter, &idlecounter, transmission_process_send);
			} else if (procmblock->sheduler == SHEDULER_RWLOCK) {
				rwlockshed_readlock(rwlockshed);
				transmission_process_send(&index, &errorcounter, &idlecounter);
				rwlockshed_readunlock(rwlockshed);
			}
			
			if (errorcounter >= ERROR_COUNT_LIMIT) {
				log_add_message(mainlog, LOG_EVENT_ERROR, "Unable to establish connection with client (ID: %i/%i)", client->pid, index);
				tmexitflag = 1;
			}
			
			if (idlecounter >= IDLE_COUNT_LIMIT) {
				log_add_message(mainlog, LOG_EVENT_ERROR, "Client not recieving transmitted data (ID: %i/%i)", client->pid, index);
				tmexitflag = 1;
			}
		}
		
		close(client->socketfd);
		
		if (remove_client(client->pid) == -1) {
			log_add_message(mainlog, LOG_EVENT_ERROR, "No client removed - seems like internal error");
		}
		
		exit(EXIT_SUCCESS);
	}
}

static int parse_request(char *request, char *buffer, size_t size, char *start, char *end)
{
	char *first, *second;
	int numsym;
	
	memset(buffer, 0, size);
	if ((first = strstr(request, start)) == NULL) return -1;
	if ((second = strstr(first+1, end)) == NULL) return -1;
	
	if ((numsym = second - first - strlen(start)) <= 0) return 0;
	if (numsym >= size - 1) numsym = size - 1;
	
	strncpy(buffer, first + strlen(start), numsym);
	buffer[numsym] = '\0';

	return numsym;
}

static int streamserver_find_pid(unsigned short pid)
{
	int i, pidfound;
	
	if ((procmblock->hostpid == 0) || (!sysutils_process_exists(procmblock->hostpid))) {
		if (procmblock->tsinfo.pids[pid].stream_type != TS_STREAM_UNKNOWN) {
			return 1;
		} else {
			return 0;
		}
	}
	
	pidfound = 0;
	for (i=0; i<256; i++) {
		if (procmblock->streampids[i].value == 0) break;
		if (procmblock->streampids[i].value == pid) {
			if (procmblock->streampids[i].enabled) {
				if (procmblock->tsinfo.pids[pid].stream_type != TS_STREAM_UNKNOWN) {
					pidfound = 1;
				} else {
					pidfound = 0;
				}
				break;
			}
		}
	}
	
	return pidfound;
}

void streamserver_stop_client(unsigned int clientid)
{
	if ((clientid < TCP_FIRST_CLIENT) || (clientid > MAXCLIENTS)) return;
	
	if (!sysutils_process_exists(procmblock->clients[clientid].pid)) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "Unable to stop client - not running");
		return;
	}
	
	kill(procmblock->clients[clientid].pid, SIGTERM);
}

static int streamserver_get_client(pid_t pid, struct sclient *client)
{
	int id, resid;
	
	if (!sysutils_process_exists(pid)) return -1;
	
	resid = -1;
	
	for (id=0; id<MAXCLIENTS; id++) {
		if (procmblock->clients[id].pid == pid) {
			if (client != NULL) {
				memcpy(client, &(procmblock->clients[id]), sizeof(struct sclient));
			}
			resid = id;
		}
	}
	
	return resid;
}

static int streamserver_fdtransport_send_message(int fd, int cmd, int ctlfd, void *data, size_t size)
{
	struct msghdr msg;
	char ccmsg[CMSG_SPACE(sizeof(int))];
	struct cmsghdr *cmsg;
	struct iovec vec[2];
	char cmdbuf[1];
	int res;
	
	cmdbuf[0] = (char)cmd;
	
	vec[0].iov_base = cmdbuf;
	vec[0].iov_len = 1;
	if ((data != NULL) && (size > 0)) {
		vec[1].iov_base = data;
		vec[1].iov_len = size;
		msg.msg_iovlen = 2;
	} else {
		msg.msg_iovlen = 1;
	}
	msg.msg_iov = vec;
	
	msg.msg_name = NULL;
	msg.msg_namelen = 0;
	
	if (ctlfd != -1) {
		msg.msg_control = ccmsg;
		msg.msg_controllen = sizeof(ccmsg);
		cmsg = CMSG_FIRSTHDR(&msg);
		cmsg->cmsg_level = SOL_SOCKET;
		cmsg->cmsg_type = SCM_RIGHTS;
		cmsg->cmsg_len = CMSG_LEN(sizeof(int));
		*(int*)CMSG_DATA(cmsg) = ctlfd;
		msg.msg_controllen = cmsg->cmsg_len;
	} else {
		msg.msg_control = NULL;
		msg.msg_controllen = 0;
	}
	msg.msg_flags = 0;
	
	res = sendmsg(fd, &msg, 0);
	
	return res;
}

static int streamserver_fdtransport_recv_message(int fd, int *cmd, int *ctlfd, void *data, size_t size)
{
	struct msghdr msg;
	struct iovec iov[2];
	char cmdbuf[1];
	char ccmsg[CMSG_SPACE(sizeof(int))];
	struct cmsghdr *cmsg;
	int res;
	
	if (cmdbuf == NULL) return -1;
	
	iov[0].iov_base = cmdbuf;
	iov[0].iov_len = 1;
	if ((data != NULL) && (size != 0)) {
		iov[1].iov_base = data;
		iov[1].iov_len = size;
		msg.msg_iovlen = 2;
	} else {
		msg.msg_iovlen = 1;
	}
	msg.msg_iov = iov;
	
	msg.msg_name = NULL;
	msg.msg_namelen = 0;
	
	if (ctlfd != NULL) {
		msg.msg_control = ccmsg;
		msg.msg_controllen = sizeof(ccmsg);
	} else {
		msg.msg_control = NULL;
		msg.msg_controllen = 0;
	}
	
	if ((res = recvmsg(fd, &msg, 0)) == -1) return res;
	
	*cmd = (int)cmdbuf[0];
	
	if (ctlfd != NULL) {
		*ctlfd = -1;
		if ((cmsg = CMSG_FIRSTHDR(&msg)) != NULL) {
			if (cmsg->cmsg_type == SCM_RIGHTS) {
				*ctlfd = *(int*)CMSG_DATA(cmsg);
			}
		}
	}
	
	return res;
}

void streamserver_tcp_listen_process_signal_handler(int sig, siginfo_t *si, void *unused)
{
	if ((sig == SIGINT) || (sig == SIGTERM)) {
		tlpexitflag = 1;
	}
}

void streamserver_init_tcp_listen_process(void)
{
	struct sigaction sa;
	sigset_t blockmask, oldmask;
	const int sflag = 1;
	int helperindex;
	//int dsignal = SIGTERM;
	int listenfdv6;
	struct pollfd sockets[1];
	int readysockets;
	socklen_t addressv6len = sizeof(struct sockaddr_in6);
	char recvbuffer[4096];
	struct sclient helperclient;
	pid_t pid;
	struct sclient client;
	char *contenttype;
	unsigned int bitrate;
	
	if ((pid = fork()) < 0) {
		//error
		log_add_message(mainlog, LOG_EVENT_ERROR, "TCP LISTEN PROCESS error in fork system call: %s", strerror(errno));
		exit(EXIT_FAILURE);
	} else if (pid == 0) {
		sa.sa_flags = SA_SIGINFO;
		sigemptyset(&sa.sa_mask);
		sa.sa_sigaction = streamserver_tcp_listen_process_signal_handler;
		
		if (sigaction(SIGTERM, &sa, NULL) == -1) {
			log_add_message(mainlog, LOG_EVENT_ERROR, "TCP LISTEN PROCESS error calling SIGCHLD handler procedure in listen thread");
		}
		
		if (sigaction(SIGINT, &sa, NULL) == -1) {
			log_add_message(mainlog, LOG_EVENT_ERROR, "TCP LISTEN PROCESS error calling SIGINT handler procedure in listen thread");
		}
		
		sigemptyset(&blockmask);
		sigaddset(&blockmask, SIGCONT);
		sigprocmask(SIG_BLOCK, &blockmask, &oldmask);
		
		/*#ifdef linux
			prctl(PR_SET_PDEATHSIG, dsignal, 0, 0, 0);
		#endif*/
				
		if ((listenfdv6 = socket(AF_INET6, SOCK_STREAM, 0)) == -1) {
			log_add_message(mainlog, LOG_EVENT_ERROR, "TCP LISTEN PROCESS failed to create listen socket for incoming TCP connections: %s", strerror(errno));
			exit(EXIT_FAILURE);
		}
			
		if (setsockopt(listenfdv6, SOL_SOCKET, SO_REUSEADDR, &sflag, sizeof(sflag)) < 0) {
			log_add_message(mainlog, LOG_EVENT_ERROR, "TCP LISTEN PROCESS failed to set socket flags: %s", strerror(errno));
			exit(EXIT_FAILURE);
		}
		
		if (bind(listenfdv6, (struct sockaddr *)&procmblock->clients[LISTEN_PROCESS_INDEX].addressv6, sizeof(struct sockaddr_in6)) == -1) {
			log_add_message(mainlog, LOG_EVENT_ERROR, "TCP LISTEN PROCESS failed to bind listen socket for incoming TCP connections: %s", strerror(errno));
			exit(EXIT_FAILURE);
		}
		
		helperclient.type = CLIENT_TYPE_TCP_LISTEN_HELPER;
		helperclient.family = CLIENT_FAMILY_HELPER;
		helperclient.pid = getpid();
		helperclient.socketfd = listenfdv6;
		memcpy(&helperclient.address, &procmblock->clients[LISTEN_PROCESS_INDEX].address, sizeof(struct sockaddr_in));
		memcpy(&helperclient.addressv6, &procmblock->clients[LISTEN_PROCESS_INDEX].addressv6, sizeof(struct sockaddr_in6));
		snprintf(helperclient.useragent, sizeof(helperclient.useragent), "TCP listen helper process");
		helperclient.pidfilter = 0;
		helperclient.transmittedbytes = 0;
		
		if ((helperindex = add_client(&helperclient)) == -1) {
			log_add_message(mainlog, LOG_EVENT_ERROR, "TCP LISTEN PROCESS failed to add helper client info into storage");
			close(helperclient.socketfd);
			exit(EXIT_FAILURE);
		}
		
		if (listen(listenfdv6, (int)procmblock->tcpqueue) == -1) {
			log_add_message(mainlog, LOG_EVENT_ERROR, "TCP LISTEN PROCESS failed to listen socket: %s", strerror(errno));
			exit(EXIT_FAILURE);
		}
				
		sockets[0].fd = listenfdv6;
		sockets[0].events = POLLIN;
		
		while (!tlpexitflag) {
			if ((readysockets = poll(sockets, 1, INFTIM)) > 0) {
				if (sockets[0].revents & POLLIN) {
					client.type = CLIENT_TYPE_TCP_UNICAST;
					client.family = CLIENT_FAMILY_V6_TRANSPORT;
					client.pid = 0;
					if ((client.socketfd = accept(listenfdv6, NULL, NULL)) == -1) {
						log_add_message(mainlog, LOG_EVENT_ERROR, "TCP LISTEN PROCESS connection closed");
						close(client.socketfd);
						continue;
					}
					
					if (getpeername(client.socketfd, (struct sockaddr *)&client.addressv6, &addressv6len) == -1) {
						log_add_message(mainlog, LOG_EVENT_ERROR, "TCP LISTEN PROCESS failed to get client address");
					}
					
					memset(recvbuffer, 0, sizeof(recvbuffer));
					if (recv(client.socketfd, recvbuffer, sizeof(recvbuffer), 0) <= 0) {
						log_add_message(mainlog, LOG_EVENT_ERROR, "TCP LISTEN PROCESS connection closed");
						close(client.socketfd);
						continue;
					} else {
						if (strstr((char *)&recvbuffer, HTTPREQUEST) != NULL) {
							if (parse_request((char *)&recvbuffer, (char *)&client.useragent, sizeof(client.useragent), HTTPREQUEST, " HTTP")) {
								client.pidfilter = atoi(client.useragent);
								if (!streamserver_find_pid(client.pidfilter)) {
									client.transmittedbytes = sprintf((char *)&recvbuffer, HTTPANSWER404);
									recvbuffer[client.transmittedbytes] = '\0';
									send(client.socketfd, recvbuffer, client.transmittedbytes, 0);
									log_add_message(mainlog, LOG_EVENT_ERROR, "TCP LISTEN PROCESS connection closed: unable to find requested resource");
									close(client.socketfd);
									continue;
								}
							} else {
								client.pidfilter = 0;
							}
							
							if (!parse_request((char *)&recvbuffer, (char *)&client.useragent, sizeof(client.useragent), "User-Agent: ", "\r\n")) {
								sprintf(client.useragent, "%s", "Unknown client");
							}
							
							if (streamserver_get_free_tcp_client_slots() == 0) {
								client.transmittedbytes = sprintf((char *)&recvbuffer, HTTPANSWER503);
								recvbuffer[client.transmittedbytes] = '\0';
								send(client.socketfd, recvbuffer, client.transmittedbytes, 0);
								log_add_message(mainlog, LOG_EVENT_ERROR, "TCP LISTEN PROCESS connection closed: server busy");
								close(client.socketfd);
								continue;
							} else {
								if (client.pidfilter == 0) {
									client.transmittedbytes = sprintf((char *)&recvbuffer, HTTPANSWER200);
									recvbuffer[client.transmittedbytes] = '\0';
								} else {
									if (procmblock->tsinfo.pids[client.pidfilter].stream_type == TS_STREAM_PES_AUDIO_STREAM) {
										bitrate = procmblock->tsinfo.pids[client.pidfilter].params.audio.bitrate;
										switch (procmblock->tsinfo.pids[client.pidfilter].stream_format) {
											case TS_STREAM_FORMAT_AUDIO_MPEG:
												contenttype = CONTENT_TYPE_AUDIO_MPEG;
												if (bitrate == 0) bitrate = 320;
												break;
											case TS_STREAM_FORMAT_AUDIO_AC3:
												contenttype = CONTENT_TYPE_AUDIO_AC3;
												if (bitrate == 0) bitrate = 640;
												break;
											case TS_STREAM_FORMAT_AUDIO_DTS:
												contenttype = CONTENT_TYPE_AUDIO_DTS;
												if (bitrate == 0) bitrate = 640;
												break;
											case TS_STREAM_FORMAT_AUDIO_AAC:
												contenttype = CONTENT_TYPE_AUDIO_AAC;
												if (bitrate == 0) bitrate = 320;
												break;
											default:
												contenttype = CONTENT_TYPE_OTHER;
												if (bitrate == 0) bitrate = 320;
												break;
										}
										client.transmittedbytes = sprintf((char *)&recvbuffer, HTTPANSWERICY, contenttype, procmblock->servicename, procmblock->nodename, ntohs((procmblock->clients[LISTEN_PROCESS_INDEX].addressv6).sin6_port), client.pidfilter, bitrate);
										recvbuffer[client.transmittedbytes] = '\0';
									} else if (procmblock->tsinfo.pids[client.pidfilter].stream_type == TS_STREAM_PES_VIDEO_STREAM) {
										client.transmittedbytes = sprintf((char *)&recvbuffer, HTTPANSWER200, CONTENT_TYPE_VIDEO_MPEG);
										recvbuffer[client.transmittedbytes] = '\0';
									} else {
										client.transmittedbytes = sprintf((char *)&recvbuffer, HTTPANSWER200, CONTENT_TYPE_TS);
										recvbuffer[client.transmittedbytes] = '\0';
									}
								}
								
								
								
								if (send(client.socketfd, recvbuffer, client.transmittedbytes, 0) < 0) {
									log_add_message(mainlog, LOG_EVENT_ERROR, "TCP LISTEN PROCESS connection closed");
									close(client.socketfd);
									continue;
								}
																
								streamserver_fdtransport_send_message(fdtransport[1], FDTRANSPORT_NEW_TCP_CLIENT, client.socketfd, &client, sizeof(struct sclient));
								
								close(client.socketfd);
							}
						}
					}
				}
			}
		}
		
		close(helperclient.socketfd);
		
		if (remove_client(helperclient.pid) == -1) {
			log_add_message(mainlog, LOG_EVENT_ERROR, "TCP LISTEN PROCESS no client removed - seems like internal error");
		}
		
		sigprocmask(SIG_SETMASK, &oldmask, NULL);
		
		exit(EXIT_SUCCESS);
	}
}

int streamserver_stop_tcp_listen_process(void)
{
	if (!sysutils_process_exists(procmblock->clients[LISTEN_PROCESS_INDEX].pid)) {
		log_add_message(mainlog, LOG_EVENT_MESSAGE, "TCP LISTEN PROCESS already stopped - need to be started first");
		return 0;
	}	
	
	kill(procmblock->clients[LISTEN_PROCESS_INDEX].pid, SIGTERM);

	return 1;
}

int streamserver_init_udp_broadcast(struct sockaddr_in *bcaddr)
{
	int bcfd;
	const int broadcast = 1;
	pid_t pid;
	
	int socketfd;
	struct ifconf ifc;
	struct ifreq  *ifr, *currentifr;
	int ifcbuflen;
	char *ifcbuf;
	unsigned int ifcount;
	int i;
	
	struct sclient client;
	
	if (procmblock->clients[UDP_BC_PROCESS_INDEX].socketfd != 0) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "UDP BROADCAST already running - need to be stopped first");
		return 0;
	}
		
	if ((bcfd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "Failed to create socket for UDP broadcast: %s", strerror(errno));
		close(bcfd);
		return 0;
	}
	
	if (strcmp(procmblock->clients[UDP_BC_PROCESS_INDEX].ifname, "lo") != 0) {
		if ((socketfd = socket(AF_INET, SOCK_DGRAM, 0)) != -1) {
			ifcbuflen = 100 * sizeof(struct ifreq);
			
			while (1) {
				ifcbuf = malloc(ifcbuflen);
				ifc.ifc_len = ifcbuflen;
				ifc.ifc_buf = ifcbuf;
				
				if (ioctl(socketfd, SIOCGIFCONF, &ifc) == -1) {
					if (errno != EINVAL) {
						log_add_message(mainlog, LOG_EVENT_ERROR, "IOCTL SIOCGIFCONF error: %s", strerror(errno));
					}
				} else {
					ifcbuflen = ifc.ifc_len;
					break;
				}		 
				ifcbuflen += 10 * sizeof(struct ifreq);
				free(ifcbuf);
			}
			
			ifr = ifc.ifc_req;
			ifcount = ifc.ifc_len / sizeof(struct ifreq);
			
			for (i=0; i<ifcount; i++) {
				currentifr = &ifr[i];
				
				if (strcmp(procmblock->clients[UDP_BC_PROCESS_INDEX].ifname, currentifr->ifr_name) == 0) {
					if (ioctl(socketfd, SIOCGIFFLAGS, currentifr) != -1) {
						if ((currentifr->ifr_flags & IFF_UP) && (currentifr->ifr_flags & IFF_BROADCAST)) {
							if (ioctl(socketfd, SIOCGIFBRDADDR, currentifr) != -1) {
								memcpy(bcaddr, (struct sockaddr_in *)&currentifr->ifr_broadaddr, sizeof(struct sockaddr_in));
							}
						}
					}
				}
				
			}
			
			free(ifcbuf);
			close(socketfd);
		}
	}
	
	if (setsockopt(bcfd, SOL_SOCKET, SO_BROADCAST, &broadcast, sizeof(broadcast)) == -1) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "Failed to set UDP socket broadcast option: %s", strerror(errno));
		close(bcfd);
		return 0;
	}
	
	client.type = CLIENT_TYPE_UDP_BROADCAST;
	client.family = CLIENT_FAMILY_V4_TRANSPORT;
	client.pid = 0;
	client.socketfd = bcfd;
	memcpy((struct sockaddr_in *)&client.address, bcaddr, sizeof(struct sockaddr_in));
	snprintf(client.useragent, sizeof(client.useragent), "UDP broadcast process");
	client.pidfilter = 0;
	client.transmittedbytes = 0;
	
	if (!sysutils_process_exists(procmblock->clients[UDP_BC_PROCESS_INDEX].pid)) {
		transmission_process(&client);
	}
	
	return 1;
}

int streamserver_stop_udp_broadcast(void)
{
	if (!sysutils_process_exists(procmblock->clients[UDP_BC_PROCESS_INDEX].pid)) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "UDP BROADCAST already stopped - need to be started first");
		return 0;
	}	
	
	kill(procmblock->clients[UDP_BC_PROCESS_INDEX].pid, SIGTERM);
	
	return 1;
}

int streamserver_init_udp_multicast(struct sockaddr_in *udpmgroup)
{
	int mcfd;
	struct in_addr udpmlocal;
	pid_t pid;
	
	int socketfd;
	struct ifconf ifc;
	struct ifreq  *ifr, *currentifr;
	int ifcbuflen;
	char *ifcbuf;
	unsigned int ifcount;
	int i;
	
	unsigned char ttl = (unsigned char)procmblock->udpmcttl;
	unsigned char loop = (unsigned char)procmblock->udpmcloop;
	
	struct sclient client;
	
	if (procmblock->clients[UDP_MC_PROCESS_INDEX].socketfd != 0) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "UDP MULTICAST already running - need to be stopped first");
		return 0;
	}
	
	if ((mcfd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "Failed to create socket for UDP multicast: %s", strerror(errno));
		close(mcfd);
		return 0;
    }
	
	udpmlocal.s_addr = INADDR_ANY;
	
	if (strcmp(procmblock->clients[UDP_MC_PROCESS_INDEX].ifname, "lo") != 0) {
		if ((socketfd = socket(AF_INET, SOCK_DGRAM, 0)) != -1) {
			ifcbuflen = 100 * sizeof(struct ifreq);
			
			while (1) {
				ifcbuf = malloc(ifcbuflen);
				ifc.ifc_len = ifcbuflen;
				ifc.ifc_buf = ifcbuf;
				
				if (ioctl(socketfd, SIOCGIFCONF, &ifc) == -1) {
					if (errno != EINVAL) {
						log_add_message(mainlog, LOG_EVENT_ERROR, "IOCTL SIOCGIFCONF error: %s", strerror(errno));
					}
				} else {
					ifcbuflen = ifc.ifc_len;
					break;
				}		 
				ifcbuflen += 10 * sizeof(struct ifreq);
				free(ifcbuf);
			}
			
			ifr = ifc.ifc_req;
			ifcount = ifc.ifc_len / sizeof(struct ifreq);
			
			for (i=0; i<ifcount; i++) {
				currentifr = &ifr[i];
				
				if (strcmp(procmblock->clients[UDP_MC_PROCESS_INDEX].ifname, currentifr->ifr_name) == 0) {
					if (ioctl(socketfd, SIOCGIFFLAGS, currentifr) != -1) {
						if ((currentifr->ifr_flags & IFF_UP) && (currentifr->ifr_flags & IFF_MULTICAST)) {
							memcpy((struct sockaddr_in *)&udpmlocal.s_addr, (struct sockaddr_in *)&currentifr->ifr_addr, sizeof(struct sockaddr_in));
						}
					}
				}
				
			}
			
			free(ifcbuf);
			close(socketfd);
		}
	}
	
	// Set the outgoing interface for multicast
	if (setsockopt(mcfd, IPPROTO_IP, IP_MULTICAST_IF, (char *)&udpmlocal, sizeof(struct in_addr)) == -1) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "Failed to set local interface for UDP multicast: %s", strerror(errno));
		close(mcfd);
		return 0;
	}
	
	// Set multicast packet TTL
	if (setsockopt(mcfd, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(unsigned char)) == -1) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "Failed to set UDP multicast TTL: %s", strerror(errno));
	}
	
	// Set multicast traffic loopback
	if (setsockopt(mcfd, IPPROTO_IP, IP_MULTICAST_LOOP, &loop, sizeof(unsigned char)) == -1) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "Failed to set UDP multicast loopback: %s", strerror(errno));
	}
	
	client.type = CLIENT_TYPE_UDP_MULTICAST;
	client.family = CLIENT_FAMILY_V4_TRANSPORT;
	client.pid = 0;
	client.socketfd = mcfd;
	memcpy(&client.address, udpmgroup, sizeof(struct sockaddr_in));
	snprintf(client.useragent, sizeof(client.useragent), "UDP multicast process");
	client.pidfilter = 0;
	client.transmittedbytes = 0;
	
	if (!sysutils_process_exists(procmblock->clients[UDP_MC_PROCESS_INDEX].pid)) {
		transmission_process(&client);
	}
	
	return 1;
}

int streamserver_stop_udp_multicast(void)
{
	if (!sysutils_process_exists(procmblock->clients[UDP_MC_PROCESS_INDEX].pid)) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "UDP MULTICAST already stopped - need to be started first");
		return 0;
	}	
	
	kill(procmblock->clients[UDP_MC_PROCESS_INDEX].pid, SIGTERM);
	
	return 1;
}

int streamserver_init_record(void)
{
	const char *homedir;
	time_t utime;
	char tmstr[32];
    struct tm *prtime;
	char filename[PATH_MAX];
	int fd;
		
	if (procmblock->recording) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "Record already running - need to be stopped first");
		return 0;
	}
	
	if (strlen(procmblock->servicename) == 0) {
		sprintf(procmblock->servicename, "%s", "Current service");
	}
	
	homedir = sysutils_gethomedir();
	
	utime = time(NULL);
	prtime = localtime(&utime);
	memset(tmstr, 0, sizeof(tmstr));
	strftime(tmstr, sizeof(tmstr), "%F_%T", prtime);
	
	memset(filename, 0, sizeof(filename));
	sprintf((char *)filename, "%s/%s_%s.ts", homedir, procmblock->servicename, tmstr);
		
	if ((fd = open((char *)filename, O_CREAT | O_WRONLY | O_NONBLOCK, S_IRWXU)) < 0) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "Failed to open file for record: %s", strerror(errno));
		return 0;
	}
		
	if (streamserver_fdtransport_send_message(fdtransport[0], FDTRANSPORT_START_RECORD, fd, NULL, 0) != -1) {
		procmblock->recording = 1;
		close(fd);
		log_add_message(mainlog, LOG_EVENT_ERROR, "Started record service: %s, file: %s", procmblock->servicename, filename);
		return 1;
	} else {
		procmblock->recording = 0;
		close(fd);
		log_add_message(mainlog, LOG_EVENT_ERROR, "Open file for recording");
		return 0;
	}
}

int streamserver_stop_record(void)
{
	if (!procmblock->recording) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "Record not running - need to be started first");
		return 0;
	}
		
	if (streamserver_fdtransport_send_message(fdtransport[0], FDTRANSPORT_STOP_RECORD, -1, NULL, 0) != -1) {
		procmblock->recording = 0;
		log_add_message(mainlog, LOG_EVENT_MESSAGE, "Record stopped");
		return 1;
	} else {
		log_add_message(mainlog, LOG_EVENT_ERROR, "FDTRANSPORT sendmsg error");
		return 0;
	}
}

int streamserver_init_csa_decoder(int caid, char *key)
{
	int i;
	char hexdigit[3];
	
	if (procmblock->encrypted) {
		log_add_message(mainlog, LOG_EVENT_MESSAGE, "Decoder already enabled, changing key");
		procmblock->encrypted = 0;
	}
		
	switch (caid) {
		case 0x2600:
			if (strlen(key) == 16) {
				memset(procmblock->csakey, 0, sizeof(procmblock->csakey));
				for (i=0; i<8; i++) {
					memset(hexdigit, 0, sizeof(hexdigit));
					strncpy(hexdigit, &key[i*2], 2);
					sscanf(hexdigit, "%x", &procmblock->csakey[i]);
					procmblock->csakey[3] = (procmblock->csakey[0] + procmblock->csakey[1] + procmblock->csakey[2]) & 0xff;
					procmblock->csakey[7] = (procmblock->csakey[4] + procmblock->csakey[5] + procmblock->csakey[6]) & 0xff;
				}				
			}
			procmblock->encrypted = 1;
			log_add_message(mainlog, LOG_EVENT_MESSAGE, "Encryption: BISS");
			break;
		
		case 0x0000: break;
		
		default: 
			log_add_message(mainlog, LOG_EVENT_MESSAGE, "Encryption system not supported");
			return 0;
	}
	
	return 1;
}

int streamserver_stop_csa_decoder(void)
{
	if (!procmblock->encrypted) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "Decoder not enabled");
		return 0;
	} else {
		procmblock->encrypted = 0;
		memset(procmblock->csakey, 0, sizeof(procmblock->csakey));
		log_add_message(mainlog, LOG_EVENT_MESSAGE, "Decoder disabled");
		return 1;
	}
}

int streamserver_open_device(int adapter, int frontend)
{
	char dev_fn[PATH_MAX];
	int fd;
		
	memset(dev_fn, 0, sizeof(dev_fn));
	sprintf(dev_fn, "/dev/dvb/adapter%i/dvr%i", adapter, frontend);
	
	if (access(dev_fn, R_OK) != 0) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "No access to DVR device %s", dev_fn);
		return -1;
	}
	
	if ((fd = open(dev_fn, O_RDONLY | O_NONBLOCK)) <= 0) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "Failed to open DVR device %s", dev_fn);
		return -1;
	}
	
	if (streamserver_fdtransport_send_message(fdtransport[0], FDTRANSPORT_CHANGE_DEVICE, fd, NULL, 0) != -1) {
		log_add_message(mainlog, LOG_EVENT_MESSAGE, "DVR device %s opened", dev_fn);
		close(fd);
		return 1;
	} else {
		log_add_message(mainlog, LOG_EVENT_ERROR, "FDTRANSPORT sendmsg error");
		close(fd);
		return 0;
	}
}

int streamserver_change_service(char *svcname, unsigned short caid, char *cakey)
{
	if (svcname == NULL) return 0;
	
	if (svcname[0] != '\0') {
		if (strcmp(svcname, procmblock->servicename) != 0) {
			strncpy((char *)&procmblock->servicename, svcname, sizeof(procmblock->servicename));
			log_add_message(mainlog, LOG_EVENT_MESSAGE, "New service name: %s", procmblock->servicename);
			transportstream_clear_statistics((struct _tsinfo *)&(procmblock->tsinfo));
		}
	}
	
	if ((caid != 0) && (cakey[0] != 0)) {
		streamserver_init_csa_decoder(caid, cakey);
	} else if (procmblock->encrypted) {
		streamserver_stop_csa_decoder();
	}
	
	return 1;
}

static int streamserver_write_to_disk(int fd, struct memblock *procmblock)
{
	int pos, bytes, wrbytes, res;
	struct timespec sleeptime = {.tv_sec = 0, .tv_nsec = 250 * 1000 * 1000 };	
	struct timeval tval;
	sigset_t blockmask, oldmask;
	unsigned long starttime, endtime;
	
	pos = 0;
	
	do {
		if ((BUFSIZE - pos) >= procmblock->fsblocksize) {
			bytes = procmblock->fsblocksize;
		} else {
			bytes = BUFSIZE - pos;
		}
		
		sigemptyset(&blockmask);
		sigaddset(&blockmask, SIGINT);
				
		res = 0;
		
		gettimeofday(&tval, NULL);
		starttime = tval.tv_sec*1000 + tval.tv_usec/1000;
		
		do {
			sigprocmask(SIG_BLOCK, &blockmask, &oldmask);
			wrbytes = write(fd, &procmblock->buffer[pos], bytes);
			sigprocmask(SIG_SETMASK, &oldmask, NULL);
			
			if (wrbytes != -1) {
				pos = pos + wrbytes;
				res = 1;
			} else if ((wrbytes == -1) && (errno == EAGAIN)) {
				gettimeofday(&tval, NULL);
				endtime = tval.tv_sec*1000 + tval.tv_usec/1000;
				
				if ((endtime - starttime) <= 1000) {
					nanosleep(&sleeptime, NULL);
					res = 0;
				} else {
					res = 1;
				}
			} else if ((wrbytes == -1) && (errno != EAGAIN)) {
				pos = -1;
				res = 1;
			}
		} while (!res);
	} while ((pos < BUFSIZE) && (res > 0)); 
	
	return pos;
}

static void streamserver_read_process_signal_handler(int sig, siginfo_t *si, void *unused)
{
	if ((sig == SIGINT) || (sig == SIGTERM)) {
		rpexitflag = 1;
	}
}

static void init_read_process(void)
{
	struct sigaction sa;
	pid_t pid;
	sigset_t blockmask, oldmask;
	struct sclient readclient;
	unsigned int readindex;
	//int dsignal = SIGTERM;
	struct pollfd sockets[2];
	int readysockets;
	struct stat statbuf;
	struct timeval startvtime, endvtime;
	int deviceready = 0;
	int bufptr = 0;
	int readb = 0;
	int i;
	int cmd, res;
	int recordfd = -1;
	int devicefd = -1;
	int tempfd = -1;
		
	if ((pid = fork()) < 0) {
		//error
		log_add_message(mainlog, LOG_EVENT_ERROR, "READ PROCESS error in fork system call: %s", strerror(errno));
		exit(EXIT_FAILURE);
	} else if (pid == 0) {
		//child
		close(fdtransport[0]);
		
		sigemptyset(&blockmask);
		sigaddset(&blockmask, SIGCONT);
		sigprocmask(SIG_BLOCK, &blockmask, &oldmask);
		
		sa.sa_flags = 0;
		sigemptyset(&sa.sa_mask);
		sa.sa_sigaction = streamserver_read_process_signal_handler;
		
		if (sigaction(SIGINT, &sa, NULL) == -1) {
			log_add_message(mainlog, LOG_EVENT_ERROR, "Error calling SIGINT handler procedure in client process");
		}
		
		if (sigaction(SIGTERM, &sa, NULL) == -1) {
			log_add_message(mainlog, LOG_EVENT_ERROR, "Error calling SIGINT handler procedure in client process");
		}
		
		/*#ifdef linux
			prctl(PR_SET_PDEATHSIG, dsignal, 0, 0, 0);
		#endif*/
		
		readclient.type = CLIENT_TYPE_READ_HELPER;
		readclient.family = CLIENT_FAMILY_HELPER;
		readclient.pid = getpid();
		readclient.socketfd = 0;
		memcpy(&readclient.address, &procmblock->clients[READ_PROCESS_INDEX].address, sizeof(struct sockaddr_in));
		memcpy(&readclient.addressv6, &procmblock->clients[READ_PROCESS_INDEX].addressv6, sizeof(struct sockaddr_in6));
		snprintf(readclient.useragent, sizeof(readclient.useragent), "Read helper process");
		readclient.pidfilter = 0;
		readclient.transmittedbytes = 0;
		
		if ((readindex = add_client(&readclient)) == -1) {
			log_add_message(mainlog, LOG_EVENT_ERROR, "Failed to add READ PROCESS info into storage");
			exit(EXIT_FAILURE);
		}
		
		sockets[0].fd = fdtransport[1];
		sockets[0].events = POLLIN | POLLERR;
				
		if ((fstat(devicefd, &statbuf) != -1) && (statbuf.st_mode & S_IFCHR)) {
			sockets[1].fd = devicefd;
			sockets[1].events = POLLIN;
			deviceready = 1;
		}
		
		transportstream_clear_statistics((struct _tsinfo *)&(procmblock->tsinfo));
		
		while (!rpexitflag) {
			if (procmblock->sheduler == SHEDULER_RWLOCK) rwlockshed_writelock(rwlockshed);
			
			//read data and send
			bufptr = 0;
			
			gettimeofday(&startvtime, 0);
			
			while (bufptr != BUFSIZE) {
				if ((readysockets = poll(sockets, deviceready ? 2 : 1, INFTIM)) > 0) {
					if ((deviceready) && (sockets[1].revents & POLLIN)) {
						if ((readb = read(devicefd, &(procmblock->buffer)[bufptr], BUFSIZE-bufptr)) != -1) {
							bufptr = bufptr + readb;
							if (procmblock->buffer[0] != 0x47) {
								for (i=0;i<bufptr;i++) {
									if (procmblock->buffer[i] == 0x47) {
										memcpy(&procmblock->buffer[0], &procmblock->buffer[i], bufptr-i);
										bufptr = bufptr - i;
										break;
									}
								
								}
								log_add_message(mainlog, LOG_EVENT_WARNING, "Lost sync, %i bytes dropped (read: %i)", i, readb);
							}
						}
					}
					
					if ((fdtransport[1] > 0) && ((sockets[0].revents & POLLIN) || (sockets[0].revents & POLLERR))) {
						
						if (sockets[0].revents & POLLERR) {
							log_add_message(mainlog, LOG_EVENT_ERROR, "READ PROCESS FDTRANSPORT error");
							break;
						}
						
						res = streamserver_fdtransport_recv_message(fdtransport[1], &cmd, &tempfd, NULL, 0);
						
						if (res == 0) {
							log_add_message(mainlog, LOG_EVENT_MESSAGE, "READ PROCESS FDTRANSPORT exit signal");
							break;
						} else if (res == -1) {
							log_add_message(mainlog, LOG_EVENT_ERROR, "READ PROCESS FDTRANSPORT error");
							break;
						}
						
						switch (cmd) {
							case FDTRANSPORT_START_RECORD:
								if ((fstat(tempfd, &statbuf) != -1) && (statbuf.st_mode & S_IFREG)) {
									if (recordfd > 0) {
										close(recordfd);
										recordfd = -1;
									}
									recordfd = tempfd;
								} else {
									procmblock->recording = 0;
									log_add_message(mainlog, LOG_EVENT_ERROR, "FDTRANSPORT wrong record descriptor");
								}
								break;
							
							case FDTRANSPORT_STOP_RECORD:
								if (recordfd > 0) {
									close(recordfd);
									recordfd = -1;
								}
								break;
								
							case FDTRANSPORT_CHANGE_DEVICE:
								if ((fstat(tempfd, &statbuf) != -1) && (statbuf.st_mode & S_IFCHR)) {
									if (deviceready) {
										close(devicefd);
										deviceready = 0;
									}
									devicefd = tempfd;
									sockets[1].fd = devicefd;
									sockets[1].events = POLLIN;
									deviceready = 1;
								} else {
									log_add_message(mainlog, LOG_EVENT_ERROR, "FDTRANSPORT wrong device descriptor");
								}
								break;
							
						}
					}
				} else if ((errno == EINTR) && (rpexitflag)) break;
			}
			
			procmblock->readyclients = 0;
			if (procmblock->sheduler == SHEDULER_SIGNAL) signalshed_wait_value(signalshed, procmblock->numclients);
			
			if (procmblock->encrypted) {
				bufptr = 0;
				while (bufptr < BUFSIZE-1) {
					csa_decrypt(&(procmblock->buffer)[bufptr], procmblock->csakey);
					bufptr = bufptr + 188;
				}
			}
			
			if (procmblock->numclients > 0) {
				for (i=UDP_BC_PROCESS_INDEX; i<MAXCLIENTS; i++) {
					procmblock->clients[i].ready = 0;
				}
				if (procmblock->sheduler == SHEDULER_SIGNAL) signalshed_continue(signalshed);
			}
			
			if ((procmblock->recording) && (recordfd >= 0)) {
				if (streamserver_write_to_disk(recordfd, procmblock) == -1) {
					log_add_message(mainlog, LOG_EVENT_ERROR, "Record stopped: %s", strerror(errno));
					procmblock->recording = 0;
					close(recordfd);
					recordfd = -1;
				}
			}
			
			if ((procmblock->sheduler == SHEDULER_SIGNAL) && (procmblock->numclients > 0) && (!rpexitflag)) signalshed_wait(signalshed);
			if (procmblock->sheduler == SHEDULER_RWLOCK) rwlockshed_writeunlock(rwlockshed);
			
			
			gettimeofday(&endvtime, 0);
			
			procmblock->tsinfo.timeframe = (endvtime.tv_sec - startvtime.tv_sec) * 1000 + (endvtime.tv_usec - startvtime.tv_usec) / 1000;
			
			transportstream_parse_buffer((unsigned char *)&procmblock->buffer, sizeof(procmblock->buffer), (struct _tsinfo *)&(procmblock->tsinfo));
		}
		
		
		close(devicefd);
		
		if (remove_client(readclient.pid) == -1) {
			log_add_message(mainlog, LOG_EVENT_ERROR, "No client removed - seems like internal error");
		}
		
		sigprocmask(SIG_SETMASK, &oldmask, NULL);
		
		exit(EXIT_SUCCESS);
	}
}

/*static void clean_lost_clients(void)
{
	int i;
	static union sigval sigvalue;
	
	for (i=MAXCLIENTS; i>=TCP_FIRST_CLIENT; i--) {
		if (process_exists(procmblock->clients[i].pid)) {
			if (procmblock->clients[i].sentbytes == procmblock->clients[i].sentbytessnap) {
				printf("[+] Stopped idle TCP UNICAST process with PID: %u\n", procmblock->clients[i].pid);
				sigvalue.sival_int = 0;
				sigqueue(procmblock->clients[i].pid, SIGTERM, sigvalue);
			}
			procmblock->clients[i].sentbytessnap = procmblock->clients[i].sentbytes;
		}
	}
}*/	

int streamserver_send_log_message(int flags, char *format, ...)
{
	va_list arg;
	struct sclient client;
	struct logmsg_t msg;
	int len;
	
	if (format == NULL) return 0;
	
	if (streamserver_get_client(getpid(), &client) == -1) return 0;
	
	memset(&msg, 0, sizeof(msg));
	
	va_start(arg, format);
	len = vsnprintf(msg.message, sizeof(msg.message), format, arg);
	va_end(arg);
	
	msg.flags = flags;
	
	streamserver_fdtransport_send_message(client.fdtransport, FDTRANSPORT_LOG_MESSAGE, -1, &msg, sizeof(msg));
	
	return len;
}

static int streamserver_recv_log_message(struct logmsg_t *msg)
{
	log_add_message_static(mainlog, msg->flags, msg->message);
}

static void streamserver_main_recv_message(int fd)
{
	struct sclient client;
	struct logmsg_t msg;
	int cmd, socketfd;
	char data[4096];
	
	memset(data, 0, sizeof(data));
	
	if (streamserver_fdtransport_recv_message(fd, &cmd, &socketfd, &data, sizeof(data)) != -1) {
		switch (cmd) {
			case FDTRANSPORT_NEW_TCP_CLIENT:
				memcpy(&client, &data, sizeof(struct sclient));
				client.socketfd = socketfd;
				transmission_process(&client);
				break;
			
			case FDTRANSPORT_LOG_MESSAGE:
				memcpy(&msg, &data, sizeof(struct logmsg_t));
				streamserver_recv_log_message(&msg);
				break;
			
			default:
				close(socketfd);
				break;
			
		}
	}
}

void streamserver_register_host_process(pid_t hostpid)
{
	if (sysutils_process_exists(hostpid)) {
		procmblock->hostpid = hostpid;
		log_add_message(mainlog, LOG_EVENT_MESSAGE, "Registered host process PID: %u, UID: %u, GID: %u", procmblock->hostpid, getuid(), getgid());
	}
}

static void main_signal_handler(int sig, siginfo_t *si, void *unused)
{
	int exit_status;
	int client;
	
	if ((sig == SIGINT) || (sig == SIGTERM)) {
		exitflag = 1;
	} else if (sig == SIGHUP) {
		readconfflag = 1;
	} else if (sig == SIGCHLD) {
		cleanupflag = 1;
		/*if ((client = remove_client(si->si_pid)) == -1) {
			printf("[-] No client removed - seems like internal error\n");
		}*/
		//wait(&exit_status);
	}/* else if (sig == SIGALRM) {
		lostcliflag = 1;
	}*/
}

static void streamserver_daemon_main_process(void)
{
	key_t mainkey;
	sigset_t blockmask, oldmask, zeromask;
	int msgid;
	struct svcmsg_t message;
	struct timespec sleeptime = {.tv_sec = 0, .tv_nsec = 500 * 1000 * 1000 };	
	struct timeval tval;
	ssize_t res;
	unsigned long starttime, endtime;
	int hostpresent;
	
	if ((mainkey = ftok(RESOURCE_BINARY_STREAMSERVER, MAIN_PROJECT_ID)) == -1) {
		printf("[-] Failed to generate main SYSVIPC key\n");
		exit(EXIT_FAILURE);
	}
	
	if ((msgid = msgget(mainkey, 0666 | IPC_CREAT)) == -1) {
		printf("[-] Failed to create message queue\n");
		exit(EXIT_FAILURE);
	}
	
	sigemptyset(&blockmask);
	sigaddset(&blockmask, SIGHUP);
	sigaddset(&blockmask, SIGCHLD);
	
	gettimeofday(&tval, NULL);
	starttime = tval.tv_sec*1000 + tval.tv_usec/1000; 
	res = 0;
	
	do {
		sigprocmask(SIG_BLOCK, &blockmask, &oldmask);
		res = msgrcv(msgid, &message, sizeof(message), SERVICE_MESSAGE_ACTIVATION, IPC_NOWAIT);
		sigprocmask(SIG_SETMASK, &oldmask, NULL);
		if (res > 0) {
			break;
		} else if ((res == -1) && (errno == EAGAIN)) {
			gettimeofday(&tval, NULL);
			endtime = tval.tv_sec*1000 + tval.tv_usec/1000;
			if ((endtime - starttime) <= 3000) {
				nanosleep(&sleeptime, NULL);
				continue;
			} else {
				res = 0;
				break;
			}
		}
	} while (res <= 0);
	
	msgctl(msgid, IPC_RMID, 0);
	
	if (getenv("HOSTPID") != NULL) hostpresent = 1; else hostpresent = 0;
	
	if (res > 0) {
		if ((!message.svcrunning) && (message.svcpid != 0) && (sysutils_process_exists(message.svcpid))) {
			if (!hostpresent) {
				printf("[+] Service PID: %u\n", message.svcpid);
			} else {
				fprintf(stdout, "%ld", message.svcpid);
			}
			exit(EXIT_SUCCESS);
		} else if ((message.svcrunning) && (message.svcpid != 0)) {
			if (!hostpresent) {
				printf("[+] Service already running with PID: %u\n", message.svcpid);
			} else {
				fprintf(stdout, "%ld", message.svcpid);
			}
			exit(EXIT_FAILURE);
		}
	} else {
		if (!hostpresent) {
			printf("[-] Service PID unknown\n");
		} else {
			fprintf(stdout, "%ld", 0);
		}
		exit(EXIT_FAILURE);
	}
	
}

static void streamserver_start_daemon(void(*main_callback)(void))
{
	int i, fd0, fd1, fd2;
	pid_t pid, sid;
	struct rlimit rlim;
	struct sigaction sa;
		
	umask(0);
	
	if (getrlimit(RLIMIT_NOFILE, &rlim) == -1) {
		printf("[-] rlimit call error\n");
	}
	
	if ((pid = fork()) < 0) {
		printf("[-] Fork call error\n");
	} else if (pid != 0) {
		if (main_callback != NULL) {
			main_callback();
		} else {
			exit(0);
		}
	}
	
	sid = setsid();
	
	sa.sa_handler = SIG_IGN;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	
	if (sigaction(SIGHUP, &sa, NULL) < 0) {
		printf("[-] Failed to block SIGHUP\n");
	}
	
	if ((pid = fork()) < 0) {
		printf("[-] Fork call error\n");
	} else if (pid != 0) {
		exit(0);
	}
	
	if (rlim.rlim_max == RLIM_INFINITY) rlim.rlim_max = 1024;
	for (i=0; i<rlim.rlim_max; i++) {
		close(i);
	}
	
	fd0 = open("/dev/null", O_RDWR);
	fd1 = dup(0);
	fd2 = dup(0);
}

static int streamserver_lock_file(const char *homedir, int *runpid, int killdaemon, int daemonize)
{
	int lockfd;
	char lockfn[PATH_MAX];
	char string[64];
	struct flock lock;
	int hostpresent;
	
	memset(lockfn, 0, sizeof(lockfn));
	memset(&lock, 0, sizeof(lock));
	
	if (getenv("HOSTPID") != NULL) hostpresent = 1; else hostpresent = 0;
	
	sprintf((char *)lockfn, STREAMCONFIG_LOCK_PATH, homedir);
	
	if ((lockfd = open(lockfn, O_RDWR|O_CREAT, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)) == -1) {
		//log_add_message(log, LOG_EVENT_ERROR, "Unable to open lock file");
		if (!hostpresent) {
			printf("[-] Unable to open lock file\n");
		} else {
			fprintf(stdout, "%ld", 0);
		}
		return -1;
	}
	
	lock.l_type = F_WRLCK;
	lock.l_whence = SEEK_SET;
	lock.l_start = 0;
	lock.l_len = 0;
	
	if (fcntl(lockfd, F_SETLK, &lock) == -1) {
		if (errno == EACCES || errno == EAGAIN) {
			if (fcntl(lockfd, F_GETLK, &lock) != -1) {
				if (killdaemon) {
					//send SIGINT to stop daemon
					kill(lock.l_pid, SIGTERM);
					//log_add_message(log, LOG_EVENT_MESSAGE, "Sent INT signal to process %i", lock.l_pid);
					if (!hostpresent) {
						printf("[+] Sent INT signal to process %i\n", lock.l_pid);
					} else {
						fprintf(stdout, "%ld", lock.l_pid);
					}
					close(lockfd);
					return -1;
				} else {
					//running process found
					if (daemonize) {
						//svcrunpid = lock.l_pid;
						if (runpid != NULL) *runpid = lock.l_pid;
						return 0;
					} else {
						//log_add_message(log, LOG_EVENT_ERROR, "MAIN PROCESS already running with PID %i", lock.l_pid);
						//printf("[-] MAIN PROCESS already running with PID %i\n", lock.l_pid);
						if (!hostpresent) {
							printf("[-] MAIN PROCESS already running with PID %i\n", lock.l_pid);
						} else {
							fprintf(stdout, "%ld", lock.l_pid);
						}
						close(lockfd);
						return -1;
					}
				}
			}
		}
	} else if (!killdaemon) {
		//write pid into lock file
		ftruncate(lockfd, 0);
		memset(string, 0, sizeof(string));
		sprintf(string, "%ld", (long)getpid());
		write(lockfd, string, strlen(string)+1);
		return lockfd;
	} else if (killdaemon) {
		//no running process found
		//log_add_message(log, LOG_EVENT_ERROR, "MAIN PROCESS not running");
		//printf("[-] MAIN PROCESS not running\n");
		if (!hostpresent) {
			printf("[-] MAIN PROCESS not running\n");
		} else {
			fprintf(stdout, "%ld", 0);
		}
		close(lockfd);
		return -1;
	}
} 

static int streamserver_send_status(key_t key, pid_t svcrunpid)
{
	int msgid;
	struct timespec msgsleeptime = {.tv_sec = 0, .tv_nsec = 500 * 1000 * 1000 };
	struct timeval tval;
	unsigned long starttime, endtime;
	struct svcmsg_t message;
	sigset_t blockmask, oldmask;
	int res;
	
	gettimeofday(&tval, NULL);
	starttime = tval.tv_sec*1000 + tval.tv_usec/1000;
	
	do {
		msgid = msgget(key, 0666);
		if (msgid >= 0) {
			break;
		} else if ((msgid == -1) && (errno == ENOENT)) {
			gettimeofday(&tval, NULL);
			endtime = tval.tv_sec*1000 + tval.tv_usec/1000;
			if ((endtime - starttime) <= 3000) {
				nanosleep(&msgsleeptime, NULL);
				continue;
			} else {
				break;
			}
		} else if ((msgid == -1) && (errno != ENOENT)) {
			break;
		}
	} while (msgid < 0);
				
	if (msgid != -1) {
		message.mtype = SERVICE_MESSAGE_ACTIVATION;
		if (svcrunpid == 0) {
			message.svcpid = getpid();
			message.svcrunning = 0;
		} else {
			message.svcpid = svcrunpid;
			message.svcrunning = 1;
		}
		
		sigemptyset(&blockmask);
		sigaddset(&blockmask, SIGHUP);
		sigaddset(&blockmask, SIGCHLD);
		sigprocmask(SIG_BLOCK, &blockmask, &oldmask);
		
		res = msgsnd(msgid, &message, sizeof(message), IPC_NOWAIT);
			
		sigprocmask(SIG_SETMASK, &oldmask, NULL);
		
		return res;
	} else {
		return -1;
	}
}

static long int streamserver_stop_clients(void)
{
	struct timespec msgsleeptime = {.tv_sec = 0, .tv_nsec = 100 * 1000 * 1000 };
	struct timeval tval;
	unsigned long starttime, endtime;
	int i, res;
	sigset_t blockmask, oldmask;
	
	sigemptyset(&blockmask);
	sigaddset(&blockmask, SIGHUP);
	sigaddset(&blockmask, SIGCHLD);
	
	gettimeofday(&tval, NULL);
	starttime = tval.tv_sec*1000 + tval.tv_usec/1000;
	
	do {
		//Cleanup client slots
		sigprocmask(SIG_BLOCK, &blockmask, &oldmask);
		if (readconfflag) {
			readconfflag = 0;
		} else if (cleanupflag) {
			streamserver_cleanup_client_slots();
			cleanupflag = 0;
		}
		sigprocmask(SIG_SETMASK, &oldmask, NULL);
		//Search for active clients
		res = 0;
		for (i=READ_PROCESS_INDEX; i<MAXCLIENTS; i++) {
			if (procmblock->clients[i].pid != 0) {
				res = 1;
				break;
			}
		}
		//Timeout if needed
		if (res == 0) {
			break;
		} else if (res > 0) {
			gettimeofday(&tval, NULL);
			endtime = tval.tv_sec*1000 + tval.tv_usec/1000;
			if ((endtime - starttime) <= 10000) {
				nanosleep(&msgsleeptime, NULL);
				continue;
			} else {
				break;
			}
		}
	} while (res > 0);
	
	gettimeofday(&tval, NULL);
	endtime = tval.tv_sec*1000 + tval.tv_usec/1000;
	
	return (endtime - starttime);
}


int main(int argc, char *argv[])
{
	int opt;
	char daemonize = 0, killdaemon = 0;
	const char *homedir;
	pid_t svcrunpid;
	int lockfd;
	key_t mainkey, shedkey;
	const char *hostpidenv;
	struct utsname hostinfo;
	struct statvfs vfsinfo;
	int ctlsockfd;
	
	unsigned short semvalues[2];
	union semun semarg;
	struct _svrsettings sp;
	struct sigaction sa;
	sigset_t blockmask, oldmask;
	struct pollfd sockets[2];
	int readysockets;
	//struct itimerval timerdelay;
	
	
	while ((opt = getopt(argc, argv, "dkh")) != -1) {
		switch (opt) {
			case 'd':
				daemonize = 1;
				break;
			case 'k':
				killdaemon = 1;
				break;
			case 'h':
				printf("gsfstrsvr [-d] [-k] [-h]\n\n"
						"COMMANDS:\n"
						"  -d		Run as daemon\n"
						"  -k		Kill daemon if running\n"
						"  -h		Print help and exit\n");
				exit(EXIT_SUCCESS);
			default:
				fprintf(stderr, "Usage: %s [-d] [-k] [-h]\n", argv[0]);
				exit(EXIT_FAILURE);
		}
	}
	
	if (daemonize) {
		streamserver_start_daemon(streamserver_daemon_main_process);
	}
	
	//Reset blocked signals
	sigfillset(&blockmask);
	sigprocmask(SIG_UNBLOCK, &blockmask, NULL);
	sigemptyset(&blockmask);
	
	homedir = sysutils_gethomedir();
	
	svcrunpid = 0;
	
	//Log initialization
	if ((mainlog = log_init_instance("gsf", "streamserver.log", 0, daemonize ? 1 : 0)) == NULL) {
		printf("[-] Failed to create log instance\n");
	}
	
	//Lock file
	if ((lockfd = streamserver_lock_file(homedir, &svcrunpid, killdaemon, daemonize)) == -1) {
		log_free_instance(mainlog);
		exit(EXIT_SUCCESS);
	}
	
	//SVIPC initialization	
	if ((mainkey = ftok(RESOURCE_BINARY_STREAMSERVER, MAIN_PROJECT_ID)) == -1) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "Failed to generate main SYSVIPC key");
		log_free_instance(mainlog);
		close(lockfd);
		exit(EXIT_FAILURE);
	}
	
	if ((shedkey = ftok(RESOURCE_BINARY_STREAMSERVER, SHED_PROJECT_ID)) == -1) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "Failed to generate sheduler SYSVIPC key");
		log_free_instance(mainlog);
		close(lockfd);
		exit(EXIT_FAILURE);
	}
	
	if (daemonize) {
		if (streamserver_send_status(mainkey, svcrunpid) == -1) {
			log_add_message(mainlog, LOG_EVENT_ERROR, "Failed to send activation message");
		}
		if (svcrunpid != 0) {
			log_free_instance(mainlog);
			close(lockfd);
			exit(EXIT_FAILURE);
		}
	}
	
	shmid = shmget(mainkey, sizeof(struct memblock), 0666 | IPC_CREAT);
	
	procmblock = (struct memblock *)shmat(shmid, 0, 0);
	
	memset(procmblock, 0, sizeof(procmblock));
	
	procmblock->hostpid = 0;
	
	if ((hostpidenv = getenv("HOSTPID")) != NULL) {
		streamserver_register_host_process((pid_t)atoi(hostpidenv));
	}
	
	if (uname(&hostinfo) == 0) {
		memcpy((char *)&procmblock->nodename, (char *)&hostinfo.nodename, _UTSNAME_NODENAME_LENGTH);
		memcpy((char *)&procmblock->osname, (char *)&hostinfo.sysname, _UTSNAME_SYSNAME_LENGTH);
		memcpy((char *)&procmblock->osrelease, (char *)&hostinfo.release, _UTSNAME_RELEASE_LENGTH);
	} else {
		sprintf((char *)&procmblock->nodename, "localhost");
		sprintf((char *)&procmblock->osname, "Unknown OS");
		sprintf((char *)&procmblock->osrelease, "Unknown release");
	}
		
	procmblock->fsblocksize = 4096;
	
	if (homedir != NULL) {
		if (statvfs(homedir, &vfsinfo) == 0) {
			if (vfsinfo.f_bsize > 0) {
				procmblock->fsblocksize = vfsinfo.f_bsize;
			}
		}
	}
	
	if ((ctlsockfd = svrsettings_init()) == -1) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "Failed to create control socket and set config options");
		log_free_instance(mainlog);
		close(lockfd);
		exit(EXIT_FAILURE);
	}
	
	log_add_message(mainlog, LOG_EVENT_MESSAGE, "GSF streaming server started (%s/%s/%s)", procmblock->nodename, procmblock->osname, procmblock->osrelease);
	
	clisemid = semget(mainkey, 2, 0666 | IPC_CREAT);
	semvalues[0] = 1;
	semvalues[1] = 1;
	
	semarg.array = (unsigned short *)&semvalues;
	//semarg.val = 1;
	if (semctl(clisemid, 0, SETALL, semarg) == -1) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "[MAIN PROCESS] semctl call: %s", strerror(errno));
	}
	
	//FDTRANSPORT socket pair initialization
	socketpair(AF_LOCAL, SOCK_STREAM, 0, fdtransport);
	
	//read default values from config file
	svrsettings_read_config_file(&sp, 1);
	
	//init sheduler
	switch (sp.sheduler) {
		case SHEDULER_SIGNAL:
			signalshed_init(&signalshed, shedkey, 100);
			log_add_message(mainlog, LOG_EVENT_MESSAGE, "Selected sheduler: SIGNAL");
			break;
		case SHEDULER_RWLOCK:
			rwlockshed_init(&rwlockshed, shedkey, 100, 500);
			log_add_message(mainlog, LOG_EVENT_MESSAGE, "Selected sheduler: RWLOCK");
			break;
		default:
			signalshed_init(&signalshed, shedkey, 100);
			log_add_message(mainlog, LOG_EVENT_MESSAGE, "Selected sheduler: SIGNAL");
			break;
	}
	
	//apply settings
	svrsettings_config_apply_settings(&sp, procmblock);
	
	//start read process main loop	
	init_read_process();
	
	//open device and send descriptor
	streamserver_open_device(procmblock->adapter, procmblock->dvr);
	
	//initialize signal handlers
	sa.sa_flags = SA_SIGINFO;
	#ifdef SA_INTERRUPT
		sa.sa_flags |= SA_INTERRUPT;
	#endif
	
	/*#ifdef SA_NOCLDWAIT
		sa.sa_flags |= SA_NOCLDWAIT;
	#endif*/
	
	sigemptyset(&sa.sa_mask);
	sa.sa_sigaction = main_signal_handler;
		
	if (sigaction(SIGINT, &sa, NULL) == -1) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "Error calling SIGINT signal handler procedure in main process");
	}
	
	if (sigaction(SIGTERM, &sa, NULL) == -1) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "Error calling SIGTERM signal handler procedure in main process");
	}
	
	if (sigaction(SIGCHLD, &sa, NULL) == -1) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "Error calling SIGCHLD signal handler procedure in main process");
	}
	
	if (sigaction(SIGHUP, &sa, NULL) == -1) {
		log_add_message(mainlog, LOG_EVENT_ERROR, "Error calling SIGHUP signal handler procedure in main process");
	}
	
	/*if (sigaction(SIGALRM, &sa, NULL) == -1) {
		printf("[-] Error calling SIGALRM signal handler procedure in main process\n");
	}*/
	
	/*timerdelay.it_value.tv_sec = 10;
	timerdelay.it_value.tv_usec = 0;
	timerdelay.it_interval.tv_sec = 10;
	timerdelay.it_interval.tv_usec = 0;
	
	if (setitimer(ITIMER_REAL, &timerdelay, NULL) > 0) {
		printf("[-] Error setting timer - no lost connection control\n");
	}*/
	
	sigemptyset(&blockmask);
	sigaddset(&blockmask, SIGHUP);
	/*sigaddset(&blockmask, SIGALRM);*/
	sigaddset(&blockmask, SIGCHLD);
	
	sockets[0].fd = ctlsockfd;
	sockets[0].events = POLLIN;
	sockets[1].fd = fdtransport[0];
	sockets[1].events = POLLIN;
	
	while (!exitflag) {
		if ((readysockets = poll(sockets, 2, INFTIM)) > 0) {
			if (sockets[0].revents & POLLIN) {
				svrsettings_read_control_socket(sockets[0].fd, procmblock);
			} else if (sockets[1].revents & POLLIN) {
				streamserver_main_recv_message(sockets[1].fd);
			}
		} else if ((readysockets == -1) && (errno == EINTR)) {
			sigprocmask(SIG_BLOCK, &blockmask, &oldmask);
			if (readconfflag) {
				svrsettings_read_config_file(&sp, 0);
				svrsettings_config_apply_settings(&sp, procmblock);
				readconfflag = 0;
			} else if (cleanupflag) {
				streamserver_cleanup_client_slots();
				cleanupflag = 0;
			}
						
			/* else if (lostcliflag) {
				clean_lost_clients();
				lostcliflag = 0;
			}*/
			
			sigprocmask(SIG_SETMASK, &oldmask, NULL);
		}
	} 
	
	log_add_message(mainlog, LOG_EVENT_MESSAGE, "Cleaning up reserved resources");
	
	//stop all client processes
	sigemptyset(&blockmask);
	sigaddset(&blockmask, SIGTERM);
	sigprocmask(SIG_BLOCK, &blockmask, &oldmask);
	
	killpg(getpgrp(), SIGTERM);
	streamserver_stop_clients();
			
	//destroy sheduler
	switch (procmblock->sheduler) {
		case SHEDULER_SIGNAL:
			signalshed_destroy(signalshed);
			break;
		case SHEDULER_RWLOCK:
			rwlockshed_destroy(rwlockshed);
			break;
		default:
			break;
	}
	
	//Inform host process about exit
	if ((procmblock->hostpid != 0) && (sysutils_process_exists(procmblock->hostpid))) {
		if (kill(procmblock->hostpid, SIGCHLD) == 0) {
			log_add_message(mainlog, LOG_EVENT_MESSAGE, "Sent signal to host process PID: %u, UID: %u, GID: %u", procmblock->hostpid, getuid(), getgid());
		}
	}
	
	//delete memory block
	shmdt((void *)procmblock);
	shmctl(shmid, IPC_RMID, 0);
	
	//delete semaphore
	semctl(clisemid, 0, IPC_RMID);
	semctl(clisemid, 1, IPC_RMID);
	
	//release lock file	
	//unlink(lockfn);
	close(lockfd);
	
	log_add_message(mainlog, LOG_EVENT_MESSAGE, "Streaming server stopped");
	
	log_free_instance(mainlog);
	
	sigprocmask(SIG_SETMASK, &oldmask, NULL);
	
	exit(EXIT_SUCCESS);
}
