/*
 *      svrsettings.c
 *      
 *      Copyright 2010-2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <limits.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "streamserver.h"
#include "svrsettings.h"
#include "transportstream.h"
#include "../common/serialize.h"
#include "../common/sysutils.h"
#include "../common/log.h"


static unsigned int cliptr = 0;
static unsigned int strptr = 0;

void svrsettings_clients_list_callback(unsigned int *params, unsigned int count, int seq, int size, char **buffer, void *data);
void svrsettings_stream_structure_callback(unsigned int *params, unsigned int count, int seq, int size, char **buffer, void *data);
void svrsettings_mediastreams_callback(unsigned int *params, unsigned int count, int seq, int size, char **message, void *data);


int svrsettings_init(void)
{
	int ctlsockfd;
	int flagsval;
	struct sockaddr_un ctlserver;
	char *fullusername;
	char username[82];
	size_t unamelength;
	char clientsockpath[100], serversockpath[100];
	const int onflag = 1;
	
	//control unix domain socket initialization
	if ((ctlsockfd = socket(AF_LOCAL, SOCK_DGRAM, 0)) != -1) {
		flagsval = fcntl(ctlsockfd, F_GETFL, 0);
		fcntl(ctlsockfd, F_SETFL, flagsval | O_NONBLOCK);
		
		fullusername = strdup(getenv("USER"));
		
		if ((unamelength = strlen(fullusername)) >= 82) {
			strncpy(username, fullusername, 81);
			username[81] = '\0';
		} else {
			strncpy(username, fullusername, unamelength);
			username[unamelength] = '\0';
		}
		
		free(fullusername);
		
		sprintf(clientsockpath, STREAMCONFIG_CLIENT_SOCKET, username);
		sprintf(serversockpath, STREAMCONFIG_SERVER_SOCKET, username);
		
		unlink(serversockpath);
		memset(&ctlserver, 0, sizeof(struct sockaddr_un));
		ctlserver.sun_family = AF_LOCAL;
		strncpy(ctlserver.sun_path, serversockpath, strlen(serversockpath));
		
		bind(ctlsockfd, (const struct sockaddr *)&ctlserver, sizeof(struct sockaddr_un));
		
		setsockopt(ctlsockfd, SOL_SOCKET, SO_PASSCRED, &onflag, sizeof(onflag));
		
		return ctlsockfd;
	}
	
	return -1;
}

void svrsettings_config_apply_settings(struct _svrsettings *sp, struct memblock *procmblock)
{
	char tmpbuffer[INET6_ADDRSTRLEN];
	
	switch(sp->request) {
		case STREAMCONFIG_SET_RECORD_STATUS: {
			if (!sp->recordstatus) {
				if (procmblock->recording) {
					streamserver_stop_record();
				}
			} else {
				if (!procmblock->recording) {
					streamserver_init_record();
				}
			}
			
			sp->result = STREAMCONFIG_RES_SUCCESS;
			break;
		}
		
		case STREAMCONFIG_SET_UDP_BC_STATUS: {
			if (!sp->udpbcstatus) {
				if (sysutils_process_exists(procmblock->clients[UDP_BC_PROCESS_INDEX].pid)) {
					streamserver_stop_udp_broadcast();
				}
			} else {
				if (!sysutils_process_exists(procmblock->clients[UDP_BC_PROCESS_INDEX].pid)) {
					streamserver_init_udp_broadcast(&procmblock->clients[UDP_BC_PROCESS_INDEX].address);
				}
			}
			
			sp->result = STREAMCONFIG_RES_SUCCESS;
			break;
		}
		
		case STREAMCONFIG_SET_UDP_MC_STATUS: {
			if (!sp->udpmcstatus) {
				if (sysutils_process_exists(procmblock->clients[UDP_MC_PROCESS_INDEX].pid)) {
					streamserver_stop_udp_multicast();
				}
			} else {
				if (!sysutils_process_exists(procmblock->clients[UDP_MC_PROCESS_INDEX].pid)) {
					streamserver_init_udp_multicast(&procmblock->clients[UDP_MC_PROCESS_INDEX].address);
				}
			}
			
			sp->result = STREAMCONFIG_RES_SUCCESS;
			break;
		}
		
		case STREAMCONFIG_SET_TCP_UC_STATUS: {
			if (!sp->tcpucstatus) {
				if (sysutils_process_exists(procmblock->clients[LISTEN_PROCESS_INDEX].pid)) {
					streamserver_stop_tcp_listen_process();
				}
			} else {
				if (!sysutils_process_exists(procmblock->clients[LISTEN_PROCESS_INDEX].pid)) {
					streamserver_init_tcp_listen_process();
				}
			}
			
			sp->result = STREAMCONFIG_RES_SUCCESS;
			break;
		}
		
		case STREAMCONFIG_SET_STREAM_DATA: {
			streamserver_change_service((char *)&sp->servicename, sp->caid, (char *)&sp->cakey);
			
			sp->result = STREAMCONFIG_RES_SUCCESS;
			break;
		}
		
		case STREAMCONFIG_SET_CONNECTION_DETAILS: {
			//UDP BROADCAST
			memset(&procmblock->clients[UDP_BC_PROCESS_INDEX].ifname, 0, sizeof(procmblock->clients[UDP_BC_PROCESS_INDEX].ifname));
			memcpy(&procmblock->clients[UDP_BC_PROCESS_INDEX].ifname, sp->udpbcint, sizeof(procmblock->clients[UDP_BC_PROCESS_INDEX].ifname));
			inet_pton(AF_INET, UDP_BROADCAST_ADDRESS, &procmblock->clients[UDP_BC_PROCESS_INDEX].address.sin_addr);
			procmblock->clients[UDP_BC_PROCESS_INDEX].address.sin_port = htons(sp->udpbcport);
						
			//UDP MULTICAST
			memset(&procmblock->clients[UDP_MC_PROCESS_INDEX].ifname, 0, sizeof(procmblock->clients[UDP_MC_PROCESS_INDEX].ifname));
			memcpy(&procmblock->clients[UDP_MC_PROCESS_INDEX].ifname, sp->udpmcint, sizeof(procmblock->clients[UDP_MC_PROCESS_INDEX].ifname));
			memset(&procmblock->clients[UDP_MC_PROCESS_INDEX].address, 0, sizeof(procmblock->clients[UDP_MC_PROCESS_INDEX].address));
			procmblock->clients[UDP_MC_PROCESS_INDEX].address.sin_family = AF_INET;
			if ((inet_pton(AF_INET, sp->udpmcgroup, &procmblock->clients[UDP_MC_PROCESS_INDEX].address.sin_addr) != 1)) {
				inet_pton(AF_INET, UDP_MULTICAST_GROUP, &procmblock->clients[UDP_MC_PROCESS_INDEX].address.sin_addr);
				procmblock->clients[UDP_MC_PROCESS_INDEX].address.sin_port = htons(UDP_MULTICAST_PORT);
			} else {
				procmblock->clients[UDP_MC_PROCESS_INDEX].address.sin_port = htons(sp->udpmcport);
			}
			procmblock->udpmcloop = sp->udpmcloop;
			procmblock->udpmcttl = sp->udpmcttl;
			
			//TCP UNICAST
			memset(&procmblock->clients[LISTEN_PROCESS_INDEX].address, 0, sizeof(procmblock->clients[LISTEN_PROCESS_INDEX].address));
			procmblock->clients[LISTEN_PROCESS_INDEX].address.sin_family = AF_INET;
			procmblock->clients[LISTEN_PROCESS_INDEX].address.sin_addr.s_addr = htonl(INADDR_ANY);
			procmblock->clients[LISTEN_PROCESS_INDEX].address.sin_port = htons(sp->tcpucport);
			
			memset(&procmblock->clients[LISTEN_PROCESS_INDEX].addressv6, 0, sizeof(procmblock->clients[LISTEN_PROCESS_INDEX].addressv6));
			procmblock->clients[LISTEN_PROCESS_INDEX].addressv6.sin6_family = AF_INET6;
			inet_pton(AF_INET6, "0::0", &procmblock->clients[LISTEN_PROCESS_INDEX].addressv6.sin6_addr);
			procmblock->clients[LISTEN_PROCESS_INDEX].addressv6.sin6_port = htons(sp->tcpucport);
			
			procmblock->tcpqueue = sp->tcpucqueue;
			procmblock->tcpclients = sp->tcpucclients;
			
			sp->result = STREAMCONFIG_RES_SUCCESS;
			break;
		}
		
		case STREAMCONFIG_SET_DEVICE: {
			procmblock->adapter = sp->adapter;
			procmblock->dvr = sp->dvr;
			streamserver_open_device(procmblock->adapter, procmblock->dvr);
			
			sp->result = STREAMCONFIG_RES_SUCCESS;
			break;
		}
		
		case STREAMCONFIG_STOP_CLIENT:
			streamserver_stop_client(sp->clientid);
			
			sp->result = STREAMCONFIG_RES_SUCCESS;
			break;
		
		case STREAMCONFIG_SET_FILECONFIG: {
			//UDP BROADCAST
			memset(&procmblock->clients[UDP_BC_PROCESS_INDEX].ifname, 0, sizeof(procmblock->clients[UDP_BC_PROCESS_INDEX].ifname));
			memcpy(&procmblock->clients[UDP_BC_PROCESS_INDEX].ifname, sp->udpbcint, sizeof(procmblock->clients[UDP_BC_PROCESS_INDEX].ifname));
			inet_pton(AF_INET, UDP_BROADCAST_ADDRESS, &procmblock->clients[UDP_BC_PROCESS_INDEX].address.sin_addr);
			procmblock->clients[UDP_BC_PROCESS_INDEX].address.sin_port = htons(sp->udpbcport);
						
			//UDP MULTICAST
			memset(&procmblock->clients[UDP_MC_PROCESS_INDEX].ifname, 0, sizeof(procmblock->clients[UDP_MC_PROCESS_INDEX].ifname));
			memcpy(&procmblock->clients[UDP_MC_PROCESS_INDEX].ifname, sp->udpmcint, sizeof(procmblock->clients[UDP_MC_PROCESS_INDEX].ifname));
			memset(&procmblock->clients[UDP_MC_PROCESS_INDEX].address, 0, sizeof(procmblock->clients[UDP_MC_PROCESS_INDEX].address));
			procmblock->clients[UDP_MC_PROCESS_INDEX].address.sin_family = AF_INET;
			if ((inet_pton(AF_INET, sp->udpmcgroup, &procmblock->clients[UDP_MC_PROCESS_INDEX].address.sin_addr) != 1)) {
				inet_pton(AF_INET, UDP_MULTICAST_GROUP, &procmblock->clients[UDP_MC_PROCESS_INDEX].address.sin_addr);
				procmblock->clients[UDP_MC_PROCESS_INDEX].address.sin_port = htons(UDP_MULTICAST_PORT);
			} else {
				procmblock->clients[UDP_MC_PROCESS_INDEX].address.sin_port = htons(sp->udpmcport);
			}
			procmblock->udpmcloop = sp->udpmcloop;
			procmblock->udpmcttl = sp->udpmcttl;
						
			//TCP UNICAST
			memset(&procmblock->clients[LISTEN_PROCESS_INDEX].address, 0, sizeof(procmblock->clients[LISTEN_PROCESS_INDEX].address));
			procmblock->clients[LISTEN_PROCESS_INDEX].address.sin_family = AF_INET;
			procmblock->clients[LISTEN_PROCESS_INDEX].address.sin_addr.s_addr = htonl(INADDR_ANY);
			procmblock->clients[LISTEN_PROCESS_INDEX].address.sin_port = htons(sp->tcpucport);
			
			memset(&procmblock->clients[LISTEN_PROCESS_INDEX].addressv6, 0, sizeof(procmblock->clients[LISTEN_PROCESS_INDEX].addressv6));
			procmblock->clients[LISTEN_PROCESS_INDEX].addressv6.sin6_family = AF_INET6;
			inet_pton(AF_INET6, "0::0", &procmblock->clients[LISTEN_PROCESS_INDEX].addressv6.sin6_addr);
			procmblock->clients[LISTEN_PROCESS_INDEX].addressv6.sin6_port = htons(sp->tcpucport);
			
			procmblock->tcpqueue = sp->tcpucqueue;
			procmblock->tcpclients = sp->tcpucclients;
			
			//DEVICE
			procmblock->adapter = sp->adapter;
			procmblock->dvr = sp->dvr;
			
			//FLAGS
			if (!sp->recordstatus) {
				if (procmblock->recording) {
					streamserver_stop_record();
				}
			} else {
				if (!procmblock->recording) {
					streamserver_init_record();
				}
			}
			
			if (!sp->udpbcstatus) {
				if (sysutils_process_exists(procmblock->clients[UDP_BC_PROCESS_INDEX].pid)) {
					streamserver_stop_udp_broadcast();
				}
			} else {
				if (!sysutils_process_exists(procmblock->clients[UDP_BC_PROCESS_INDEX].pid)) {
					streamserver_init_udp_broadcast(&procmblock->clients[UDP_BC_PROCESS_INDEX].address);
				}
			}
			
			if (!sp->udpmcstatus) {
				if (sysutils_process_exists(procmblock->clients[UDP_MC_PROCESS_INDEX].pid)) {
					streamserver_stop_udp_multicast();
				}
			} else {
				if (!sysutils_process_exists(procmblock->clients[UDP_MC_PROCESS_INDEX].pid)) {
					streamserver_init_udp_multicast(&procmblock->clients[UDP_MC_PROCESS_INDEX].address);
				}
			}
			
			if (!sp->tcpucstatus) {
				if (sysutils_process_exists(procmblock->clients[LISTEN_PROCESS_INDEX].pid)) {
					streamserver_stop_tcp_listen_process();
				}
			} else {
				if (!sysutils_process_exists(procmblock->clients[LISTEN_PROCESS_INDEX].pid)) {
					streamserver_init_tcp_listen_process();
				}
			}
			
			procmblock->sheduler = sp->sheduler;
			
			sp->result = STREAMCONFIG_RES_SUCCESS;
			break;
		}
		
		case STREAMCONFIG_GET_SERVICE_FLAGS: {
			if (procmblock->recording) sp->recordstatus = 1; else sp->recordstatus = 0;
			if (sysutils_process_exists(procmblock->clients[UDP_BC_PROCESS_INDEX].pid)) sp->udpbcstatus = 1; else sp->udpbcstatus = 0;
			if (sysutils_process_exists(procmblock->clients[UDP_MC_PROCESS_INDEX].pid)) sp->udpmcstatus = 1; else sp->udpmcstatus = 0;
			if (sysutils_process_exists(procmblock->clients[LISTEN_PROCESS_INDEX].pid)) sp->tcpucstatus = 1; else sp->tcpucstatus = 0;
			
			sp->result = STREAMCONFIG_RES_SUCCESS;
			break;
		}
		
		case STREAMCONFIG_GET_CONNECTION_DETAILS: {
			//UDP BROADCAST
			memset(&sp->udpbcint, 0, sizeof(sp->udpbcint));
			if (procmblock->clients[UDP_BC_PROCESS_INDEX].ifname[0] != '\0') {
				memcpy(&sp->udpbcint, procmblock->clients[UDP_BC_PROCESS_INDEX].ifname, sizeof(sp->udpbcint));
			} else {
				sprintf((char *)&sp->udpbcint, UDP_MULTICAST_INTERFACE);
			}
			sp->udpbcport = ntohs(procmblock->clients[UDP_BC_PROCESS_INDEX].address.sin_port);
						
			//UDP MULTICAST
			memset(&sp->udpmcint, 0, sizeof(sp->udpmcint));
			if (procmblock->clients[UDP_MC_PROCESS_INDEX].ifname[0] != '\0') {
				memcpy(&sp->udpmcint, procmblock->clients[UDP_MC_PROCESS_INDEX].ifname, sizeof(sp->udpmcint));
			} else {
				sprintf((char *)&sp->udpmcint, UDP_MULTICAST_INTERFACE);
			}
			if (inet_ntop(AF_INET, &procmblock->clients[UDP_MC_PROCESS_INDEX].address.sin_addr, sp->udpmcgroup, sizeof(sp->udpmcgroup)) != NULL) {
				sp->udpmcport = ntohs(procmblock->clients[UDP_MC_PROCESS_INDEX].address.sin_port);
			} else {
				sprintf((char *)&sp->udpmcgroup, UDP_MULTICAST_GROUP);
				sp->udpmcport = UDP_MULTICAST_PORT;
			}
			sp->udpmcloop = procmblock->udpmcloop;
			sp->udpmcttl = procmblock->udpmcttl;
			
			//TCP UNICAST
			if (inet_ntop(AF_INET, &procmblock->clients[LISTEN_PROCESS_INDEX].address.sin_addr, tmpbuffer, sizeof(tmpbuffer)) != NULL) {
				sp->tcpucport = ntohs(procmblock->clients[LISTEN_PROCESS_INDEX].address.sin_port);
			} else {
				sp->tcpucport = TCP_UNICAST_PORT;
			}
			sp->tcpucclients = procmblock->tcpclients;
			sp->tcpucqueue = procmblock->tcpqueue;
			
			sp->result = STREAMCONFIG_RES_SUCCESS;
			break;
		}
		
		case STREAMCONFIG_GET_SERVER_STATISTICS:
			sp->udpmcclientsnum = sysutils_process_exists(procmblock->clients[UDP_MC_PROCESS_INDEX].pid);
			sp->udpbcclientsnum = sysutils_process_exists(procmblock->clients[UDP_BC_PROCESS_INDEX].pid);
			sp->tcpucclientsnum = (unsigned int)procmblock->numtcpclients;
			sp->tcpucclientslimit = (unsigned int)procmblock->tcpclients;
			sp->lastsessiontime = (time_t)procmblock->tsinfo.starttime;
			
			sp->result = STREAMCONFIG_RES_SUCCESS;
			break;
		
			
		default: break;
	}
	
}

static char *svrsettings_control_message_execute(char *inbuffer, size_t *outbufsize, int execute, struct memblock *procmblock)
{
	struct _svrsettings recvparams;
	char formatstr[1024];
	char *outbuffer;
	int clicount, i;
	
	if ((inbuffer == NULL) || (outbufsize == NULL) || (procmblock == NULL)) return NULL;
	
	outbuffer = NULL;
	
	if (!execute) {
		outbuffer = serialize_form_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_ANSWER_REQ, sizeof(formatstr)),
										0, outbufsize, SERIALIZE_REQUEST(inbuffer), STREAMCONFIG_RES_FAILURE);
		return outbuffer;
	}
	
	switch (SERIALIZE_REQUEST(inbuffer)) {
		case STREAMCONFIG_SET_RECORD_STATUS: 
			serialize_parse_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, SET_RECORD_STATUS_REQ, sizeof(formatstr)),
										inbuffer, 0, &recvparams.request, &recvparams.recordstatus);
			break;
		
		case STREAMCONFIG_SET_UDP_BC_STATUS: 
			serialize_parse_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, SET_UDPBC_STATUS_REQ, sizeof(formatstr)),
										inbuffer, 0, &recvparams.request, &recvparams.udpbcstatus);
			break; 
		
		case STREAMCONFIG_SET_UDP_MC_STATUS: 
			serialize_parse_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, SET_UDPMC_STATUS_REQ, sizeof(formatstr)),
										inbuffer, 0, &recvparams.request, &recvparams.udpmcstatus);
			break;
		
		case STREAMCONFIG_SET_TCP_UC_STATUS: 
			serialize_parse_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, SET_TCPUC_STATUS_REQ, sizeof(formatstr)),
										inbuffer, 0, &recvparams.request, &recvparams.tcpucstatus);
			break;
		
		case STREAMCONFIG_SET_STREAM_DATA: 
			memset(&procmblock->streampids, 0, sizeof(procmblock->streampids));
			serialize_parse_message_static(SERIALIZE_FORM_REQSTR((char *)&formatstr, SET_STREAM_DATA_REQ, sizeof(formatstr)),
										inbuffer, 0, &recvparams.request, &recvparams.servicename, &recvparams.cakey, &recvparams.caid);
			break;
			
		case STREAMCONFIG_SET_STREAM_CONTENT:
			serialize_parse_array_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, SET_STREAM_CONTENT_ARRAY_REQ, sizeof(formatstr)),
										NULL, inbuffer, NULL, svrsettings_mediastreams_callback, (void *)procmblock);
			break;
			
		case STREAMCONFIG_SET_CONNECTION_DETAILS: 
			serialize_parse_message_static(SERIALIZE_FORM_REQSTR((char *)&formatstr, SET_CONNECTION_DETAILS_REQ, sizeof(formatstr)),
										inbuffer, 0, &recvparams.request, &recvparams.udpbcint, &recvparams.udpbcport, &recvparams.udpmcint,
										&recvparams.udpmcgroup, &recvparams.udpmcport, &recvparams.udpmcloop, &recvparams.udpmcttl, &recvparams.tcpucport,
										&recvparams.tcpucqueue, &recvparams.tcpucclients);
			break;
			
		case STREAMCONFIG_SET_DEVICE: 
			serialize_parse_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, SET_DEVICE_REQ, sizeof(formatstr)),
										inbuffer, 0, &recvparams.request, &recvparams.adapter, &recvparams.dvr);
			break;
			
		case STREAMCONFIG_STOP_CLIENT: 
			serialize_parse_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, STOP_CLIENT_REQ, sizeof(formatstr)),
										inbuffer, 0, &recvparams.request, &recvparams.clientid);
			break;
		
		case STREAMCONFIG_GET_CLIENTS: 
			serialize_parse_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_SIMPLE_REQ, sizeof(formatstr)),
										inbuffer, 0, &recvparams.request);
			break;
			
		case STREAMCONFIG_GET_CONNECTION_DETAILS: 
			serialize_parse_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_SIMPLE_REQ, sizeof(formatstr)),
										inbuffer, 0, &recvparams.request);
			break;
			
		case STREAMCONFIG_GET_SERVICE_FLAGS: 
			serialize_parse_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_SIMPLE_REQ, sizeof(formatstr)),
										inbuffer, 0, &recvparams.request);
			break;	
		
		case STREAMCONFIG_GET_STREAM_STRUCTURE:
			serialize_parse_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_SIMPLE_REQ, sizeof(formatstr)),
										inbuffer, 0, &recvparams.request);
			break;
		
		case STREAMCONFIG_GET_SERVER_STATISTICS:
			serialize_parse_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_SIMPLE_REQ, sizeof(formatstr)),
										inbuffer, 0, &recvparams.request);
			break;
		
		default: return NULL;
	}
	
	svrsettings_config_apply_settings(&recvparams, procmblock);
			
	switch (SERIALIZE_REQUEST(inbuffer)) {
		case STREAMCONFIG_SET_RECORD_STATUS: 
			outbuffer = serialize_form_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_ANSWER_REQ, sizeof(formatstr)),
										0, outbufsize, recvparams.request, recvparams.result);
			break;
		
		case STREAMCONFIG_SET_UDP_BC_STATUS: 
			outbuffer = serialize_form_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_ANSWER_REQ, sizeof(formatstr)),
										0, outbufsize, recvparams.request, recvparams.result);
			break; 
		
		case STREAMCONFIG_SET_UDP_MC_STATUS: 
			outbuffer = serialize_form_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_ANSWER_REQ, sizeof(formatstr)),
										0, outbufsize, recvparams.request, recvparams.result);
			break;
		
		case STREAMCONFIG_SET_TCP_UC_STATUS: 
			outbuffer = serialize_form_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_ANSWER_REQ, sizeof(formatstr)),
										0, outbufsize, recvparams.request, recvparams.result);
			break;
		
		case STREAMCONFIG_SET_STREAM_DATA: 
			outbuffer = serialize_form_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_ANSWER_REQ, sizeof(formatstr)),
										0, outbufsize, recvparams.request, recvparams.result);
			break;
			
		case STREAMCONFIG_SET_STREAM_CONTENT:
			outbuffer = serialize_form_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_ANSWER_REQ, sizeof(formatstr)),
										0, outbufsize, recvparams.request, recvparams.result);
			break;
			
		case STREAMCONFIG_SET_CONNECTION_DETAILS: 
			outbuffer = serialize_form_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_ANSWER_REQ, sizeof(formatstr)),
										0, outbufsize, recvparams.request, recvparams.result);
			break;
			
		case STREAMCONFIG_SET_DEVICE: 
			outbuffer = serialize_form_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_ANSWER_REQ, sizeof(formatstr)),
										0, outbufsize, recvparams.request, recvparams.result);
			break;
			
		case STREAMCONFIG_STOP_CLIENT: 
			outbuffer = serialize_form_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_ANSWER_REQ, sizeof(formatstr)),
										0, outbufsize, recvparams.request, recvparams.result);
			break;
		
		case STREAMCONFIG_GET_CLIENTS: 
			cliptr = 0;
			
			clicount = 0;
			for (i = 0; i<MAXCLIENTS; i++)
				if (procmblock->clients[i].pid > 0) 
					clicount++;
			
			outbuffer = serialize_form_array_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_CLIENTS_ARRAY_REQ, sizeof(formatstr)),
										STREAMCONFIG_GET_CLIENTS, clicount, outbufsize, svrsettings_clients_list_callback, (void *)procmblock);
			break;
			
		case STREAMCONFIG_GET_CONNECTION_DETAILS: 
			outbuffer = serialize_form_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_CONNECTION_DETAILS_REQ, sizeof(formatstr)),
									0, outbufsize, recvparams.request, recvparams.result, recvparams.udpbcint, recvparams.udpbcport, recvparams.udpmcint,
									recvparams.udpmcgroup, recvparams.udpmcport, recvparams.udpmcloop, recvparams.udpmcttl, recvparams.tcpucport, recvparams.tcpucqueue, recvparams.tcpucclients);
			break;
			
		case STREAMCONFIG_GET_SERVICE_FLAGS: 
			outbuffer = serialize_form_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_SERVICE_FLAGS_REQ, sizeof(formatstr)),
									0, outbufsize, recvparams.request, recvparams.result, recvparams.recordstatus, recvparams.udpbcstatus, recvparams.udpmcstatus, recvparams.tcpucstatus);
			break;	
		
		case STREAMCONFIG_GET_STREAM_STRUCTURE:
			strptr = 0;
			
			clicount = 0;
			for (i = 0; i<TS_NUMBER_PIDS; i++)
				if (procmblock->tsinfo.pids[i].stream_type != TS_STREAM_UNKNOWN) 
					clicount++;
			
			outbuffer = serialize_form_array_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_STREAM_STRUCTURE_ARRAY_REQ, sizeof(formatstr)),
									STREAMCONFIG_GET_STREAM_STRUCTURE, clicount, outbufsize, svrsettings_stream_structure_callback, (void *)procmblock);
			break;
		
		case STREAMCONFIG_GET_SERVER_STATISTICS:
			outbuffer = serialize_form_message(SERIALIZE_FORM_REQSTR((char *)&formatstr, GET_SERVER_STATISTICS_REQ, sizeof(formatstr)),
												0, outbufsize, recvparams.request, recvparams.result, 
												recvparams.udpmcclientsnum, recvparams.udpbcclientsnum, recvparams.tcpucclientsnum,
												recvparams.tcpucclientslimit, recvparams.lastsessiontime);
			break;
		
		default: return NULL;
	}
		
	return outbuffer;	
}

void svrsettings_read_control_socket(int ctlsockfd, struct memblock *procmblock)
{
	struct sockaddr_un client;
	static struct msghdr msg;
	struct iovec iov[1];
	union {
        struct cmsghdr cmh;
        char   control[CMSG_SPACE(sizeof(struct ucred))];
    } control_un;
    struct cmsghdr *msgch;
	struct ucred *cred;
	
	char inbuffer[4096];
	size_t inbufsize = 4096;
	char *outbuffer = NULL;
	size_t outbufsize;
	int bytes;
	
	
	while (1) {
		msg.msg_name = &client;
		msg.msg_namelen = sizeof(struct sockaddr_un);
		msg.msg_iov = iov;
		msg.msg_iovlen = 1;
		iov[0].iov_base = &inbuffer;
		iov[0].iov_len = inbufsize;
		
		msg.msg_control = control_un.control;
		msg.msg_controllen = sizeof(control_un.control);
		
		control_un.cmh.cmsg_len = CMSG_LEN(sizeof(struct ucred));
		control_un.cmh.cmsg_level = SOL_SOCKET;
		control_un.cmh.cmsg_type = SCM_CREDENTIALS;
		
		if ((bytes = recvmsg(ctlsockfd, &msg, 0)) == -1) break;
		
		msgch = CMSG_FIRSTHDR(&msg);
		if ((msgch != NULL) && (msgch->cmsg_type == SCM_CREDENTIALS)) {
			cred = (struct ucred *)CMSG_DATA(msgch);
			if ((cred->uid == getuid()) && (cred->gid == getgid()) && (sysutils_process_exists(cred->pid))) {
				if ((procmblock->hostpid == cred->pid) || (procmblock->hostpid == 0) || ((procmblock->hostpid != 0) && (!sysutils_process_exists(procmblock->hostpid)))) {
					if ((outbuffer = svrsettings_control_message_execute((char *)&inbuffer, &outbufsize, 1, procmblock)) == NULL) continue;
					if ((procmblock->hostpid == 0) || (procmblock->hostpid != cred->pid)) {
						streamserver_register_host_process(cred->pid);
					}
				} else {
					if ((outbuffer = svrsettings_control_message_execute((char *)&inbuffer, &outbufsize, 0, procmblock)) == NULL) continue;
					streamserver_send_log_message(LOG_EVENT_WARNING, "Unknown host process PID: %u, UID: %u, GID: %u message not executed", cred->pid, cred->uid, cred->gid);
				}	
				
				msg.msg_name = &client;
				msg.msg_namelen = sizeof(struct sockaddr_un);
				msg.msg_iov = iov;
				msg.msg_iovlen = 1;
				iov[0].iov_base = outbuffer;
				iov[0].iov_len = outbufsize;
				
				msg.msg_control = control_un.control;
				msg.msg_controllen = sizeof(control_un.control);
				
				msgch = CMSG_FIRSTHDR(&msg);
				msgch->cmsg_len = CMSG_LEN(sizeof(struct ucred));
				msgch->cmsg_level = SOL_SOCKET;
				msgch->cmsg_type = SCM_CREDENTIALS;
				cred = (struct ucred *)CMSG_DATA(msgch);
				
				cred->pid = getpid();
				cred->uid = getuid();
				cred->gid = getgid();
				
				bytes = sendmsg(ctlsockfd, &msg, 0);
				
				if (outbuffer != NULL) free(outbuffer);
			}
			
		}
		
	}
}

void svrsettings_read_config_file(struct _svrsettings *sp, int firsttime)
{
	const char *homedir;
	char filename[PATH_MAX];
	FILE *config;
	char string[4096];
	char *del;
	char *parameter;
	char *argument;
	int chr, len;
	int flag = 0;
	int length;
	
	memset(sp, 0, sizeof(struct _svrsettings));
	sp->request = STREAMCONFIG_SET_FILECONFIG;
	sprintf((char *)&sp->udpbcint, UDP_BROADCAST_INTERFACE);
	sp->udpbcport = UDP_BROADCAST_PORT;
	sprintf((char *)&sp->udpmcint, UDP_MULTICAST_INTERFACE);
	sprintf((char *)&sp->udpmcgroup, UDP_MULTICAST_GROUP);
	sp->udpmcport = UDP_MULTICAST_PORT;
	sp->udpmcloop = UDP_MULTICAST_LOOPBACK;
	sp->udpmcttl = UDP_MULTICAST_TTL;
	sp->tcpucport = TCP_UNICAST_PORT;
	sp->tcpucclients = TCP_UNICAST_CLIENTS;
	sp->tcpucqueue = TCP_UNICAST_QUEUE;
	sp->recordstatus = 0;
	sp->udpbcstatus = 0;
	sp->udpmcstatus = 0;
	sp->tcpucstatus = 0;
	sprintf((char *)&sp->servicename, "Service");
	sp->caid = 0;
	sp->adapter = 0;
	sp->dvr = 0;
	sp->sheduler = SHEDULER_SIGNAL;
	
	homedir = sysutils_gethomedir();
	memset(filename, 0, sizeof(filename));
	sprintf((char *)&filename, STREAMCONFIG_CONFIG_PATH, homedir);
	
	if ((config = fopen(filename, "r")) == NULL) {
		streamserver_send_log_message(LOG_EVENT_WARNING, "Unable to open config file, using default settings");
		//printf("[-] Unable to open config file, using default settings\n");
		return;
	}
	
	while (fgets((char *)&string, 4096, config) != NULL) {
		
		del = strchr((char *)&string, '=');
		
		if (del != NULL) {
			len = strlen(string) - strlen(del);
			chr=0;
			while ((chr <= len-1) && (string[chr] == ' ')) chr++;
			
			if (string[chr] == ';') continue;
			
			if (len <= chr) continue;
			
			parameter = (char *)malloc(len - chr + 1);
			memcpy(parameter, &string[chr], len - chr);
			parameter[len] = '\0';
			
			if ((len = strlen(del)-1-1) == 0) continue;
			argument = (char *)malloc(len+1);
			memcpy(argument, del+1, len);
			argument[len] = '\0';	
			
			
			if (strcmp(parameter, "udp_broadcast_interface") == 0) {
				if ((length = strlen(argument)) >= IFNAMSIZ) {
					strncpy(sp->udpbcint, argument, IFNAMSIZ);
					sp->udpbcint[IFNAMSIZ] = '\0';
				} else {
					strncpy(sp->udpbcint, argument, length);
					sp->udpbcint[length] = '\0';
				}
			} else if (strcmp(parameter, "udp_multicast_interface") == 0) {
				if ((length = strlen(argument)) >= IFNAMSIZ) {
					strncpy(sp->udpmcint, argument, IFNAMSIZ);
					sp->udpmcint[IFNAMSIZ] = '\0';
				} else {
					strncpy(sp->udpmcint, argument, length);
					sp->udpmcint[length] = '\0';
				}	
			} else if (strcmp(parameter, "udp_multicast_address") == 0) {
				if ((length = strlen(argument)) >= INET_ADDRSTRLEN) {
					strncpy(sp->udpmcgroup, argument, INET_ADDRSTRLEN);
					sp->udpmcgroup[INET_ADDRSTRLEN] = '\0';
				} else {
					strncpy(sp->udpmcgroup, argument, length);
					sp->udpmcgroup[length] = '\0';
				}
			} else if (strcmp(parameter, "udp_broadcast_port") == 0) {
				flag = atoi(argument);
				if ((flag >= 1024) && (flag <= 65535)) {
					sp->udpbcport = flag;
				}
			} else if (strcmp(parameter, "udp_multicast_port") == 0) {
				flag = atoi(argument);
				if ((flag >= 1024) && (flag <= 65535)) {
					sp->udpmcport = flag;
				}
			} else if (strcmp(parameter, "udp_multicast_loopback") == 0) {
				flag = atoi(argument);
				if ((flag >= 0) && (flag < 2)) {
					sp->udpmcloop = flag;
				}
			} else if (strcmp(parameter, "udp_multicast_ttl") == 0) {
				flag = atoi(argument);
				if ((flag >= 0) && (flag <= 255)) {
					sp->udpmcttl = flag;
				}
			} else if (strcmp(parameter, "tcp_unicast_port") == 0) {
				flag = atoi(argument);
				if ((flag >= 1024) && (flag <= 65535)) {
					sp->tcpucport = flag;
				}
			} else if (strcmp(parameter, "tcp_unicast_queue") == 0) {
				flag = atoi(argument);
				if ((flag >= 1) && (flag <= 10)) {
					sp->tcpucqueue = flag;
				}
			} else if (strcmp(parameter, "tcp_unicast_clients") == 0) {
				flag = atoi(argument);
				if ((flag >= 1) && (flag <= 90)) {
					sp->tcpucclients = flag;
				}
			}  else if (strcmp(parameter, "record_status") == 0) {
				flag = atoi(argument);
				if ((flag >= 0) && (flag < 2) && (!firsttime)) {
					sp->recordstatus = flag;
				}
			} else if (strcmp(parameter, "udp_broadcast_status") == 0) {
				flag = atoi(argument);
				if ((flag >= 0) && (flag < 2)) {
					sp->udpbcstatus = flag;
				}
			} else if (strcmp(parameter, "udp_multicast_status") == 0) {
				flag = atoi(argument);
				if ((flag >= 0) && (flag < 2)) {
					sp->udpmcstatus = flag;
				}
			} else if (strcmp(parameter, "tcp_unicast_status") == 0) {
				flag = atoi(argument);
				if ((flag >= 1) && (flag <= 90)) {
					sp->tcpucstatus = flag;
				}
			} else if (strcmp(parameter, "adapter") == 0) {
				flag = atoi(argument);
				if ((flag >= 0) && (flag <= 15)) {
					sp->adapter = flag;
				}
			} else if (strcmp(parameter, "dvr") == 0) {
				flag = atoi(argument);
				if ((flag >= 0) && (flag <= 15)) {
					sp->dvr = flag;
				}
			} else if (strcmp(parameter, "sheduler") == 0) {
				if (firsttime) {
					if (strcmp(argument, "signal") == 0) {
						sp->sheduler = SHEDULER_SIGNAL;
					} else if (strcmp(argument, "rwlock") == 0) {
						sp->sheduler = SHEDULER_RWLOCK;
					}
				}
			}
			
			free(parameter);
			free(argument);
		}
	}
	
	fclose(config);
}

void svrsettings_clients_list_callback(unsigned int *params, unsigned int count, int seq, int size, char **buffer, void *data)
{
	int i;
	char abuf[55];
	int alength;
	struct memblock *procmblock;
		
	procmblock = (struct memblock *)data;
	
	memset(abuf, 0, sizeof(abuf));
	
	if (procmblock->clients[cliptr].family == CLIENT_FAMILY_HELPER) {
		alength = sprintf((char *)&abuf, "%s", procmblock->nodename);
	} else if (procmblock->clients[cliptr].family == CLIENT_FAMILY_V6_TRANSPORT) {
		alength = sprintf((char *)&abuf, "%s:%d", inet_ntop(AF_INET6, (struct in_addr6 *)&(procmblock->clients[cliptr].addressv6.sin6_addr), abuf, sizeof(abuf)), ntohs(procmblock->clients[cliptr].addressv6.sin6_port));
	} else if (procmblock->clients[cliptr].family == CLIENT_FAMILY_V4_TRANSPORT) {
		alength = sprintf((char *)&abuf, "%s:%d", inet_ntop(AF_INET, (struct in_addr *)&(procmblock->clients[cliptr].address.sin_addr), abuf, sizeof(abuf)), ntohs(procmblock->clients[cliptr].address.sin_port));
	}
	abuf[alength] = '\0';
		
	serialize_form_message_buffer(params, count, buffer, MSGOFFSET(seq, size), procmblock->clients[cliptr].family, procmblock->clients[cliptr].type, procmblock->clients[cliptr].pid, cliptr, procmblock->clients[cliptr].pidfilter,
									&abuf, &procmblock->clients[cliptr].useragent, procmblock->clients[cliptr].starttime, procmblock->clients[cliptr].transmittedbytes);
	
	cliptr++;
	
	for (i = cliptr; i<MAXCLIENTS; i++) {
		if (procmblock->clients[i].pid > 0) {
			cliptr = i;
			break;
		}
	}
		
}

void svrsettings_stream_structure_callback(unsigned int *params, unsigned int count, int seq, int size, char **buffer, void *data)
{
	int i;
	struct memblock *procmblock;
	char description[64];
	unsigned char type, format, value, percentage;
	unsigned short pid;
	unsigned int bandwidth, revision, param1, param2, param3, param4;
		
	procmblock = (struct memblock *)data;
	
	for (i = strptr; i<TS_NUMBER_PIDS; i++) {
		if (procmblock->tsinfo.pids[i].stream_type != TS_STREAM_UNKNOWN) {
			strptr = i;
			break;
		}
	}
	
	type = (unsigned char)procmblock->tsinfo.pids[i].stream_type;
	format = (unsigned char)procmblock->tsinfo.pids[i].stream_format;
	value = (unsigned char)procmblock->tsinfo.pids[i].stream_value;
	pid = (unsigned short)strptr;
	percentage = (unsigned char)procmblock->tsinfo.pids[i].percentage;
	bandwidth = procmblock->tsinfo.pids[i].bandwidth;
	revision = procmblock->tsinfo.pids[i].stream_revision;
	
	if (procmblock->tsinfo.pids[i].stream_type == TS_STREAM_PES_AUDIO_STREAM) {
		param1 = (procmblock->tsinfo.pids[i].params.audio.version << 16) | procmblock->tsinfo.pids[i].params.audio.revision;
		param2 = (procmblock->tsinfo.pids[i].params.audio.profile << 16) | procmblock->tsinfo.pids[i].params.audio.mode;
		param3 = procmblock->tsinfo.pids[i].params.audio.bitrate;
		param4 = procmblock->tsinfo.pids[i].params.audio.samplerate;
	} else if (procmblock->tsinfo.pids[i].stream_type == TS_STREAM_PES_VIDEO_STREAM) {
		param1 = procmblock->tsinfo.pids[i].params.video.picwidth;
		param2 = procmblock->tsinfo.pids[i].params.video.picheight;
		param3 = procmblock->tsinfo.pids[i].params.video.aspect;
		param4 = procmblock->tsinfo.pids[i].params.video.framerate;
	} else {
		param1 = param2 = param3 = param4 = 0;
	}
	
	transportstream_print_stream_params((struct _tspid *)&procmblock->tsinfo.pids[i], (char *)&description, sizeof(description));
	
	serialize_form_message_buffer(params, count, buffer, MSGOFFSET(seq, size), type, format, value, percentage, pid, bandwidth, revision, param1, param2, param3, param4, (char *)&description);
	
	strptr++;
}

void svrsettings_mediastreams_callback(unsigned int *params, unsigned int count, int seq, int size, char **message, void *data)
{
	struct memblock *procmblock;
	
	if (seq > 255) return;
	
	procmblock = (struct memblock *)data;
	
	serialize_parse_message_buffer(params, count, message, MSGOFFSET(seq, size), &procmblock->streampids[seq].type, &procmblock->streampids[seq].value, &procmblock->streampids[seq].enabled);
}
