/*
 *      transportstream.c
 *      
 *      Copyright 2009-2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */     


#include <stdio.h>
#include <string.h>
#include <time.h>

#include "streamserver.h"
#include "transportstream.h"

#include "../common/log.h"

static unsigned int transportstream_get_bit(char *data, int *bitptr);
static unsigned int transportstream_get_bitstream(char *data, int *bitptr, int number);
static unsigned int transportstream_decode_golomb(char *data, int *bitptr);

static unsigned int transportstream_get_bit(char *data, int *bitptr)
{
	int mask, index;
	
	mask = 1 << (7 - (*bitptr & 7));
	index = *bitptr >> 3;
	(*bitptr)++;
	
	return ((data[index] & mask) == 0) ? 0 : 1;
}

static unsigned int transportstream_get_bitstream(char *data, int *bitptr, int number)
{
	int res, i;
	
	res = 0;
	
	for (i = 0; i < number; i++) {
		res = res * 2 + transportstream_get_bit(data, bitptr);
	}
	
	return res;
}

static unsigned int transportstream_decode_golomb(char *data, int *bitptr)
{
	int bitcount, res, i, bit;
	
	bitcount = 0;
	
	while (transportstream_get_bit(data, bitptr) == 0) {
		bitcount++;
	}
	
	res = 1;
	
	for (i = 0; i < bitcount; i++) {
		bit = transportstream_get_bit(data, bitptr);
		res = res * 2 + bit;
	}
	
	res--;
	
	return res;
}

int transportstream_decode_header(unsigned char *packet, struct _tsheader *header)
{
	if ((packet == NULL) || (header == NULL)) return 0;
	
	unsigned int pid;
	
	pid = ((((unsigned char)packet[1] & 0x1f) << 8) | (unsigned char)packet[2]) & 0x1FFF;
	
	if (pid <= 8191) {
		header->transport_error_indicator = (packet[1] & 0x80) >> 7;
		header->payload_unit_start_indicator = (packet[1] & 0x40) >> 6;
		header->transport_priority = (packet[1] & 0x20) >> 5;
		header->pid = pid; 
		header->transport_scrambling_control = (packet[3] & 0xC0) >> 7;
		header->adaptation_field_control = (packet[3] & 0x30) >> 4;
		header->continuity_counter = (packet[3] & 0x0f);
		return 1;
	} else {
		return 0;
	}
}

int transportstream_print_stream_params(struct _tspid *tspid, char *buffer, size_t bufsize)
{
	static const char * audio_modes[] = {"--", "Mono", "Stereo", "C", "L+R", "L+C+R", "L+R+S", "L+C+R+S", "L+R+SL+SR", "L+C+R+SL+SR", NULL};
	static const char * audio_profiles[] = {"--", "Default", "Main", "Low complexity", "Scalable sample rate", "Reserved", "LTP", NULL};
	
	int len;
	
	if ((buffer == NULL) || (bufsize == 0)) return 0;
	
	if (tspid->stream_format == TS_STREAM_FORMAT_SPECIAL) {
		switch (tspid->stream_type) {
			case TS_STREAM_PES_PROGRAM_STREAM_MAP:
				len = snprintf(buffer, bufsize, "Program stream map");
				break;
			case TS_STREAM_PES_PRIVATE_STREAM_1:
				len = snprintf(buffer, bufsize, "Private stream 1");
				break;
			case TS_STREAM_PES_PADDING_STREAM:
				len = snprintf(buffer, bufsize, "Padding stream");
				break;
			case TS_STREAM_PES_PRIVATE_STREAM_2:
				len = snprintf(buffer, bufsize, "Private stream 2");
				break;
			case TS_STREAM_PES_ECM_STREAM:
				len = snprintf(buffer, bufsize, "ECM stream");
				break;
			case TS_STREAM_PES_EMM_STREAM:
				len = snprintf(buffer, bufsize, "EMM stream");
				break;
			case TS_STREAM_PES_DSMCC_STREAM:
				len = snprintf(buffer, bufsize, "DSMCC stream");
				break;
			case TS_STREAM_PES_IEC_13522_STREAM:
				len = snprintf(buffer, bufsize, "ISO/IEC 13522 stream");
				break;
			case TS_STREAM_PES_H_222_1_TYPE_A:
				len = snprintf(buffer, bufsize, "ITU-T Rec. H.222.1 type A");
				break;
			case TS_STREAM_PES_H_222_1_TYPE_B:
				len = snprintf(buffer, bufsize, "ITU-T Rec. H.222.1 type B");
				break;
			case TS_STREAM_PES_H_222_1_TYPE_C:
				len = snprintf(buffer, bufsize, "ITU-T Rec. H.222.1 type C");
				break;
			case TS_STREAM_PES_H_222_1_TYPE_D:
				len = snprintf(buffer, bufsize, "ITU-T Rec. H.222.1 type D");
				break;
			case TS_STREAM_PES_H_222_1_TYPE_E:
				len = snprintf(buffer, bufsize, "ITU-T Rec. H.222.1 type E");
				break;
			case TS_STREAM_PES_ANCILLARY_STREAM:
				len = snprintf(buffer, bufsize, "Ancillary stream");
				break;
			case TS_STREAM_PES_SL_PACKETIZED_STREAM:
				len = snprintf(buffer, bufsize, "ISO/IEC14496-1 SL-packetized stream");
				break;
			case TS_STREAM_PES_FLEXMUX_STREAM:
				len = snprintf(buffer, bufsize, "ISO/IEC14496-1 FlexMux stream");
				break;
			case TS_STREAM_PES_PROGRAM_STREAM_DIRECTORY:
				len = snprintf(buffer, bufsize, "Program stream directory");
				break;
			default:
				len = snprintf(buffer, bufsize, "Reserved data stream");
				break;
		}
	} else if (tspid->stream_format == TS_STREAM_FORMAT_TABLE) {
		switch (tspid->stream_type) {
			case TS_STREAM_TABLE_PAT:
				len = snprintf(buffer, bufsize, "Program Association Table");
				break;
			case TS_STREAM_TABLE_CAT:
				len = snprintf(buffer, bufsize, "Conditional Access Table");
				break;
			case TS_STREAM_TABLE_TSDT:
				len = snprintf(buffer, bufsize, "Transport Stream Description Table");
				break;
			case TS_STREAM_TABLE_NIT:
				len = snprintf(buffer, bufsize, "Network Information Table");
				break;
			case TS_STREAM_TABLE_SDT_BAT:
				len = snprintf(buffer, bufsize, "Service Description Table, Bouquet Association Table");
				break;
			case TS_STREAM_TABLE_EIT:
				len = snprintf(buffer, bufsize, "Event Information Table");
				break;
			case TS_STREAM_TABLE_RST:
				len = snprintf(buffer, bufsize, "Running Status Table");
				break;
			case TS_STREAM_TABLE_TDT_TOT:
				len = snprintf(buffer, bufsize, "Time and Date Table, Time Offset Table");
				break;
			case TS_STREAM_TABLE_NETWORK_SYNCHRONIZATION:
				len = snprintf(buffer, bufsize, "Network synchronization data");
				break;
			case TS_STREAM_TABLE_INBAND_SIGNALLING:
				len = snprintf(buffer, bufsize, "Inband signalling data");
				break;
			case TS_STREAM_TABLE_MEASUREMENT:
				len = snprintf(buffer, bufsize, "Measurement data");
				break;
			/*case TS_STREAM_TABLE_EIT:
				len = snprintf(buffer, bufsize, "Event Information Table");
				break;*/
			case TS_STREAM_TABLE_SIT:
				len = snprintf(buffer, bufsize, "Selection Information Table");
				break;
			default:
				len = snprintf(buffer, bufsize, "Reserved data stream");
				break;
		}
	} else if (tspid->stream_format == TS_STREAM_FORMAT_UNKNOWN) {
		switch (tspid->stream_type) {
			case TS_STREAM_PES_PRIVATE_STREAM_1:
				len = snprintf(buffer, bufsize, "Private stream 1");
				break;
			case TS_STREAM_PES_AUDIO_STREAM:
				len = snprintf(buffer, bufsize, "Audio stream (%u)", tspid->stream_value);
				break;
			case TS_STREAM_PES_VIDEO_STREAM:
				len = snprintf(buffer, bufsize, "Video stream (%u)", tspid->stream_value);
				break;
			default:
				len = snprintf(buffer, bufsize, "Unknown data stream");
				break;
		}
	} else if (tspid->stream_format == TS_STREAM_FORMAT_ENCRYPTED) {
		len = snprintf(buffer, bufsize, "Encrypted data stream");
	} else if (tspid->stream_format == TS_STREAM_FORMAT_VIDEO_H262) {
		len = snprintf(buffer, bufsize, "H.262 video %ux%u, aspect %1.2f, %1.2f fps (%u)", tspid->params.video.picwidth, tspid->params.video.picheight,
																	tspid->params.video.aspect/1000.0, tspid->params.video.framerate/1000.0,
																	tspid->stream_value);
	} else if (tspid->stream_format == TS_STREAM_FORMAT_VIDEO_H264) {
		len = snprintf(buffer, bufsize, "H.264 video %ux%u, aspect %1.2f, %1.2f fps (%u)", tspid->params.video.picwidth, tspid->params.video.picheight,
																	tspid->params.video.aspect/1000.0, tspid->params.video.framerate/1000.0,
																	tspid->stream_value);
	} else if (tspid->stream_format == TS_STREAM_FORMAT_AUDIO_MPEG) {
		len = snprintf(buffer, bufsize, "MPEG-%u layer %u audio, %u kbps, %u Hz, %s (%u)", tspid->params.audio.version, tspid->params.audio.revision,
																	tspid->params.audio.bitrate, tspid->params.audio.samplerate,
																	audio_modes[tspid->params.audio.mode], tspid->stream_value);
	} else if (tspid->stream_format == TS_STREAM_FORMAT_AUDIO_AC3) {
		len = snprintf(buffer, bufsize, "AC3 v%u audio, %u kbps, %u Hz, %s (%u)", tspid->params.audio.version, tspid->params.audio.bitrate,
																	tspid->params.audio.samplerate, audio_modes[tspid->params.audio.mode],
																	tspid->stream_value);
	} else if (tspid->stream_format == TS_STREAM_FORMAT_AUDIO_AAC) {
		len = snprintf(buffer, bufsize, "MPEG %u AAC audio, profile \"%s\", %u Hz, %s (%u)", tspid->params.audio.revision,
																	audio_profiles[tspid->params.audio.profile], tspid->params.audio.samplerate,
																	audio_modes[tspid->params.audio.mode], tspid->stream_value);
	} else if (tspid->stream_format == TS_STREAM_FORMAT_AUDIO_DTS) {
		len = snprintf(buffer, bufsize, "DTS audio (%u)", tspid->stream_value);
	}
	
	buffer[len] = '\0';
	len++;
	
	return len;
}

int transportstream_decode_video_header(unsigned char *buffer, size_t length, struct _tspid *tspid)
{
	int ptr; 
	int hsize, vsize, aspectid, framerateid, bitrate;
	char nal_ref_idc = 0;
	char nal_unit_type = 0;
	
	int bitptr;
	int profile_idc, level_idc;
	int chroma_format_idc, separate_colour_plane_flag;
	int idx, sl_n, sl_i;
	int seq_parameter_set_id, log2_max_frame_num_minus4, pict_order_cnt_type;
	int num_ref_frames_in_pic_order_cnt_cycle;
	int num_ref_frames, pic_width, pic_height, frame_mbs_only_flag;
	int mb_adaptive_frame_field_flag;
	int frame_crop_left_offset, frame_crop_right_offset, frame_crop_top_offset, frame_crop_bottom_offset;
	int aspect_ratio_idc, sar_width, sar_height;
	int num_units_in_tick, time_scale, fixed_frame_rate_flag;
	float aspect = 0.0;
	float framerate = 0.0;
	
	char buf[1024];
		
	for (ptr=0; ptr<=length; ptr++) {
		if ((buffer[ptr] == 0x00) && (buffer[ptr+1] == 0x00) && (buffer[ptr+2] == 0x01)) {
			if ((buffer[ptr+3] & 0x80) && (buffer[ptr+3] == 0xb3)) {
				if (length-ptr >= 11) {
					hsize = (((buffer[ptr+4] & 0xff) << 4) | ((buffer[ptr+5] & 0xf0) >> 4));
					vsize = (((buffer[ptr+5] & 0x0f) << 8) | ((buffer[ptr+6] & 0xff)));
					aspectid = (buffer[ptr+7] & 0xf0) >> 4;
					framerateid = buffer[ptr+7] & 0x0f;
					bitrate = (((buffer[ptr+8] & 0xff) << 18) | ((buffer[ptr+9] & 0xff) << 10) | ((buffer[ptr+10] & 0xc0) >> 6));
					
					switch (aspectid) {
						case 1:
							//1:1
							aspect = 1.0;
							break;
						case 2:
							//4:3
							aspect = 1.333333333;
							break;
						case 3:
							//16:9
							aspect = 1.777777778;
							break;
						case 4:
							//2.21:1
							aspect = 2.21;
							break;
						default:
							//Unknown
							aspect = 0;
							break;
					}
					
					switch (framerateid) {
						case 1:
							framerate = 23.976;
							break;
						case 2:
							framerate = 24.0;
							break;
						case 3:
							framerate = 25.0;
							break;
						case 4:
							framerate = 29.97;
							break;
						case 5:
							framerate = 30.0;
							break;
						case 6:
							framerate = 50.0;
							break;
						case 7:
							framerate = 59.94;
							break;
						case 8:
							framerate = 60.0;
							break;
						default:
							framerate = 0;
							break;
					}
					
					tspid->params.video.picwidth = hsize;
					tspid->params.video.picheight = vsize;
					tspid->params.video.aspect = (unsigned int)(aspect * 1000);
					tspid->params.video.framerate = (unsigned int)(framerate * 1000);
					tspid->stream_format = TS_STREAM_FORMAT_VIDEO_H262;
					tspid->stream_revision++;
					
					//printf("-->H.262 video start h: %u, v: %u, a: %1.2f, f: %1.2f\n", hsize, vsize, aspect, framerate);
					
					transportstream_print_stream_params(tspid, (char *)&buf, sizeof(buf));
					streamserver_send_log_message(LOG_EVENT_MESSAGE, "Video: %s", buf);
					//printf("[+] Video: %s\n", buf);
					
					return TS_STREAM_FORMAT_VIDEO_H262;
				}
			} else if ((buffer[ptr+3] & 0x80) == 0) {
				nal_ref_idc = (buffer[ptr+3] & 0x60) >> 5;
				nal_unit_type = (buffer[ptr+3] & 0x1F);  
				
				if (nal_unit_type < 12) {
					if (((nal_ref_idc != 0) && ((nal_unit_type == 5) || (nal_unit_type == 7) || (nal_unit_type == 8))) ||
						((nal_ref_idc == 0) && ((nal_unit_type == 6) || (nal_unit_type == 9) || (nal_unit_type == 10) ||
							(nal_unit_type == 11) || (nal_unit_type == 12)))) {
						
						if ((nal_ref_idc == 1) && (nal_unit_type == 7)) {
							bitptr = 0;
														
							profile_idc = transportstream_get_bitstream(buffer+ptr+4, &bitptr, 8);
							bitptr = bitptr+8;
							level_idc = transportstream_get_bitstream(buffer+ptr+4, &bitptr, 8);
							
							if ((profile_idc == 100) || (profile_idc == 110) || (profile_idc == 122) || (profile_idc == 244) ||
									(profile_idc == 44) || (profile_idc == 83) || (profile_idc == 86)) {
								if ((chroma_format_idc = transportstream_decode_golomb(buffer+ptr+4, &bitptr)) == 3) {
									separate_colour_plane_flag = transportstream_get_bitstream(buffer+ptr+4, &bitptr, 1);
								}
								
								transportstream_decode_golomb(buffer+ptr+4, &bitptr); /*bit_depth_luma_minus8*/
								transportstream_decode_golomb(buffer+ptr+4, &bitptr); /*bit_depth_chroma_minus8*/
								bitptr++; /*qpprime_y_zero_transform_bypass_flag*/
								
								if (transportstream_get_bitstream(buffer+ptr+4, &bitptr, 1)) { /*seq_scaling_matrix_present_flag*/
									for (idx = 0; idx < ((chroma_format_idc != 3) ? 8 : 12); idx++) {
										if (transportstream_get_bitstream(buffer+ptr+4, &bitptr, 1)) { /*seq_scaling_list_present_flag*/
											sl_n = ((idx < 6) ? 16 : 64);
											for (sl_i = 0; sl_i < sl_n; sl_i++) {
												transportstream_decode_golomb(buffer+ptr+4, &bitptr);
											}
										}
									}
								}
							}
							
							seq_parameter_set_id = transportstream_decode_golomb(buffer+ptr+4, &bitptr);
							log2_max_frame_num_minus4 = transportstream_decode_golomb(buffer+ptr+4, &bitptr);
							pict_order_cnt_type = transportstream_decode_golomb(buffer+ptr+4, &bitptr);
							
							if (pict_order_cnt_type == 0) {
								transportstream_decode_golomb(buffer+ptr+4, &bitptr); /*log2_max_pic_order_cnt_lsb_minus4*/
							} else if (pict_order_cnt_type == 1) {
								bitptr++; /*delta_pic_order_always_zero_flag*/
								transportstream_decode_golomb(buffer+ptr+4, &bitptr); /*offset_for_non_ref_pic*/
								transportstream_decode_golomb(buffer+ptr+4, &bitptr); /*offset_for_top_to_bottom_field*/
								num_ref_frames_in_pic_order_cnt_cycle = transportstream_decode_golomb(buffer+ptr+4, &bitptr);
								for (idx = 0; idx < num_ref_frames_in_pic_order_cnt_cycle; idx++) {
									transportstream_decode_golomb(buffer+ptr+4, &bitptr); /*offset_for_ref_frame[ i ]*/
								}
							}
							
							num_ref_frames = transportstream_decode_golomb(buffer+ptr+4, &bitptr);
							bitptr++; /*gaps_in_frame_num_value_allowed_flag*/
							pic_width = (transportstream_decode_golomb(buffer+ptr+4, &bitptr) + 1) * 16;
							pic_height = (transportstream_decode_golomb(buffer+ptr+4, &bitptr) + 1) * 16;
							frame_mbs_only_flag = transportstream_get_bitstream(buffer+ptr+4, &bitptr, 1);
							
							if (!frame_mbs_only_flag) {
								pic_height *= 2;
								mb_adaptive_frame_field_flag = transportstream_get_bitstream(buffer+ptr+4, &bitptr, 1);
							}
							
							bitptr++; /*direct_8x8_inference_flag*/
							
							if (transportstream_get_bitstream(buffer+ptr+4, &bitptr, 1)) { /*frame_cropping_flag*/
								frame_crop_left_offset = transportstream_decode_golomb(buffer+ptr+4, &bitptr);
								frame_crop_right_offset = transportstream_decode_golomb(buffer+ptr+4, &bitptr);
								frame_crop_top_offset = transportstream_decode_golomb(buffer+ptr+4, &bitptr);
								frame_crop_bottom_offset = transportstream_decode_golomb(buffer+ptr+4, &bitptr);
							}
							
							if (transportstream_get_bitstream(buffer+ptr+4, &bitptr, 1)) { /*vui_parameters_present_flag*/
								if (transportstream_get_bitstream(buffer+ptr+4, &bitptr, 1)) { /*aspect_ratio_info_present_flag*/
									aspect_ratio_idc = transportstream_get_bitstream(buffer+ptr+4, &bitptr, 8);
									if (aspect_ratio_idc == 255) {
										sar_width = transportstream_get_bitstream(buffer+ptr+4, &bitptr, 16);
										sar_height = transportstream_get_bitstream(buffer+ptr+4, &bitptr, 16);
									}
								}
								
								if (transportstream_get_bitstream(buffer+ptr+4, &bitptr, 1)) /*overscan_info_present_flag*/
									transportstream_get_bitstream(buffer+ptr+4, &bitptr, 1); /*overscan_appropriate_flag*/
								
								if (transportstream_get_bitstream(buffer+ptr+4, &bitptr, 1)) { /*video_signal_type_present_flag*/
									transportstream_get_bitstream(buffer+ptr+4, &bitptr, 3); /*video_format*/
									transportstream_get_bitstream(buffer+ptr+4, &bitptr, 1); /*video_full_range_flag*/
									
									if (transportstream_get_bitstream(buffer+ptr+4, &bitptr, 1)) { /*colour_description_present_flag*/
										transportstream_get_bitstream(buffer+ptr+4, &bitptr, 8); /*colour_primaries*/
										transportstream_get_bitstream(buffer+ptr+4, &bitptr, 8); /*transfer_characteristics*/
										transportstream_get_bitstream(buffer+ptr+4, &bitptr, 8); /*matrix_coefficients*/
									}
								}
								
								if (transportstream_get_bitstream(buffer+ptr+4, &bitptr, 1)) { /*chroma_loc_info_present_flag*/
									transportstream_decode_golomb(buffer+ptr+4, &bitptr); /*chroma_sample_loc_type_top_field*/
									transportstream_decode_golomb(buffer+ptr+4, &bitptr); /*chroma_sample_loc_type_bottom_field*/
								}
								
								if (transportstream_get_bitstream(buffer+ptr+4, &bitptr, 1)) { /*timing_info_present_flag*/
									num_units_in_tick = transportstream_get_bitstream(buffer+ptr+4, &bitptr, 32);
									time_scale = transportstream_get_bitstream(buffer+ptr+4, &bitptr, 32);
									fixed_frame_rate_flag = transportstream_get_bitstream(buffer+ptr+4, &bitptr, 1);
									
									if (fixed_frame_rate_flag) {
										framerate = (2 * num_units_in_tick) / (float)time_scale;
									} else {
										framerate = 0;
									}
								}
									
							}
							
							if (pic_height > 0) {
								aspect = pic_width / (float)pic_height;
								
								switch (aspect_ratio_idc) {
									case 0:
										// Unspecified
										break;
									case 1:
										// 1:1
										break;
									case 2:
										// 12:11
										aspect *= 1.0909090909090908;
										break;
									case 3:
										// 10:11
										aspect *= 0.90909090909090906;
										break;
									case 4:
										// 16:11
										aspect *= 1.4545454545454546;
										break;
									case 5:
										// 40:33
										aspect *= 1.2121212121212122;
										break;
									case 6:
										// 24:11
										aspect *= 2.1818181818181817;
										break;
									case 7:
										// 20:11
										aspect *= 1.8181818181818181;
										break;
									case 8:
										// 32:11
										aspect *= 2.9090909090909092;
										break;
									case 9:
										// 80:33
										aspect *= 2.4242424242424243;
										break;
									case 10:
										// 18:11
										aspect *= 1.6363636363636365;
										break;
									case 11:
										// 15:11
										aspect *= 1.3636363636363635;
										break;
									case 12:
										// 64:33
										aspect *= 1.9393939393939394;
										break;
									case 13:
										// 160:99
										aspect *= 1.6161616161616161;
										break;
									case 255:
										if (sar_height > 0) {
											aspect *= sar_width / (float)sar_height;
										} else {
											aspect = 0.0;
										}
										break;
								}
							}	
							
							tspid->params.video.picwidth = pic_width;
							tspid->params.video.picheight = pic_height;
							tspid->params.video.aspect = (unsigned int)(aspect * 1000);
							tspid->params.video.framerate = (unsigned int)(framerate * 1000);
							tspid->stream_format = TS_STREAM_FORMAT_VIDEO_H264;
							tspid->stream_revision++;
							
							//printf("-->H.264 video start w: %u, h: %u, a: %1.2f, f: %1.2f\n", pic_width, pic_height, aspect, framerate);
							
							transportstream_print_stream_params(tspid, (char *)&buf, sizeof(buf));
							streamserver_send_log_message(LOG_EVENT_MESSAGE, "Video: %s", buf);
							//printf("[+] Video: %s\n", buf);
							
							return TS_STREAM_FORMAT_VIDEO_H264;
						}
						
						
					}
				}
			}
			
		}
	}
	
	tspid->stream_revision++;
	
	return TS_STREAM_FORMAT_UNKNOWN;
}

int transportstream_decode_audio_header(unsigned char *buffer, size_t length, int private, struct _tspid *tspid)
{
	int ptr, version, layer, abitrate, profile, samplerate, mode;
	
	static const int mpeg_br_1_1[] = {0,  32,  64,  96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448,  0};
	static const int mpeg_br_1_2[] = {0,  32,  48,  56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 384,  0};
	static const int mpeg_br_1_3[] = {0,  32,  40,  48,  56,  64,  80, 96,  112, 128, 160, 192, 224, 256, 320,  0};
	static const int mpeg_br_2_1[] = {0,  32,  48,  56,  64,  80,  96, 112, 128, 144, 160, 176, 192, 224, 256,  0};
	static const int mpeg_br_2_2[] = {0,  8,   16,  24,  32,  40,  48, 56,  64,  80,  96,  112, 128, 144, 160,  0};
	static const int * const mpeg_br_table[3][3] = {{ mpeg_br_1_1, mpeg_br_1_2, mpeg_br_1_3 }, 
													{ mpeg_br_2_1, mpeg_br_2_2, mpeg_br_2_2 },
													{ mpeg_br_2_1, mpeg_br_2_2, mpeg_br_2_2 }};
	static const int ac3_br_8[] =    {32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384, 448, 512, 576, 640};
	static const int mpeg_sr_1[] =   {44100, 48000, 32000};
	static const int mpeg_sr_2[] =   {22050, 24000, 16000};
	static const int mpeg_sr_3[] =   {11025, 12000,  8000};
	static const int aac_sr[] =      {96000, 88200, 64000, 48000, 44100, 32000, 24000, 22050, 16000, 12000, 11025, 8000, 0, 0, 0, 0};
	static const int ac3_sr_8[] =    {48000, 44100, 32000};
	
	
	char buf[1024];
		
	
	for (ptr=0; ptr<=length; ptr++) {
		if (!private) {
			if ((buffer[ptr] == 0xff) && ((buffer[ptr+1] & 0xe0) == 0xe0)) {
				if (length-ptr >= 4) {
					version = (buffer[ptr+1] & 0x18) >> 3;
					switch (version) {
						case 0:
							version = 3;
							break;
						case 1:
							version = 4;
							break;
						case 2:
							version = 2;
							break;
						case 3:
							version = 1;
							break;
					}
					
					layer = (buffer[ptr+1] & 0x06) >> 1;
					switch (layer) {
						case 0:
							layer = 4;
							if (version == 3) {
								version = 2;
							} else {
								version = 4;
							} 
							break;
						case 1:
							layer = 3;
							break;
						case 2:
							layer = 2;
							break;
						case 3:
							layer = 1;
							break;
					}
					
					if (layer == 4) {
						profile = (buffer[ptr+2] & 0xc0) >> 6;
						switch(profile) {
							case 0:
								profile = TS_STREAM_AUDIO_PROFILE_AAC_MAIN;
								break;
							case 1:
								profile = TS_STREAM_AUDIO_PROFILE_AAC_LOW_COMPLEXITY;
								break;
							case 2:
								profile = TS_STREAM_AUDIO_PROFILE_AAC_SCALABLE_SAMPLE_RATE;
								break;
							case 3:
								if (version == 2) {
									profile = TS_STREAM_AUDIO_PROFILE_AAC_RESERVED;
								} else {
									profile = TS_STREAM_AUDIO_PROFILE_AAC_LTP;
								}
								break;
							default:
								profile = TS_STREAM_AUDIO_PROFILE_UNKNOWN;
								break;
						}
						
						samplerate = (buffer[ptr+2] & 0x3c) >> 2;
						samplerate = aac_sr[samplerate];
						
						mode = ((buffer[ptr+2] & 0x01) << 2) | ((buffer[ptr+3] & 0xc0) >> 6);
						if (mode == 3) {
							mode = TS_STREAM_AUDIO_MODE_MONO;
						} else {
							mode = TS_STREAM_AUDIO_MODE_STEREO;
						}
						
						tspid->params.audio.version = version;
						tspid->params.audio.revision = layer;
						tspid->params.audio.profile = profile;
						tspid->params.audio.mode = mode;
						tspid->params.audio.bitrate = 0;
						tspid->params.audio.samplerate = samplerate;
						tspid->stream_format = TS_STREAM_FORMAT_AUDIO_AAC;
						tspid->stream_revision++;
						
						//printf("-->AAC audio start v: %u, l: %u, p: %u, s: %u, m: %u\n", version, layer, profile, samplerate, mode);
						
						transportstream_print_stream_params(tspid, (char *)&buf, sizeof(buf));
						streamserver_send_log_message(LOG_EVENT_MESSAGE, "Audio: %s", buf);
						//printf("[+] Audio: %s\n", buf);
						
						return TS_STREAM_FORMAT_AUDIO_AAC;
					} else {
						abitrate = (buffer[ptr+2] & 0xf0) >> 4;
						abitrate = (mpeg_br_table[version-1][layer-1])[abitrate];
						
						samplerate = (buffer[ptr+2] & 0x0c) >> 2;
						switch (version) {
							case 1:
								samplerate = mpeg_sr_1[samplerate];
								break;
							case 2:
								samplerate = mpeg_sr_2[samplerate];
								break;
							case 3:
								samplerate = mpeg_sr_3[samplerate];
								break;
							case 4:
								samplerate = 0;
								break;
						}
						
						mode = (buffer[ptr+3] & 0xc0) >> 6;
						if (mode == 3) {
							mode = TS_STREAM_AUDIO_MODE_MONO;
						} else {
							mode = TS_STREAM_AUDIO_MODE_STEREO;
						}
						
						tspid->params.audio.version = version;
						tspid->params.audio.revision = layer;
						tspid->params.audio.profile = TS_STREAM_AUDIO_PROFILE_DEFAULT;
						tspid->params.audio.mode = mode;
						tspid->params.audio.bitrate = abitrate;
						tspid->params.audio.samplerate = samplerate;
						tspid->stream_format = TS_STREAM_FORMAT_AUDIO_MPEG;
						tspid->stream_revision++;
						
						//printf("-->MPEG audio start v: %u, l: %u, b: %u, s: %u, m: %u\n", version, layer, abitrate, samplerate, mode);
						
						transportstream_print_stream_params(tspid, (char *)&buf, sizeof(buf));
						streamserver_send_log_message(LOG_EVENT_MESSAGE, "Audio: %s", buf);
						//printf("[+] Audio: %s\n", buf);
						
						return TS_STREAM_FORMAT_AUDIO_MPEG;
					}
				}
			}
		} else {
			if ((buffer[ptr] == 0x0b) && (buffer[ptr+1]== 0x77)) {
				samplerate = (buffer[ptr+4] & 0xc0) >> 6;
				samplerate = ac3_sr_8[samplerate];
				
				abitrate = buffer[ptr+4] & 0x3f;
				abitrate = ac3_br_8[abitrate];
				
				version = (buffer[ptr+5] & 0xf8) >> 3;
				layer = buffer[ptr+5] & 0x07;
				
				mode = (buffer[ptr+6] & 0xE0) >> 5;
				switch (mode) {
					case 0:
						mode = TS_STREAM_AUDIO_MODE_STEREO;
						break;
					case 1:
						mode = TS_STREAM_AUDIO_MODE_C;
						break;
					case 2:
						mode = TS_STREAM_AUDIO_MODE_L_R;
						break;
					case 3:
						mode = TS_STREAM_AUDIO_MODE_L_C_R;
						break;
					case 4:
						mode = TS_STREAM_AUDIO_MODE_L_R_S;
						break;
					case 5:
						mode = TS_STREAM_AUDIO_MODE_L_C_R_S;
						break;
					case 6:
						mode = TS_STREAM_AUDIO_MODE_L_R_SL_SR;
						break;
					case 7:
						mode = TS_STREAM_AUDIO_MODE_L_C_R_SL_SR;
						break;
					default:
						mode = TS_STREAM_AUDIO_MODE_UNKNOWN;
						break;
				}
				
				tspid->params.audio.version = version;
				tspid->params.audio.revision = layer;
				tspid->params.audio.profile = TS_STREAM_AUDIO_PROFILE_DEFAULT;
				tspid->params.audio.mode = mode;
				tspid->params.audio.bitrate = abitrate;
				tspid->params.audio.samplerate = samplerate;
				tspid->stream_format = TS_STREAM_FORMAT_AUDIO_AC3;
				tspid->stream_revision++;
				
				//printf("-->AC3 audio start v: %u, l: %u, b: %u, s: %u, m: %u\n", version, layer, abitrate, samplerate, mode);
				
				transportstream_print_stream_params(tspid, (char *)&buf, sizeof(buf));
				streamserver_send_log_message(LOG_EVENT_MESSAGE, "Audio: %s", buf);
				//printf("[+] Audio: %s\n", buf);
				
				return TS_STREAM_FORMAT_AUDIO_AC3;
			}
		}
	}
	
	tspid->stream_revision++;
	
	return TS_STREAM_FORMAT_UNKNOWN;
}

int transportstream_parse_buffer(unsigned char *buffer, size_t bufsize, struct _tsinfo *tsinfo)
{
	int pos, ptr, length, peslength, peshdrlength, i;
	time_t timestamp;
	struct _tsheader tsheader;
	struct _tspid *tspid;
	
	pos = 0;
	tsinfo->packets = 0;
	
	timestamp = time(NULL);
	
	if (difftime(timestamp, tsinfo->starttime) <= TS_STREAM_SWITCH_TIMEOUT) goto finalize;
	
	while ((pos + 1) < bufsize) {
		while (buffer[pos] != 0x47) {
			if ((pos + 1) >= bufsize) goto finalize;
			pos++;
		}
		
		if ((bufsize - pos) < 188) goto finalize;
		
		if (!transportstream_decode_header(buffer+pos, &tsheader)) {
			length = 0;
			ptr = pos + 4;
			//printf("CORRUPTED PACKET\n");
			tsinfo->pids[0x1FFF].packets++;
			tsinfo->packets++;
		} else if (tsheader.pid == 0x1FFF) {
			length = 0;
			ptr = pos + 4;
			//printf("NULL PACKET (%u)\n", tsheader.pid);
			tsinfo->pids[0x1FFF].packets++;
			tsinfo->packets++;
		} else {
			switch (tsheader.adaptation_field_control) {
				case 0x00:
					length = 0;
					ptr = pos + 4;
					//printf("EMPTY PACKET (%u)\n", tsheader.pid);
					break;
				
				case 0x01:
					length = 188 - 4;
					ptr = pos + 4;
					break;
				
				case 0x02:
					length = 0;
					ptr = pos + 4;
					//printf("ADAPTATION FIELD PACKET (%u)\n", tsheader.pid);
					break;
				
				case 0x03:
					length = 188 - buffer[pos+4];
					ptr = pos + 4 + buffer[pos+4] + 1;
					break;
				
				default:
					break;
			}
			
			tspid = (struct _tspid *)&tsinfo->pids[tsheader.pid];
			
			tspid->packets++;
			tsinfo->packets++;
		}
		
		if ((length > 0) && ((ptr - pos) >= 4) && (tspid != NULL)) {
			if ((tsheader.payload_unit_start_indicator) && !((tsheader.transport_scrambling_control) || (tsheader.transport_error_indicator))) {
				if ((buffer[ptr+0] == 0x00) && (buffer[ptr+1] == 0x00) && (buffer[ptr+2] == 0x01)) {
					if ((tspid->stream_type == TS_STREAM_UNKNOWN) || ((tspid->stream_format == TS_STREAM_FORMAT_UNKNOWN) && (tspid->stream_revision <= TS_ANALYZE_PACKETS_LIMIT))) {
						peslength = (((buffer[ptr+4] & 0xff) << 8) | buffer[ptr+5]) & 0xffff;
						peshdrlength = buffer[ptr+8];
						
						switch (buffer[ptr+3]) {
							case 0xBC:
								//printf("PES (%u) program_stream_map\n", tsheader.pid);
								tspid->stream_type = TS_STREAM_PES_PROGRAM_STREAM_MAP;
								tspid->stream_format = TS_STREAM_FORMAT_SPECIAL;
								break;
							case 0xBD:
								//printf("PES (%u) private_stream_1\n", tsheader.pid);
								tspid->stream_type = TS_STREAM_PES_PRIVATE_STREAM_1;
								tspid->stream_format = transportstream_decode_audio_header((unsigned char *)&buffer[ptr+9+peshdrlength], (size_t)length-9-peshdrlength, 1, &tsinfo->pids[tsheader.pid]);
								tspid->stream_value = 0;
								break;
							case 0xBE:
								//printf("PES (%u) padding_stream\n", tsheader.pid);
								tspid->stream_type = TS_STREAM_PES_PADDING_STREAM;
								tspid->stream_format = TS_STREAM_FORMAT_SPECIAL;
								break;
							case 0xBF:
								//printf("PES (%u) private_stream_2\n", tsheader.pid);
								tspid->stream_type = TS_STREAM_PES_PRIVATE_STREAM_2;
								tspid->stream_format = TS_STREAM_FORMAT_SPECIAL;
								break;
							case 0xC0 ... 0xDF:
								//printf("PES (%u) audio stream number %u\n", tsheader.pid, buffer[ptr+3] & 0x1F);
								tspid->stream_type = TS_STREAM_PES_AUDIO_STREAM;
								tspid->stream_format = transportstream_decode_audio_header((unsigned char *)&buffer[ptr+9+peshdrlength], (size_t)length-9-peshdrlength, 0, &tsinfo->pids[tsheader.pid]);
								tspid->stream_value = buffer[ptr+3] & 0x1F;
								break;
							case 0xE0 ... 0xEF:
								//printf("PES (%u) video stream number %u\n", tsheader.pid, buffer[ptr+3] & 0x0F);
								tspid->stream_type = TS_STREAM_PES_VIDEO_STREAM;
								tspid->stream_format = transportstream_decode_video_header((unsigned char *)&buffer[ptr+9+peshdrlength], (size_t)length-9-peshdrlength, &tsinfo->pids[tsheader.pid]);
								tspid->stream_value = buffer[ptr+3] & 0x0F;
								break;
							case 0xF0:
								//printf("PES (%u) ECM_stream\n", tsheader.pid);
								tspid->stream_type = TS_STREAM_PES_ECM_STREAM;
								tspid->stream_format = TS_STREAM_FORMAT_SPECIAL;
								break;
							case 0xF1:
								//printf("PES (%u) EMM_stream\n", tsheader.pid);
								tspid->stream_type = TS_STREAM_PES_EMM_STREAM;
								tspid->stream_format = TS_STREAM_FORMAT_SPECIAL;
								break;
							case 0xF2:
								//printf("PES (%u) DSMCC_stream\n", tsheader.pid);
								tspid->stream_type = TS_STREAM_PES_DSMCC_STREAM;
								tspid->stream_format = TS_STREAM_FORMAT_SPECIAL;
								break;
							case 0xF3:
								//printf("PES (%u) ISO/IEC_13522_stream\n", tsheader.pid);
								tspid->stream_type = TS_STREAM_PES_IEC_13522_STREAM;
								tspid->stream_format = TS_STREAM_FORMAT_SPECIAL;
								break;
							case 0xF4:
								//printf("PES (%u) ITU-T Rec. H.222.1 type A\n", tsheader.pid);
								tspid->stream_type = TS_STREAM_PES_H_222_1_TYPE_A;
								tspid->stream_format = TS_STREAM_FORMAT_SPECIAL;
								break;
							case 0xF5:
								//printf("PES (%u) ITU-T Rec. H.222.1 type B\n", tsheader.pid);
								tspid->stream_type = TS_STREAM_PES_H_222_1_TYPE_B;
								tspid->stream_format = TS_STREAM_FORMAT_SPECIAL;
								break;
							case 0xF6:
								//printf("PES (%u) ITU-T Rec. H.222.1 type C\n", tsheader.pid);
								tspid->stream_type = TS_STREAM_PES_H_222_1_TYPE_C;
								tspid->stream_format = TS_STREAM_FORMAT_SPECIAL;
								break;
							case 0xF7:
								//printf("PES (%u) ITU-T Rec. H.222.1 type D\n", tsheader.pid);
								tspid->stream_type = TS_STREAM_PES_H_222_1_TYPE_D;
								tspid->stream_format = TS_STREAM_FORMAT_SPECIAL;
								break;
							case 0xF8:
								//printf("PES (%u) ITU-T Rec. H.222.1 type E\n", tsheader.pid);
								tspid->stream_type = TS_STREAM_PES_H_222_1_TYPE_E;
								tspid->stream_format = TS_STREAM_FORMAT_SPECIAL;
								break;
							case 0xF9:
								//printf("PES (%u) ancillary_stream\n", tsheader.pid);
								tspid->stream_type = TS_STREAM_PES_ANCILLARY_STREAM;
								tspid->stream_format = TS_STREAM_FORMAT_SPECIAL;
								break;
							case 0xFA:
								//printf("PES (%u) ISO/IEC14496-1_SL-packetized_stream\n", tsheader.pid);
								tspid->stream_type = TS_STREAM_PES_SL_PACKETIZED_STREAM;
								tspid->stream_format = TS_STREAM_FORMAT_SPECIAL;
								break;
							case 0xFB:
								//printf("PES (%u) ISO/IEC14496-1_FlexMux_stream\n", tsheader.pid);
								tspid->stream_type = TS_STREAM_PES_FLEXMUX_STREAM;
								tspid->stream_format = TS_STREAM_FORMAT_SPECIAL;
								break;
							case 0xFF:
								//printf("PES (%u) program_stream_directory\n", tsheader.pid);
								tspid->stream_type = TS_STREAM_PES_PROGRAM_STREAM_DIRECTORY;
								tspid->stream_format = TS_STREAM_FORMAT_SPECIAL;
								break;
							default:
								//printf("PES (%u) reserved data stream\n", tsheader.pid);
								tspid->stream_type = TS_STREAM_PES_RESERVED_DATA_STREAM;
								tspid->stream_format = TS_STREAM_FORMAT_SPECIAL;
								break;
						}
					}
				} else if (tspid->stream_type == TS_STREAM_UNKNOWN) {
					switch (tsheader.pid) {
						case 0x0000:
							//printf("TABLE PAT (%u)\n", tsheader.pid);
							tspid->stream_type = TS_STREAM_TABLE_PAT;
							tspid->stream_format = TS_STREAM_FORMAT_TABLE;
							break;
						case 0x0001:
							//printf("TABLE CAT (%u)\n", tsheader.pid);
							tspid->stream_type = TS_STREAM_TABLE_CAT;
							tspid->stream_format = TS_STREAM_FORMAT_TABLE;
							break;
						case 0x0002:
							//printf("TABLE TSDT (%u)\n", tsheader.pid);
							tspid->stream_type = TS_STREAM_TABLE_TSDT;
							tspid->stream_format = TS_STREAM_FORMAT_TABLE;
							break;
						case 0x0010:
							//printf("TABLE NIT (%u)\n", tsheader.pid);
							tspid->stream_type = TS_STREAM_TABLE_NIT;
							tspid->stream_format = TS_STREAM_FORMAT_TABLE;
							break;
						case 0x0011:
							//printf("TABLE SDT, BAT (%u)\n", tsheader.pid, buffer[ptr+3] & 0x1F);
							tspid->stream_type = TS_STREAM_TABLE_SDT_BAT;
							tspid->stream_format = TS_STREAM_FORMAT_TABLE;
							break;
						case 0x0012:
							//printf("TABLE EIT (%u)\n", tsheader.pid, buffer[ptr+3] & 0x0F);
							tspid->stream_type = TS_STREAM_TABLE_EIT;
							tspid->stream_format = TS_STREAM_FORMAT_TABLE;
							break;
						case 0x0013:
							//printf("TABLE RST (%u)\n", tsheader.pid);
							tspid->stream_type = TS_STREAM_TABLE_RST;
							tspid->stream_format = TS_STREAM_FORMAT_TABLE;
							break;
						case 0x0014:
							//printf("TABLE TDT, TOT (%u)\n", tsheader.pid);
							tspid->stream_type = TS_STREAM_TABLE_TDT_TOT;
							tspid->stream_format = TS_STREAM_FORMAT_TABLE;
							break;
						case 0x0015:
							//printf("TABLE NETWORK SYNCHRONIZATION (%u)\n", tsheader.pid);
							tspid->stream_type = TS_STREAM_TABLE_NETWORK_SYNCHRONIZATION;
							tspid->stream_format = TS_STREAM_FORMAT_TABLE;
							break;
						case 0x001C:
							//printf("TABLE INBAND SIGNALLING (%u)\n", tsheader.pid);
							tspid->stream_type = TS_STREAM_TABLE_INBAND_SIGNALLING;
							tspid->stream_format = TS_STREAM_FORMAT_TABLE;
							break;
						case 0x001D:
							//printf("TABLE MEASUREMENT (%u)\n", tsheader.pid);
							tspid->stream_type = TS_STREAM_TABLE_MEASUREMENT;
							tspid->stream_format = TS_STREAM_FORMAT_TABLE;
							break;
						case 0x001E:
							//printf("TABLE EIT (%u)\n", tsheader.pid);
							tspid->stream_type = TS_STREAM_TABLE_EIT;
							tspid->stream_format = TS_STREAM_FORMAT_TABLE;
							break;
						case 0x001F:
							//printf("TABLE SIT (%u)\n", tsheader.pid);
							tspid->stream_type = TS_STREAM_TABLE_SIT;
							tspid->stream_format = TS_STREAM_FORMAT_TABLE;
							break;
						default:
							//printf("TABLE RESERVED (%u)\n", tsheader.pid);
							tspid->stream_type = TS_STREAM_TABLE_RESERVED;
							tspid->stream_format = TS_STREAM_FORMAT_TABLE;
							break;
					}
				}
			} else if (tsheader.transport_scrambling_control) {
				tspid->stream_type = TS_STREAM_UNKNOWN;
				tspid->stream_format = TS_STREAM_FORMAT_ENCRYPTED;
			}
		}
		pos = pos + 188;
	}
	
	finalize:
	
	if (tsinfo->packets > 0) {
		for (i=0; i<TS_NUMBER_PIDS; i++) {
			if (tsinfo->pids[i].packets > 0) {
				if (tsinfo->timeframe > 10) {
					tsinfo->pids[i].bandwidth = tsinfo->pids[i].packets * 8 * 1000 / tsinfo->timeframe * 188 / 1000;
				} else {
					tsinfo->pids[i].bandwidth = 0;
				}
				tsinfo->pids[i].percentage = (int)(((float)tsinfo->pids[i].packets / (float)tsinfo->packets) * 100);
				tsinfo->pids[i].lasttime = timestamp;
				tsinfo->pids[i].packets = 0;
				//printf("VAL: PID %u, BW: %u, PR: %u, TF %u\n", i, tsinfo->pids[i].bandwidth, tsinfo->pids[i].percentage, tsinfo->timeframe);
			} else if ((tsinfo->pids[i].packets == 0) && ((tsinfo->pids[i].bandwidth > 0) || (tsinfo->pids[i].percentage > 0))) {
				if (difftime(timestamp, tsinfo->pids[i].lasttime) >= TS_STREAM_PID_LIFETIME) {
					tsinfo->pids[i].stream_type = TS_STREAM_UNKNOWN;
					tsinfo->pids[i].stream_format = TS_STREAM_FORMAT_UNKNOWN;
					tsinfo->pids[i].stream_value = 0;
					tsinfo->pids[i].stream_revision = 0;
					tsinfo->pids[i].bandwidth = 0;
					tsinfo->pids[i].percentage = 0;
					tsinfo->pids[i].packets = 0;
					tsinfo->pids[i].lasttime = 0;
				}
			}
		}
	}
	
	return 1;
}

void transportstream_clear_statistics(struct _tsinfo *tsinfo)
{
	int i;
	
	memset(tsinfo, 0, sizeof(struct _tsinfo));
	
	tsinfo->starttime = time(NULL);
	
	for (i=0; i<TS_NUMBER_PIDS; i++) {
		tsinfo->pids[i].stream_type = TS_STREAM_UNKNOWN;
		tsinfo->pids[i].stream_format = TS_STREAM_FORMAT_UNKNOWN;
		tsinfo->pids[i].stream_value = 0;
		tsinfo->pids[i].stream_revision = 0;
		tsinfo->pids[i].bandwidth = 0;
		tsinfo->pids[i].percentage = 0;
		tsinfo->pids[i].packets = 0;
		tsinfo->pids[i].lasttime = 0;
	}
	
}

unsigned int transportstream_demux_pes(unsigned char *tsbuf, size_t tsbufsize, unsigned char *dmxbuf, size_t dmxbufsize, unsigned short pid)
{
	int pos, ptr, length, pesbytes;
	struct _tsheader tsheader;
	
	if ((tsbuf == NULL) || (dmxbuf == NULL) || (tsbufsize == 0) || (dmxbufsize == 0) || (pid >= 8192)) return 0;
	
	pos = 0;
	pesbytes = 0;
		
	while ((pos + 1) < tsbufsize) {
		while (tsbuf[pos] != 0x47) {
			if ((pos + 1) >= tsbufsize) goto finalize;
			pos++;
		}
				
		if (!transportstream_decode_header(tsbuf+pos, &tsheader)) {
			//corrupted packet
			length = 0;
			ptr = pos + 4;
		} else if (tsheader.pid == 0x1FFF) {
			//null packet
			length = 0;
			ptr = pos + 4;
		} else if (tsheader.pid == pid) {
			switch (tsheader.adaptation_field_control) {
				case 0x00:
					//empty packet
					length = 0;
					ptr = pos + 4;
					break;
				
				case 0x01:
					//full payload packet
					length = 188 - 4;
					ptr = pos + 4;
					break;
				
				case 0x02:
					//adaptation field packet
					length = 0;
					ptr = pos + 4;
					break;
				
				case 0x03:
					//packet with adaptation field
					length = 188 - tsbuf[pos+4];
					ptr = pos + 4 + tsbuf[pos+4] + 1;
					break;
			}
			
			if (length > 0) {
				if (pesbytes + length > dmxbufsize) 
					if ((length = dmxbufsize - pesbytes) <= 0)
						goto finalize;
				memcpy(dmxbuf+pesbytes, tsbuf+ptr, length);
				pesbytes = pesbytes + length;
			} /*else {
				goto finalize;
			}*/	
		}
		pos = pos + 188;
	}
	
	finalize:
	
	return pesbytes;
}
