/*
 *      settings.c
 *      
 *      Copyright 2009 - 2012 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#define CONFDIR_PERMISSIONS 0744
#define CONFDIR ".config/gsf"
#define CONFFILE ".config/gsf/settings.conf"
#define CONFIG_SECTION "settings"

struct ssettings {
	gboolean opened;
	GKeyFile *config;
	GMappedFile *mmap;
	gchar *filename;
	gchar *content;
	gsize bytes;
	GError *error;
	gboolean autoflush;
};
struct ssettings settings;



gboolean settings_open(gboolean autoflush)
{
	gint fd;
		
	if (settings.error != NULL) settings.error = NULL;
	
	if ((g_mkdir_with_parents(g_build_filename(g_get_home_dir(), CONFDIR, NULL), CONFDIR_PERMISSIONS)) == -1) {
		settings.opened = FALSE;
		return FALSE;
	}
	
	settings.filename = g_build_filename(g_get_home_dir(), CONFFILE, NULL);
	settings.config = g_key_file_new();
	
	if (!g_file_test(settings.filename, G_FILE_TEST_EXISTS)) {
		if ((fd = g_creat(settings.filename, S_IRUSR | S_IWUSR)) != -1) {
			close(fd);
		} else {
			settings.opened = FALSE;
			return FALSE;
		}
	}
		
	if ((settings.mmap = g_mapped_file_new(settings.filename, FALSE, &settings.error)) != NULL) {
		if (g_mapped_file_get_length(settings.mmap) > 0) {
			g_key_file_load_from_data(settings.config, g_mapped_file_get_contents(settings.mmap), g_mapped_file_get_length(settings.mmap), G_KEY_FILE_NONE, &settings.error);
		}
		#if GLIB_CHECK_VERSION(2, 22, 0)
			g_mapped_file_unref(settings.mmap);
		#else
			g_mapped_file_free(settings.mmap);
		#endif
	}
	
	settings.opened = TRUE;
	settings.autoflush = autoflush;
	
	return TRUE;
}

gboolean settings_close(void)
{
	gboolean result = FALSE;
	
	if (settings.error != NULL) settings.error = NULL;
	
	if (settings.opened) {
		settings.content = g_key_file_to_data(settings.config, &settings.bytes, NULL);
		result = g_file_set_contents(settings.filename, settings.content, -1, &settings.error);
		g_free(settings.filename); g_free(settings.content); g_key_file_free(settings.config);
		settings.opened = FALSE;
	}
	
	return result;
}

gboolean settings_flush(void)
{
	gboolean result = FALSE;
	
	if (settings.error != NULL) settings.error = NULL;
	
	if (settings.opened) {
		settings.content = g_key_file_to_data(settings.config, &settings.bytes, NULL);
		result = g_file_set_contents(settings.filename, settings.content, settings.bytes, &settings.error);
	}
	
	return result;
} 

gboolean settings_append_string(gchar *name, gchar *value)
{
	gboolean result = FALSE;
	
	if (settings.opened) {
		g_key_file_set_string(settings.config, CONFIG_SECTION, name, value);
		if (settings.autoflush) settings_flush();
		result = TRUE;
	}
	
	return result;
}

gchar *settings_read_string(gchar *name, gchar *def)
{
	gchar *result = NULL;
	
	if (settings.error != NULL) settings.error = NULL;
	
	if (settings.opened) {
		if (g_key_file_has_key(settings.config, CONFIG_SECTION, name, &settings.error)) {
			if (settings.error != NULL) settings.error = NULL;
			result = g_key_file_get_string(settings.config, CONFIG_SECTION, name, &settings.error);
		}
	}
	
	if (result == NULL) result = g_strdup(def);
	
	return result;
}

gboolean settings_append_int(gchar *name, gint value)
{
	gboolean result = FALSE;
	
	if (settings.opened) {
		g_key_file_set_integer(settings.config, CONFIG_SECTION, name, value);
		if (settings.autoflush) settings_flush();
		result = TRUE;
	}
	
	return result;
}

gint settings_read_int(gchar *name, gint def)
{
	gint result = def;
	
	if (settings.error != NULL) settings.error = NULL;
	if (settings.opened) {
		if (g_key_file_has_key(settings.config, CONFIG_SECTION, name, &settings.error)) {
			if (settings.error != NULL) settings.error = NULL;
			result = g_key_file_get_integer(settings.config, CONFIG_SECTION, name, &settings.error);
		}
	}
	
	return result;
}

gboolean settings_append_float(gchar *name, gdouble value)
{
	gboolean result = FALSE;
	
	if (settings.opened) {
		g_key_file_set_double(settings.config, CONFIG_SECTION, name, value);
		if (settings.autoflush) settings_flush();
		result = TRUE;
	}
	
	return result;
}

gdouble settings_read_float(gchar *name, gdouble def)
{
	gdouble result = def;
	
	if (settings.error != NULL) settings.error = NULL;
	if (settings.opened) {
		if (g_key_file_has_key(settings.config, CONFIG_SECTION, name, &settings.error)) {
			if (settings.error != NULL) settings.error = NULL;
			result = g_key_file_get_double(settings.config, CONFIG_SECTION, name, &settings.error);
		}
	}
	
	return result;
}

gboolean settings_append_boolean(gchar *name, gboolean value)
{
	gboolean result = FALSE;
	
	if (settings.opened) {
		g_key_file_set_boolean(settings.config, CONFIG_SECTION, name, value);
		if (settings.autoflush) settings_flush();
		result = TRUE;
	}
	
	return result;
}

gboolean settings_read_boolean(gchar *name, gboolean def)
{
	gboolean result = def;
	
	if (settings.error != NULL) settings.error = NULL;
	if (settings.opened) {
		if (g_key_file_has_key(settings.config, CONFIG_SECTION, name, &settings.error)) {
			if (settings.error != NULL) settings.error = NULL;
			result = g_key_file_get_boolean(settings.config, CONFIG_SECTION, name, &settings.error);
		}
	}
	
	return result;
}
