/*
 *      sound.c
 *      
 *      Copyright 2009 - 2012 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <gmodule.h>
#include <glib.h>
#include <string.h>
#include <stdio.h>
#include <limits.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "sound.h"
#include "gsfcore.h"
#include "settings.h"
#include "soundmodules/soundmodules.h"
#include "common/iso639.h"
#include "resources.h"

#define WAV_UNCOMPRESSED_PCM   1

typedef gboolean (*sound_module_init_func)(guint *moduleid, gchar *modulename, gint *moduleversion);
typedef gboolean (*sound_module_play_file_func)(gchar *filename, glong (*file_handler)(wavheader_t header, gint command, gpointer buffer, gsize size));
typedef gboolean (*sound_module_close_func)(void);

GSList *soundlanguages = NULL;
GSList *soundmodules = NULL;	

sound_module_init_func sound_module_init;
sound_module_play_file_func sound_module_play_file;
sound_module_close_func sound_module_close;

GModule *module;
guint currentmoduleid;
struct _soundlanguage *currentlanguage;
gchar *systemlocale;
guchar soundparameter;
guchar soundlimit;

static gulong sound_wav_file_read_value(guchar *data, gsize size);
static void sound_modules_free_foreach(gpointer data, gpointer user_data);
static gboolean sound_modules_free(GSList *modules);
static void sound_langauages_free_foreach(gpointer data, gpointer user_data);
static gint sound_languages_set_valid_finder(gconstpointer a, gconstpointer b);
static gint sound_languages_set_finder(gconstpointer a, gconstpointer b);
static gchar **sound_languages_path(struct _frontend *frontend, guchar parameter, guchar limit);
static gint sound_languages_get_finder(gconstpointer a, gconstpointer b);


static gulong sound_wav_file_read_value(guchar *data, gsize size)
{
	gint testnum, bytes, i;
	gulong value;
		
	testnum = 1;
	value = 0;
	
	if (size > sizeof(value)) {
		bytes = sizeof(value);
	} else {
		bytes = size;
	}
	
	if(*(gchar *)&testnum == 1) {
		//Little-Endian value
		for (i=0; i<bytes; i++) {
			value = value | (data[i] << (i * 8));
		}
	} else {
		//Big-Endian value
		for (i=0; i<bytes; i++) {
			value = value | (data[i] << ((bytes - i - 1) * 8));
		}
	}
	
	return value;
}

glong sound_wav_file_handler(wavheader_t header, gint command, gpointer buffer, gsize size)
{
	guchar data[20];
	glong bufsize, readsize;
	
	switch (command) {
		case SOUND_WAV_FILE_OPEN:
			if ((header == NULL) || (buffer == NULL)) return 0;
			
			memset(header, 0, sizeof(struct _wavheader));
			
			header->fd = open(buffer, O_RDONLY, 0);
			
			if (header->fd == -1) {
				g_warning("Failed to open sound file");
				return 0;
			}
			
			if (lseek(header->fd, 0, SEEK_CUR) == -1) {
				g_warning("Sound file is empty");
				close(header->fd);
				return 0;
			}
			
			//test if file is wav - RIFF is first 4 bytes WAVE is third 4 bytes
			if (read(header->fd, &data, 12) != 12) {
				g_warning("Sound file truncated");
				close(header->fd);
				return 0;
			}
			
			if ((strncmp(&data[0], "RIFF", 4) != 0) || (strncmp(&data[8], "WAVE", 4) != 0)) {
				g_warning("Sound format not supported: no RIFF and WAVE signature");
				close(header->fd);
				return 0;
			}
			
			//find format chunk
			while(1) {
				if (read(header->fd, &data, 4) != 4) {
					g_warning("Sound file corrupted: unexpected end of file");
					close(header->fd);
					return 0;
				}
				if (strncmp("fmt ", data, 4) == 0) break;
			}
			
			//read stream parameters
			if (read(header->fd, &data, 20) != 20) {
				g_warning("Sound file corrupted: format section truncated from file");
				close(header->fd);
				return 0;
			}
			
			header->len = (gulong)sound_wav_file_read_value(&data[0], 4);
			header->format = (gshort)sound_wav_file_read_value(&data[4], 2);
			header->channels = (gshort)sound_wav_file_read_value(&data[6], 2);
			header->samples_per_sec = (gulong)sound_wav_file_read_value(&data[8], 4);
			header->avg_bytes_per_sec = (gulong)sound_wav_file_read_value(&data[12], 4);
			header->align = (gshort)sound_wav_file_read_value(&data[16], 2);
			header->bits = (gshort)sound_wav_file_read_value(&data[18], 2);
			
			//test format chunk size - must be more than 16 bytes
			if (header->len < 16) {
				g_warning("Sound file corrupted: wrong size of fmt section");
				close(header->fd);
				return 0;
			}
			
			//test format - only uncompressed PCM supported
			if (header->format != WAV_UNCOMPRESSED_PCM) {
				g_warning("Sound format not supported: only uncompressed PCM supported");
				close(header->fd);
				return 0;
			}
			
			//supported bps values - 8 and 16 bps
			if ((header->bits != 8) && (header->bits != 16)) {
				g_warning("Sound format not supported: only 8 and 16 bps supported");
				close(header->fd);
				return 0;
			}
			
			//supported channels values - 1 and 2
			if ((header->channels < 1) || (header->channels > 2)) {
				g_warning("Sound format not supported: only 1 and 2 channels supported");
				close(header->fd);
				return 0;
			}
			
			//extra bytes at the end of chunk - must be ignored
			header->len -= 16;
			if (header->len) {
				if (lseek(header->fd, header->len, SEEK_CUR) == -1) {
					g_warning("Sound file corrupted: no data after header");
					close(header->fd);
					return 0;
				}
			}
			
			//find data chunk
			while(1) {
				if (read(header->fd, &data, 4) != 4) {
					g_warning("Sound file corrupted: unexpected end of file");
					close(header->fd);
					return 0;
				}
				if (strncmp("data", data, 4) == 0) break;
			}
			
			//size of data to read
			if (read(header->fd, &data, 4) != 4) {
				g_warning("Sound file corrupted: no data found");
				close(header->fd);
				return 0;
			}
			
			header->len = (gulong)sound_wav_file_read_value(&data[0], 4);
			
			if (header->len == 0) {
				g_warning("Sound file corrupted: no data found");
				close(header->fd);
				return 0;
			}
			
			return 1;
		
		case SOUND_WAV_FILE_READ:
			if ((header == NULL) || (buffer == NULL) || (size == 0)) return 0;
			
			if ((header->len - header->pos) >= size) {
				bufsize = size;
			} else {
				bufsize = header->len - header->pos;
			}
			
			if (bufsize <= 0) return 0;
			
			readsize = read(header->fd, buffer, bufsize);
			
			if (readsize > 0) {
				header->pos += readsize;
				return readsize;
			} else {
				return 0;
			}
			
		case SOUND_WAV_FILE_REWIND:
			if (header == NULL) return 0;
			
			if (header->pos > 0) {
				//move back by read bytes value
				if (lseek(header->fd, -(header->pos), SEEK_CUR) == -1) {
					return 0;
				}
				//clear position
				header->pos = 0;
			}
			
			return 1;
		
		case SOUND_WAV_FILE_CLOSE:
			if (header == NULL) return 0;
			
			//close file and clear data
			close(header->fd);
			memset(header, 0, sizeof(struct _wavheader));
			
			return 1;
		
		default:
			return 0;
	}
}

guint sound_modules_enum(void)
{
	GDir *moddir;
	const gchar *modname;
	gsize length;
	const gchar modext[] = ".so";
	gchar *modpath;
	guint number = 0;
	struct _soundmodule *soundmodule;
	GModule *module;
	guint moduleid;
	gchar modulename[32];
	gint moduleversion;
	
	
	if ((moddir = g_dir_open(RESOURCE_DIR_SOUND_MODULES, 0, NULL)) != NULL) {
		printf("Sound modules enumeration\n");
		while ((modname = g_dir_read_name(moddir)) != NULL) {
			if (((length = strlen(modname)) > 3) && (strncmp((const char *)modname + length - 3, (const char *)&modext, 3) == 0)) {
				modpath = g_strconcat(RESOURCE_DIR_SOUND_MODULES, modname, NULL);
				if (g_file_test(modpath, G_FILE_TEST_EXISTS)) {
					if ((module = g_module_open(modpath, G_MODULE_BIND_LAZY)) != NULL) {
						if (g_module_symbol(module, "sound_module_init", (gpointer*)&sound_module_init)) {
							if (sound_module_init(&moduleid, (gchar *)&modulename, &moduleversion)) {
								soundmodule = g_malloc0(sizeof(struct _soundmodule));
								soundmodule->path = g_strdup(modpath);
								soundmodule->id = moduleid;
								soundmodule->name = g_strdup(modulename);
								soundmodule->version = moduleversion;
								soundmodules = g_slist_prepend(soundmodules, soundmodule);
								printf(" %u. %s [ID %u]\n", number+1, modulename, moduleid);
								number++;
							}
						}
						if (!g_module_close(module)) g_warning("Error unloading sound module: %s\n", (gchar*)g_module_error()); 
					}
				}
				g_free(modpath);
			}
		}
		g_dir_close(moddir);
	} else {
		g_warning("No sound modules found");
	}
	
	if (number > 0) soundmodules = g_slist_reverse(soundmodules);
	
	return number;
} 

static void sound_modules_free_foreach(gpointer data, gpointer user_data)
{
	struct _soundmodule *soundmodule;
	
	soundmodule = (struct _soundmodule *)data;
	
	if (soundmodule->path != NULL) g_free(soundmodule->path);
	if (soundmodule->name != NULL) g_free(soundmodule->name);
	
}

static gboolean sound_modules_free(GSList *modules)
{
	if (modules == NULL) return FALSE;
	
	g_slist_foreach(modules, sound_modules_free_foreach, NULL);
	
	g_slist_free(modules);
	
	return TRUE;
}


gboolean sound_modules_set_module(guint moduleid)
{
	gint i;
	struct _soundmodule *soundmodule;
	gsize length;
	gboolean modulefound = FALSE;
		
	if (currentmoduleid == moduleid) return FALSE;
	
	if ((length = g_slist_length(soundmodules)) > 0) {
		for (i=0; i<length; i++) {
			soundmodule = (struct _soundmodule *)g_slist_nth_data(soundmodules, i);
			if (soundmodule->id == moduleid) {
				//clean module unload
				if (module != NULL) {
					sound_module_close();
					if (!g_module_close(module)) g_warning("Error unloading sound module: %s\n", (gchar*)g_module_error());
				} 
				//new module
				if ((module = g_module_open(soundmodule->path, G_MODULE_BIND_LAZY)) != NULL) {
					if (g_module_symbol(module, "sound_module_play_file", (gpointer*)&sound_module_play_file)) {
						g_module_symbol(module, "sound_module_close", (gpointer*)&sound_module_close);
						currentmoduleid = moduleid;
						modulefound = TRUE;
					}
				}
				break;
			}
		}
	} else return FALSE;
	
	if (!modulefound) {
		soundmodule = (struct _soundmodule *)g_slist_nth_data(soundmodules, 0);
		//clean module unload
		if (module != NULL) {
			sound_module_close();
			if (!g_module_close(module)) g_warning("Error unloading sound module: %s\n", (gchar*)g_module_error());
		} 
		//new module
		if ((module = g_module_open(soundmodule->path, G_MODULE_BIND_LAZY)) != NULL) {
			if (g_module_symbol(module, "sound_module_play_file", (gpointer*)&sound_module_play_file)) {
				g_module_symbol(module, "sound_module_close", (gpointer*)&sound_module_close);
				currentmoduleid = soundmodule->id;
				modulefound = TRUE;
			}
		}
	}
	
	if (modulefound) {
		settings_append_int("sounds_module_id", currentmoduleid);
	}
	
	return modulefound;
}

static void sound_langauages_free_foreach(gpointer data, gpointer user_data)
{
	struct _soundlanguage *soundlanguage;
	
	soundlanguage = (struct _soundlanguage *)data;
	g_free(soundlanguage->name);
	g_free(soundlanguage->nativename);
	g_free(soundlanguage->code);
}

gboolean sound_languages_enum(void)
{
	GDir *langdir;
	const gchar *langname;
	struct _soundlanguage *soundlanguage;
	struct _iso639lang isolang;
	gchar **localev;
	gboolean aslocale = FALSE;
	const gchar *enlanguage = "en_US";
	gchar *setlanguage;
	gchar path[PATH_MAX];
	gchar filepath[PATH_MAX];
	gint langid = 0;
	gint langlocaleid = -1;
	gint langsetid =-1;
	gint langenid =-1;
		
	currentlanguage = NULL;
	g_slist_foreach(soundlanguages, sound_langauages_free_foreach, NULL);
	g_slist_free(soundlanguages); soundlanguages = NULL;
	if (systemlocale != NULL) {
		g_free(systemlocale);
		systemlocale = NULL;
	}
	
	localev = g_strsplit(g_getenv("LANG"), ".", 2);
	if (localev[0] != NULL) systemlocale = g_strdup(localev[0]); else systemlocale = g_strdup(enlanguage);
	g_strfreev(localev);
	printf("System locale: %s\n", systemlocale); 
	
	setlanguage = settings_read_string("sounds_language", "locale");
	if (g_str_equal(setlanguage, "locale")) {
		g_free(setlanguage);
		setlanguage = g_strdup(systemlocale);
		aslocale = TRUE;
	}
	
	if ((langdir = g_dir_open(RESOURCE_DIR_SOUND_LANGUAGES, 0, NULL)) != NULL) {
		printf("Languages sound packs enumeration\n");
		while ((langname = g_dir_read_name(langdir)) != NULL) {
			g_sprintf(path, "%s%s", RESOURCE_DIR_SOUND_LANGUAGES, langname);
			if (g_file_test(path, G_FILE_TEST_EXISTS)) {
				soundlanguage = (struct _soundlanguage *)g_malloc0(sizeof(struct _soundlanguage));
				soundlanguage->code = g_strdup(langname);
				soundlanguage->multiplex = SOUND_LANGUAGE_MULTIPLEX_VOID;
				soundlanguage->islocale = FALSE;
				soundlanguage->flagexists = FALSE;
				
				if (iso639_get_lang_by_locale((char *)langname, &isolang) > 0) {
					soundlanguage->name = g_strdup(isolang.name);
					soundlanguage->nativename = g_strdup(isolang.nativename);
				} else {
					soundlanguage->name = g_strdup_printf("Unknown %s", langname);
					soundlanguage->nativename = g_strdup_printf("Unknown %s", langname);
				}
								
				g_sprintf(filepath, "%s/10.wav", path);
				if (g_file_test(filepath, G_FILE_TEST_EXISTS))
					soundlanguage->multiplex = SOUND_LANGUAGE_MULTIPLEX_10;
				g_sprintf(filepath, "%s/100.wav", path);
				if (g_file_test(filepath, G_FILE_TEST_EXISTS)) 
					soundlanguage->multiplex = SOUND_LANGUAGE_MULTIPLEX_100_SIMPLE;
				g_sprintf(filepath, "%s/21.wav", path);
				if (g_file_test(filepath, G_FILE_TEST_EXISTS)) 
					soundlanguage->multiplex = SOUND_LANGUAGE_MULTIPLEX_100_FULL;
				
				g_sprintf(filepath, "%s/flag.png", path);
				if (g_file_test(filepath, G_FILE_TEST_EXISTS)) 
					soundlanguage->flagexists = TRUE;
				
				if (soundlanguage->multiplex == SOUND_LANGUAGE_MULTIPLEX_VOID) {
					g_free(soundlanguage->code);
					g_free(soundlanguage->name);
					g_free(soundlanguage->nativename);
					g_free(soundlanguage);
				} else {
					soundlanguages = g_slist_append(soundlanguages, soundlanguage);
					if ((langlocaleid == -1) && (g_str_equal(systemlocale, langname))) {
						langlocaleid = langid;
						soundlanguage->islocale = TRUE;
						//g_free(soundlanguage->code);
						//soundlanguage->code = g_strdup("locale");
					}
					if ((langsetid == -1) && (g_str_equal(setlanguage, langname))) {
						langsetid = langid;
					}
					if ((langenid == -1) && (g_str_equal(enlanguage, langname))) {
						langenid = langid;
					}
					printf(" %u. %s [%s]\n", langid+1, soundlanguage->name, langname);
					langid++;
				}
			}
		}
		g_dir_close(langdir);
	}
	
	if ((soundlanguages == NULL) || (langlocaleid == -1)) {
		//Add system language entry at least
		soundlanguage = (struct _soundlanguage *)g_malloc0(sizeof(struct _soundlanguage));
		//soundlanguage->code = g_strdup("locale");
		soundlanguage->code = g_strdup(systemlocale);
		soundlanguage->multiplex = SOUND_LANGUAGE_MULTIPLEX_VOID;
		soundlanguage->islocale = TRUE;
		if (iso639_get_lang_by_locale((char *)systemlocale, &isolang) > 0) {
			soundlanguage->name = g_strdup(isolang.name);
			soundlanguage->nativename = g_strdup(isolang.nativename);
		} else {
			soundlanguage->name = g_strdup_printf("Unknown %s", langname);
			soundlanguage->nativename = g_strdup_printf("Unknown %s", langname);
		}
		soundlanguages = g_slist_append(soundlanguages, soundlanguage);
		
		if (langid == 0) {
			printf(" No languages found\n");
			return FALSE;
		}
	}
	
	if ((aslocale) && (langlocaleid != -1)) {
		currentlanguage = g_slist_nth_data(soundlanguages, langlocaleid);
		return TRUE;
	}
	
	if (langsetid != -1) {
		currentlanguage = g_slist_nth_data(soundlanguages, langsetid);
		return TRUE;
	} else if (langenid != -1) {
		currentlanguage = g_slist_nth_data(soundlanguages, langenid);
		return TRUE;
	} else {
		currentlanguage = g_slist_nth_data(soundlanguages, 0);
		return TRUE;
	}
	
	return FALSE;
}

static gint sound_languages_set_valid_finder(gconstpointer a, gconstpointer b)
{
	struct _soundlanguage *language;
	
	language = (struct _soundlanguage *)a;
	
	if (language->multiplex != SOUND_LANGUAGE_MULTIPLEX_VOID) return 0; else return 1;
}

static gint sound_languages_set_finder(gconstpointer a, gconstpointer b)
{
	struct _soundlanguage *language;
	gchar *code;
	
	language = (struct _soundlanguage *)a;
	code = (gchar *)b;
	
	if ((g_str_equal(language->code, code)) && 
		(language->multiplex != SOUND_LANGUAGE_MULTIPLEX_VOID)) return 0; else return 1;
}

gboolean sound_languages_set_language(gchar *code)
{
	gchar *searchcode;
	gboolean localeoption;
	gint i;
	GSList *item;
	struct _soundlanguage *language;
	
	
	if ((soundlanguages == NULL) || (code == NULL)) return FALSE;
	
	printf("Requested language pack code: %s\n", code);
	
	if (g_slist_length(soundlanguages) == 1) {
		language = g_slist_nth_data(soundlanguages, 0);
		if (language != NULL) {
			if ((language->multiplex == SOUND_LANGUAGE_MULTIPLEX_VOID) && (language->islocale)) {
				printf("Language pack %s not installed\n", language->code);
				return FALSE;
			}
		}
	}
	
	language = NULL;
	
	for (i=0; i<=1; i++) {
		if (i == 0) {
			if (g_str_equal(code, "locale")) {
				searchcode = g_strdup(systemlocale);
				localeoption = TRUE;
			} else {
				searchcode = g_strdup(code);
				localeoption = FALSE;
			}
		} else if (i == 1) {
			searchcode = g_strdup("en_US");
			//localeoption = FALSE;
		}
		
		item = g_slist_find_custom(soundlanguages, searchcode, sound_languages_set_finder);
				
		if (item == NULL) {
			g_free(searchcode);
			continue;
		} else {
			language = (struct _soundlanguage *)item->data;
			printf("Selected language pack: %s [%s]\n", language->name, language->code);
			g_free(searchcode);
			break;
		}
		
	}
	
	if (language == NULL) {
		item = g_slist_find_custom(soundlanguages, NULL, sound_languages_set_valid_finder);
		if (item == NULL) {
			printf("No valid language packs found\n");
			return FALSE;
		} else {
			language = item->data;
			//localeoption = FALSE;
			printf("Selected first valid language pack: %s [%s]\n", language->name, language->code);
		}
	}
	
	currentlanguage = language;
	if ((localeoption) || (currentlanguage->islocale)) {
		settings_append_string("sounds_language", "locale");
	} else {
		settings_append_string("sounds_language", language->code);
	}
		
	return TRUE;
}

static gchar **sound_languages_path(struct _frontend *frontend, guchar parameter, guchar limit)
{
	guint reallevel;
	guint multiplex;
	guint limitlevel;
	gchar **pathv;
	
	if (currentlanguage != NULL) {
		if ((currentlanguage->multiplex == SOUND_LANGUAGE_MULTIPLEX_100_SIMPLE) ||
			(currentlanguage->multiplex == SOUND_LANGUAGE_MULTIPLEX_100_FULL)) {
			multiplex = 100;
		} else if (currentlanguage->multiplex == SOUND_LANGUAGE_MULTIPLEX_10) {
			multiplex = 10;
		} else if (currentlanguage->multiplex == SOUND_LANGUAGE_MULTIPLEX_VOID) {
			return NULL;
		}
	} else {
		return NULL;
	}
	
	if (frontend->tuned) {
		switch (parameter) {
			case SOUND_PARAMETER_SNR: {
				reallevel = (frontend->snr/settings_read_float("scale_snr", 100.0)) * multiplex;
				break;
			}
			
			case SOUND_PARAMETER_AGC: {
				reallevel = (frontend->signal/settings_read_float("scale_agc", 1000.0)) * multiplex;
				break;
			}
		
			case SOUND_PARAMETER_BER: {
				reallevel = (frontend->ber/settings_read_float("scale_ber", 4294967296.0)) * multiplex;
				break;
			}
			
			default: {
				reallevel = 0;
				break;
			}
		}
		if (reallevel > multiplex) reallevel = multiplex;
	} else {
		reallevel = 0;
	}
	
	if (currentlanguage->multiplex == SOUND_LANGUAGE_MULTIPLEX_10) {
		if ((reallevel * 10) < limit) return NULL;
	} else if ((currentlanguage->multiplex == SOUND_LANGUAGE_MULTIPLEX_100_SIMPLE) ||
				(currentlanguage->multiplex == SOUND_LANGUAGE_MULTIPLEX_100_FULL)) {
		if (reallevel < limit) return NULL;
	}
	
	if (currentlanguage->multiplex == SOUND_LANGUAGE_MULTIPLEX_10) {
		pathv = g_malloc(sizeof(gchar *)*2);
		pathv[0] = g_strdup_printf("%s%s/%u.wav", RESOURCE_DIR_SOUND_LANGUAGES, currentlanguage->code, reallevel);
		pathv[1] = NULL;
	} else if (currentlanguage->multiplex == SOUND_LANGUAGE_MULTIPLEX_100_SIMPLE) {
		if ((reallevel <= 20) || (reallevel % 10 == 0)) {
			pathv = g_malloc(sizeof(gchar *)*2);
			pathv[0] = g_strdup_printf("%s%s/%u.wav", RESOURCE_DIR_SOUND_LANGUAGES, currentlanguage->code, reallevel);
			pathv[1] = NULL;
		} else {
			pathv = g_malloc(sizeof(gchar *)*3);
			pathv[0] = g_strdup_printf("%s%s/%u.wav", RESOURCE_DIR_SOUND_LANGUAGES, currentlanguage->code, reallevel - reallevel % 10);
			pathv[1] = g_strdup_printf("%s%s/%u.wav", RESOURCE_DIR_SOUND_LANGUAGES, currentlanguage->code, reallevel % 10);
			pathv[2] = NULL;
		}
	} else if (currentlanguage->multiplex == SOUND_LANGUAGE_MULTIPLEX_100_FULL) {
		pathv = g_malloc(sizeof(gchar *)*2);
		pathv[0] = g_strdup_printf("%s%s/%u.wav", RESOURCE_DIR_SOUND_LANGUAGES, currentlanguage->code, reallevel);
		pathv[1] = NULL;
	}
	
	if (!g_file_test(pathv[0], G_FILE_TEST_EXISTS)) {
		g_strfreev(pathv);
		return NULL;
	}
	
	return pathv;
}

gchar *sound_languages_flag_path(gchar *code)
{
	gchar *searchcode;
	gchar *path;
	
	path = NULL;
	
	if (code != NULL) {
		if (g_str_equal(code, "locale")) {
			searchcode = g_strdup(systemlocale);
		} else {
			searchcode = g_strdup(code);
		}
		path = g_strdup_printf("%s%s/flag.png", RESOURCE_DIR_SOUND_LANGUAGES, searchcode);
		g_free(searchcode);
	}
		
	return path;
}

void sound_limit_set(guchar limit)
{
	if (limit <= 100) {
		soundlimit = limit;
		settings_append_int("sounds_limit", soundlimit);
	}
}

guint sound_limit_get(void)
{
	return soundlimit;
}

void sound_parameter_set(guchar parameter)
{
	switch (parameter) {
		case 0: soundparameter = SOUND_PARAMETER_SNR; break;
		case 1: soundparameter = SOUND_PARAMETER_AGC; break;
		case 2: soundparameter = SOUND_PARAMETER_BER; break;
		default: soundparameter = SOUND_PARAMETER_SNR; break;
	}
	
	settings_append_int("sounds_parameter", (guint)soundparameter);
}

guint sound_parameter_get(void)
{
	return soundparameter;
}

GSList *sound_modules_get(guint *index)
{
	gint i;
	struct _soundmodule *soundmodule;
	gsize length;
		
	if (index != NULL) {
		if ((length = g_slist_length(soundmodules)) > 0) {
			for (i=0; i<length; i++) {
				soundmodule = (struct _soundmodule *)g_slist_nth_data(soundmodules, i);
				if (soundmodule->id == currentmoduleid) {
					*index = i;
					break;
				}
			}
		}
	}
		
	return soundmodules;
}

static gint sound_languages_get_finder(gconstpointer a, gconstpointer b)
{
	struct _soundlanguage *language;
	gchar *code;
	
	language = (struct _soundlanguage *)a;
	code = (gchar *)b;
	
	if (g_str_equal(language->code, code)) return 0; else return 1;
}

GSList *sound_languages_get(gint *index)
{
	gint localindex;
	gchar *setlang;
	GSList *item;
	
	if (soundlanguages == NULL) return NULL;
	
	setlang = settings_read_string("sounds_language", "locale");
	
	if (g_str_equal(setlang, "locale")) {
		g_free(setlang);
		setlang = g_strdup(systemlocale);
	}
	
	if (index != NULL) {
		if (currentlanguage != NULL) {
			localindex = g_slist_index(soundlanguages, currentlanguage);
			if (localindex == -1) {
				*index = 0;
			} else {
				*index = localindex;
			} 
		} else {
			item = g_slist_find_custom(soundlanguages, setlang, sound_languages_get_finder);
			if (item != NULL) {
				*index = g_slist_position(soundlanguages, item);
			} else {
				*index = 0;
			}
		}
	}
	
	if (setlang != NULL) g_free(setlang);
	
	return soundlanguages;
}

gboolean sound_init(void)
{
	gboolean langres;
	guint modules;
	guint setmoduleid;
	gboolean modres = FALSE;
	
	langres = sound_languages_enum();
	modules = sound_modules_enum();
		
	setmoduleid = settings_read_int("sounds_module_id", 1);
	
	if (modules > 0) {
		if (sound_modules_set_module(setmoduleid)) {
			modres = TRUE;
		}
	}
	
	soundparameter = settings_read_int("sounds_parameter", 0);
	soundlimit = settings_read_int("sounds_limit", 0);
	
	return langres && modres;
}

gboolean sound_play(struct _frontend *frontend, gulong timeout)
{
	gchar **sndpathv;
	guint i;
	
	if ((sndpathv = sound_languages_path(frontend, soundparameter, soundlimit)) != NULL) {
		i = 0;
		while (sndpathv[i] != NULL) {
			//printf("--> %s\n", sndpathv[i]);
			//sound_module_play_file(sndpathv[i], sound_wav_header_parser_callback);
			sound_module_play_file(sndpathv[i], sound_wav_file_handler);
			i++;
		}
		g_strfreev(sndpathv);
		if (timeout != 0) g_usleep(timeout);
		return TRUE;
	}
	
	if (timeout != 0) g_usleep(timeout);
	
	return FALSE;
}
