/*
 *      alsasound.c
 *      
 *      Copyright 2010-2012 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#define ALSA_PCM_NEW_HW_PARAMS_API
#define ALSA_PCM_NEW_SW_PARAMS_API

#include <glib.h>
#include <gmodule.h>
#include <alsa/asoundlib.h>
#include <sys/time.h>
#include "soundmodules.h"


#define SOUND_MODULE_ID       1
#define SOUND_MODULE_NAME     "ALSA"
#define SOUND_MODULE_VERSION  0x100

#define	ALSA_BUFFER_LEN	      2048
#define	ALSA_DEVICE	          "default"

static snd_pcm_t *device;
static gboolean deviceopened;
static gshort channels;
static gshort bits;
static gulong samples;

static gboolean alsa_open(struct _wavheader *wavhdr);
static gboolean alsa_close(void);
static int alsa_write_data(snd_pcm_t *alsa_dev, short *data, int frames, int channels);


G_MODULE_EXPORT gboolean sound_module_play_file(gchar *filename, glong (*file_handler)(wavheader_t header, gint command, gpointer buffer, gsize size))
{	
	struct _wavheader header;
	gshort buffer[ALSA_BUFFER_LEN];
	glong readbytes;
	gint multiplex = sizeof(short);
	gboolean res;
	
	if (file_handler == NULL) return FALSE;
	
	if ((file_handler)(&header, SOUND_WAV_FILE_OPEN, filename, 0) == 0) return FALSE;
	
	res = FALSE;
	
	if (alsa_open(&header)) {
		res = TRUE;
		//Correction for 8 bps format
		switch(header.bits) {
			case 8:  multiplex = sizeof(char); break;
			case 16: multiplex = sizeof(short); break;
			default: multiplex = sizeof(short); break;
		}
		//Play sound	
		while ((readbytes = (file_handler)(&header, SOUND_WAV_FILE_READ, &buffer, ALSA_BUFFER_LEN*multiplex)) > 0) {
			if (alsa_write_data(device, buffer, readbytes/((header.channels)*multiplex), header.channels) <= 0) {
				res = FALSE;
				break;
			}
		}
	}
	
	(file_handler)(&header, SOUND_WAV_FILE_CLOSE, NULL, 0);
	
	return res;
}

static gboolean alsa_open(struct _wavheader *wavhdr)
{	
	const char *devname = "default";
	snd_pcm_hw_params_t *hw_params;
	snd_pcm_sw_params_t *sw_params;
	snd_pcm_uframes_t period_size, buffer_frames;

	int err;
	
	if (!deviceopened) {
		if ((err = snd_pcm_open(&device, devname, SND_PCM_STREAM_PLAYBACK, 0)) < 0) {
			g_warning("[ALSA] Audio device %s open error %s\n", devname, snd_strerror(err));
			if (device != NULL) snd_pcm_close(device);
			deviceopened = FALSE;
			return FALSE;
		}
		deviceopened = TRUE;
	}
	
	if ((channels == wavhdr->channels) || (bits == wavhdr->bits) || (samples == wavhdr->samples_per_sec)) return TRUE;
	
	snd_pcm_nonblock(device, 0);

	if ((err = snd_pcm_hw_params_malloc(&hw_params)) < 0) {
		g_warning("[ALSA] Hardware parameters allocation error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	if ((err = snd_pcm_hw_params_any(device, hw_params)) < 0) {
		g_warning("[ALSA] Hardware parameters initialization error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	if ((err = snd_pcm_hw_params_set_access(device, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED)) < 0) {
		g_warning("[ALSA] Access type setting error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}
	
	int format = 0;
	switch(wavhdr->bits) {
		case 8:  format = SND_PCM_FORMAT_U8; break;
		case 16: format = SND_PCM_FORMAT_S16_LE; break;
		default: format = SND_PCM_FORMAT_S16_LE; break;
	}

	if ((err = snd_pcm_hw_params_set_format(device, hw_params, format)) < 0) {
		g_warning("[ALSA] Sample format setting error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	if ((err = snd_pcm_hw_params_set_rate_near(device, hw_params, (unsigned int *)&wavhdr->samples_per_sec, 0)) < 0) {
		g_warning("[ALSA] Sample rate setting error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	if ((err = snd_pcm_hw_params_set_channels(device, hw_params, wavhdr->channels)) < 0) {
		g_warning("[ALSA] Channel count setting error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}
	
	period_size = 1024;
	buffer_frames = 4*period_size;

	if ((err = snd_pcm_hw_params_set_buffer_size_near(device, hw_params, &buffer_frames)) < 0) {
		g_warning("[ALSA] Buffer size setting error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	if ((err = snd_pcm_hw_params_set_period_size_near(device, hw_params, &period_size, 0)) < 0) {
		g_warning("[ALSA] Period size setting error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	if ((err = snd_pcm_hw_params(device, hw_params)) < 0) {
		g_warning("[ALSA] Hardware parameters setting error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	snd_pcm_hw_params_free(hw_params) ;

	if ((err = snd_pcm_sw_params_malloc(&sw_params)) != 0) {
		g_warning("[ALSA] Software parameters allocation error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	if ((err = snd_pcm_sw_params_current(device, sw_params)) != 0) {
		g_warning("[ALSA] Default software parameters setting setting error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	if ((err = snd_pcm_sw_params(device, sw_params)) != 0) {
		g_warning("[ALSA] Software parameters setting error %s\n", snd_strerror(err));
		if (device != NULL) snd_pcm_close(device);
		deviceopened = FALSE;
		return FALSE;
	}

	snd_pcm_sw_params_free(sw_params);

	snd_pcm_reset(device);
	
	channels = wavhdr->channels;
	bits = wavhdr->bits;
	samples = wavhdr->samples_per_sec;
	
	return TRUE;
}

static gboolean alsa_close(void)
{
	if ((device != NULL) && (deviceopened)) {
		snd_pcm_drain(device);
		snd_pcm_close(device);
		device = NULL;
		deviceopened = FALSE;
		channels = 0;
		bits = 0;
		samples = 0;
	}
	
	return TRUE;
}

static int alsa_write_data(snd_pcm_t *alsa_dev, short *data, int frames, int channels)
{	
	int wrbytes = 0;
	int wrnow;

	while (wrbytes < frames) {
		wrnow = snd_pcm_writei(alsa_dev, data + wrbytes * channels, frames - wrbytes);

		if (wrnow >= 0) {
			wrbytes += wrnow;
			if (wrbytes == frames) return wrbytes;
			continue;
		}

		switch (wrnow) {
			case -EAGAIN: {
				g_warning("[ALSA] Error: EAGAIN\n");
				continue;
				break;
			}

			case -EPIPE: {
				g_warning("[ALSA] Error: EPIPE\n");
				snd_pcm_prepare(alsa_dev) ;
				break ;
			}

			case -EBADFD: {
				g_warning("[ALSA] Error: BAD FD\n");
				return 0;
				break;
			}

			case -ESTRPIPE: {
				g_warning("[ALSA] Error: ESTRPIPE\n");
				return 0;
				break;
			}

			case -EIO: {
				g_warning("[ALSA] Error: EIO\n");
				return 0;
			}

			default: {
				g_warning("[ALSA] Error: %d\n", wrnow);
				return 0;
				break;
			}
		}
	}
	return wrbytes;
}

G_MODULE_EXPORT gboolean sound_module_close(void)
{
	alsa_close();
	
	return TRUE;
}

G_MODULE_EXPORT gboolean sound_module_init(guint *moduleid, gchar *modulename, gint *moduleversion)
{
	gsize namelen;
	
	if (access("/proc/asound/cards", R_OK) != 0) {
		g_warning("[ALSA] Sound cards not found\n");
		return FALSE;
	}
		
	if (moduleid != NULL) {
		*moduleid = SOUND_MODULE_ID;
	}
	
	if (modulename != NULL) {
		namelen = strlen(SOUND_MODULE_NAME);
		memset(modulename, 0, 32);
		if (namelen > 31) {
			strncpy(modulename, SOUND_MODULE_NAME, 31);
		} else {
			strncpy(modulename, SOUND_MODULE_NAME, namelen);
		}
	}
	
	if (moduleversion != NULL) {
		*moduleversion = SOUND_MODULE_VERSION;
	}
	
	device = NULL;
	deviceopened = FALSE;
	channels = 0;
	bits = 0;
	samples = 0;
	
	return TRUE;
}
