/*
 *      osssound.c
 *      
 *      Copyright 2010-2012 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib.h>
#include <gmodule.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include "soundmodules.h"

#define SOUND_MODULE_ID       3
#define SOUND_MODULE_NAME     "OpenSound (OSS)"
#define SOUND_MODULE_VERSION  0x100

#define OPENSOUND_DSP_DEVICE  "/dev/dsp"
#define OPENSOUND_BUFFER_LEN  1024

static gint  device;
static gboolean deviceopened;
static gshort channels;
static gshort bits;
static gulong samples;

gboolean opensound_open(struct _wavheader *wavhdr)
{
	int soundprofile = APF_NORMAL;
	int sampleformat;
	int newsample;
	
	if (!deviceopened) {
		if ((device = open(OPENSOUND_DSP_DEVICE, O_WRONLY, 0)) == -1) {
			g_warning("[OpenSound] Failed to open sound device\n");
			return FALSE;
		}
		deviceopened = TRUE;
	}
	
	if ((channels == wavhdr->channels) || (bits == wavhdr->bits) || (samples == wavhdr->samples_per_sec)) return TRUE;
	
	switch(wavhdr->bits) {
		case 8:  sampleformat = AFMT_U8; break;
		case 16: sampleformat = AFMT_S16_LE; break;
		default: sampleformat = AFMT_S16_LE; break;
	}
	
	newsample = sampleformat;
	if (ioctl(device, SNDCTL_DSP_SETFMT, &newsample) == -1) {
		g_warning("[OpenSound] Failed to set sample format\n");
		close(device);
		deviceopened = FALSE;
		return FALSE;
	}
	
	if (newsample != sampleformat) {
		g_warning("[OpenSound] Sample format not supported\n");
		close(device);
		deviceopened = FALSE;
		return FALSE;
	}
	
	ioctl(device, SNDCTL_DSP_PROFILE, &soundprofile);
	
	if (ioctl(device, SNDCTL_DSP_CHANNELS, &wavhdr->channels) == -1) {
		g_warning("[OpenSound] Failed to set channels number\n");
		close(device);
		deviceopened = FALSE;
		return FALSE;
	}
	
	if (ioctl(device, SNDCTL_DSP_SPEED, &wavhdr->samples_per_sec) == -1) {
		g_warning("[OpenSound] Failed to set sample rate\n");
		close(device);
		deviceopened = FALSE;
		return FALSE;
	}
	
	channels = wavhdr->channels;
	bits = wavhdr->bits;
	samples = wavhdr->samples_per_sec;
	
	return TRUE;	
}

G_MODULE_EXPORT gboolean sound_module_close(void)
{
	if (deviceopened) {
		ioctl(device, SNDCTL_DSP_SYNC, NULL);
		close(device);
		device = 0;
		deviceopened = FALSE;
		channels = 0;
		bits = 0;
		samples = 0;
	}
	
	return TRUE;
}

G_MODULE_EXPORT gboolean sound_module_play_file(gchar *filename,  glong (*file_handler)(wavheader_t header, gint command, gpointer buffer, gsize size))
{	
	struct _wavheader header;
	guchar buffer[OPENSOUND_BUFFER_LEN];
	glong readbytes;
	gboolean res;
	
	if (file_handler == NULL) return FALSE;
	
	if ((file_handler)(&header, SOUND_WAV_FILE_OPEN, filename, 0) == 0) return FALSE;
	
	res = FALSE;
	
	if (opensound_open(&header)) {
		res = TRUE;
		while ((readbytes = (file_handler)(&header, SOUND_WAV_FILE_READ, &buffer, OPENSOUND_BUFFER_LEN)) > 0) {
			if (write(device, buffer, readbytes) != readbytes) {
				g_warning("[OpenSound] Write error\n");
				res = FALSE;
				break;
			}
		}
		ioctl(device, SNDCTL_DSP_SYNC, NULL);
	}
	
	(file_handler)(&header, SOUND_WAV_FILE_CLOSE, NULL, 0);
	
	return res;
}

G_MODULE_EXPORT gboolean sound_module_init(guint *moduleid, gchar *modulename, gint *moduleversion)
{
	gint fd;
	gint ver = 0;
	gint namelen;
	
	if ((fd = open(OPENSOUND_DSP_DEVICE, O_RDONLY, 0)) == -1) {
		g_warning("[OpenSound] Failed to open sound device\n");
		return FALSE;
	}

	if (ioctl(fd, OSS_GETVERSION, &ver) == -1) {
		close(fd);
		g_warning("[OpenSound] Failed to get version\n");
		return FALSE;
	}
	
	if (ver < 0x030900) {
		close(fd);
		g_warning("[OpenSound] Wrong version\n");
		return FALSE;
	}
	
	close(fd);
	
	if (moduleid != NULL) {
		*moduleid = SOUND_MODULE_ID;
	}
	
	if (modulename != NULL) {
		namelen = strlen(SOUND_MODULE_NAME);
		memset(modulename, 0, 32);
		if (namelen > 31) {
			strncpy(modulename, SOUND_MODULE_NAME, 31);
		} else {
			strncpy(modulename, SOUND_MODULE_NAME, namelen);
		}
	}
	
	if (moduleversion != NULL) {
		*moduleversion = SOUND_MODULE_VERSION;
	}
	
	device = 0;
	deviceopened = FALSE;
	channels = 0;
	bits = 0;
	samples = 0;
	
	return TRUE;
}
