/*
 *      pulsesound.c
 *      
 *      Copyright 2010-2012 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


#include <glib.h>
#include <gmodule.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h>
#include <pulse/simple.h>
#include <pulse/error.h>
#include <pulse/gccmacro.h>
#include "soundmodules.h"

#define SOUND_MODULE_ID       2
#define SOUND_MODULE_NAME     "PulseAudio"
#define SOUND_MODULE_VERSION  0x100

#define PULSE_BUFFER_LEN      1024

static pa_simple *device;
static gshort channels;
static gshort bits;
static gulong samples;

static gboolean pulseaudio_open(struct _wavheader *wavhdr);
static gboolean pulseaudio_close(void);


static gboolean pulseaudio_open(struct _wavheader *wavhdr)
{
	pa_sample_spec ss;
	int error;
	
	if ((channels != wavhdr->channels) || (bits != wavhdr->bits) || (samples != wavhdr->samples_per_sec)) {
		switch(wavhdr->bits) {
			case 8:  ss.format = PA_SAMPLE_U8; break;
			case 16: ss.format = PA_SAMPLE_S16LE; break;
			default: ss.format = PA_SAMPLE_S16LE; break;
		}
		
		ss.rate = wavhdr->samples_per_sec;
		ss.channels = wavhdr->channels;
		
		pulseaudio_close();
		
		if ((device = pa_simple_new(NULL, "GSF", PA_STREAM_PLAYBACK, NULL, "playback", &ss, NULL, NULL, &error)) == NULL) {
			g_warning("[PulseAudio] Audio server open error %s", pa_strerror(error));
			return FALSE;
		}
		
		channels = wavhdr->channels;
		bits = wavhdr->bits;
		samples = wavhdr->samples_per_sec;
	}
	
	return TRUE;	
}

static gboolean pulseaudio_close(void)
{
	gint error;
	
	if (device != NULL) {
		//drain and close connection
		if (pa_simple_drain(device, &error) < 0) {
			g_warning("[PulseAudio] Audio server drain error %s", pa_strerror(error));
		}
		pa_simple_free(device);
		device = NULL;
	}
	
	return TRUE;
}

G_MODULE_EXPORT gboolean sound_module_play_file(gchar *filename, glong (*file_handler)(wavheader_t header, gint command, gpointer buffer, gsize size))
{	
	struct _wavheader header;
	uint8_t buffer[PULSE_BUFFER_LEN];
	glong readbytes;
	gint error;
	gboolean res;
	
	if (file_handler == NULL) return FALSE;
	
	if ((file_handler)(&header, SOUND_WAV_FILE_OPEN, filename, 0) == 0) return FALSE;
	
	res = FALSE;
	
	if (pulseaudio_open(&header)) {
		res = TRUE;
		while ((readbytes = (file_handler)(&header, SOUND_WAV_FILE_READ, &buffer, PULSE_BUFFER_LEN)) > 0) {
			if (pa_simple_write(device, &buffer, (size_t)readbytes, &error) < 0) {
				g_warning("[PulseAudio] Audio server write error %s\n", pa_strerror(error));
				res = FALSE;
				break;
			}
		}
	}
	
	(file_handler)(&header, SOUND_WAV_FILE_CLOSE, NULL, 0);
	
	return res;
}

G_MODULE_EXPORT gboolean sound_module_close(void)
{
	pulseaudio_close();
	
	channels = 0;
	bits = 0;
	samples = 0;
	
	return TRUE;
}

G_MODULE_EXPORT gboolean sound_module_init(guint *moduleid, gchar *modulename, gint *moduleversion)
{
	gchar path[PATH_MAX];
	gchar *tmp;
	gboolean pathfound = FALSE;
	gchar buf[16];
	gint uid = -1;
	DIR *d;
	struct dirent *entry;
	struct stat status;
	gint pidfd;
	gboolean pidres = FALSE;
	gsize namelen;
	
	uid = getuid();
	if ((d = opendir("/tmp")) == NULL) {
		g_warning("[PulseAudio] Directory listing read error");
		return FALSE;
	}
	
	while (entry = readdir(d)) {
		if (strstr(entry->d_name, "pulse") != NULL) {
			g_sprintf(path, "/tmp/%s", entry->d_name);
			if (stat(path, &status) != -1) {
				if (status.st_uid == uid) {
					pathfound = TRUE;
					break;
				}
			}
		}
	}
	
	closedir(d);
	
	if (pathfound) {
		tmp = g_strdup(path);
		g_sprintf(path, "%s/pid", tmp);
		g_free(tmp);
		if ((pidfd = open(path, O_RDWR, 0)) != -1) {
			if (read(pidfd, &buf, 15) != 0) {
				if (kill((pid_t)atoi(buf), 0) == 0) {
					pidres = TRUE;
				}
			}
			close(pidfd);
		}
	}
	
	if (pidres) {
		if (moduleid != NULL) {
			*moduleid = SOUND_MODULE_ID;
		}
		
		if (modulename != NULL) {
			namelen = strlen(SOUND_MODULE_NAME);
			memset(modulename, 0, 32);
			if (namelen > 31) {
				strncpy(modulename, SOUND_MODULE_NAME, 31);
			} else {
				strncpy(modulename, SOUND_MODULE_NAME, namelen);
			}
		}
		
		if (moduleversion != NULL) {
			*moduleversion = SOUND_MODULE_VERSION;
		}
		
		device = NULL;
		channels = 0;
		bits = 0;
		samples = 0;
		
		return TRUE;
	}
	
	return FALSE;
}
