/*
 *      streamconfig.c
 *      
 *      Copyright 2009 - 2012 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <net/route.h>
#include <arpa/inet.h>
#include <stdint.h>
#include <sys/un.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/poll.h>
#include <signal.h>
#include "streamconfig.h"
#include "gsfcore.h"
#include "resources.h"
#include "./common/serialize.h"
#include "./common/sysutils.h"
#include "./gsfprivileged/gsfshm.h"
#include "./gsfstrsvr/streamserver.h"
#include "./gsfstrsvr/transportstream.h"

#define SHORT_ANSWER_BUFFER   16
#define ANSWER_BUFFER         1024

#ifndef INFTIM
	#define INFTIM -1
#endif

static GSList *netinterfaces = NULL;

static gboolean streamconfig_add_multicast_route(struct _streamconfig *sc, gchar *intname);
static gboolean streamconfig_remove_multicast_route(struct _streamconfig *sc);

//static gboolean streamconfig_service_restart(struct _streamconfig *sc);
static gboolean streamconfig_service_stop(struct _streamconfig *sc);

static gboolean streamconfig_request(struct _streamconfig *sc, char *reqbuf, size_t *reqsize, char *ansbuf, size_t *anssize);
gboolean streamconfig_file_flush(struct _streamconfig *sc);
gboolean streamconfig_file_append_string(struct _streamconfig *sc, gchar *name, gchar *value);
gchar *streamconfig_file_read_string(struct _streamconfig *sc, gchar *name, gchar *def);
gboolean streamconfig_file_append_int(struct _streamconfig *sc, gchar *name, gint value);
gint streamconfig_file_read_int(struct _streamconfig *sc, gchar *name, gint def);


gchar *streamconfig_sockaddr_to_string(struct sockaddr_in *sa, gchar *buffer, gsize length)
{
	switch (sa->sin_family) {
			case AF_INET: {
				inet_ntop(AF_INET, &sa->sin_addr, buffer, length);
				break;
			}
			
			case AF_INET6: {
				inet_ntop(AF_INET6, &sa->sin_addr, buffer, length);
				break;
			}
			
			default: {
				inet_ntop(AF_INET, &sa->sin_addr, buffer, length);
				break;
			}
		}
	
	return buffer;
}

void streamconfig_interfaces_get_with_multicast_foreach(gpointer data, GSList **interfaces)
{
	struct _netinterface *interface;
	
	interface = (struct _netinterface *)data;
	if ((interface->ifup) && ((interface->ifmc) || (interface->ifloop))) {
		*interfaces = g_slist_prepend(*interfaces, interface);
	}
}

GSList *streamconfig_interfaces_get_with_multicast(void)
{
	GSList *interfaces = NULL;
	
	if (netinterfaces == NULL) return NULL;
	
	g_slist_foreach(netinterfaces, (GFunc)streamconfig_interfaces_get_with_multicast_foreach, &interfaces);
	interfaces = g_slist_reverse(interfaces);
	
	return interfaces;
}

void streamconfig_interfaces_get_with_broadcast_foreach(gpointer data, GSList **interfaces)
{
	struct _netinterface *interface;
	
	interface = (struct _netinterface *)data;
	if ((interface->ifup) && ((interface->ifbc) || (interface->ifloop))) {
		*interfaces = g_slist_prepend(*interfaces, interface);
	}
}

GSList *streamconfig_interfaces_get_with_broadcast(void)
{
	GSList *interfaces = NULL;
	
	if (netinterfaces == NULL) return NULL;
	
	g_slist_foreach(netinterfaces, (GFunc)streamconfig_interfaces_get_with_broadcast_foreach, &interfaces);
	interfaces = g_slist_reverse(interfaces);
	
	return interfaces;
}

void streamconfig_interfaces_list_destroy_foreach(gpointer data, gpointer user_data)
{
	g_slice_free(struct _netinterface, data);
}

void streamconfig_interfaces_list_destroy(void)
{
	g_slist_foreach(netinterfaces, streamconfig_interfaces_list_destroy_foreach, NULL);
	g_slist_free(netinterfaces);
	netinterfaces = NULL;
}

gint streamconfig_interfaces_list_build(void)
{
	gint socketfd;
	struct ifconf ifc;
	struct ifreq  *ifr, *currentifr;
	gint ifcbuflen;
	gchar *ifcbuf;
	guint ifcount;
	gint i;
	struct _netinterface *interface;
	
	
	if ((socketfd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
		perror("Socket creation error");
		return -1;
	}
	
	if (netinterfaces != NULL)
		streamconfig_interfaces_list_destroy();
	
	ifcbuflen = 100 * sizeof(struct ifreq);
	
	while (1) {
		ifcbuf = malloc(ifcbuflen);
		ifc.ifc_len = ifcbuflen;
		ifc.ifc_buf = ifcbuf;
		
		if (ioctl(socketfd, SIOCGIFCONF, &ifc) == -1) {
			if (errno != EINVAL) perror("IOCTL SIOCGIFCONF error");
		} else {
			ifcbuflen = ifc.ifc_len;
			break;
		}		 
		ifcbuflen += 10 * sizeof(struct ifreq);
		free(ifcbuf);
	}
	
	ifr = ifc.ifc_req;
	ifcount = ifc.ifc_len / sizeof(struct ifreq);
	
	for (i=0; i<ifcount; i++) {
		currentifr = &ifr[i];
		
		interface = g_slice_new0(struct _netinterface);
				
		memcpy(interface->ifname, currentifr->ifr_name, IFNAMSIZ);
		
		streamconfig_sockaddr_to_string((struct sockaddr_in *)&currentifr->ifr_addr, (char *)&interface->ifipaddr, INET6_ADDRSTRLEN);
		
		interface->ifindex = if_nametoindex((gchar *)&interface->ifname);
		
		if (ioctl(socketfd, SIOCGIFHWADDR, currentifr) == -1) {
			perror("IOCTL SIOCGIFHWADDR error");
		} else {
			memcpy(interface->ifmacaddr, &currentifr->ifr_hwaddr.sa_data[0], IFHWADDRLEN);
		}
		
		if (ioctl(socketfd, SIOCGIFFLAGS, currentifr) == -1) {
			interface->ifup = interface->ifbc = interface->ifmc = interface->ifloop = interface->ifptp = 0;
			perror("IOCTL SIOCGIFFLAGS error");
		} else {
			if (currentifr->ifr_flags & IFF_UP) interface->ifup = 1; else interface->ifup = 0;
			if (currentifr->ifr_flags & IFF_BROADCAST) interface->ifbc = 1; else interface->ifbc = 0;
			if (currentifr->ifr_flags & IFF_MULTICAST) interface->ifmc = 1; else interface->ifmc = 0;
			if (currentifr->ifr_flags & IFF_LOOPBACK) interface->ifloop = 1; else interface->ifloop = 0;
			if (currentifr->ifr_flags & IFF_POINTOPOINT) interface->ifptp = 1; else interface->ifptp = 0;
		}
		
		if (ioctl(socketfd, SIOCGIFBRDADDR, currentifr) == -1) {
			perror("IOCTL SIOCGIFBRDADDR error");
		} else {
			streamconfig_sockaddr_to_string((struct sockaddr_in *)&currentifr->ifr_addr, (char *)&interface->ifbcipaddr, INET_ADDRSTRLEN);
		}
		
		if (ioctl(socketfd, SIOCGIFNETMASK, currentifr) == -1) {
			perror("IOCTL SIOCGIFBRDADDR error");
		} else {
			streamconfig_sockaddr_to_string((struct sockaddr_in *)&currentifr->ifr_netmask, (char *)&interface->ifnetmask, INET6_ADDRSTRLEN);
		}
		
		netinterfaces = g_slist_prepend(netinterfaces, interface);
		
		printf("%s <%u>:\n", interface->ifname, interface->ifindex);
		printf("  IP: %s, NETMASK: %s, BC IP: %s, MAC: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n", interface->ifipaddr, interface->ifnetmask, interface->ifbcipaddr, interface->ifmacaddr[0], interface->ifmacaddr[1], interface->ifmacaddr[2], interface->ifmacaddr[3], interface->ifmacaddr[4], interface->ifmacaddr[5]);
		printf("  [UP: %i, BC: %i, MC: %i, LOOP: %i, PTP: %i]\n", interface->ifup, interface->ifbc, interface->ifmc, interface->ifloop, interface->ifptp);
	}
	
	free(ifcbuf);
	close(socketfd);
	
	netinterfaces = g_slist_reverse(netinterfaces);
		
	return ifcount;
}


static gboolean streamconfig_add_multicast_route(struct _streamconfig *sc, gchar *intname)
{
	key_t shmkey;
	int shmid;
	struct srequest *shmblock;
	gchar **ppargv;
	gint ppstatus;
	GError *pperror = NULL;
	gboolean result = FALSE; 
	
	if (strncmp(sc->mcintname, intname, IFNAMSIZ) == 0) return TRUE;
	
	shmkey = ftok(PRIVILEGED_TOKEN_FILE, PRIVILEGED_PROJID);
	shmid = shmget(shmkey, sizeof(struct srequest), 0666 | IPC_CREAT);
	shmblock = (struct srequest *)shmat(shmid, 0, 0);
	
	memset(shmblock, 0, sizeof(struct srequest));
	shmblock->action = PRIVILEGED_ACTION_ADD_MULTICAST_ROUTE;
	strncpy(shmblock->intname, intname, IFNAMSIZ);
	strncpy(shmblock->remintname, sc->mcintname, IFNAMSIZ);
	
	ppargv = (gchar **)g_malloc(sizeof(gchar *));
	
	ppargv[0] = g_strdup(RESOURCE_BINARY_PRIVILEGED); 
	ppargv[1] = NULL;
	if (g_spawn_sync(NULL, ppargv, NULL, G_SPAWN_STDOUT_TO_DEV_NULL | G_SPAWN_STDERR_TO_DEV_NULL | G_SPAWN_FILE_AND_ARGV_ZERO, NULL, NULL, NULL, NULL, &ppstatus, &pperror)) {
		ppstatus = WEXITSTATUS(ppstatus);
		switch(ppstatus) {
			case PRIVILEGED_STATUS_OK:
				strncpy(sc->mcintname, intname, IFNAMSIZ);
				printf("[+] Set multicast route for interface %s\n", intname);
				result = TRUE;
				break;
			case PRIVILEGED_STATUS_ROUTE_ERROR:
				printf("[-] Failed to set multicast route for interface %s\n", intname);
				//strncpy(sc->mcintname, intname, IFNAMSIZ);
				result = FALSE;
				break;
			default: 
				printf("[-] Failed to set multicast route for interface %s (status %i)\n", intname, ppstatus);
				result = FALSE;
				break;
		}
	} else {
		printf("[-] Error while executing privileged process: %s\n", pperror->message);
		g_error_free(pperror);
		result = FALSE;
	}
	
	g_strfreev(ppargv);
	
	shmdt(shmblock);
	shmctl(shmid, IPC_RMID, 0);
		
	return result;
}

static gboolean streamconfig_remove_multicast_route(struct _streamconfig *sc)
{
	key_t shmkey;
	int shmid;
	struct srequest *shmblock;
	gchar **ppargv;
	gint ppstatus;
	GError *pperror = NULL;
	gboolean result = FALSE; 
	
	if (sc->mcintname[0] == '\0') return FALSE;
	
	shmkey = ftok(PRIVILEGED_TOKEN_FILE, PRIVILEGED_PROJID);
	shmid = shmget(shmkey, sizeof(struct srequest), 0666 | IPC_CREAT);
	shmblock = (struct srequest *)shmat(shmid, 0, 0);
	
	memset(shmblock, 0, sizeof(struct srequest));
	shmblock->action = PRIVILEGED_ACTION_REMOVE_MULTICAST_ROUTE;
	strncpy(shmblock->intname, sc->mcintname, IFNAMSIZ);
	
	ppargv = (gchar **)g_malloc(sizeof(gchar *));
	
	ppargv[0] = g_strdup(RESOURCE_BINARY_PRIVILEGED); 
	ppargv[1] = NULL;
	if (g_spawn_sync(NULL, ppargv, NULL, G_SPAWN_STDOUT_TO_DEV_NULL | G_SPAWN_STDERR_TO_DEV_NULL | G_SPAWN_FILE_AND_ARGV_ZERO, NULL, NULL, NULL, NULL, &ppstatus, &pperror)) {
		ppstatus = WEXITSTATUS(ppstatus);
		switch(ppstatus) {
			case PRIVILEGED_STATUS_OK:
				printf("[+] Removed multicast route for interface %s\n", sc->mcintname);
				memset(sc->mcintname, 0, IFNAMSIZ);
				result = TRUE;
				break;
			case PRIVILEGED_STATUS_ROUTE_ERROR:
				printf("[-] Failed to remove multicast route for interface %s\n", sc->mcintname);
				result = FALSE;
				break;
			default: 
				printf("[-] Failed to remove multicast route for interface %s (status %i)\n", sc->mcintname, ppstatus);
				result = FALSE;
				break;
		}
	} else {
		printf("[-] Error while executing privileged process: %s\n", pperror->message);
		g_error_free(pperror);
		result = FALSE;
	}
	
	g_strfreev(ppargv);
	
	shmdt(shmblock);
	shmctl(shmid, IPC_RMID, 0);
		
	return result;
}

/* -1 - stopped, 0 - no change, 1 - restarted*/

gint streamconfig_service_state(struct _streamconfig *sc)
{
	gint lockfd;
	const char *homedir;
	gchar *lockfn;
	struct flock lock;
	
	if (sysutils_process_exists(sc->processpid)) return 0;
	
	homedir = sysutils_gethomedir();
	lockfn = g_strdup_printf(STREAMCONFIG_LOCK_PATH, homedir);
	
	if ((lockfd = open(lockfn, O_RDWR, 0)) == -1) {
		g_free(lockfn);
		return -1;
	}
	g_free(lockfn);
	
	lock.l_type = F_WRLCK;
	lock.l_whence = SEEK_SET;
	lock.l_start = 0;
	lock.l_len = 0;
	
	if (fcntl(lockfd, F_SETLK, &lock) == -1) {
		if (errno == EACCES || errno == EAGAIN) {
			if (fcntl(lockfd, F_GETLK, &lock) != -1) {
				if (lock.l_pid != sc->processpid) {
					sc->processpid = lock.l_pid;
					close(lockfd);
					return 1;
				} else {
					close(lockfd);
					return 0;
				}
			}
			close(lockfd);
			return -1;
		}
	}
	
	close(lockfd);
	return -1;
}

gboolean streamconfig_service_restart(struct _streamconfig *sc)
{
	gchar *mcintname;
	sigset_t blockmask, oldmask;
	gchar **processargv, **envp;
	gchar *stdout;
	GError *process_error;
	gint status;
	gboolean result;
		
	
	if (streamconfig_service_state(sc) == 1) {
		printf("[+] Streaming server already started (PID: %ld)\n", sc->processpid);
		return TRUE;
	}
	
	if (!g_file_test(RESOURCE_BINARY_STREAMSERVER, G_FILE_TEST_IS_EXECUTABLE)) {
		printf("[-] Streaming server executable (%s) not found\n", RESOURCE_BINARY_STREAMSERVER);
		return FALSE;
	}
		
	if (sc->netintlist) {
		mcintname = streamconfig_file_read_string(sc, "udp_multicast_interface", "lo");
		sc->netintctl = streamconfig_add_multicast_route(sc, mcintname);
		g_free(mcintname);
	}
	
	processargv = (gchar **)g_malloc(sizeof(gchar *)*2);
	processargv[0] = g_strdup(RESOURCE_BINARY_STREAMSERVER);
	processargv[1] = g_strdup("-d");
	processargv[2] = NULL;
	
	envp = (gchar **)g_malloc(sizeof(gchar *)*4);
	envp[0] = g_strdup_printf("USER=%s", g_get_user_name());
	envp[1] = g_strdup_printf("HOME=%s", g_get_home_dir());
	envp[2] = g_strdup_printf("HOSTPID=%i", (gint)getpid());
	envp[3] = NULL;
	
	sigemptyset(&blockmask);
	sigaddset(&blockmask, SIGCHLD);
	sigprocmask(SIG_BLOCK, &blockmask, &oldmask);
	
	process_error = NULL;
	stdout = NULL;
	
	if (!g_spawn_sync(NULL, processargv, envp, 0, NULL, NULL, &stdout, NULL, &status, &process_error)) {
		printf("[-] Error while running streaming server: %s\n", process_error->message);
		g_error_free(process_error);
		result = FALSE;
	} else {
		status = WEXITSTATUS(status);
		sc->processpid = atoi(stdout);
		if (sc->processpid != 0) {
			if (status == EXIT_SUCCESS) {
				printf("[+] Started streaming server (PID: %ld)\n", sc->processpid);
				result = TRUE;
			} else {
				printf("[+] Streaming server already started (PID: %ld)\n", sc->processpid);
				result = TRUE;
			}
		} else {
			printf("[-] Streaming server not started (status: %i)\n", status);
			result = FALSE;
		}
	}
	
	g_strfreev(processargv);
	g_strfreev(envp);
	if (stdout != NULL) g_free(stdout);
	
	sigprocmask(SIG_SETMASK, &oldmask, NULL);
	
	return result;
}

static gboolean streamconfig_service_stop(struct _streamconfig *sc)
{
	gboolean result;
	
	result = FALSE;
	
	if ((sc->netintlist) && (sc->netintctl)) {
		if (streamconfig_remove_multicast_route(sc)) {
			sc->netintctl = FALSE;
		}
	}
	
	if ((streamconfig_service_state(sc) != -1) && (sc->serverctl)) {
		if (sysutils_process_exists(sc->processpid)) {
			sc->serverctl = FALSE;
			if (kill(sc->processpid, SIGTERM) == 0) {
				printf("[+] Stopped streaming server (PID: %ld)\n", sc->processpid);
				sc->processpid = 0;
				result = TRUE;
			}
		} else {
			sc->serverctl = FALSE;
			printf("[-] Failed to stop streaming server - no process (PID: %ld)\n", sc->processpid);
			sc->processpid = 0;
		}
	}
	
	return result;
}

gboolean streamconfig_init(struct _streamconfig *sc)
{
	const gchar *fullusername;
	gchar username[82];
	gsize unamelength;
	const int onflag = 1;
			
	
	sc->serverctl = FALSE;
	sc->netintctl = FALSE;
	sc->netintlist = FALSE;
	
	memset(&sc->mcintname, 0, sizeof(sc->mcintname));
	
	if ((fullusername = getenv("USER")) == NULL) fullusername = g_get_user_name();
	
	if ((unamelength = strlen(fullusername)) >= 82) {
		strncpy(username, fullusername, 81);
		username[81] = '\0';
	} else {
		strncpy(username, fullusername, unamelength);
		username[unamelength] = '\0';
	}
	
	sc->clientsockpath = g_strdup_printf(STREAMCONFIG_CLIENT_SOCKET, username);
	sc->serversockpath = g_strdup_printf(STREAMCONFIG_SERVER_SOCKET, username);
			
	if ((sc->clientsock = socket(AF_LOCAL, SOCK_DGRAM, 0)) == -1) {
		perror("[-] Creating stream server control socket");
		return FALSE;
	}
	
	unlink(sc->clientsockpath);
	
	memset(&sc->clientaddr, 0, sizeof(struct sockaddr_un));
	sc->clientaddr.sun_family = AF_LOCAL;
	strncpy(sc->clientaddr.sun_path, sc->clientsockpath, strlen(sc->clientsockpath));
	
	if (bind(sc->clientsock, (struct sockaddr *)&sc->clientaddr, sizeof(struct sockaddr_un)) == -1) {
		perror("[-] Binding stream server control socket");
		close(sc->clientsock);
		return FALSE;
	}
	
	setsockopt(sc->clientsock, SOL_SOCKET, SO_PASSCRED, &onflag, sizeof(onflag));
	
	memset(&sc->serveraddr, 0, sizeof(struct sockaddr_un));
	sc->serveraddr.sun_family = AF_LOCAL;
	strncpy(sc->serveraddr.sun_path, sc->serversockpath, strlen(sc->serversockpath));
	
	streamconfig_file_open(sc);
	
	if (streamconfig_interfaces_list_build() > 0) {
		sc->netintlist = TRUE;
	}
	
	sc->serverctl = streamconfig_service_restart(sc);
		
	return sc->serverctl;
}

gboolean streamconfig_close(struct _streamconfig *sc)
{
	gboolean result;
		
	if (sc->clientsock != -1) {
		close(sc->clientsock);
		sc->clientsock = -1;
	}
	
	if (sc->clientsockpath != NULL) {
		unlink(sc->clientsockpath);
		g_free(sc->clientsockpath);
		sc->clientsockpath = NULL;
	}
	
	if (sc->serversockpath != NULL) {
		g_free(sc->serversockpath);
		sc->serversockpath = NULL;
	}
		
	result = streamconfig_file_close(sc);
	
	result = result && streamconfig_service_stop(sc);
	
	sc->serverctl = FALSE;
	sc->netintctl = FALSE;
	sc->netintlist = FALSE;
	
	return result;
}

gboolean streamconfig_file_open(struct _streamconfig *sc)
{
	gint fd;
	gchar *data;
	gsize size;
	GError *error;
	gboolean justcreated;
	struct stat statbuf;
	
	sc->fileopened = TRUE;
	
	sc->filename = g_build_filename(g_get_home_dir(), STREAMCONFIG_CONFFILE_PATH, NULL);
	sc->configfile = g_key_file_new();
		
	if ((g_mkdir_with_parents(g_build_filename(g_get_home_dir(), STREAMCONFIG_CONFDIR_PATH, NULL), STREAMCONFIG_DIR_PERMISSIONS)) == -1) {
		printf("[-] Failed to create stream server config path\n");
		sc->fileopened = FALSE;
	}
	
	justcreated = FALSE;
	
	if (!g_file_test(sc->filename, G_FILE_TEST_EXISTS)) {
		if ((fd = g_creat(sc->filename, S_IRUSR | S_IWUSR)) != -1) {
			justcreated = TRUE;
			close(fd);
		} else {
			printf("[-] Stream server settings file not created: %s\n", sc->filename);
			sc->fileopened = FALSE;
			return FALSE;
		}
	}
	
	if ((sc->fileopened) && (!justcreated)) {
		if (g_stat(sc->filename, &statbuf) == -1) {
			printf("[-] No access to stream server settings file\n");
			sc->fileopened = FALSE;
			return FALSE;
		}
		
		if (statbuf.st_size > 0) {
			error = NULL;
			sc->fileopened = TRUE;
			if (g_file_get_contents(sc->filename, &data, &size, &error)) {
				error = NULL;
				if (!g_key_file_load_from_data(sc->configfile, data, size, G_KEY_FILE_NONE, &error)) {
					printf("[-] Error while opening stream server settings file: %s\n", error->message);
					g_error_free(error); error = NULL;
				}
				g_free(data);
			} else {
				printf("[-] Error while opening stream server settings file: %s\n", error->message);
				g_error_free(error); error = NULL;
			}	
		}
	}
	
	return TRUE;
}

gboolean streamconfig_file_close(struct _streamconfig *sc)
{
	gchar *data;
	gsize size;
	GError *error;
	gboolean result;
	
	result = TRUE;
	
	if (sc->fileopened) {
		data = g_key_file_to_data(sc->configfile, &size, NULL);
		g_key_file_free(sc->configfile);
		
		error = NULL;
		if (!g_file_set_contents(sc->filename, data, -1, &error)) {
			printf("[-] Error while saving stream server settings: %s\n", error->message);
			g_error_free(error);
			result = FALSE;
		}
		
		g_free(sc->filename);
		g_free(data);
		sc->fileopened = FALSE;
	}
	
	return result;
}

gboolean streamconfig_file_flush(struct _streamconfig *sc)
{
	gchar *data;
	gsize size;
	GError *error;
	gboolean result;
	
	error = NULL;
	result = FALSE;
	
	if (sc->fileopened) {
		data = g_key_file_to_data(sc->configfile, &size, NULL);
		if ((result = g_file_set_contents(sc->filename, data, size, &error)) == FALSE) {
			printf("[-] Error while saving stream server settings: %s\n", error->message);
			g_error_free(error);
		}
		g_free(data);
	}
	
	return result;
} 

gboolean streamconfig_file_append_string(struct _streamconfig *sc, gchar *name, gchar *value)
{
	gboolean result;
	
	result = FALSE;
	
	if (sc->fileopened) {
		g_key_file_set_string(sc->configfile, STREAMCONFIG_FILE_SECTION, name, value);
		result = TRUE;
	}
	
	return result;
}

gchar *streamconfig_file_read_string(struct _streamconfig *sc, gchar *name, gchar *def)
{
	GError *error;
	gchar *resstring;
	
	error = NULL;
	resstring = NULL;
	
	if (sc->fileopened) {
		if (g_key_file_has_key(sc->configfile, STREAMCONFIG_FILE_SECTION, name, &error)) {
			if (error != NULL) error = NULL;
			resstring = g_key_file_get_string(sc->configfile, STREAMCONFIG_FILE_SECTION, name, &error);
		}
	}
	
	if (resstring == NULL) resstring = g_strdup(def);
	
	return resstring;
}

gboolean streamconfig_file_append_int(struct _streamconfig *sc, gchar *name, gint value)
{
	gboolean result;
	
	result = FALSE;
	
	if (sc->fileopened) {
		g_key_file_set_integer(sc->configfile, STREAMCONFIG_FILE_SECTION, name, value);
		result = TRUE;
	}
	
	return result;
}

gint streamconfig_file_read_int(struct _streamconfig *sc, gchar *name, gint def)
{
	GError *error;
	gint result;
	
	error = NULL;
	result = def;
	
	if (sc->fileopened) {
		if (g_key_file_has_key(sc->configfile, STREAMCONFIG_FILE_SECTION, name, &error)) {
			if (error != NULL) error = NULL;
			result = g_key_file_get_integer(sc->configfile, STREAMCONFIG_FILE_SECTION, name, &error);
		}
	}
	
	return result;
}

gboolean streamconfig_set_record_status(struct _streamconfig *sc, gboolean status)
{
	const char set_record_status_req[] = SET_RECORD_STATUS_REQ;
	char reqid = STREAMCONFIG_SET_RECORD_STATUS;
	char reqstatus = (char)status;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[SHORT_ANSWER_BUFFER];
	size_t anssize = SHORT_ANSWER_BUFFER;
	
	if ((reqbuf = serialize_form_message((char *)&set_record_status_req, 0, &reqsize, reqid, reqstatus)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			streamconfig_file_append_int(sc, "record_status", (gint)status);
			streamconfig_file_flush(sc);
			return TRUE;
		}
	}
	
	return FALSE;
}

gboolean streamconfig_set_udp_broadcast_status(struct _streamconfig *sc, gboolean status)
{
	const char set_udpbc_status_req[] = SET_UDPBC_STATUS_REQ;
	char reqid = STREAMCONFIG_SET_UDP_BC_STATUS;
	char reqstatus = (char)status;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[SHORT_ANSWER_BUFFER];
	size_t anssize = SHORT_ANSWER_BUFFER;
	
	if ((reqbuf = serialize_form_message((char *)&set_udpbc_status_req, 0, &reqsize, reqid, reqstatus)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			streamconfig_file_append_int(sc, "udp_broadcast_status", (gint)status);
			streamconfig_file_flush(sc);
			return TRUE;
		}
	}
	
	return FALSE;
}

gboolean streamconfig_set_udp_multicast_status(struct _streamconfig *sc, gboolean status)
{
	const char set_udpmc_status_req[] = SET_UDPMC_STATUS_REQ;
	char reqid = STREAMCONFIG_SET_UDP_MC_STATUS;
	char reqstatus = (char)status;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[SHORT_ANSWER_BUFFER];
	size_t anssize = SHORT_ANSWER_BUFFER;
	
	if ((reqbuf = serialize_form_message((char *)&set_udpmc_status_req, 0, &reqsize, reqid, reqstatus)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			streamconfig_file_append_int(sc, "udp_multicast_status", (gint)status);
			streamconfig_file_flush(sc);
			return TRUE;
		}
	}
	
	return FALSE;
}

gboolean streamconfig_set_tcp_unicast_status(struct _streamconfig *sc, gboolean status)
{
	const char set_tcpuc_status_req[] = SET_TCPUC_STATUS_REQ;
	char reqid = STREAMCONFIG_SET_TCP_UC_STATUS;
	char reqstatus = (char)status;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[SHORT_ANSWER_BUFFER];
	size_t anssize = SHORT_ANSWER_BUFFER;
	
	if ((reqbuf = serialize_form_message((char *)&set_tcpuc_status_req, 0, &reqsize, reqid, reqstatus)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			streamconfig_file_append_int(sc, "tcp_unicast_status", (gint)status);
			streamconfig_file_flush(sc);
			return TRUE;
		}
	}
	
	return FALSE;
}

gboolean streamconfig_set_connection_details(struct _streamconfig *sc, struct _streamconnparams *sendparams)
{
	const char set_connections_details_req[] = SET_CONNECTION_DETAILS_REQ;
	char reqid = STREAMCONFIG_SET_CONNECTION_DETAILS;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[SHORT_ANSWER_BUFFER];
	size_t anssize = SHORT_ANSWER_BUFFER;
		
	if ((sc == NULL) || (sendparams == NULL)) return FALSE;
	
	if ((reqbuf = serialize_form_message((char *)&set_connections_details_req, 0, &reqsize, reqid, sendparams->udpbcint, sendparams->udpbcport, 
					sendparams->udpmcint, sendparams->udpmcgroup, sendparams->udpmcport, sendparams->udpmcloop, sendparams->udpmcttl,
					sendparams->tcpucport, sendparams->tcpucqueue, sendparams->tcpucclients)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			if (sendparams->udpbcint[0] != '\0') {
				streamconfig_file_append_string(sc, "udp_broadcast_interface", (gchar *)sendparams->udpbcint);
			}
			
			if (sendparams->udpmcint[0] != '\0') {
				streamconfig_file_append_string(sc, "udp_multicast_interface", (gchar *)sendparams->udpmcint);
				if (sc->netintlist) {
					if ((sc->netintctl) && (strncmp(sc->mcintname, (gchar *)&sendparams->udpmcint, IFNAMSIZ) != 0)) {
						if (streamconfig_remove_multicast_route(sc)) {
							sc->netintctl = streamconfig_add_multicast_route(sc, (gchar *)&sendparams->udpmcint);
						}
					} else if (strncmp(sc->mcintname, (gchar *)&sendparams->udpmcint, IFNAMSIZ) != 0) {
						sc->netintctl = streamconfig_add_multicast_route(sc, (gchar *)&sendparams->udpmcint);
					}
				}
			}
			
			if (sendparams->udpmcgroup[0] != '\0') {
				streamconfig_file_append_string(sc, "udp_multicast_group", (gchar *)sendparams->udpmcgroup);
			}
			
			streamconfig_file_append_int(sc, "udp_broadcast_port", (gint)sendparams->udpbcport);
			streamconfig_file_append_int(sc, "udp_multicast_port", (gint)sendparams->udpmcport);
			streamconfig_file_append_int(sc, "udp_multicast_loopback", (gint)sendparams->udpmcloop);
			streamconfig_file_append_int(sc, "udp_multicast_ttl", (gint)sendparams->udpmcttl);
			streamconfig_file_append_int(sc, "tcp_unicast_port", (gint)sendparams->tcpucport);
			streamconfig_file_append_int(sc, "tcp_unicast_queue", (gint)sendparams->tcpucqueue);
			streamconfig_file_append_int(sc, "tcp_unicast_clients", (gint)sendparams->tcpucclients);
			streamconfig_file_flush(sc);
			
			return TRUE;
		}
	}
	
	return FALSE;
}

void streamconfig_stream_content_foreach(unsigned int *params, unsigned int count, int seq, int size, char **buffer, void *data)
{
	GSList *mediastreams;
	struct _mediastream *mediastream;
	
	mediastreams = (GSList *)data;
	mediastream = (struct _mediastream *)g_slist_nth_data(mediastreams, seq);
	
	serialize_form_message_buffer(params, count, buffer, MSGOFFSET(seq, size), mediastream->type, mediastream->value, mediastream->enabled);
}

gboolean streamconfig_set_stream_data(struct _streamconfig *sc, gchar *svcname, GSList *mediastreams, guint16 caid, gchar *cakey)
{
	gint i;
	struct _mediastream *mediastream;
	gboolean result = FALSE;
	const char set_stream_data_req[] = SET_STREAM_DATA_REQ;
	char reqid = STREAMCONFIG_SET_STREAM_DATA;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[SHORT_ANSWER_BUFFER];
	size_t anssize = SHORT_ANSWER_BUFFER;
	const char set_stream_content_array_req[] = SET_STREAM_CONTENT_ARRAY_REQ;
	guint numstreams;
	
	
	if ((sc == NULL) || (svcname == NULL)) return FALSE;
			
	if ((reqbuf = serialize_form_message((char *)&set_stream_data_req, 0, &reqsize, reqid, svcname, cakey, caid)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			streamconfig_file_append_string(sc, "service_name", svcname);
			streamconfig_file_append_int(sc, "service_caid", caid); 
			streamconfig_file_append_string(sc, "service_cakey", cakey);
			streamconfig_file_flush(sc);
			result |= TRUE;
		}
	}
	
	free(reqbuf);
	
	if (mediastreams == NULL) return result;
	
	numstreams = g_slist_length(mediastreams);
	if (numstreams > 256) {
		numstreams = 256;
	} else if (numstreams == 0) {
		return result;
	}
	
	reqid = STREAMCONFIG_SET_STREAM_CONTENT;
	
	if ((reqbuf = serialize_form_array_message((char *)&set_stream_content_array_req, reqid, numstreams, &reqsize, streamconfig_stream_content_foreach, mediastreams)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			result |= TRUE;
		}
	}
	
	free(reqbuf);
	
	return result;
}

gboolean streamconfig_set_device(struct _streamconfig *sc, gchar adapter, gchar dvr)
{
	const char set_device_req[] = SET_DEVICE_REQ;
	char reqid = STREAMCONFIG_SET_DEVICE;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[SHORT_ANSWER_BUFFER];
	size_t anssize = SHORT_ANSWER_BUFFER;
	
	if (sc == NULL) return FALSE;
	
	if ((reqbuf = serialize_form_message((char *)&set_device_req, 0, &reqsize, reqid, adapter, dvr)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			streamconfig_file_append_int(sc, "adapter", (gint)adapter);
			streamconfig_file_append_int(sc, "dvr", (gint)dvr);
			streamconfig_file_flush(sc);
			return TRUE;
		}
	}
	
	return FALSE;
}

void streamconfig_get_clients_callback(unsigned int *params, unsigned int count, int seq, int size, char **message, void *data)
{
	GSList *clients;
	struct _strclient *strclient;
	
	clients = *((GSList **)data);
	strclient = (struct _strclient *)g_malloc0(sizeof(struct _strclient));
		
	serialize_parse_message_buffer(params, count, message, MSGOFFSET(seq, size), &strclient->family, &strclient->type,
									&strclient->pid, &strclient->seq, &strclient->pidfilter, &strclient->address, &strclient->useragent,
									&strclient->starttime, &strclient->transmittedbytes);
	//printf("--> pid: %i, address: %s, ua: %s\n", strclient->pid, strclient->address, strclient->useragent);
	strclient->sticky = FALSE;
	*((GSList **)data) = g_slist_prepend(clients, strclient);
}

GSList *streamconfig_get_clients(struct _streamconfig *sc)
{
	const char get_simple_req[] = GET_SIMPLE_REQ;
	const char get_clients_array_req[] = GET_CLIENTS_ARRAY_REQ;
	char reqid;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[4096];
	size_t anssize;
	char result;
	GSList *clients = NULL;
	
	if (sc == NULL) return NULL;
	
	reqid = STREAMCONFIG_GET_CLIENTS;
	anssize = 4096;
	
	if ((reqbuf = serialize_form_message((char *)&get_simple_req, 0, &reqsize, reqid)) != NULL) {
		if (streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) {
			if (serialize_parse_array_message((char *)&get_clients_array_req, NULL, (char *)&ansbuf, NULL, streamconfig_get_clients_callback, &clients)) {
				clients = g_slist_reverse(clients);
				return clients;
			}
		}
	}
	
	return NULL;
}

void streamconfig_free_clients(GSList *clients)
{
	if (clients == NULL) return;
	
	g_slist_free(clients);
}

void streamconfig_get_tsstruct_callback(unsigned int *params, unsigned int count, int seq, int size, char **message, void *data)
{
	GSList *tsstruct;
	struct _strtspid *strtspid;
	unsigned int param1, param2, param3, param4;
	
	tsstruct = *((GSList **)data);
	strtspid = (struct _strtspid *)g_malloc0(sizeof(struct _strtspid));
	
	serialize_parse_message_buffer(params, count, message, MSGOFFSET(seq, size), &strtspid->type, &strtspid->format, &strtspid->value,
									&strtspid->percentage, &strtspid->pid, &strtspid->bandwidth, &strtspid->revision,
									&param1, &param2, &param3, &param4, &strtspid->description);
	
	if (strtspid->type == TS_STREAM_PES_AUDIO_STREAM) {
		strtspid->params.audio.version = (param1 & 0xffff0000) >> 16;
		strtspid->params.audio.revision = param1 & 0x0000ffff;
		strtspid->params.audio.profile = (param2 & 0xffff0000) >> 16;
		strtspid->params.audio.mode = param2 & 0x0000ffff;
		strtspid->params.audio.bitrate = param3;
		strtspid->params.audio.samplerate = param4;
	} else if (strtspid->type == TS_STREAM_PES_VIDEO_STREAM) {
		strtspid->params.video.picwidth = param1;
		strtspid->params.video.picheight = param2;
		strtspid->params.video.aspect = param3;
		strtspid->params.video.framerate = param4;
	}
	
	//printf("-->type: %i, pid: %i, params %i, %i, %i, %i\n", type, pid, param1, param2, param3, param4);
	strtspid->sticky = FALSE;
	*((GSList **)data) = g_slist_prepend(tsstruct, strtspid);
}

GSList *streamconfig_get_tsstruct(struct _streamconfig *sc)
{
	const char get_simple_req[] = GET_SIMPLE_REQ;
	const char get_clients_array_req[] = GET_STREAM_STRUCTURE_ARRAY_REQ;
	char reqid;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[4096];
	size_t anssize;
	char result;
	GSList *tsstruct = NULL;
	
	if (sc == NULL) return NULL;
	
	reqid = STREAMCONFIG_GET_STREAM_STRUCTURE;
	anssize = 4096;
	
	if ((reqbuf = serialize_form_message((char *)&get_simple_req, 0, &reqsize, reqid)) != NULL) {
		if (streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) {
			if (serialize_parse_array_message((char *)&get_clients_array_req, NULL, (char *)&ansbuf, NULL, streamconfig_get_tsstruct_callback, &tsstruct)) {
				tsstruct = g_slist_reverse(tsstruct);
				return tsstruct;
			}
		}
	}
	
	return NULL;
}

void streamconfig_free_tsstruct(GSList *tsstruct)
{
	if (tsstruct == NULL) return;
	
	g_slist_free(tsstruct);
}

gboolean streamconfig_get_statistics(struct _streamconfig *sc, guint *udpmccli, guint *udpbccli, guint *tcpuccli, guint *tcpucclimax, glong *sessiontime)
{
	const char get_simple_req[] = GET_SIMPLE_REQ;
	const char get_server_statistics_req[] = GET_SERVER_STATISTICS_REQ;
	char reqid;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[ANSWER_BUFFER];
	size_t anssize;
	
	char result;
	char udpmcclientsnum;
	char udpbcclientsnum;
	unsigned int tcpucclientsnum;
	unsigned int tcpucclientslimit;
	unsigned long long lastsessiontime;
	
	
	if (sc == NULL) return FALSE;
		
	reqid = STREAMCONFIG_GET_SERVER_STATISTICS;
	anssize = ANSWER_BUFFER;
	
	if ((reqbuf = serialize_form_message((char *)&get_simple_req, 0, &reqsize, reqid)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			if (serialize_parse_message((char *)&get_server_statistics_req, (char *)&ansbuf, 0, &reqid, &result, &udpmcclientsnum,
										&udpbcclientsnum, &tcpucclientsnum, &tcpucclientslimit, &lastsessiontime) > 0) {
				
				if (udpmccli != NULL) *udpmccli = (guint)udpmcclientsnum;
				if (udpbccli != NULL) *udpbccli = (guint)udpbcclientsnum;
				if (tcpuccli != NULL) *tcpuccli = (guint)tcpucclientsnum;
				if (tcpucclimax != NULL) *tcpucclimax = (guint)tcpucclientslimit;
				if (sessiontime != NULL) *sessiontime = (glong)lastsessiontime;
				
				return TRUE;
			}
		}
	}
	
	if (udpmccli != NULL) *udpmccli = 0;
	if (udpbccli != NULL) *udpbccli = 0;
	if (tcpuccli != NULL) *tcpuccli = 0;
	if (tcpucclimax != NULL) *tcpucclimax = (guint)streamconfig_file_read_int(sc, "tcp_unicast_clients", 20);
	if (sessiontime != NULL) *sessiontime = (glong)time(NULL);
	
	return FALSE;
}

gboolean streamconfig_get_service_flags(struct _streamconfig *sc, gboolean *record, gboolean *udpbc, gboolean *udpmc, gboolean *tcpuc)
{
	const char get_simple_req[] = GET_SIMPLE_REQ;
	const char get_service_flags_req[] = GET_SERVICE_FLAGS_REQ;
	char reqid;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[SHORT_ANSWER_BUFFER];
	size_t anssize;
	
	char result;
	char recordstatus;
	char udpbcstatus;
	char udpmcstatus;
	char tcpucstatus;
	
	
	if (sc == NULL) return FALSE;
		
	reqid = STREAMCONFIG_GET_SERVICE_FLAGS;
	anssize = SHORT_ANSWER_BUFFER;
	
	if ((reqbuf = serialize_form_message((char *)&get_simple_req, 0, &reqsize, reqid)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			if (serialize_parse_message((char *)&get_service_flags_req, (char *)&ansbuf, 0, &reqid, &result, &recordstatus, &udpbcstatus, &udpmcstatus, &tcpucstatus) > 0) {
				if (record != NULL) *record = (gboolean)recordstatus;
				if (udpbc != NULL) *udpbc = (gboolean)udpbcstatus;
				if (udpmc != NULL) *udpmc = (gboolean)udpmcstatus;
				if (tcpuc != NULL) *tcpuc = (gboolean)tcpucstatus;
				return TRUE;
			}
		}
	}
	
	if (record != NULL) *record = (gboolean)streamconfig_file_read_int(sc, "record_status", 0);
	if (udpbc != NULL) *udpbc = (gboolean)streamconfig_file_read_int(sc, "udp_broadcast_status", 0);
	if (udpmc != NULL) *udpmc = (gboolean)streamconfig_file_read_int(sc, "udp_multicast_status", 0);
	if (tcpuc != NULL) *tcpuc = (gboolean)streamconfig_file_read_int(sc, "tcp_unicast_status", 0);
	return FALSE;
}

gboolean streamconfig_get_connection_details(struct _streamconfig *sc, struct _streamconnparams *recvparams)
{
	const char get_simple_req[] = GET_SIMPLE_REQ;
	const char get_connection_details_req[] = GET_CONNECTION_DETAILS_REQ;
	char reqid = STREAMCONFIG_GET_CONNECTION_DETAILS;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[ANSWER_BUFFER];
	size_t anssize = ANSWER_BUFFER;
	char result;
	gchar *str;
		
	if ((sc == NULL) || (recvparams == NULL)) return FALSE;
	
	memset(recvparams, 0, sizeof(struct _streamconnparams));
		
	if ((reqbuf = serialize_form_message((char *)&get_simple_req, 0, &reqsize, reqid)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			if (serialize_parse_message_static((char *)&get_connection_details_req, (char *)&ansbuf, 0, &reqid, &result, &(recvparams->udpbcint), &(recvparams->udpbcport), 
					&(recvparams->udpmcint), &(recvparams->udpmcgroup), &(recvparams->udpmcport), &(recvparams->udpmcloop), &(recvparams->udpmcttl),
					&(recvparams->tcpucport), &(recvparams->tcpucqueue), &(recvparams->tcpucclients)) > 0) {
				return TRUE;
			}
		}
	}
	
	str = streamconfig_file_read_string(sc, "udp_broadcast_interface", "lo");
	strncpy(recvparams->udpbcint, str, IFNAMSIZ);
	g_free(str);
	
	str = streamconfig_file_read_string(sc, "udp_multicast_interface", "lo");
	strncpy(recvparams->udpmcint, str, IFNAMSIZ);
	g_free(str);
	
	str = streamconfig_file_read_string(sc, "udp_multicast_group", "225.0.0.0");
	strncpy(recvparams->udpmcgroup, str, INET_ADDRSTRLEN);
	g_free(str);
	
	recvparams->udpbcport = streamconfig_file_read_int(sc, "udp_broadcast_port", 4000);
	recvparams->udpmcport = streamconfig_file_read_int(sc, "udp_multicast_port", 4000);
	recvparams->udpmcloop = streamconfig_file_read_int(sc, "udp_multicast_loopback", 1);
	recvparams->udpmcttl = streamconfig_file_read_int(sc, "udp_multicast_ttl", 1);
	recvparams->tcpucport = streamconfig_file_read_int(sc, "tcp_unicast_port", 4000);
	recvparams->tcpucqueue = streamconfig_file_read_int(sc, "tcp_unicast_queue", 10);
	recvparams->tcpucclients = streamconfig_file_read_int(sc, "tcp_unicast_clients", 20);
	
	return TRUE;
}


gboolean streamconfig_stop_client(struct _streamconfig *sc, guint clientid)
{
	const char stop_client_req[] = STOP_CLIENT_REQ;
	char reqid = STREAMCONFIG_STOP_CLIENT;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[SHORT_ANSWER_BUFFER];
	size_t anssize = SHORT_ANSWER_BUFFER;
	
	if ((reqbuf = serialize_form_message((char *)&stop_client_req, 0, &reqsize, reqid, clientid)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			return TRUE;
		}
	}
	
	return FALSE;
}


static gboolean streamconfig_request(struct _streamconfig *sc, char *reqbuf, size_t *reqsize, char *ansbuf, size_t *anssize)
{
	gint ready;
	struct pollfd sockets[1];
	struct sockaddr_un server;
	static struct msghdr msg;
	struct iovec iov[1];
	union {
        struct cmsghdr cmh;
        char   control[CMSG_SPACE(sizeof(struct ucred))];
    } control_un;
    struct cmsghdr *msgch;
	struct ucred *cred;
	gsize bytes;
	guchar request;
	
	if (!sc->serverctl) return FALSE;
	
	if (streamconfig_service_state(sc) == -1) {
		//printf("[+] Streaming server not started\n");
		raise(SIGCHLD);
		return FALSE;
	}
	
	request = SERIALIZE_REQUEST(reqbuf);
	
	msg.msg_name = &sc->serveraddr;
	msg.msg_namelen = sizeof(struct sockaddr_un);
	msg.msg_iov = iov;
	msg.msg_iovlen = 1;
	iov[0].iov_base = reqbuf;
	iov[0].iov_len = *reqsize;
	
	msg.msg_control = control_un.control;
	msg.msg_controllen = sizeof(control_un.control);
	
	msgch = CMSG_FIRSTHDR(&msg);
	msgch->cmsg_len = CMSG_LEN(sizeof(struct ucred));
	msgch->cmsg_level = SOL_SOCKET;
	msgch->cmsg_type = SCM_CREDENTIALS;
	cred = (struct ucred *)CMSG_DATA(msgch);
	
	cred->pid = getpid();
	cred->uid = getuid();
	cred->gid = getgid();
	
	if ((bytes = sendmsg(sc->clientsock, &msg, 0)) == -1) {
		perror("[-] STREAMCONFIG sendmsg");
		return FALSE;
	}
	
	sockets[0].fd = sc->clientsock;
	sockets[0].events = POLLIN;
	
	while (1) {
		if ((ready = poll(sockets, 1, 500)) > 0) {
			if (sockets[0].revents & POLLIN) {
				msg.msg_name = &server;
				msg.msg_namelen = sizeof(struct sockaddr_un);
				msg.msg_iov = iov;
				msg.msg_iovlen = 1;
				iov[0].iov_base = ansbuf;
				iov[0].iov_len = *anssize;
				
				msg.msg_control = control_un.control;
				msg.msg_controllen = sizeof(control_un.control);
				
				control_un.cmh.cmsg_len = CMSG_LEN(sizeof(struct ucred));
				control_un.cmh.cmsg_level = SOL_SOCKET;
				control_un.cmh.cmsg_type = SCM_CREDENTIALS;
				
				memset(ansbuf, 0, *anssize);
				
				if ((bytes = recvmsg(sc->clientsock, &msg, 0)) <= 0) {
					perror("[-] STREAMCONFIG recvmsg:");
					return FALSE;
				}
				
				msgch = CMSG_FIRSTHDR(&msg);
				if ((msgch != NULL) && (msgch->cmsg_type == SCM_CREDENTIALS)) {
					cred = (struct ucred *)CMSG_DATA(msgch);
					if ((sc->processpid == cred->pid) && (getuid() == cred->uid) && (getgid() == cred->gid)) {
						if (SERIALIZE_REQUEST(ansbuf) == request) {
							return TRUE;
						}
					}
				}
			}
		} else break;
	}
	
	return FALSE;
}
