/*
 *      teletext.c
 *      
 *      Copyright 2012 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */    

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdint.h>
#include <errno.h>
#include <poll.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <linux/dvb/dmx.h>
#include <glib.h>

#include "teletext.h"

#define TT_BIT_SET(a,b) ((a) |= (1<<(b)))
#define TT_BIT_CLEAR(a,b) ((a) &= ~(1<<(b)))
#define TT_BIT_FLIP(a,b) ((a) ^= (1<<(b)))
#define TT_BIT_CHECK(a,b) ((a) & (1<<(b)))
#define TT_BITMAP_FULL(a) (((a) & 0x7FFFFF) == 0x7FFFFF)

enum {
	TT_CACHE_CURRENT = 0,
	TT_CACHE_LONGTERM = 1
};

enum {
	TT_CHARSET_LATIN = 0,
	TT_CHARSET_CYRILLIC1,
	TT_CHARSET_CYRILLIC2,
	TT_CHARSET_CYRILLIC3,
	TT_CHARSET_GREEK,
	TT_CHARSET_ARABIC,
	TT_CHARSET_HEBREW
};

enum {
	TT_COMMAND_NULL = 0,
	TT_COMMAND_STOP
};

enum {
	TT_ANSWER_SUCCESS = 0,
	TT_ANSWER_FAILURE
};


static const unsigned char reverse[256] = {
	0x00, 0x80, 0x40, 0xC0, 0x20, 0xA0, 0x60, 0xE0, 0x10, 0x90, 0x50, 0xD0, 0x30, 0xB0, 0x70, 0xF0,
	0x08, 0x88, 0x48, 0xC8, 0x28, 0xA8, 0x68, 0xE8, 0x18, 0x98, 0x58, 0xD8, 0x38, 0xB8, 0x78, 0xF8,
	0x04, 0x84, 0x44, 0xC4, 0x24, 0xA4, 0x64, 0xE4, 0x14, 0x94, 0x54, 0xD4, 0x34, 0xB4, 0x74, 0xF4,
	0x0C, 0x8C, 0x4C, 0xCC, 0x2C, 0xAC, 0x6C, 0xEC, 0x1C, 0x9C, 0x5C, 0xDC, 0x3C, 0xBC, 0x7C, 0xFC,
	0x02, 0x82, 0x42, 0xC2, 0x22, 0xA2, 0x62, 0xE2, 0x12, 0x92, 0x52, 0xD2, 0x32, 0xB2, 0x72, 0xF2,
	0x0A, 0x8A, 0x4A, 0xCA, 0x2A, 0xAA, 0x6A, 0xEA, 0x1A, 0x9A, 0x5A, 0xDA, 0x3A, 0xBA, 0x7A, 0xFA,
	0x06, 0x86, 0x46, 0xC6, 0x26, 0xA6, 0x66, 0xE6, 0x16, 0x96, 0x56, 0xD6, 0x36, 0xB6, 0x76, 0xF6,
	0x0E, 0x8E, 0x4E, 0xCE, 0x2E, 0xAE, 0x6E, 0xEE, 0x1E, 0x9E, 0x5E, 0xDE, 0x3E, 0xBE, 0x7E, 0xFE,
	0x01, 0x81, 0x41, 0xC1, 0x21, 0xA1, 0x61, 0xE1, 0x11, 0x91, 0x51, 0xD1, 0x31, 0xB1, 0x71, 0xF1,
	0x09, 0x89, 0x49, 0xC9, 0x29, 0xA9, 0x69, 0xE9, 0x19, 0x99, 0x59, 0xD9, 0x39, 0xB9, 0x79, 0xF9,
	0x05, 0x85, 0x45, 0xC5, 0x25, 0xA5, 0x65, 0xE5, 0x15, 0x95, 0x55, 0xD5, 0x35, 0xB5, 0x75, 0xF5,
	0x0D, 0x8D, 0x4D, 0xCD, 0x2D, 0xAD, 0x6D, 0xED, 0x1D, 0x9D, 0x5D, 0xDD, 0x3D, 0xBD, 0x7D, 0xFD,
	0x03, 0x83, 0x43, 0xC3, 0x23, 0xA3, 0x63, 0xE3, 0x13, 0x93, 0x53, 0xD3, 0x33, 0xB3, 0x73, 0xF3,
	0x0B, 0x8B, 0x4B, 0xCB, 0x2B, 0xAB, 0x6B, 0xEB, 0x1B, 0x9B, 0x5B, 0xDB, 0x3B, 0xBB, 0x7B, 0xFB,
	0x07, 0x87, 0x47, 0xC7, 0x27, 0xA7, 0x67, 0xE7, 0x17, 0x97, 0x57, 0xD7, 0x37, 0xB7, 0x77, 0xF7,
	0x0F, 0x8F, 0x4F, 0xCF, 0x2F, 0xAF, 0x6F, 0xEF, 0x1F, 0x9F, 0x5F, 0xDF, 0x3F, 0xBF, 0x7F, 0xFF
};

static const unsigned char hamming[256] = {
	0x01, 0xff, 0x01, 0x01, 0xff, 0x00, 0x01, 0xff, 0xff, 0x02, 0x01, 0xff, 0x0a, 0xff, 0xff, 0x07,
	0xff, 0x00, 0x01, 0xff, 0x00, 0x00, 0xff, 0x00, 0x06, 0xff, 0xff, 0x0b, 0xff, 0x00, 0x03, 0xff,
	0xff, 0x0c, 0x01, 0xff, 0x04, 0xff, 0xff, 0x07, 0x06, 0xff, 0xff, 0x07, 0xff, 0x07, 0x07, 0x07,
	0x06, 0xff, 0xff, 0x05, 0xff, 0x00, 0x0d, 0xff, 0x06, 0x06, 0x06, 0xff, 0x06, 0xff, 0xff, 0x07,
	0xff, 0x02, 0x01, 0xff, 0x04, 0xff, 0xff, 0x09, 0x02, 0x02, 0xff, 0x02, 0xff, 0x02, 0x03, 0xff,
	0x08, 0xff, 0xff, 0x05, 0xff, 0x00, 0x03, 0xff, 0xff, 0x02, 0x03, 0xff, 0x03, 0xff, 0x03, 0x03,
	0x04, 0xff, 0xff, 0x05, 0x04, 0x04, 0x04, 0xff, 0xff, 0x02, 0x0f, 0xff, 0x04, 0xff, 0xff, 0x07,
	0xff, 0x05, 0x05, 0x05, 0x04, 0xff, 0xff, 0x05, 0x06, 0xff, 0xff, 0x05, 0xff, 0x0e, 0x03, 0xff,
	0xff, 0x0c, 0x01, 0xff, 0x0a, 0xff, 0xff, 0x09, 0x0a, 0xff, 0xff, 0x0b, 0x0a, 0x0a, 0x0a, 0xff,
	0x08, 0xff, 0xff, 0x0b, 0xff, 0x00, 0x0d, 0xff, 0xff, 0x0b, 0x0b, 0x0b, 0x0a, 0xff, 0xff, 0x0b,
	0x0c, 0x0c, 0xff, 0x0c, 0xff, 0x0c, 0x0d, 0xff, 0xff, 0x0c, 0x0f, 0xff, 0x0a, 0xff, 0xff, 0x07,
	0xff, 0x0c, 0x0d, 0xff, 0x0d, 0xff, 0x0d, 0x0d, 0x06, 0xff, 0xff, 0x0b, 0xff, 0x0e, 0x0d, 0xff,
	0x08, 0xff, 0xff, 0x09, 0xff, 0x09, 0x09, 0x09, 0xff, 0x02, 0x0f, 0xff, 0x0a, 0xff, 0xff, 0x09,
	0x08, 0x08, 0x08, 0xff, 0x08, 0xff, 0xff, 0x09, 0x08, 0xff, 0xff, 0x0b, 0xff, 0x0e, 0x03, 0xff,
	0xff, 0x0c, 0x0f, 0xff, 0x04, 0xff, 0xff, 0x09, 0x0f, 0xff, 0x0f, 0x0f, 0xff, 0x0e, 0x0f, 0xff,
	0x08, 0xff, 0xff, 0x05, 0xff, 0x0e, 0x0d, 0xff, 0xff, 0x0e, 0x0f, 0xff, 0x0e, 0x0e, 0xff, 0x0e
};

static const unsigned int tt_charset_latin_g0[96] = {
	0x000020, 0x000021, 0x000022, 0x000023, 0x000024, 0x000025, 0x000026, 0x000027,
	0x000028, 0x000029, 0x00002a, 0x00002b, 0x00002c, 0x00002d, 0x00002e, 0x00002f,
	0x000030, 0x000031, 0x000032, 0x000033, 0x000034, 0x000035, 0x000036, 0x000037,
	0x000038, 0x000039, 0x00003a, 0x00003b, 0x00003c, 0x00003d, 0x00003e, 0x00003f,
	0x000040, 0x000041, 0x000042, 0x000043, 0x000044, 0x000045, 0x000046, 0x000047,
	0x000048, 0x000049, 0x00004a, 0x00004b, 0x00004c, 0x00004d, 0x00004e, 0x00004f,
	0x000050, 0x000051, 0x000052, 0x000053, 0x000054, 0x000055, 0x000056, 0x000057,
	0x000058, 0x000059, 0x00005a, 0x00005b, 0x00005c, 0x00005d, 0x00005e, 0x00005f,
	0x000060, 0x000061, 0x000062, 0x000063, 0x000064, 0x000065, 0x000066, 0x000067,
	0x000068, 0x000069, 0x00006a, 0x00006b, 0x00006c, 0x00006d, 0x00006e, 0x00006f,
	0x000070, 0x000071, 0x000072, 0x000073, 0x000074, 0x000075, 0x000076, 0x000077,
	0x000078, 0x000079, 0x00007a, 0x00007b, 0x00007c, 0x00007d, 0x00007e, 0x00007f
};

static const unsigned int tt_charset_latin_g2[96] = {
	0x000020, 0x00c2a1, 0x00c2a2, 0x00c2a3, 0x000024, 0x00c2a5, 0x000023, 0x00c2a7,
	0x00c2a4, 0x00cabb, 0x00caba, 0x00c2ab, 0xe28690, 0xe28691, 0xe28692, 0xe28693,
	0x00c2b0, 0x00c2b1, 0x00c2b2, 0x00c2b3, 0x00c397, 0x00c2b5, 0x00c2b6, 0x00c2b7,
	0x00c3b7, 0x00c2b4, 0x000022, 0x00c2bb, 0x00c2bc, 0x00c2bd, 0x00c2be, 0x00c2bf,
	0x000020, 0x000060, 0x00c2b4, 0x00005e, 0x00007e, 0x00c2af, 0x00cb98, 0x00cb99,
	0x00c2a8, 0x00002e, 0x00cb9a, 0x00cca7, 0x00005f, 0x00cb9d, 0x00cca8, 0x00cc8c,
	0x00002d, 0x00c2b9, 0x00c2ae, 0x00c2a9, 0xe284a2, 0xe29886, 0xe282a0, 0xe280b0,
	0x00c991, 0x000020, 0x000020, 0x000020, 0xe2859b, 0xe2859c, 0xe2859d, 0xe2859e,
	0x00cea9, 0x00c386, 0x00c390, 0x00c2aa, 0x00c4a6, 0x000020, 0x00c4b2, 0x00c4bf,
	0x00c581, 0x00c398, 0x00c592, 0x00c2ba, 0x00c39e, 0x00c5a6, 0x00c58a, 0x00c589,
	0x00c4b8, 0x00c3a6, 0x00c491, 0x00c3b0, 0x00c4a7, 0x00c4b1, 0x00c4b3, 0x00c580,
	0x00c582, 0x00c3b8, 0x00c593, 0x00c39f, 0x00c3be, 0x00c5a7, 0x00c58b, 0x000020
};

static const unsigned int tt_charset_cyrillic_1_g0[96] = {
	0x000020, 0x000021, 0x000022, 0x000023, 0x000024, 0x000025, 0x000026, 0x000027,
	0x000028, 0x000029, 0x00002a, 0x00002b, 0x00002c, 0x00002d, 0x00002e, 0x00002f,
	0x000030, 0x000031, 0x000032, 0x000033, 0x000034, 0x000035, 0x000036, 0x000037,
	0x000038, 0x000039, 0x00003a, 0x00003b, 0x00003c, 0x00003d, 0x00003e, 0x00003f,
	0x00d0a7, 0x00d090, 0x00d091, 0x00d0a6, 0x00d094, 0x00d095, 0x00d0a4, 0x00d093,
	0x00d0a5, 0x00d098, 0x00d088, 0x00d09a, 0x00d09b, 0x00d09c, 0x00d09d, 0x00d09e,
	0x00d09f, 0x00d08c, 0x00d0a0, 0x00d0a1, 0x00d0a2, 0x00d0a3, 0x00d092, 0x00d083,
	0x00d089, 0x00d08a, 0x00d097, 0x00d08b, 0x00d096, 0x00d082, 0x00d0a8, 0x00d08f,
	0x00d187, 0x00d0b0, 0x00d0b1, 0x00d186, 0x00d0b4, 0x00d0b5, 0x00d184, 0x00d0b3,
	0x00d185, 0x00d0b8, 0x00d198, 0x00d0ba, 0x00d0bb, 0x00d0bc, 0x00d0bd, 0x00d0be,
	0x00d0bf, 0x00d19c, 0x00d180, 0x00d181, 0x00d182, 0x00d183, 0x00d0b2, 0x00d193,
	0x00d199, 0x00d19a, 0x00d0b7, 0x00d19b, 0x00d0b6, 0x00d082, 0x00d188, 0x000020
};

static const unsigned int tt_charset_cyrillic_2_g0[96] = {
	0x000020, 0x000021, 0x000022, 0x000023, 0x000024, 0x000025, 0x00d18b, 0x000027,
	0x000028, 0x000029, 0x00002a, 0x00002b, 0x00002c, 0x00002d, 0x00002e, 0x00002f,
	0x000030, 0x000031, 0x000032, 0x000033, 0x000034, 0x000035, 0x000036, 0x000037,
	0x000038, 0x000039, 0x00003a, 0x00003b, 0x00003c, 0x00003d, 0x00003e, 0x00003f,
	0x00d0ae, 0x00d090, 0x00d091, 0x00d0a6, 0x00d094, 0x00d095, 0x00d0a4, 0x00d093,
	0x00d0a5, 0x00d098, 0x00d099, 0x00d09a, 0x00d09b, 0x00d09c, 0x00d09d, 0x00d09e,
	0x00d09f, 0x00d0af, 0x00d0a0, 0x00d0a1, 0x00d0a2, 0x00d0a3, 0x00d096, 0x00d092,
	0x00d0ac, 0x00d0aa, 0x00d097, 0x00d0a8, 0x00d0ad, 0x00d0a9, 0x00d0a7, 0x00d0ab,
	0x00d18e, 0x00d0b0, 0x00d0b1, 0x00d186, 0x00d0b4, 0x00d0b5, 0x00d184, 0x00d0b3,
	0x00d185, 0x00d0b8, 0x00d0b9, 0x00d0ba, 0x00d0bb, 0x00d0bc, 0x00d0bd, 0x00d0be,
	0x00d0bf, 0x00d18f, 0x00d180, 0x00d181, 0x00d182, 0x00d183, 0x00d0b6, 0x00d0b2,
	0x00d18c, 0x00d18a, 0x00d0b7, 0x00d188, 0x00d18d, 0x00d189, 0x00d187, 0x000020
};

static const unsigned int tt_charset_cyrillic_3_g0[96] = {
	0x000020, 0x000021, 0x000022, 0x000023, 0x000024, 0x000025, 0x00d197, 0x000027,
	0x000028, 0x000029, 0x00002a, 0x00002b, 0x00002c, 0x00002d, 0x00002e, 0x00002f,
	0x000030, 0x000031, 0x000032, 0x000033, 0x000034, 0x000035, 0x000036, 0x000037,
	0x000038, 0x000039, 0x00003a, 0x00003b, 0x00003c, 0x00003d, 0x00003e, 0x00003f,
	0x00d0ae, 0x00d090, 0x00d091, 0x00d0a6, 0x00d094, 0x00d095, 0x00d0a4, 0x00d093,
	0x00d0a5, 0x00d098, 0x00d099, 0x00d09a, 0x00d09b, 0x00d09c, 0x00d09d, 0x00d09e,
	0x00d09f, 0x00d0af, 0x00d0a0, 0x00d0a1, 0x00d0a2, 0x00d0a3, 0x00d096, 0x00d092,
	0x00d0ac, 0x00d086, 0x00d097, 0x00d0a8, 0x00d084, 0x00d0a9, 0x00d0a7, 0x00d087,
	0x00d18e, 0x00d0b0, 0x00d0b1, 0x00d186, 0x00d0b4, 0x00d0b5, 0x00d184, 0x00d0b3,
	0x00d185, 0x00d0b8, 0x00d0b9, 0x00d0ba, 0x00d0bb, 0x00d0bc, 0x00d0bd, 0x00d0be,
	0x00d0bf, 0x00d18f, 0x00d180, 0x00d181, 0x00d182, 0x00d183, 0x00d0b6, 0x00d0b2,
	0x00d18c, 0x00d196, 0x00d0b7, 0x00d188, 0x00d194, 0x00d189, 0x00d187, 0x000020
};

static const unsigned int tt_charset_cyrillic_g2[96] = {
	0x000020, 0x00c2a1, 0x00c2a2, 0x00c2a3, 0x000024, 0x00c2a5, 0x000020, 0x00c2a7,
	0x000020, 0x00cabb, 0x00caba, 0x00c2ab, 0xe28690, 0xe28691, 0xe28692, 0xe28693,
	0x00c2b0, 0x00c2b1, 0x00c2b2, 0x00c2b3, 0x00c397, 0x00c2b5, 0x00c2b6, 0x00c2b7,
	0x00c3b7, 0x00c2b4, 0x000022, 0x00c2bb, 0x00c2bc, 0x00c2bd, 0x00c2be, 0x00c2bf,
	0x000020, 0x000060, 0x00c2b4, 0x00005e, 0x00007e, 0x00c2af, 0x00cb98, 0x00cb99,
	0x00c2a8, 0x00002e, 0x00cb9a, 0x00cca7, 0x00005f, 0x00cb9d, 0x00cca8, 0x00cc8c,
	0x00002d, 0x00c2b9, 0x00c2ae, 0x00c2a9, 0xe284a2, 0xe29886, 0xe282a0, 0xe280b0,
	0x00c991, 0x00C581, 0x00C582, 0x00C39F, 0xe2859b, 0xe2859c, 0xe2859d, 0xe2859e,
	0x000044, 0x000045, 0x000046, 0x000047, 0x000049, 0x00004A, 0x00004B, 0x00004C,
	0x00004E, 0x000051, 0x000052, 0x000053, 0x000055, 0x000056, 0x000057, 0x00005A,
	0x000064, 0x000065, 0x000066, 0x000067, 0x000069, 0x00006A, 0x00006b, 0x00006c,
	0x00006E, 0x000071, 0x000072, 0x000073, 0x000075, 0x000076, 0x000077, 0x00007A
};
     
static const unsigned int tt_charset_greek_g0[96] = {
	0x000020, 0x000021, 0x000022, 0x000023, 0x000024, 0x000025, 0x000026, 0x000027,
	0x000028, 0x000029, 0x00002a, 0x00002b, 0x00002c, 0x00002d, 0x00002e, 0x00002f,
	0x000030, 0x000031, 0x000032, 0x000033, 0x000034, 0x000035, 0x000036, 0x000037,
	0x000038, 0x000039, 0x00003a, 0x00003b, 0x00003c, 0x00003d, 0x00003e, 0x00003f,
	0x00ce90, 0x00ce91, 0x00ce92, 0x00ce93, 0x00ce94, 0x00ce95, 0x00ce96, 0x00ce97,
	0x00ce98, 0x00ce99, 0x00ce9a, 0x00ce9b, 0x00ce9c, 0x00ce9d, 0x00ce9e, 0x00ce9f,
	0x00cea0, 0x00cea1, 0x00ce84, 0x00cea3, 0x00cea4, 0x00cea5, 0x00cea6, 0x00cea7,
	0x00cea8, 0x00cea9, 0x00ceaa, 0x00ceab, 0x00ceac, 0x00cead, 0x00ceae, 0x00ceaf,
	0x00ceb0, 0x00ceb1, 0x00ceb2, 0x00ceb3, 0x00ceb4, 0x00ceb5, 0x00ceb6, 0x00ceb7,
	0x00ceb8, 0x00ceb9, 0x00ceba, 0x00cebb, 0x00cebc, 0x00cebd, 0x00cebe, 0x00cebf,
	0x00cf80, 0x00cf81, 0x00cf82, 0x00cf83, 0x00cf84, 0x00cf85, 0x00cf86, 0x00cf87,
	0x00cf88, 0x00cf89, 0x00cf8a, 0x00cf8b, 0x00cf8c, 0x00cf8d, 0x00cf8e, 0x000020
};

static const unsigned int tt_charset_greek_g2[96] = {
	0x000020, 0x000061, 0x000062, 0x00c2a3, 0x000065, 0x000068, 0x000069, 0x00c2a7,
	0x00003a, 0x00cabb, 0x00caba, 0x00006b, 0xe28690, 0xe28691, 0xe28692, 0xe28693,
	0x00c2b0, 0x00c2b1, 0x00c2b2, 0x00c2b3, 0x00c397, 0x00006d, 0x00006e, 0x000070,
	0x00c3b7, 0x00c2b4, 0x000022, 0x000074, 0x00c2bc, 0x00c2bd, 0x00c2be, 0x000078,
	0x000020, 0x000060, 0x00c2b4, 0x00005e, 0x00007e, 0x00c2af, 0x00cb98, 0x00cb99,
	0x00c2a8, 0x00002e, 0x00cb9a, 0x00cca7, 0x00005f, 0x00cb9d, 0x00cca8, 0x00cc8c,
	0x00003f, 0x00c2b9, 0x00c2ae, 0x00c2a9, 0xe284a2, 0xe29886, 0xe282a0, 0xe280b0,
	0x00c991, 0x00ce8a, 0x00ce8e, 0x00ce8f, 0xe2859b, 0xe2859c, 0xe2859d, 0xe2859e,
	0x000043, 0x000044, 0x000046, 0x000047, 0x00004a, 0x00004c, 0x000051, 0x000052,
	0x000053, 0x000055, 0x000056, 0x000057, 0x000059, 0x00005a, 0x00ce86, 0x00ce89,
	0x000063, 0x000064, 0x000066, 0x000067, 0x00006a, 0x00006c, 0x000071, 0x000072,
	0x000073, 0x000075, 0x000076, 0x000077, 0x000079, 0x00007a, 0x00ce88, 0x000020
};

static const unsigned int tt_charset_arabic_g0[96] = {
	0x000020, 0x000021, 0x000022, 0x00c2a3, 0x000024, 0x000025, 0x00d888, 0x00d98a,
	0x000028, 0x000029, 0x00002a, 0x00002b, 0x00002c, 0x00002d, 0x00002e, 0x00002f,
	0x000030, 0x000031, 0x000032, 0x000033, 0x000034, 0x000035, 0x000036, 0x000037,
	0x000038, 0x000039, 0x00003a, 0x00d89b, 0x00003c, 0x00003d, 0x00003e, 0x00d89f,
	0xefba94, 0x00d8a1, 0xefba92, 0x00d8a8, 0xefba98, 0x00d8aa, 0xefba8e, 0x00d8a7,
	0xefba91, 0xefba93, 0xefba97, 0xefba9b, 0xefba9f, 0xefbaa3, 0xefbaa7, 0x00d8af,
	0x00d8b0, 0x00d8b1, 0x00d8b2, 0xefbab4, 0xefbab8, 0xefbabb, 0xefbabf, 0xefbb83,
	0xefbb85, 0xefbb8b, 0xefbb8f, 0xefba9c, 0xefbaa0, 0xefbaa4, 0xefbaa8, 0x000023,
	0x00002d, 0xefbb93, 0xefbb97, 0xefbb9b, 0xefbb9f, 0xefbba1, 0xefbaab, 0xefbb80,
	0xefbbad, 0xefbbb0, 0xefae82, 0x00d8ab, 0x00d8ac, 0x00d8ad, 0x00d8ae, 0xefbbb4,
	0xefbbaf, 0xefbb8c, 0xefbb90, 0xefbb94, 0xefbb91, 0xefbb98, 0xefbb95, 0xefbb9a,
	0xefbba0, 0xefbb9d, 0xefbba4, 0xefbba1, 0xefbba8, 0xefbba5, 0xefbbbc, 0x000020
};

static const unsigned int tt_charset_arabic_g2[96] = {
	0x000020, 0xefbb89, 0x00d8a2, 0x00d8a3, 0x00d8a4, 0x00d8a5, 0x00da88, 0x00d8a6,
	0xefadbc, 0xefadbd, 0xefadbe, 0xefad98, 0xefad99, 0xefad96, 0xefae8a, 0xefae94,
	0x00dbb0, 0x00dbb1, 0x00dbb2, 0x00dbb3, 0x00dbb4, 0x00dbb5, 0x00dbb6, 0x00dbb7,
	0x00dbb8, 0x00dbb9, 0xefbaa5, 0xefbb8d, 0xefbbbc, 0xefbbac, 0xefbbaa, 0xefbba9,
	0x00c3a0, 0x000041, 0x000042, 0x000043, 0x000044, 0x000045, 0x000046, 0x000047,
	0x000048, 0x000049, 0x00004a, 0x00004b, 0x00004c, 0x00004d, 0x00004e, 0x00004f,
	0x000050, 0x000051, 0x000052, 0x000053, 0x000054, 0x000055, 0x000056, 0x000057,
	0x000058, 0x000059, 0x00005a, 0x00c3ab, 0x00c3aa, 0x00c3b9, 0x00c3ae, 0xefbaa1,
	0x00c3a9, 0x000061, 0x000062, 0x000063, 0x000064, 0x000065, 0x000066, 0x000067,
	0x000068, 0x000069, 0x00006a, 0x00006b, 0x00006c, 0x00006d, 0x00006e, 0x00006f,
	0x000070, 0x000071, 0x000072, 0x000073, 0x000074, 0x000075, 0x000076, 0x000077,
	0x000078, 0x000079, 0x00007a, 0x00c3a2, 0x00c3b4, 0x00c3bb, 0x00c3a7, 0x000020
};

static const unsigned int tt_charset_hebrew_g0[96] = {
	0x000020, 0x000021, 0x000022, 0x00c2a3, 0x000024, 0x000025, 0x000026, 0x000027,
	0x000028, 0x000029, 0x00002a, 0x00002b, 0x00002c, 0x00002d, 0x00002e, 0x00002f,
	0x000030, 0x000031, 0x000032, 0x000033, 0x000034, 0x000035, 0x000036, 0x000037,
	0x000038, 0x000039, 0x00003a, 0x00003b, 0x00003c, 0x00003d, 0x00003e, 0x00003f,
	0x000040, 0x000041, 0x000042, 0x000043, 0x000044, 0x000045, 0x000046, 0x000047,
	0x000048, 0x000049, 0x00004a, 0x00004b, 0x00004c, 0x00004d, 0x00004e, 0x00004f,
	0x000050, 0x000051, 0x000052, 0x000053, 0x000054, 0x000055, 0x000056, 0x000057,
	0x000058, 0x000059, 0x00005a, 0xe28690, 0x00c2bd, 0xe28692, 0xe28691, 0x000023,
	0x00d790, 0x00d791, 0x00d792, 0x00d793, 0x00d794, 0x00d795, 0x00d796, 0x00d797,
	0x00d798, 0x00d799, 0x00d79a, 0x00d79b, 0x00d79c, 0x00d79d, 0x00d79e, 0x00d79f,
	0x00d7a0, 0x00d7a1, 0x00d7a2, 0x00d7a3, 0x00d7a4, 0x00d7a5, 0x00d7a6, 0x00d7a7,
	0x00d7a8, 0x00d7a9, 0x00d7aa, 0x00d7b1, 0x00d7b0, 0x00c2be, 0x00c3b7, 0x000020
};

static const unsigned int tt_mosaics[96] = {
	0x000020/*0xe29692*/, 0xe29698, 0xe2969d, 0xe29680, 0xe29698, 0xe29698, 0xe2969e, 0xe2969b,
	0xe2969d, 0xe2969a, 0xe2969d, 0xe2969c, 0xe296a0, 0xe29699, 0xe2969f, 0xe29680,
	0xe29696, 0xe2968b, 0xe2969e, 0xe2969b, 0xe29696, 0xe2968c, 0xe2969b, 0xe2969b,
	0xe2969e, 0xe2969e, 0xe2969f, 0xe2969c, 0xe2969b, 0xe2969b, 0xe2969e, 0xe2969b,
	0xe29692, 0xe29691, 0xe29691, 0xe29691, 0xe29691, 0xe29691, 0xe29691, 0xe29691,
	0xe29691, 0xe29691, 0xe29691, 0xe29691, 0xe29691, 0xe29691, 0xe29691, 0xe29691,
	0xe29691, 0xe29691, 0xe29691, 0xe29691, 0xe29691, 0xe29691, 0xe29691, 0xe29691,
	0xe29691, 0xe29691, 0xe29691, 0xe29692, 0xe29692, 0xe29692, 0xe29692, 0xe29692,
	0xe29697, 0xe2969a, 0xe29690, 0xe2969c, 0xe2969a, 0xe2969a, 0xe29699, 0xe2969b,
	0xe29690, 0xe2969c, 0xe29690, 0xe2969c, 0xe2969c, 0xe2969a, 0xe2969c, 0xe2969c,
	0xe296a0, 0xe29699, 0xe2969f, 0xe296a0, 0xe29699, 0xe29699, 0xe2969e, 0xe2969b,
	0xe2969f, 0xe2969f, 0xe2969f, 0xe2969c, 0xe296a0, 0xe296a0, 0xe296a0, 0xe29688
};

static const unsigned int latinsubsets[13][13] = {
	{0x0023, 0xC5AF, 0xC48D, 0xC5A5, 0xC5BE, 0xc3bd, 0xc3ad, 0xC599, 0xc3a9, 0xc3a1, 0xC49B, 0xC3BA, 0xC5A1}, /*Czech/Slovak*/
	{0xc2a3, 0x0024, 0x0040, 0x003c, 0xc2bd, 0x003e, 0x005e, 0x0023, 0x002d, 0xc2bc, 0xC781, 0xc2be, 0xc3b7}, /*English*/
	{0x0023, 0xc3b5, 0xC5A0, 0xc384, 0xc396, 0xC5BD, 0xc39c, 0xc395, 0xC5A1, 0xc3a4, 0xC3B6, 0xC5BE, 0xC3BC}, /*Estonian*/
	{0xc3a9, 0xc3af, 0xc3a0, 0xc3ab, 0xc3aa, 0xC3B9, 0xc3ae, 0x0023, 0xC3A8, 0xc3a2, 0xC3B4, 0xC3BB, 0xC3A7}, /*French*/
	{0x0023, 0x0024, 0xc2a7, 0xc384, 0xc396, 0xc39c, 0x005e, 0x005f, 0xc2b0, 0xc3a4, 0xC3B6, 0xC3BC, 0xC39F}, /*German*/
	{0xc2a3, 0x0024, 0xc3a9, 0xc2b0, 0xC3A7, 0x003e, 0x005e, 0x0023, 0xc3b9, 0xc3a0, 0xC3B2, 0xC3A8, 0xC3AC}, /*Italian*/
	{0x0023, 0x0024, 0xC5A0, 0xC497, 0xC499, 0xC5BD, 0xC48D, 0xC5AB, 0xC5A1, 0xC485, 0xC5B3, 0xC5BE, 0xC4AF}, /*Lettish/Lithuanian*/
	{0x0023, 0xC584, 0xC485, 0xC6B5, 0xC59A, 0xC581, 0xC487, 0xC3B3, 0xC499, 0xC5BC, 0xC59B, 0xC582, 0xC5BA}, /*Polish*/
	{0xC3A7, 0x0024, 0x0069, 0xc3a1, 0xC3A9, 0xC3AD, 0xC3B3, 0xC3BA, 0xc2bf, 0xC3BC, 0xC3B1, 0xC3A8, 0xC3A0}, /*Portuguese/Spanish*/
	{0x0023, 0xc2a4, 0xC5A2, 0xc382, 0xC59E, 0xC482, 0xC38E, 0xC4B1, 0xC5A3, 0xC3A2, 0xC59F, 0xC483, 0xC3AE}, /*Romanian*/
	{0x0023, 0xD081, 0xC48C, 0xC48C, 0xC5BD, 0xC490, 0xC5A0, 0xD191, 0xC48D, 0xC487, 0xC5BE, 0xC491, 0xC5A1}, /*Serbian/Croatian/Slovenian*/
	{0x0023, 0xc2a4, 0xc389, 0xc384, 0xc396, 0xC385, 0xc39c, 0x005f, 0xc3a9, 0xc3a4, 0xC3B6, 0xC3A5, 0xC3BC}, /*Swedish/Finnish/Hungarian*/
	{0x0054, 0xC49F, 0xC4B0, 0xC59E, 0xc396, 0xC387, 0xc39c, 0xC49E, 0xC4B1, 0xC59F, 0xC3B6, 0xC3A7, 0xC3BC}  /*Turkish*/
};

static const unsigned char defcolormap[32][3] = {
	{ 0, 0, 0}, {15, 0, 0}, {0, 15, 0}, {15, 15,  0}, {0,  0, 15}, {15, 0, 15}, {0, 15, 15}, {15, 15, 15}, //CLUT 0
	{ 0, 0, 0}, { 7, 0, 0}, {0,  7, 0}, { 7,  7,  0}, {0,  0,  7}, { 7, 0,  7}, {0,  7,  7}, { 7,  7,  7}, //CLUT 1
	{15, 0, 5}, {15, 7, 0}, {0, 15, 7}, {15, 15, 11}, {0, 12, 10}, { 5, 0,  0}, {6,  5,  2}, {12,  7,  7}, //CLUT 2
	{ 3, 3, 3}, {15, 7, 7}, {7, 15, 7}, {15, 15,  7}, {7,  7, 15}, {15, 7, 15}, {7, 15, 15}, {13, 13, 13}  //CLUT 3
};

static const unsigned int *tt_charsets[7][2] = {
	{tt_charset_latin_g0,      tt_charset_latin_g2},
	{tt_charset_cyrillic_1_g0, tt_charset_cyrillic_g2},
	{tt_charset_cyrillic_2_g0, tt_charset_cyrillic_g2},
	{tt_charset_cyrillic_3_g0, tt_charset_cyrillic_g2},
	{tt_charset_greek_g0,      tt_charset_greek_g2},
	{tt_charset_arabic_g0,     tt_charset_arabic_g2},
	{tt_charset_hebrew_g0,     tt_charset_arabic_g2}
};


static gboolean tt_start_filter(ttparser_t parser);
static gint tt_send_command(ttparser_t parser, gint command);
static gint tt_read_packet(ttparser_t parser, gchar *buffer, gsize size, guint timeout);
static guchar tt_unham_8_4(guchar sym1, guchar sym2);
static guint tt_unham24_18(guchar sym1, guchar sym2, guchar sym3);
static void tt_page_clear_cache(ttpage_t page);
static guint tt_get_page_id(gushort pagenum, gushort subpagenum);

static ttpage_t tt_get_page(ttparser_t parser, guint pageid, gboolean cancreate);
static gushort tt_get_tag_id(guchar tag, guchar value);
static guchar tt_get_tag(gushort tagid);
static guchar tt_get_tag_value(gushort tagid);
static void tt_set_charset(ttparser_t parser, ttpage_t page);
static void tt_page_parse(ttparser_t parser, ttpage_t page);

static void tt_page_finalize(ttparser_t parser, guint32 pageid);
static void tt_page_add_line(ttparser_t parser, ttpage_t pagecache, gushort line, guchar *data, guint length);
static ttpage_t tt_parse_packet_0(ttparser_t parser, gint mag, gchar *buffer, gint lastpage);
static gint tt_parse_packet_1_25(ttparser_t parser, ttpage_t page, gushort line, gchar *buffer);
static gint tt_parse_packet_26(ttparser_t parser, ttpage_t page, guchar *buffer);
static gint tt_parse_packet_28_29(ttparser_t parser, ttpage_t page, gushort pack, gchar *buffer);
static gint tt_parse_packet_30(ttparser_t parser, guchar *buffer);

static gpointer ttwindow_scan_thread(gpointer data);
static void tt_page_cache_destroy(gpointer data);


static gboolean tt_start_filter(ttparser_t parser)
{
	gchar *demuxfn;
	struct dmx_pes_filter_params f;
	
	if (parser == NULL) return FALSE;
	
	demuxfn = g_strdup_printf("/dev/dvb/adapter%u/demux%u", parser->adapter, parser->demux, NULL);
	
	if ((parser->fd = open(demuxfn, O_RDONLY)) < 0) {
		g_free(demuxfn);
		return FALSE;
	}
	
	g_free(demuxfn);
	
	f.pid = (uint16_t)parser->pid;
	f.input = DMX_IN_FRONTEND;
	f.output = DMX_OUT_TAP;
	f.pes_type = DMX_PES_OTHER;
	f.flags = DMX_IMMEDIATE_START;
	
	if (ioctl(parser->fd, DMX_SET_PES_FILTER, &f) == -1) {
		perror("ioctl DMX_SET_PES_FILTER");
		close(parser->fd);
		return FALSE;
	}
	
	return TRUE;
}

static gint tt_send_command(ttparser_t parser, gint command)
{
	struct pollfd pollfds[1];
		
	if (parser == NULL) return TT_ANSWER_FAILURE;
	
	if (write(parser->ctlfd[0], &command, sizeof(command)) > 0) {
		pollfds[0].fd = parser->ctlfd[0];
		pollfds[0].events = POLLIN;
		
		if (poll(&pollfds[0], 1, 5000) > 0) {
			if (pollfds[0].revents & POLLIN) {
				if (read(parser->ctlfd[0], &command, sizeof(command)) > 0) {
					return command;
				}
			}
		}
	}
	
	return TT_ANSWER_FAILURE;
}

static gint tt_read_packet(ttparser_t parser, gchar *buffer, gsize size, guint timeout)
{
	struct pollfd pollfds[2];
	ssize_t readb;
	gint ptr, answer;
	
	
	if ((parser == NULL) || (buffer == NULL) || (size == 0)) return -1;
	
	parser->command = TT_COMMAND_NULL;
	ptr = 0;
	
	do {
		pollfds[0].fd = parser->fd;
		pollfds[0].events = POLLIN;
		pollfds[1].fd = parser->ctlfd[1];
		pollfds[1].events = POLLIN;
		
		if (poll(&pollfds[0], 2, timeout*1000) > 0) {
			if (pollfds[1].revents & POLLIN) {
				readb = read(parser->ctlfd[1], &(parser->command), sizeof(parser->command));
				if (readb > 0) {
					if (parser->command == TT_COMMAND_STOP) {
						answer = TT_ANSWER_SUCCESS;
						write(parser->ctlfd[1], &answer, sizeof(answer));
						readb = 0;
						return readb;
					}
				}
			}
			
			if (pollfds[0].revents & POLLIN) {
				readb = read(parser->fd, buffer+ptr, size-ptr);
				if (readb <= 0) {
					break;
				} else {
					ptr += readb;
				}
			}
		}
		
	} while (ptr < size);
	
	
	return ptr;
}

static guchar tt_unham_8_4(guchar sym1, guchar sym2)
{
	guchar v1, v2;
	
	v1 = hamming[sym1];
	v2 = hamming[sym2];
	
	return (v2 << 4) | (v1 & 0x0f);
}

static guint tt_unham24_18(guchar sym1, guchar sym2, guchar sym3)
{
	guint value;
	
	value = 0;
	value |= (((sym1 & 0x04) >> 2) | ((sym1 & 0x70) >> 3)) & 0x0f;
	value |= (sym2 & 0x7f) << 4;
	value |= (sym3 & 0x7f) << 11;
		
	return value;
}

static void tt_page_clear_cache(ttpage_t page)
{
	memset(page->text[TT_CACHE_CURRENT], 0, sizeof(page->text[TT_CACHE_CURRENT]));
	memset(page->tags[TT_CACHE_CURRENT], 0, sizeof(page->tags[TT_CACHE_CURRENT]));
	memcpy(page->colormap[TT_CACHE_CURRENT], defcolormap, sizeof(page->colormap[TT_CACHE_CURRENT]));
	memset(&page->bitmap, 0, sizeof(page->bitmap));
}

static guint tt_get_page_id(gushort pagenum, gushort subpagenum)
{
	guint pageid;
	
	if ((pagenum < 100) || (pagenum > 899)) return 0;
	
	/*24 bytes: 11 - page, 13 - subpage, */
	pageid = ((pagenum & 0x07ff) << 13) | (subpagenum & 0x1fff); 
	
	return pageid;
}

gushort tt_get_pagenum(guint pageid)
{
	return ((pageid & 0xffe000) >> 13);
}

gushort tt_get_subpagenum(guint pageid)
{
	return (pageid & 0x1fff);
}

static ttpage_t tt_get_page(ttparser_t parser, guint pageid, gboolean cancreate)
{
	ttpage_t page;
		
	page = g_hash_table_lookup(parser->pages, &pageid);
	
	if ((page == NULL) && (cancreate)) {
		page = g_malloc(sizeof(struct _ttpage));
		page->pageid = pageid;
		page->lastupdate = 0;
		page->visible = 0;
		
		tt_page_clear_cache(page);
		
		g_hash_table_insert(parser->pages, &page->pageid, page);
	}
	
	return page;
}

static gushort tt_get_tag_id(guchar tag, guchar value)
{
	return (((value & 0xff) << 8) | (tag & 0xff));
}

static guchar tt_get_tag(gushort tagid)
{
	return (tagid & 0x00ff);
}

static guchar tt_get_tag_value(gushort tagid)
{
	return ((tagid & 0xff00) >> 8);
}

static void tt_set_charset(ttparser_t parser, ttpage_t page)
{
	struct {
		guchar language;
		guchar pagebits;
		guchar primary;
		guchar secondary;
		guchar subset;
		const gchar *name;
	} const charsets[] = {
		{TT_LANGUAGE_ENGLISH,    0x00, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 1,  "English"},
		{TT_LANGUAGE_FRENCH,     0x01, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 3,  "French"},
		{TT_LANGUAGE_SWEDISH,    0x02, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 11, "Swedish"},
		{TT_LANGUAGE_FINNISH,    0x02, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 11, "Finnish"},
		{TT_LANGUAGE_HUNGARIAN,  0x02, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 11, "Hungarian"},
		{TT_LANGUAGE_CZECH,      0x03, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 0,  "Czech"},
		{TT_LANGUAGE_SLOVAK,     0x03, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 0,  "Slovak"},
		{TT_LANGUAGE_GERMAN,     0x04, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 4,  "German"},
		{TT_LANGUAGE_PORTUGUESE, 0x05, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 8,  "Portuguese"},
		{TT_LANGUAGE_SPANISH,    0x05, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 8,  "Spanish"},
		{TT_LANGUAGE_ITALIAN,    0x06, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 5,  "Italian"},
		{TT_LANGUAGE_POLISH,     0x00, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 7,  "Polish"},
		{TT_LANGUAGE_TURKISH,    0x03, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 12, "Turkish"},
		{TT_LANGUAGE_SERBIAN,    0x00, TT_CHARSET_CYRILLIC1, TT_CHARSET_LATIN, 1,  "Serbian"},
		{TT_LANGUAGE_CROATIAN,   0x00, TT_CHARSET_CYRILLIC1, TT_CHARSET_LATIN, 1,  "Croatian"},
		{TT_LANGUAGE_SERBIAN_L,  0x05, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 10, "Serbian latin"},
		{TT_LANGUAGE_CROATIAN_L, 0x05, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 10, "Croatian latin"},
		{TT_LANGUAGE_SLOVENIAN,  0x05, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 10, "Slovenian"},
		{TT_LANGUAGE_RUSSIAN,    0x01, TT_CHARSET_CYRILLIC2, TT_CHARSET_LATIN, 1,  "Russian"},
		{TT_LANGUAGE_BULGARIAN,  0x01, TT_CHARSET_CYRILLIC2, TT_CHARSET_LATIN, 1,  "Bulgarian"},
		{TT_LANGUAGE_ESTONIAN,   0x02, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 2,  "Estonian"},
		{TT_LANGUAGE_UKRAINIAN,  0x05, TT_CHARSET_CYRILLIC3, TT_CHARSET_LATIN, 1,  "Ukrainian"},
		{TT_LANGUAGE_LETTISH,    0x06, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 6,  "Lettish"},
		{TT_LANGUAGE_LITHUANIAN, 0x06, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 6,  "Lithuanian"},
		{TT_LANGUAGE_RUMANIAN,   0x07, TT_CHARSET_LATIN,     TT_CHARSET_LATIN, 9,  "Rumanian"},
		{TT_LANGUAGE_GREEK,      0x07, TT_CHARSET_GREEK,     TT_CHARSET_LATIN, 1,  "Greek"},
		{TT_LANGUAGE_ARABIC,     0x07, TT_CHARSET_ARABIC,    TT_CHARSET_LATIN, 1,  "Arabic"},
		{TT_LANGUAGE_HEBREW,     0x05, TT_CHARSET_HEBREW,    TT_CHARSET_LATIN, 1,  "Hebrew"}
	};
	
	struct {
		guchar noid;
		guchar language;
		guchar charset;
		guchar subset;
	} const notable[] = {
		{0x00, TT_LANGUAGE_ENGLISH,    TT_CHARSET_LATIN,      1},
		{0x40, TT_LANGUAGE_ENGLISH,    TT_CHARSET_LATIN,      1},
		{0x04, TT_LANGUAGE_FRENCH,     TT_CHARSET_LATIN,      3},
		{0x0C, TT_LANGUAGE_FRENCH,     TT_CHARSET_LATIN,      3},
		{0x14, TT_LANGUAGE_FRENCH,     TT_CHARSET_LATIN,      3},
		{0x44, TT_LANGUAGE_FRENCH,     TT_CHARSET_LATIN,      3},
		{0x02, TT_LANGUAGE_SWEDISH,    TT_CHARSET_LATIN,     11},
		{0x0a, TT_LANGUAGE_SWEDISH,    TT_CHARSET_LATIN,     11},
		{0x12, TT_LANGUAGE_SWEDISH,    TT_CHARSET_LATIN,     11},
		{0x02, TT_LANGUAGE_FINNISH,    TT_CHARSET_LATIN,     11},
		
		{0x0a, TT_LANGUAGE_FINNISH,    TT_CHARSET_LATIN,     11},
		{0x12, TT_LANGUAGE_FINNISH,    TT_CHARSET_LATIN,     11},
		{0x02, TT_LANGUAGE_HUNGARIAN,  TT_CHARSET_LATIN,     11},
		{0x0a, TT_LANGUAGE_HUNGARIAN,  TT_CHARSET_LATIN,     11},
		{0x12, TT_LANGUAGE_HUNGARIAN,  TT_CHARSET_LATIN,     11},
		{0x06, TT_LANGUAGE_CZECH,      TT_CHARSET_LATIN,      0},
		{0x0e, TT_LANGUAGE_CZECH,      TT_CHARSET_LATIN,      0},
		{0x26, TT_LANGUAGE_CZECH,      TT_CHARSET_LATIN,      0},
		{0x06, TT_LANGUAGE_SLOVAK,     TT_CHARSET_LATIN,      0},
		{0x0e, TT_LANGUAGE_SLOVAK,     TT_CHARSET_LATIN,      0},
		
		{0x26, TT_LANGUAGE_SLOVAK,     TT_CHARSET_LATIN,      0},
		{0x01, TT_LANGUAGE_GERMAN,     TT_CHARSET_LATIN,      4},
		{0x09, TT_LANGUAGE_GERMAN,     TT_CHARSET_LATIN,      4},
		{0x11, TT_LANGUAGE_GERMAN,     TT_CHARSET_LATIN,      4},
		{0x21, TT_LANGUAGE_GERMAN,     TT_CHARSET_LATIN,      4},
		{0x05, TT_LANGUAGE_PORTUGUESE, TT_CHARSET_LATIN,      8},
		{0x15, TT_LANGUAGE_PORTUGUESE, TT_CHARSET_LATIN,      8},
		{0x05, TT_LANGUAGE_SPANISH,    TT_CHARSET_LATIN,      8},
		{0x15, TT_LANGUAGE_SPANISH,    TT_CHARSET_LATIN,      8},
		{0x03, TT_LANGUAGE_ITALIAN,    TT_CHARSET_LATIN,      5},
		
		{0x0b, TT_LANGUAGE_ITALIAN,    TT_CHARSET_LATIN,      5},
		{0x13, TT_LANGUAGE_ITALIAN,    TT_CHARSET_LATIN,      5},
		{0x08, TT_LANGUAGE_POLISH,     TT_CHARSET_LATIN,      7},
		{0x16, TT_LANGUAGE_TURKISH,    TT_CHARSET_LATIN,     12},
		{0x36, TT_LANGUAGE_TURKISH,    TT_CHARSET_LATIN,     12},
		{0x20, TT_LANGUAGE_SERBIAN,    TT_CHARSET_CYRILLIC1,  1},
		{0x20, TT_LANGUAGE_CROATIAN,   TT_CHARSET_CYRILLIC1,  1},
		{0x1d, TT_LANGUAGE_SERBIAN_L,  TT_CHARSET_LATIN,     10},
		{0x1d, TT_LANGUAGE_CROATIAN_L, TT_CHARSET_LATIN,     10},
		{0x1d, TT_LANGUAGE_SLOVENIAN,  TT_CHARSET_LATIN,     10},
		
		{0x24, TT_LANGUAGE_RUSSIAN,    TT_CHARSET_CYRILLIC2,  1},
		{0x24, TT_LANGUAGE_BULGARIAN,  TT_CHARSET_CYRILLIC2,  1},
		{0x22, TT_LANGUAGE_ESTONIAN,   TT_CHARSET_LATIN,      2},
		{0x25, TT_LANGUAGE_UKRAINIAN,  TT_CHARSET_CYRILLIC3,  1},
		{0x23, TT_LANGUAGE_LETTISH,    TT_CHARSET_LATIN,      6},
		{0x23, TT_LANGUAGE_LITHUANIAN, TT_CHARSET_LATIN,      6},
		{0x1f, TT_LANGUAGE_RUMANIAN,   TT_CHARSET_LATIN,      9},
		{0x37, TT_LANGUAGE_GREEK,      TT_CHARSET_GREEK,      1},
		{0x47, TT_LANGUAGE_ARABIC,     TT_CHARSET_ARABIC,     1},
		{0x57, TT_LANGUAGE_ARABIC,     TT_CHARSET_ARABIC,     1},
		
		{0x55, TT_LANGUAGE_HEBREW,     TT_CHARSET_HEBREW,     1}
	};
	
	
	gint i, j;
	
	page->charset[0] = TT_CHARSET_LATIN;
	page->charset[1] = TT_CHARSET_LATIN;
	page->charset[2] = 1;
	
	if ((parser->detectlang) && (TT_BIT_CHECK(page->bitmap, 28) || TT_BIT_CHECK(page->bitmap, 29))) {
		for (j=0; j<2; j++) {
			for (i=0; i<51; i++) {
				if (j == 0) {
					if (page->primno == notable[i].noid) {
						page->charset[0] = notable[i].charset;
						page->charset[2] = notable[i].subset;
						break;
					}
				} else {
					if (page->secno == notable[i].noid) {
						page->charset[1] = notable[i].charset;
						break;
					}
				}
			}
		}
	} else {
		for (i=0; i<28; i++) {
			if (parser->language == charsets[i].language) {
				if ((!parser->detectlang) || ((parser->detectlang) && (page->language == charsets[i].pagebits))) {
					page->charset[0] = charsets[i].primary;
					page->charset[1] = charsets[i].secondary;
					page->charset[2] = charsets[i].subset;
					break;
				}
			}
		}
	}
}



static void tt_page_parse(ttparser_t parser, ttpage_t page)
{
	int i, line, len, charsetid;
	
	const unsigned int *chartable[3];
	unsigned int utfchar;
	
	unsigned int flashflag;
	unsigned int boxflag;
	unsigned int concealflag;
	unsigned int contiguousflag;
	unsigned int escapeflag;
	unsigned int holdflag;
	unsigned int graphflag;
	unsigned int fgcolour, moscolour, bgcolour;
	
	
	if ((parser == NULL) || (page == NULL)) return;
	
	tt_set_charset(parser, page);
	
	chartable[0] = tt_charsets[page->charset[0]][0];
	chartable[1] = tt_charsets[page->charset[1]][0];
	chartable[2] = latinsubsets[page->charset[2]];
	
	for (line=0; line<26; line++) {
		flashflag = 0;
		boxflag = 0;
		concealflag = 0;
		contiguousflag = 0;
		escapeflag = 0;
		holdflag = 0;
		graphflag = 0;
		fgcolour = TT_COLOUR_WHITE;
		moscolour = TT_COLOUR_WHITE;
		bgcolour = TT_COLOUR_BLACK;
		
		if (line == 0) len = 32; else len = 40;
		
		if ((line <= 23) || ((line > 23) && (TT_BIT_CHECK(page->bitmap, line)))) {
			for (i=0; i<len; i++) {
				if ((page->text[TT_CACHE_CURRENT][line][i] >= 0x20) && (page->text[TT_CACHE_CURRENT][line][i] < 0x80)) {
					if ((!concealflag) && (!(graphflag && holdflag))){
						if (!graphflag) {
							if (!escapeflag) {
								charsetid = 0;
							} else {
								charsetid = 1;
							}
							
							utfchar = page->text[TT_CACHE_CURRENT][line][i];
							
							if (page->charset[charsetid] == TT_CHARSET_LATIN) {
								if ((utfchar == 0x23) || (utfchar == 0x24)) {
									utfchar = chartable[2][utfchar-0x23];
								} else if (utfchar == 0x40) {
									utfchar = chartable[2][2];
								} else  if ((utfchar >= 0x5b) && (utfchar <= 0x60)) {
									utfchar = chartable[2][utfchar-0x5b+3];
								} else if ((utfchar >= 0x7b) && (utfchar <= 0x7e)) {
									utfchar = chartable[2][utfchar-0x7b+9];
								} else {
									utfchar = chartable[charsetid][utfchar-0x20];
								}
							} else {
								utfchar = chartable[charsetid][utfchar-0x20];
							}
							
							page->text[TT_CACHE_CURRENT][line][i] = utfchar;
						} else {
							utfchar = tt_mosaics[page->text[TT_CACHE_CURRENT][line][i]-0x20];
							page->text[TT_CACHE_CURRENT][line][i] = utfchar;
						}
					} else {
						page->text[TT_CACHE_CURRENT][line][i] = chartable[0][0];
					}
				} else {
					if ((page->text[TT_CACHE_CURRENT][line][i] >= 0x00) && (page->text[TT_CACHE_CURRENT][line][i] <= 0x07)) {
						concealflag = 0;
						graphflag = 0;
						switch (page->text[TT_CACHE_CURRENT][line][i]) {
							case 0x00:
								fgcolour = TT_COLOUR_BLACK;
								break;
							case 0x01:
								fgcolour = TT_COLOUR_RED;
								break;
							case 0x02:
								fgcolour = TT_COLOUR_GREEN;
								break;
							case 0x03:
								fgcolour = TT_COLOUR_YELLOW;
								break;
							case 0x04:
								fgcolour = TT_COLOUR_BLUE;
								break;
							case 0x05:
								fgcolour = TT_COLOUR_MAGENTA;
								break;
							case 0x06:
								fgcolour = TT_COLOUR_CYAN;
								break;
							case 0x07:
								fgcolour = TT_COLOUR_WHITE;
								break;
							default:
								fgcolour = TT_COLOUR_WHITE;
								break;
						}
						page->tags[TT_CACHE_CURRENT][line][i] = tt_get_tag_id(TT_TAG_COLOUR, fgcolour);
						//if (tt_get_pagenum(page->pageid) == 153) printf("---> Foreground color tag 0x%2x\n", fgcolour);
					} else if ((page->text[TT_CACHE_CURRENT][line][i] >= 0x08) && (page->text[TT_CACHE_CURRENT][line][i] <= 0x09)) {
						if (page->text[TT_CACHE_CURRENT][line][i] == 0x08) {
							flashflag = 1;
						} else {
							flashflag = 0;
						}
						page->tags[TT_CACHE_CURRENT][line][i] = tt_get_tag_id(TT_TAG_FLASH, flashflag);
						//if (tt_get_pagenum(page->pageid) == 153) printf("---> Flash tag 0x%2x\n", flashflag);
					} else if ((page->text[TT_CACHE_CURRENT][line][i] >= 0x0a) && (page->text[TT_CACHE_CURRENT][line][i] <= 0x0b)) {
						if (page->text[TT_CACHE_CURRENT][line][i] == 0x0b) {
							boxflag = 1;
						} else {
							boxflag = 0;
						}
						page->tags[TT_CACHE_CURRENT][line][i] = tt_get_tag_id(TT_TAG_BOX, boxflag);
						//if (tt_get_pagenum(page->pageid) == 153) printf("---> Box tag 0x%2x\n", boxflag);
					} else if ((page->text[TT_CACHE_CURRENT][line][i] >= 0x0c) && (page->text[TT_CACHE_CURRENT][line][i] <= 0x0f)) {
						if (page->text[TT_CACHE_CURRENT][line][i] == 0x0c) {
							page->tags[TT_CACHE_CURRENT][line][i] = tt_get_tag_id(TT_TAG_SIZE, TT_SIZE_NORMAL);
						} else if (page->text[TT_CACHE_CURRENT][line][i] == 0x0d) {
							page->tags[TT_CACHE_CURRENT][line][i] = tt_get_tag_id(TT_TAG_SIZE, TT_SIZE_DOUBLE_HEIGHT);
						} else if (page->text[TT_CACHE_CURRENT][line][i] == 0x0e) {
							page->tags[TT_CACHE_CURRENT][line][i] = tt_get_tag_id(TT_TAG_SIZE, TT_SIZE_DOUBLE_WIDTH);
						} else if (page->text[TT_CACHE_CURRENT][line][i] == 0x0f) {
							page->tags[TT_CACHE_CURRENT][line][i] = tt_get_tag_id(TT_TAG_SIZE, TT_SIZE_DOUBLE_SIZE);
						}
						//if (tt_get_pagenum(page->pageid) == 153) printf("---> Size tag 0x%2x\n", data[i]);
					} else if ((page->text[TT_CACHE_CURRENT][line][i] >= 0x10) && (page->text[TT_CACHE_CURRENT][line][i] <= 0x17)) {
						concealflag = 0;
						graphflag = 1;
						switch (page->text[TT_CACHE_CURRENT][line][i]) {
							case 0x10:
								moscolour = TT_COLOUR_BLACK;
								break;
							case 0x11:
								moscolour = TT_COLOUR_RED;
								break;
							case 0x12:
								moscolour = TT_COLOUR_GREEN;
								break;
							case 0x13:
								moscolour = TT_COLOUR_YELLOW;
								break;
							case 0x14:
								moscolour = TT_COLOUR_BLUE;
								break;
							case 0x15:
								moscolour = TT_COLOUR_MAGENTA;
								break;
							case 0x16:
								moscolour = TT_COLOUR_CYAN;
								break;
							case 0x17:
								moscolour = TT_COLOUR_WHITE;
								break;
							default:
								moscolour = TT_COLOUR_WHITE;
								break;
						}
						page->tags[TT_CACHE_CURRENT][line][i] = tt_get_tag_id(TT_TAG_COLOUR, moscolour);
						//if (tt_get_pagenum(page->pageid) == 153) printf("---> Graphics color tag 0x%2x\n", moscolour);
					} else if (page->text[TT_CACHE_CURRENT][line][i] == 0x18) {
						concealflag = 1;
						//if (tt_get_pagenum(page->pageid) == 153) printf("---> Conceal tag 0x%2x\n", concealflag);
					} else if ((page->text[TT_CACHE_CURRENT][line][i] >= 0x19) && (page->text[TT_CACHE_CURRENT][line][i] <= 0x1a)) {
						if (page->text[TT_CACHE_CURRENT][line][i] == 0x19) {
							contiguousflag = 1;
						} else {
							contiguousflag = 0;
							graphflag = 1;
						}
						//if (tt_get_pagenum(page->pageid) == 153) printf("---> Contiguous tag 0x%2x\n", contiguousflag);
					} else if (page->text[TT_CACHE_CURRENT][line][i] == 0x1b) {
						escapeflag = !escapeflag;
						//if (tt_get_pagenum(page->pageid) == 153) printf("---> Escape tag 0x%2x\n", escapeflag);
					} else if (page->text[TT_CACHE_CURRENT][line][i] == 0x1c) {
						bgcolour = TT_COLOUR_BLACK;
						page->tags[TT_CACHE_CURRENT][line][i] = tt_get_tag_id(TT_TAG_BG_COLOUR, bgcolour);
						//if (tt_get_pagenum(page->pageid) == 153) printf("---> Black background tag 0x%2x\n", bgcolour);
					} else if (page->text[TT_CACHE_CURRENT][line][i] == 0x1d) {
						bgcolour = fgcolour;
						holdflag = 0;
						page->tags[TT_CACHE_CURRENT][line][i] = tt_get_tag_id(TT_TAG_BG_COLOUR, bgcolour);
						//if (tt_get_pagenum(page->pageid) == 153) printf("---> New background tag 0x%2x\n", bgcolour);
					} else if ((page->text[TT_CACHE_CURRENT][line][i] >= 0x1e) && (page->text[TT_CACHE_CURRENT][line][i] <= 0x1f)) {
						if (page->text[TT_CACHE_CURRENT][line][i] == 0x1e) {
							holdflag = 1;
						} else {
							holdflag = 0;
						}
						//if (tt_get_pagenum(page->pageid) == 153) printf("---> Hold tag 0x%2x\n", holdflag);
					}
					
					page->text[TT_CACHE_CURRENT][line][i] = chartable[0][0];
				
				}
			}
		}
	}
	
}

gchar *tt_page_get_string(ttparser_t parser, guint32 pageid, guint *length, GSList **tags)
{
	ttpage_t page;
	gchar *text, *stext;
	gint line, sym, len, i;
	guint ptr;
	tttag_t currenttag, fccoltag, bgcoltag, sizetag, flashtag, boxtag;
	
	page = tt_get_page(parser, pageid, FALSE);
	
	g_mutex_lock(parser->expmutex);
	
	if ((page == NULL) || ((page != NULL) && (page->lastupdate == 0))) return NULL;
	
	text = g_malloc0(TT_LINE_UTF_MAX_LENGTH + 1);
	
	ptr = 0;
	sym = 0;
	
	if (tags != NULL) *tags = NULL;
	
	for (line=0; line<25; line++) {
		if (line == 0) len = 32; else len = 40;
		
		if ((line <= 23) || ((line > 23) && (TT_BIT_CHECK(page->bitmap, line)))) {
			if (tags != NULL) {
				fccoltag = bgcoltag = sizetag = flashtag = boxtag = NULL;
			}
			
			for (i=0; i<len; i++) {
				if ((tags != NULL) && (page->tags[TT_CACHE_LONGTERM][line][i] != 0)) {
					currenttag = g_malloc0(sizeof(struct _tttag));
					currenttag->tag = tt_get_tag(page->tags[TT_CACHE_LONGTERM][line][i]);
					currenttag->value = tt_get_tag_value(page->tags[TT_CACHE_LONGTERM][line][i]);
					currenttag->start = sym;
					
					switch (currenttag->tag) {
						case TT_TAG_COLOUR:
							if (fccoltag != NULL) fccoltag->end = sym;
							fccoltag = currenttag;
							break;
						case TT_TAG_BG_COLOUR:
							if (bgcoltag != NULL) bgcoltag->end = sym;
							bgcoltag = currenttag;
							break;
						case TT_TAG_SIZE:
							if (sizetag != NULL) sizetag->end = sym;
							sizetag = currenttag;
							break;
						case TT_TAG_FLASH:
							if (flashtag != NULL) flashtag->end = sym;
							flashtag = currenttag;
							break;
						case TT_TAG_BOX:
							if (boxtag != NULL) boxtag->end = sym;
							boxtag = currenttag;
							break;
						default:
							break;
					}
					
					*tags = g_slist_prepend(*tags, currenttag);
				}
				
				if (page->text[TT_CACHE_LONGTERM][line][i] & 0xff0000) {
					text[ptr+0] = (gchar)((page->text[TT_CACHE_LONGTERM][line][i] & 0xff0000) >> 16);
					text[ptr+1] = (gchar)((page->text[TT_CACHE_LONGTERM][line][i] & 0x00ff00) >> 8);
					text[ptr+2] = (gchar)(page->text[TT_CACHE_LONGTERM][line][i] & 0x0000ff);
					ptr = ptr+3;
				} else if (page->text[TT_CACHE_LONGTERM][line][i] & 0x00ff00) {
					text[ptr+0] = (gchar)((page->text[TT_CACHE_LONGTERM][line][i] & 0x00ff00) >> 8);
					text[ptr+1] = (gchar)(page->text[TT_CACHE_LONGTERM][line][i] & 0x0000ff);
					ptr = ptr+2;
				} else {
					text[ptr+0] = (gchar)(page->text[TT_CACHE_LONGTERM][line][i] & 0x0000ff);
					ptr = ptr+1;
				}
				
				sym++;
			}
			
			if (tags != NULL) {
				if (fccoltag != NULL) fccoltag->end = sym;
				if (bgcoltag != NULL) bgcoltag->end = sym;
				if (sizetag != NULL) sizetag->end = sym;
				if (flashtag != NULL) flashtag->end = sym;
				if (boxtag != NULL) boxtag->end = sym;
			}
		}
			
		text[ptr+0] = '\n';
		ptr = ptr+1;
		sym++;
	}
	
	text[ptr] = '\0';
	
	g_mutex_unlock(parser->expmutex);
	
	if (length != NULL) *length = ptr;
	
	if (ptr < TT_LINE_UTF_MAX_LENGTH + 1) {
		stext = g_realloc(text, ptr+1);
		if (stext != NULL) {
			text = stext;
		}
	}
	
	if (tags != NULL) *tags = g_slist_reverse(*tags);
	
	return text;
}

ttcolour_t tt_page_get_colormap(ttparser_t parser, guint32 pageid)
{
	ttcolour_t colours;
	ttpage_t page;
	gint i;
	
	page = NULL;
	
	colours = g_malloc0(sizeof(struct _ttcolour) * 16);
	
	if ((parser != NULL) && ((pageid >= 100) && (pageid <= 899))) {
		page = tt_get_page(parser, pageid, FALSE);
	}
	
	if ((page == NULL) || ((page != NULL) && (page->lastupdate == 0))) {
		for (i=0; i<16; i++) {
			colours[i].r = defcolormap[i][0];
			colours[i].g = defcolormap[i][1];
			colours[i].b = defcolormap[i][2];
		}
	} else {
		g_mutex_lock(parser->expmutex);
		for (i=0; i<16; i++) {
			colours[i].r = page->colormap[TT_CACHE_LONGTERM][i][0];
			colours[i].g = page->colormap[TT_CACHE_LONGTERM][i][1];
			colours[i].b = page->colormap[TT_CACHE_LONGTERM][i][2];
		}
		g_mutex_unlock(parser->expmutex);
	}
		
	return colours;	
}

static void tt_page_finalize(ttparser_t parser, guint32 pageid)
{
	ttpage_t page;
	gint i, newpage;
	
	if (parser == NULL) return;
	
	page = tt_get_page(parser, pageid, FALSE);
	
	if (page == NULL) return;
	
	if (TT_BITMAP_FULL(page->bitmap)) {
		tt_page_parse(parser, page);
		
		g_mutex_lock(parser->expmutex);
		
		memcpy(page->colormap[TT_CACHE_LONGTERM], page->colormap[TT_CACHE_CURRENT], sizeof(page->colormap[TT_CACHE_LONGTERM]));
		memcpy(page->text[TT_CACHE_LONGTERM], page->text[TT_CACHE_CURRENT], sizeof(page->text[TT_CACHE_LONGTERM]));
		memcpy(page->tags[TT_CACHE_LONGTERM], page->tags[TT_CACHE_CURRENT], sizeof(page->tags[TT_CACHE_LONGTERM]));
		
		page->lastupdate = time(NULL);
		
		g_mutex_unlock(parser->expmutex);
		
		
		if (parser->callback != NULL) {
			(parser->callback)(page->visible ? TT_EVENT_PAGE_REFRESH : TT_EVENT_PAGE_NEW , page);
			page->visible = 1;
		}
	}
}

static void tt_page_add_line(ttparser_t parser, ttpage_t pagecache, gushort line, guchar *data, guint length)
{
	gint ptr;
	
	if (line > 25) {
		printf("[-] Wrong line number: %i (0-25 lines allowed), skipping\n", line);
		return;
	}
	
	if (length > 40) {
		printf("[-] Data segment is too long: %i bytes (40 bytes is limit), truncated\n", length);
		length = 40;
	}
	
	for (ptr=0; ptr<length; ptr++) {
		pagecache->text[TT_CACHE_CURRENT][line][ptr] = data[ptr];
	}
	
	/*if (tt_get_pagenum(pagecache->pageid) == 100) {
		printf("--> Add line: %u\n", line);
	}*/
	
	TT_BIT_SET(pagecache->bitmap, line);
}

static ttpage_t tt_parse_packet_0(ttparser_t parser, gint mag, gchar *buffer, gint lastpage)
{
	unsigned int sym, pagenum, subpagenum, language, flags, pageid;
	unsigned char pagedata[33]; 
	unsigned char hdrdata[8];
	unsigned char eraseflag;
	unsigned char newsflashflag;
	unsigned char subtitleflag;
	unsigned char supresshdrflag;
	unsigned char updateflag;
	unsigned char intseqflag;
	unsigned char inhibitflag;
	unsigned char magserialflag;
	ttpage_t page;
	
	pagenum = tt_unham_8_4(buffer[0], buffer[1]);
	
	if ((pagenum == 0xfd) || (pagenum == 0xfe) || (pagenum == 0xff)) return NULL;
	
	pagenum = mag*100 + (pagenum / 16)*10 + (pagenum % 16);
	if (pagenum < 100) pagenum += 800;
	
	if ((pagenum == 0x01be) || (pagenum == 0x01f0)) return NULL;
	
	for (sym=2; sym<8; sym++) {
		hdrdata[sym] = hamming[buffer[sym]];
	}
	
	subpagenum = ((hdrdata[2] & 0x0f) | ((hdrdata[3] & 0x07) << 4) | ((hdrdata[4] & 0x0f) << 7) | ((hdrdata[5] & 0x03) << 11));
	language =   (hdrdata[7] >> 1) & 0x07;
	flags =      ((hdrdata[7] & 0x01) << 7) | ((hdrdata[3] & 0x08) << 3) | ((hdrdata[5] & 0x0C) << 2) | hdrdata[6];
		
	eraseflag = (hdrdata[3] & 0x08) ? 1 : 0;
	newsflashflag = (hdrdata[5] & 0x04) ? 1 : 0; 
	subtitleflag = (hdrdata[5] & 0x08) ? 1 : 0;
	supresshdrflag = (hdrdata[6] & 0x01) ? 1 : 0;
	updateflag = (hdrdata[6] & 0x02) ? 1 : 0;
	intseqflag = (hdrdata[6] & 0x04) ? 1 : 0;
	inhibitflag = (hdrdata[6] & 0x08) ? 1 : 0;
	magserialflag = (hdrdata[7] & 0x01) ? 1 : 0;
	
	memset(pagedata, 0x20, sizeof(pagedata));
	pagedata[sizeof(pagedata)-1] = '\0';
							
	for (sym=8; sym<40; sym++) {
		pagedata[sym-8] = buffer[sym] & 0x7F;
	}
	
	if ((pagenum >= 100) && (pagenum <= 899) && (!supresshdrflag) && (!inhibitflag)) {
		pageid = tt_get_page_id(pagenum, subpagenum);
		page = tt_get_page(parser, pageid, TRUE);
		
		if (pageid != lastpage) {
			tt_page_clear_cache(page);
			if (parser->callback != NULL) {
				(parser->callback)(TT_EVENT_PAGE_PARSING, page);
			}
		}
		
		page->language = language;
		page->eraseflag = eraseflag;
		page->newsflashflag = newsflashflag;
		page->subtitleflag = subtitleflag;
		page->updateflag = updateflag;
		page->intseqflag = intseqflag;
		page->magserialflag = magserialflag;
		
		tt_page_add_line(parser, page, 0, &pagedata[0], 32);
		
		//printf("--> %u, %u, %u, (%u, %u, %u)\n", pagenum, subpagenum, pageid, mag, hdrdata[0], hdrdata[1]); 
		
		//TVC
		/*Page 888, subpage 2128, e: 1, n: 1, s: 1, s: 0, u: 1, i: 0, i: 0, m: 1
		Page 155, subpage 0, e: 0, n: 0, s: 0, s: 0, u: 0, i: 1, i: 0, m: 1
		Page 156, subpage 0, e: 0, n: 0, s: 0, s: 0, u: 0, i: 1, i: 0, m: 1
		Page 156, subpage 0, e: 0, n: 0, s: 0, s: 0, u: 0, i: 1, i: 0, m: 1
		Page 157, subpage 0, e: 0, n: 0, s: 0, s: 0, u: 0, i: 1, i: 0, m: 1
		Page 158, subpage 0, e: 0, n: 0, s: 0, s: 0, u: 0, i: 1, i: 0, m: 1
		Page 159, subpage 0, e: 0, n: 0, s: 0, s: 0, u: 0, i: 1, i: 0, m: 1*/
		
		//Channel one
		/*Page 459, subpage 0, e: 0, n: 0, s: 0, s: 0, u: 0, i: 0, i: 0, m: 1
		Page 460, subpage 0, e: 0, n: 0, s: 0, s: 0, u: 0, i: 0, i: 0, m: 1
		Page 461, subpage 0, e: 0, n: 0, s: 0, s: 0, u: 0, i: 0, i: 0, m: 1
		Page 462, subpage 0, e: 0, n: 0, s: 0, s: 0, u: 0, i: 0, i: 0, m: 1
		Page 463, subpage 0, e: 0, n: 0, s: 0, s: 0, u: 0, i: 0, i: 0, m: 1
		Page 463, subpage 0, e: 0, n: 0, s: 0, s: 0, u: 0, i: 0, i: 0, m: 1
		Page 463, subpage 0, e: 0, n: 0, s: 0, s: 0, u: 0, i: 0, i: 0, m: 1*/
		
		//if ((pagenum == 101) || (pagenum == 140)) {
		
		/*printf("Page %i, subpage %i, e: %i, n: %i, s: %i, s: %i, u: %i, i: %i, i: %i, m: %i\n", pagenum, subpagenum, eraseflag, newsflashflag, subtitleflag,
				supresshdrflag, updateflag, intseqflag, inhibitflag, magserialflag);
		*/
		//}
		
		return page;
	}
	
	return NULL;
}

static int tt_parse_packet_1_25(ttparser_t parser, ttpage_t page, gushort line, gchar *buffer)
{
	unsigned int sym;
	
	if (page == NULL) return 0;
	
	for (sym=0; sym<40; sym++) {
		buffer[sym] = buffer[sym] & 0x7F;
	}
	
	tt_page_add_line(parser, page, line, buffer, 40);
	
	return page->pageid;
}

static int tt_parse_packet_26(ttparser_t parser, ttpage_t page, guchar *buffer)
{
	unsigned int sym;
	unsigned char descode;
	unsigned int triplets[13];
	unsigned int address, mode, data;
	unsigned int row, column, posset;
	
	if (page == NULL) return 0;
	
	descode = hamming[buffer[0]];
	//printf("---> Designation code: x%2x\n", descode);
	
	for (sym=0; sym<13; sym++) {
		triplets[sym] = tt_unham24_18(buffer[1+(sym*3)], buffer[2+(sym*3)], buffer[3+(sym*3)]);
		//printf("---> Triplet %i: x%6x\n", sym, triplets[sym]);
	}
	
	posset = 0;
	
	for (sym=0; sym<13; sym++) {
		address = triplets[sym] & 0x3f;
		mode = (triplets[sym] >> 6) & 0x1f;
		data = (triplets[sym] >> 11) & 0x7f;
		//printf("---> Address %i, mode %i, data 0x%2x\n", address, mode, data);
		
		if (address <= 39) {
			if (posset) {
				column = address;
				if (mode == 15) {
					printf("---> Replace G2 %i:%i with %i\n", column, row, data);
				} else if (mode == 16) {
					printf("---> Replace G0 %i:%i with %i\n", column, row, data);
				}
			}
		} else if ((address >= 40) && (address <= 63)) {
			if ((address == 63) && (mode == 31)) {
				posset = 0;
				break;
			} else if ((mode == 1) || (mode == 4)) {
				if (mode == 1) {
					column = 0;
				} else {
					column = data;
				}
				row = address - 40;
				posset = 1;
			}
		}
	}
	
	TT_BIT_SET(page->bitmap, 26);
		
	return page->pageid;
}

static gint tt_parse_packet_28_29(ttparser_t parser, ttpage_t page, gushort pack, gchar *buffer)
{
	unsigned int sym, ptr, color, remap;
	unsigned char primno, secno;
	unsigned char descode;
	unsigned int triplets[13];
	
	static const unsigned char colorremap[8][2] = {
		{0, 0}, {0, 1}, {0, 2}, {1, 1}, {1,  2}, {2, 1}, {2, 2}, {2, 3}
	};
	
	if (page == NULL) return 0;
	
	descode = hamming[buffer[0]];
	//printf("---> Designation code: x%2x\n", descode);
	
	for (sym=0; sym<13; sym++) {
		triplets[sym] = tt_unham24_18(buffer[1+(sym*3)], buffer[2+(sym*3)], buffer[3+(sym*3)]);
		//printf("---> Triplet %i: x%6x\n", sym, triplets[sym]);
	}
	
	if ((descode == 0x00) || (descode == 0x04)) {
		primno = (triplets[0] >> 7) & 0x7f;
		secno = ((triplets[1] << 4) & 0x70) | ((triplets[0] >> 14) & 0x0f) & 0x7f;
		
		if (secno == 0x7f) secno = primno;
		
		if ((pack == 28) || ((pack == 29) && (!TT_BIT_CHECK(page->bitmap, 28)))) {
			page->primno = primno;
			page->secno = secno;
			//printf("[+] Primary/secondary national options: 0x%2x/0x%2x\n", page->primno, page->secno);
		}
					
		if (descode == 0x00) {
			ptr = 16;
			//printf("[+] Remapping CLUT 2-3\n");
		} else {
			ptr = 0;
			//printf("[+] Remapping CLUT 0-1\n");
		}
		
		color = ((triplets[2] << 8) & 0xf00) | ((triplets[1] >> 10) & 0xff) & 0xfff;
		page->colormap[TT_CACHE_CURRENT][ptr][0] = color & 0x0f;
		page->colormap[TT_CACHE_CURRENT][ptr][1] = (color >> 4) & 0x0f;
		page->colormap[TT_CACHE_CURRENT][ptr][2] = (color >> 8) & 0x0f;
		//printf("[+] Color %i: R:%i G:%i B:%i\n", ptr, page->colormap[ptr][0], page->colormap[ptr][1], page->colormap[ptr][2]); 
		ptr++;
		
		for (sym=2; sym<12; sym=sym+2) {
			color = (triplets[sym] >> 4) & 0xfff;
			page->colormap[TT_CACHE_CURRENT][ptr][0] = color & 0x0f;
			page->colormap[TT_CACHE_CURRENT][ptr][1] = (color >> 4) & 0x0f;
			page->colormap[TT_CACHE_CURRENT][ptr][2] = (color >> 8) & 0x0f;
			//printf("[+] Color %i: R:%i G:%i B:%i\n", ptr, page->colormap[ptr][0], page->colormap[ptr][1], page->colormap[ptr][2]);
			ptr++;
			color = ((triplets[sym+1] << 2) & 0xffc) | ((triplets[sym] >> 16) & 0x03) & 0xfff;
			page->colormap[TT_CACHE_CURRENT][ptr][0] = color & 0x0f;
			page->colormap[TT_CACHE_CURRENT][ptr][1] = (color >> 4) & 0x0f;
			page->colormap[TT_CACHE_CURRENT][ptr][2] = (color >> 8) & 0x0f;
			//printf("[+] Color %i: R:%i G:%i B:%i\n", ptr, page->colormap[ptr][0], page->colormap[ptr][1], page->colormap[ptr][2]);
			ptr++;
			color = ((triplets[sym+2] << 8) & 0xf00) | ((triplets[sym+1] >> 10) & 0xff) & 0xfff;
			page->colormap[TT_CACHE_CURRENT][ptr][0] = color & 0x0f;
			page->colormap[TT_CACHE_CURRENT][ptr][1] = (color >> 4) & 0x0f;
			page->colormap[TT_CACHE_CURRENT][ptr][2] = (color >> 8) & 0x0f;
			//printf("[+] Color %i: R:%i G:%i B:%i\n", ptr, page->colormap[ptr][0], page->colormap[ptr][1], page->colormap[ptr][2]);
			ptr++;
		}
		
		page->defscrcolor = (triplets[12] >> 4) & 0x1f;
		page->defrowcolor = (triplets[12] >> 9) & 0x1f;
		page->bgcolorsub = (triplets[12] >> 14) & 0x01;
		
		remap = (triplets[12] >> 15) & 0x07;
		
		page->defscrcolor += colorremap[remap][1]*8;
		page->defrowcolor += colorremap[remap][0]*8;
		
		//printf("[+] Screen/row color: %i/%i\n", page->defscrcolor, page->defrowcolor);
		
		TT_BIT_SET(page->bitmap, pack);
	}
	
	return page->pageid;
}

static gint tt_parse_packet_30(ttparser_t parser, guchar *buffer)
{
	guint sym;
	gchar hdrdata[8];
	gint timeoffset;
	guint MJD, h, m, s;
	
	for (sym=0; sym<7; sym++) {
		hdrdata[sym] = hamming[buffer[sym]];
	}
	
	if (!(hdrdata[0] & 0x0e)) {
		parser->initialpage = hdrdata[1] + (hdrdata[2] * 10) + ((((hdrdata[4] & 0x08) >> 3) | ((hdrdata[6] & 0x0C) >> 1)) * 100);
		parser->initialsubpage = (hdrdata[3] & 0x0F) | ((hdrdata[4] & 0x07) << 4) | ((hdrdata[5] & 0x0F) << 8) | ((hdrdata[6] & 0x03) << 12);
		if (parser->initialsubpage == 0x3f7f) parser->initialsubpage = 0;
		
		parser->networkident = (buffer[7] << 8) | buffer[8];
		
		timeoffset = (((buffer[9] & 0x3E) >> 1) & 0x0F) * 30 * 60;
		if (buffer[9] & 0x40) timeoffset = -timeoffset;
		
		MJD = (((buffer[10] & 0x0F) - 1) * 10000) +  ((((buffer[11] >> 4) & 0x0F) - 1) * 1000) + (((buffer[11] & 0x0F) - 1) * 100) + ((((buffer[12] >> 4) & 0x0F) - 1) * 10) + ((buffer[12] & 0x0F) - 1);
		h = (((((buffer[13] >> 4) & 0x0F) - 1) * 10) + ((buffer[13] & 0x0F) - 1));
		m = (((((buffer[14] >> 4) & 0x0F) - 1) * 10) + ((buffer[14] & 0x0F) - 1));
		s = (((((buffer[15] >> 4) & 0x0F) - 1) * 10) + ((buffer[15] & 0x0F) - 1));
		
		parser->svctime = ((((((MJD - 40587) * 24) + h) * 60) + m) * 60) + s + timeoffset;
		
		for (sym=20; sym<40; sym++) {
			parser->networkname[sym-20] = buffer[sym] & 0x7F;
		}
		parser->networkname[sizeof(parser->networkname)-1] = '\0';
		
		sym = sizeof(parser->networkname)-2;
		
		while ((sym >= 0) && (parser->networkname[sym] == 0x20)) {
			parser->networkname[sym] = '\0';
			sym--;
		}
		
		if (parser->callback != NULL) (parser->callback)(TT_EVENT_NETWORK, parser);
		
		return parser->initialpage;
	}
	
	return 0;
}

static gpointer tt_parser_thread(gpointer data)
{
	const guchar peshdr[4] = {0x00, 0x00, 0x01, 0xbd};
	
	guchar buffer[4096];
	gint readbytes;
	gint startptr, ptr, localptr;
	guint existingdata;
	
	guint peslen, peshdrlen;
	guint sid;
	guint numlines, line, sym;
	guint byte0, byte1, mag, pack;
	
	guint lastpageid;
	
	gint command;
	
	guint timeoutcount, brokenlines, emptypackets;
	
	ttparser_t parser;
	ttpage_t page;
	
	
	parser = (ttparser_t)data;
	
	if (parser == NULL) return NULL;
	
	printf("[+] Teletext parser started\n");
	
	startptr = 0;
	existingdata = 0;
	
	lastpageid = 0;
	
	timeoutcount = 0;
	emptypackets = 0;
	
	page = NULL;
	
	while (1) {
		readbytes = tt_read_packet(parser, &buffer[existingdata], sizeof(buffer)-existingdata, 5);
		
		if (readbytes > 0) {
			timeoutcount = 0;
		} else if ((readbytes == 0) && (parser->command == TT_COMMAND_NULL)) {
			if (timeoutcount <= 10) {
				timeoutcount++;
				continue;
			} else {
				printf("[-] Teletext parser no data timeout\n");
				parser->error = TT_ERROR_TIMEOUT;
				break;
			}
		} else if ((readbytes == 0) && (parser->command == TT_COMMAND_STOP)) {
			printf("[+] Teletext parser recieved stop signal\n");
			parser->error = TT_ERROR_NULL;
			break;
		}
		
		readbytes += existingdata;
		
		startptr = 0;
		
		ptr = -1;
		
		while (startptr < (readbytes - 4)) {
			if (memcmp(&buffer[startptr], &peshdr[0], sizeof(peshdr)) == 0) {
				ptr = startptr;
				break;
			} else {
				startptr++;
			}
		}
		
		if (ptr == -1) {
			printf("[-] PES header sequence not found, reading new data\n");
			existingdata = 0;
			continue;
		}
		
		do {
			if (memcmp(&buffer[ptr], &peshdr[0], sizeof(peshdr)) == 0) {
				ptr += 4;
			} else {
				printf("[-] PES header sequence not found - buffer data corrupted\n");
				ptr = readbytes + 4;
				break;
			}
						
			peslen = (buffer[ptr] << 8) | (buffer[ptr+1]);
			
			if ((readbytes - ptr) < peslen) {
				//printf("PES length: %i, segmented\n", peslen);
				break;
			}
			
			peshdrlen = buffer[ptr+4];
			
			//printf("PES length: %i, header: %i, ptr: %i\n", peslen, peshdrlen, ptr);
			
			localptr = ptr + 5 + peshdrlen;
			
			sid = buffer[localptr];
			
			//printf("PES SID: %i\n", sid);
			
			numlines = ((peslen - 6) / 46) - 1;
			
			brokenlines = 0;
			
			//printf("Lines number: %i\n", numlines);
			
			localptr++;
			
			for (line=0; line<numlines; line++) {
				if (buffer[localptr+1] == 44) {
					for (sym=2; sym<46; sym++) {
						buffer[localptr+sym] = reverse[buffer[localptr+sym]];
					}
					
					if (buffer[localptr+3] == 0x27) {
						byte0 = hamming[buffer[localptr+4]];
						byte1 = hamming[buffer[localptr+5]];
						mag   = byte0 & 0x07;
						pack  = (byte0 >> 3) | (byte1 << 1);
						//printf("Decoding line: %i (mag: %i, pack: %i)\n", line, mag, pack);
						
						if (pack == 0) {
							page = tt_parse_packet_0(parser, mag, buffer+localptr+6, lastpageid);
							if (page != NULL) {
								if (lastpageid != page->pageid) {
									tt_page_finalize(parser, lastpageid);
									lastpageid = page->pageid;
								}
							}
						} else if ((pack > 0) && (pack < 26)) {
							tt_parse_packet_1_25(parser, page, pack, buffer+localptr+6);
						} else if ((pack == 28) || (pack == 29)) {
							tt_parse_packet_28_29(parser, page, pack, buffer+localptr+6);
						} else if (pack == 26) {
							tt_parse_packet_26(parser, page, buffer+localptr+6);
						} else if (pack == 30) {
							tt_parse_packet_30(parser, buffer+localptr+6);
						} else {
							printf("packet %u\n", pack);
						}
					} else {
						brokenlines++;
					}
				
				}
				localptr += 46;
			}
			
			ptr += peslen + 2;
			
			if (brokenlines == numlines) {
				emptypackets++;
			} else {
				emptypackets = 0;
			}
						
		} while (ptr <= (readbytes - 6));
		
		if (emptypackets > 1000) {
			printf("[-] No teletext data in the stream\n");
			parser->error = TT_ERROR_NO_DATA;
			break;
		}	
		
		startptr = ptr - 4;
		
		existingdata = readbytes - startptr;
		
		if (existingdata) {
			memmove(&buffer[0], &buffer[startptr], existingdata);
		}
		
	}
	
	parser->running = FALSE;
	if (parser->callback != NULL) (parser->callback)(TT_EVENT_STOP, parser);
	
	return NULL;
}

static void tt_page_cache_destroy(gpointer data)
{
	ttpage_t page;
	
	page = (ttpage_t)data;
	
	g_free(page);
}

ttparser_t tt_parser_new(guint language, gboolean detectlang, guint datasrc[3], tt_parser_event_callback callback)
{
	ttparser_t instance;
	gint sym;
		
	instance = g_malloc0(sizeof(struct _ttparser));
	
	if (instance != NULL) {
		instance->adapter = datasrc[0];
		instance->demux = datasrc[1];
		instance->pid = datasrc[2];
		
		if (!tt_start_filter(instance)) {
			g_free(instance);
			return NULL;
		}
		
		if (socketpair(AF_LOCAL, SOCK_STREAM, 0, instance->ctlfd) == -1) {
			close(instance->fd);
			g_free(instance);
			return NULL;
		}
				
		if (language > TT_LANGUAGE_HEBREW) language = TT_LANGUAGE_ENGLISH;
		
		instance->language = language;
		instance->detectlang = detectlang;
		
		instance->initialpage = 100;
		instance->initialsubpage = 0;
		instance->networkident = 0;
		instance->svctime = 0;
		sym = sprintf(instance->networkname, "Default network teletext");
		instance->networkname[sym] = '\0';
		
		instance->callback = callback;
		
		instance->expmutex = g_mutex_new();
		
		instance->pages = g_hash_table_new_full(g_int_hash, g_int_equal, NULL, tt_page_cache_destroy);
		
		instance->error = TT_ERROR_NULL;
		instance->running = TRUE;
		
		instance->thread = g_thread_create(tt_parser_thread, instance, TRUE, NULL);
	}
	
	return instance;
}

void tt_parser_close(ttparser_t parser)
{
	if (parser != NULL) {
		parser->callback = NULL;
		
		if (parser->running) {
			if (tt_send_command(parser, TT_COMMAND_STOP) == TT_ANSWER_SUCCESS) {
				printf("[+] Teletext parser stopped\n");
			}
		}
				
		close(parser->ctlfd[0]);
		close(parser->ctlfd[1]);
		
		close(parser->fd);
		
		g_hash_table_destroy(parser->pages);
		g_mutex_free(parser->expmutex);
		g_free(parser);
	}
}

void tt_parser_set_params(ttparser_t parser, guint language, gboolean detectlang)
{
	if (parser == NULL) return;
	
	if (language > TT_LANGUAGE_HEBREW) language = TT_LANGUAGE_ENGLISH;
	
	parser->language = language;
	parser->detectlang = detectlang;
	
}

void tt_parser_get_params(ttparser_t parser, guint *language, gboolean *detectlang)
{
	if (parser == NULL) {
		if (language != NULL) *language = TT_LANGUAGE_ENGLISH;
		if (detectlang != NULL) *detectlang = TRUE;
	} else {
		if (language != NULL) *language = parser->language;
		if (detectlang != NULL) *detectlang = parser->detectlang;
	}
	
}

