/*
 *      udpclient.c
 *      
 *      Copyright 2009 - 2012 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib.h>
#include <glib/gi18n.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/utsname.h>
#include <sys/stat.h>
#include "common/gost.h"
#include "common/serialize.h"
#include "common/compression.h"
#include "udpclient.h"
#include "settings.h"
#include "udpintctl.h"
#include "gsfcore.h"
#include "dbstorage.h"
//#include "servicelist.h"

#define UDP_DATAGRAM_MAX_SIZE  1400

#define NETREQ_SIGSTAT_GET_XML "<netrequest ip=\"%s\">\n\t<request>SIGSTAT_GET</request>\n</netrequest>\n\n"
#define NETREQ_SVCLIST_GET_XML "<netrequest ip=\"%s\">\n\t<request>SVCLIST_GET</request>\n</netrequest>\n\n"
#define NETREQ_SVCLIST_MPT_XML "<netrequest ip=\"%s\">\n\t<request>SVCLIST_MPT</request>\n\t<fragment>%u</fragment>\n</netrequest>\n\n"
#define NETREQ_SVCTUNE_GET_XML "<netrequest ip=\"%s\">\n\t<request>SVCTUNE_GET</request>\n\t<svcid>%u</svcid>\n</netrequest>\n\n"

enum {
	UDP_FLAG_ENCRYPTED = 0x00000001
};

enum {
	XML_PARAM_REQUEST = 0,
	XML_PARAM_RESULT,
	XML_PARAM_DATASIZE,
	XML_PARAM_NAME,
	XML_PARAM_SNR,
	XML_PARAM_AGC,
	XML_PARAM_BER,
	XML_PARAM_LOCK,
	XML_PARAM_CARRIER,
	XML_PARAM_SERVICE,
	XML_PARAM_ERROR,
	XML_PARAM_NULL
};

static gint udpclient_xml_parameter = XML_PARAM_NULL;
static gchar udpclient_error_string[] = "GSF_UDPCLIENT_ERROR";
static GQuark udpclient_error_quark = 0;
static udpclient_status_callback udpclient_callback = NULL;

enum {
	REQUEST_SIGSTAT_GET = 0,
	REQUEST_SVCLIST_GET,
	REQUEST_SVCLIST_MPT,
	REQUEST_SVCTUNE_GET,
	REQUEST_UNKNOWN,
};

struct _answer {
	gchar ip[32];
	gchar nodename[256];
	guint request;
	gboolean result;
	guint datasize;
	gchar name[256];
	guint16 snr;
	guint16 agc;
	guint32 ber;
	gboolean lock;
	gboolean carrier;
	guint32 serviceid;
	gchar servicename[256];
	gchar error[256];
} answer;


gboolean udpclient_get_channels_list(struct _netgate *netgate);
gchar *simple_data_request(struct _netgate *netgate, guint requestid, guint requestdata);
gchar *multipart_data_request(struct _netgate *netgate, gsize datasize);
void udpclient_xml_get_element(GMarkupParseContext *context, const gchar *element, const gchar **attr_names, const gchar **attr_values, gpointer data, GError **error);
void udpclient_xml_get_value(GMarkupParseContext *context, const gchar *text, gsize size, gpointer data, GError **error);
void udpclient_xml_end_element(GMarkupParseContext *context, const gchar *element, gpointer data, GError **error);


gchar *simple_data_request(struct _netgate *netgate, guint requestid, guint requestdata)
{
	gint ready;
	struct sockaddr_in client;
	//Select
	fd_set readset;
	struct timeval tv;
	//UDP integrity control
	static struct msghdr msgsend, msgrecv;
	struct iovec iovsend[2], iovrecv[2];
	//Serialization
	const gchar udpheader[] = UDP_HEADER_FORMAT;
	guint headersize;
	gchar *header;
	gchar hdrflags;
	guint hdrseq;
	guint64 hdrtimestamp;
	guint hdrimito;
	guint64 hdrsyncro;
	//Buffers
	gchar sendbuf[UDP_DATAGRAM_MAX_SIZE], recvbuf[UDP_DATAGRAM_MAX_SIZE], cryptbuf[UDP_DATAGRAM_MAX_SIZE];
	gulong comprlen;
	gsize sendbytes, recvbytes;
	gssize bytes;
	//RTT
	struct _rtt rtt;
	
	headersize = serialize_get_size((char *)&udpheader);
	
	udpintctl_new(&rtt); //initialize RTT
	udpintctl_packet_new(&rtt); //transmissions count = 0
		
	while (udpintctl_timeout(&rtt) >= 0) {
		memset(sendbuf, 0, sizeof(sendbuf));
		
		switch (requestid) {
			case REQUEST_SVCLIST_GET:
				sendbytes = snprintf(sendbuf, sizeof(sendbuf), NETREQ_SVCLIST_GET_XML, netgate->clientipstr);
				break;
			case REQUEST_SIGSTAT_GET:
				sendbytes = snprintf(sendbuf, sizeof(sendbuf), NETREQ_SIGSTAT_GET_XML, netgate->clientipstr);
				break;
			case REQUEST_SVCTUNE_GET:
				sendbytes = snprintf(sendbuf, sizeof(sendbuf), NETREQ_SVCTUNE_GET_XML, netgate->clientipstr, requestdata);
				break;
			default: 
				return NULL; 
		}
		
		sendbytes = compression_zlib_compress((char *)&sendbuf, sizeof(sendbuf), sendbytes);
		
		hdrflags = 0;
		hdrseq = 0;
		hdrtimestamp = udpintctl_timestamp(&rtt);
		
		if (gost_segment_encrypt((gchar *)&sendbuf, (gchar *)&cryptbuf, sendbytes, (guint *)&netgate->clienthash, &hdrimito)) {
			g_memmove(sendbuf, cryptbuf, sendbytes);
			hdrflags |= UDP_FLAG_ENCRYPTED;
		}
		
		header = serialize_form_message((char *)&udpheader, TRUE, NULL, hdrflags, hdrseq, hdrtimestamp, hdrimito, hdrsyncro);
				
		msgsend.msg_name = NULL;
		msgsend.msg_namelen = 0;
		msgsend.msg_iov = iovsend;
		msgsend.msg_iovlen = 2;
		iovsend[0].iov_base = header;
		iovsend[0].iov_len = headersize;
		iovsend[1].iov_base = &sendbuf;
		iovsend[1].iov_len = sendbytes;
					
		sendmsg(netgate->clientsock, &msgsend, 0);
		
		memset(header, 0, headersize);
		memset(recvbuf, 0, sizeof(recvbuf));
		
		msgrecv.msg_name = &client;
		msgrecv.msg_namelen = sizeof(struct sockaddr_in);
		msgrecv.msg_iov = iovrecv;
		msgrecv.msg_iovlen = 2;
		iovrecv[0].iov_base = header;
		iovrecv[0].iov_len = headersize;
		iovrecv[1].iov_base = &recvbuf;
		iovrecv[1].iov_len = sizeof(recvbuf);
				
		FD_ZERO(&readset);
		FD_SET(netgate->clientsock, &readset);
		tv.tv_sec = udpintctl_packet_start(&rtt);
		tv.tv_usec = 0;
		
		ready = select(netgate->clientsock+1, &readset, NULL, NULL, &tv);
		
		if ((ready) && (FD_ISSET(netgate->clientsock, &readset))) {
			if ((bytes = recvmsg(netgate->clientsock, &msgrecv, 0)) > 0) {
				
				serialize_parse_message((char *)&udpheader, header, TRUE, &hdrflags, &hdrseq, &hdrtimestamp, &hdrimito, &hdrsyncro);
				
				udpintctl_packet_stop(&rtt, udpintctl_timestamp(&rtt) - hdrtimestamp);
				
				if (hdrflags & UDP_FLAG_ENCRYPTED) {
					if (gost_segment_decrypt((gchar *)&recvbuf, (gchar *)&cryptbuf, bytes - headersize, (guint *)&netgate->clienthash, (guint *)&hdrimito)) {
						g_memmove(recvbuf, cryptbuf, bytes - headersize);	
					} else {
						free(header);
						return NULL;
					}
				}
				
				bytes = compression_zlib_decompress((char *)&recvbuf, sizeof(recvbuf), bytes - headersize);
				
				free(header);
				return g_strdup(recvbuf);
			}
			
		}
		
	}
	
	free(header);
	return NULL;
}


char *multipart_data_request(struct _netgate *netgate, gsize datasize)
{
	gint ready;
	gchar tmres = 0;
	guint fragments, fragmentid;
	struct sockaddr_in client;
	//Select
	fd_set readset;
	struct timeval tv;
	//UDP integrity control
	static struct msghdr msgsend, msgrecv;
	struct iovec iovsend[2], iovrecv[2];
	//Serialization
	const gchar udpheader[] = UDP_HEADER_FORMAT;
	guint headersize;
	gchar *header;
	gchar hdrflags;
	guint hdrseq;
	guint64 hdrtimestamp;
	guint hdrimito;
	guint64 hdrsyncro;
	//Buffers	
	gchar sendbuf[UDP_DATAGRAM_MAX_SIZE], recvbuf[UDP_DATAGRAM_MAX_SIZE], cryptbuf[UDP_DATAGRAM_MAX_SIZE];
	gulong comprlen;
	size_t sendbytes, recvbytes;
	ssize_t bytes;
	gchar *res;
	//RTT
	struct _rtt rtt;
		
	if (datasize == 0) return NULL;
	
	if (datasize % UDP_DATAGRAM_MAX_SIZE > 0) {
		fragments = (unsigned int)((datasize / UDP_DATAGRAM_MAX_SIZE) + 1);
	} else {
		fragments = (unsigned int)(datasize / UDP_DATAGRAM_MAX_SIZE);
	}
	
	res = g_malloc0(datasize+1);
	res[datasize] = '\0';
	
	headersize = serialize_get_size((char *)&udpheader);
	
	for (fragmentid = 0; fragmentid < fragments; fragmentid++) {
		
		udpintctl_new(&rtt); //initialize RTT
		udpintctl_packet_new(&rtt); //transmissions count = 0
		tmres = 0;
		
		while (udpintctl_timeout(&rtt) >= 0) {
			memset(sendbuf, 0, sizeof(sendbuf));
			sendbytes = snprintf(sendbuf, sizeof(sendbuf), NETREQ_SVCLIST_MPT_XML, netgate->clientipstr, fragmentid); 
			
			sendbytes = compression_zlib_compress((char *)&sendbuf, sizeof(sendbuf), sendbytes);
			
			hdrflags = 0;
			hdrseq = fragmentid;
			hdrtimestamp = udpintctl_timestamp(&rtt);
			
			if (gost_segment_encrypt((gchar *)&sendbuf, (gchar *)&cryptbuf, sendbytes, (guint *)&netgate->clienthash, &hdrimito)) {
				memmove(sendbuf, cryptbuf, sendbytes);
				hdrflags |= UDP_FLAG_ENCRYPTED;
			}
			
			header = serialize_form_message((char *)&udpheader, TRUE, NULL, hdrflags, hdrseq, hdrtimestamp, hdrimito, hdrsyncro);
			
			msgsend.msg_name = NULL;
			msgsend.msg_namelen = 0;
			msgsend.msg_iov = iovsend;
			msgsend.msg_iovlen = 2;
			iovsend[0].iov_base = header;
			iovsend[0].iov_len = headersize;
			iovsend[1].iov_base = &sendbuf;
			iovsend[1].iov_len = sendbytes;
						
			bytes = sendmsg(netgate->clientsock, &msgsend, 0);
			
			memset(header, 0, headersize);
			memset(recvbuf, 0, sizeof(recvbuf));
			
			msgrecv.msg_name = &client;
			msgrecv.msg_namelen = sizeof(struct sockaddr_in);
			msgrecv.msg_iov = iovrecv;
			msgrecv.msg_iovlen = 2;
			iovrecv[0].iov_base = header;
			iovrecv[0].iov_len = headersize;
			iovrecv[1].iov_base = &recvbuf;
			iovrecv[1].iov_len = sizeof(recvbuf);
			
			
			FD_ZERO(&readset);
			FD_SET(netgate->clientsock, &readset);
			tv.tv_sec = udpintctl_packet_start(&rtt);
			tv.tv_usec = 0;
			
			ready = select(netgate->clientsock+1, &readset, NULL, NULL, &tv);
			
			if ((ready) && (FD_ISSET(netgate->clientsock, &readset))) {
				if ((bytes = recvmsg(netgate->clientsock, &msgrecv, 0)) > 0) {
					
					serialize_parse_message((char *)&udpheader, header, TRUE, &hdrflags, &hdrseq, &hdrtimestamp, &hdrimito, &hdrsyncro);
					
					if (hdrseq == fragmentid) {
						udpintctl_packet_stop(&rtt, udpintctl_timestamp(&rtt) - hdrtimestamp);
						tmres = 1;
						
						if (hdrflags & UDP_FLAG_ENCRYPTED) {
							if (gost_segment_decrypt((gchar *)&recvbuf, (gchar *)&cryptbuf, bytes - headersize, (guint *)&netgate->clienthash, (guint *)&hdrimito)) {
								g_memmove(recvbuf, cryptbuf, bytes - headersize);	
							} else {
								free(header);
								g_free(res);
								return NULL;
							}
						}
						
						bytes = compression_zlib_decompress((char *)&recvbuf, sizeof(recvbuf), bytes);
						
						if (fragmentid == (fragments - 1)) {
							memcpy(&res[fragmentid*UDP_DATAGRAM_MAX_SIZE], recvbuf, datasize-(fragmentid*UDP_DATAGRAM_MAX_SIZE));
						} else {
							memcpy(&res[fragmentid*UDP_DATAGRAM_MAX_SIZE], recvbuf, UDP_DATAGRAM_MAX_SIZE);
						}
						break;
					}
				}
			}
		}
		
		if (tmres == 0) {
			printf("Transmission failed in %u fragment\n", fragmentid);
			g_free(res);
			
			free(header);
			return NULL;
		}

	}
	
	free(header);
	return res;
}

gboolean udpclient_init(struct _netgate *netgate)
{
	struct hostent *serverhp, *clienthp;
	struct sockaddr_in client;
	struct utsname hostparams;
	gchar *password;
	
	if (udpclient_error_quark == 0) udpclient_error_quark = g_quark_from_static_string(udpclient_error_string);
	
	netgate->clientconnected = FALSE;
	netgate->serveraddrstr = NULL;
	
	netgate->serveraddrstr = settings_read_string("network_server_address", "127.0.0.1");
	netgate->serveraddrport = (gushort)settings_read_int("network_server_port", 5555);
	password = settings_read_string("network_server_password", "defaultpassword");
	gost_hash_new(password, (guchar *)&netgate->clienthash);
	g_free(password);
	
	if ((netgate->clientsock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		perror("[-] Error opening clent socket");
		return FALSE;
	}
	
	sprintf(netgate->clientipstr, "0.0.0.0");
	
	if (uname(&hostparams) == 0) {
		if ((clienthp = gethostbyname(hostparams.nodename)) != NULL) {
			client.sin_family = AF_INET;
			bcopy((char *)clienthp->h_addr, (char *)&client.sin_addr, clienthp->h_length);
			client.sin_port = htons(5555);
			memset(netgate->clientipstr, 0, sizeof(netgate->clientipstr));
			inet_ntop(AF_INET, &(client.sin_addr), netgate->clientipstr, sizeof(netgate->clientipstr));
		}
	}
	
	if ((serverhp = gethostbyname(netgate->serveraddrstr)) != NULL) {
		netgate->server.sin_family = AF_INET;
		bcopy((char *)serverhp->h_addr, (char *)&netgate->server.sin_addr, serverhp->h_length);
		netgate->server.sin_port = htons(netgate->serveraddrport);
		if (connect(netgate->clientsock, (struct sockaddr *)&netgate->server, sizeof(struct sockaddr_in)) == -1) {
			printf("[-] Server unreachable in initialization\n");
		}
	} else {
		perror("[-] Server address not recognized in initialization");
	}
	
	return TRUE;
}

gboolean udpclient_connect(struct _netgate *netgate)
{
	struct stat statbuf;
	struct hostent *serverhp;
	
	if (fstat(netgate->clientsock, &statbuf) >= 0) {
		if ((statbuf.st_mode & S_IFMT) != S_IFSOCK) {
			if (!udpclient_init(netgate)) {
				g_clear_error(&netgate->error);
				g_set_error(&netgate->error, udpclient_error_quark, GSF_UDPCLIENT_ERROR, _("Client socket error"));
				return FALSE;
			}
		}
	}
	
	if (netgate->serveraddrstr == NULL) {
		g_clear_error(&netgate->error);
		g_set_error(&netgate->error, udpclient_error_quark, GSF_UDPCLIENT_ERROR, _("Wrong server address"));
		return FALSE;
	}
	
	if ((serverhp = gethostbyname(netgate->serveraddrstr)) == NULL) {
		g_clear_error(&netgate->error);
		g_set_error(&netgate->error, udpclient_error_quark, GSF_UDPCLIENT_ERROR, _("Server address not recognized"));
		return FALSE;
	}
	
	//if (netgate->clientconnected) udpclient_disconnect(netgate);
	
	netgate->server.sin_family = AF_INET;
	bcopy((char *)serverhp->h_addr, (char *)&netgate->server.sin_addr, serverhp->h_length);
	netgate->server.sin_port = htons(netgate->serveraddrport);
	
	if (connect(netgate->clientsock, (struct sockaddr *)&netgate->server, sizeof(struct sockaddr_in)) == -1) {
		g_clear_error(&netgate->error);
		g_set_error(&netgate->error, udpclient_error_quark, GSF_UDPCLIENT_ERROR, _("Server unreachable"));
		return FALSE;
	}
		
	netgate->clientconnected = TRUE;
	
	if (!udpclient_get_channels_list(netgate)) {
		return FALSE;
	}
	
	return TRUE;
}

gboolean udpclient_set_parameters(struct _netgate *netgate, gchar *address, gushort port, gchar *password)
{
	if (netgate == NULL) return FALSE;
	
	if ((address == NULL) || (password == NULL) || (port < 1024)) {
		g_clear_error(&netgate->error);
		g_set_error(&netgate->error, udpclient_error_quark, GSF_UDPCLIENT_ERROR, _("Connection parameters are not suitable"));
		return FALSE;
	}
	
	if (netgate->serveraddrstr != NULL) g_free(netgate->serveraddrstr);
	netgate->serveraddrstr = g_strdup(address);
	netgate->serveraddrport = port;
	gost_hash_new(password, (guchar *)&netgate->clienthash);
	
	return TRUE;
}

gboolean udpclient_disconnect(struct _netgate *netgate)
{
	memset(&netgate->server, 0, sizeof(netgate->server));
	
	netgate->server.sin_family = AF_UNSPEC;
		
	connect(netgate->clientsock, (struct sockaddr *)&netgate->server, sizeof(struct sockaddr_in));
	
	service_net_clear();
	
	netgate->clientconnected = FALSE;
	
	return TRUE;
}

gboolean udpclient_get_channels_list(struct _netgate *netgate)
{
	gchar *answerstr;
	//XML Parser
	GMarkupParser mp;
	GMarkupParseContext *mpc;
	GError *error = NULL;
	
	if ((answerstr = simple_data_request(netgate, REQUEST_SVCLIST_GET, 0)) == NULL) {
		g_clear_error(&netgate->error);
		g_set_error(&netgate->error, udpclient_error_quark, GSF_UDPCLIENT_ERROR, _("Failed to get answer from server"));
		return FALSE;
	}
	
	service_net_clear();
	
	memset(&answer, 0, sizeof(answer));
	
	mp.start_element = udpclient_xml_get_element;
	mp.end_element = udpclient_xml_end_element;
	mp.text = udpclient_xml_get_value;
	mp.passthrough = NULL;
	mp.error = NULL; 
		
	mpc = g_markup_parse_context_new(&mp, 0, &answer, NULL);
	g_markup_parse_context_parse(mpc, answerstr, -1, &error);
	if (error != NULL) {
		g_warning("%s", error->message);
		g_error_free(error);
		g_markup_parse_context_free(mpc);
		answer.request = REQUEST_UNKNOWN;
	} else {
		g_markup_parse_context_free(mpc);
	}
	g_free(answerstr);
	
	if (answer.request == REQUEST_UNKNOWN) {
		g_clear_error(&netgate->error);
		g_set_error(&netgate->error, udpclient_error_quark, GSF_UDPCLIENT_ERROR, "%s", answer.error);
		return FALSE;
	}
	
	if ((answer.result) && (answer.datasize > 0)) {
		
		if ((answerstr = multipart_data_request(netgate, answer.datasize)) == NULL) {
			printf("[-] Failed to get multipart answer from server\n");
		} else {
			memset(&answer, 0, sizeof(answer));
			
			mp.start_element = udpclient_xml_get_element;
			mp.end_element = udpclient_xml_end_element;
			mp.text = udpclient_xml_get_value;
			mp.passthrough = NULL;
			mp.error = NULL; 
			
			mpc = g_markup_parse_context_new(&mp, 0, &answer, NULL);
			g_markup_parse_context_parse(mpc, answerstr, -1, &error);
			if (error != NULL) {
				g_warning("%s", error->message);
				g_error_free(error);
				g_markup_parse_context_free(mpc);
				answer.request = REQUEST_UNKNOWN;
			} else {
				g_markup_parse_context_free(mpc);
			}
			
			g_free(answerstr);
		}
		
	}
		
	return TRUE;
}

gboolean udpclient_get_signal_status(struct _netgate *netgate, struct _service *svc, struct _frontend *frontend)
{
	gchar *answerstr;
	//XML Parser
	GMarkupParser mp;
	GMarkupParseContext *mpc;
	GError *error = NULL;
	
	if ((answerstr = simple_data_request(netgate, REQUEST_SIGSTAT_GET, 0)) == NULL) {
		g_clear_error(&netgate->error);
		g_set_error(&netgate->error, udpclient_error_quark, GSF_UDPCLIENT_ERROR, _("Failed to get answer from server"));
		return FALSE;
	}
	
	mp.start_element = udpclient_xml_get_element;
	mp.end_element = udpclient_xml_end_element;
	mp.text = udpclient_xml_get_value;
	mp.passthrough = NULL;
	mp.error = NULL; 
		
	mpc = g_markup_parse_context_new(&mp, 0, &answer, NULL);
	g_markup_parse_context_parse(mpc, answerstr, -1, &error);
	if (error != NULL) {
		g_warning("%s", error->message);
		g_error_free(error);
		g_markup_parse_context_free(mpc);
		answer.request = REQUEST_UNKNOWN;
	} else {
		g_markup_parse_context_free(mpc);
	}
	g_free(answerstr);
	
	if (answer.request == REQUEST_UNKNOWN) {
		g_clear_error(&netgate->error);
		g_set_error(&netgate->error, udpclient_error_quark, GSF_UDPCLIENT_ERROR, "%s", answer.error);
		return FALSE;
	}
	
	if (svc != NULL) {
		if ((svc->name == NULL) || ((svc->name != NULL) && (!g_str_equal(svc->name, answer.name)))) {
			if (svc->name != NULL) g_free(svc->name);
			svc->name = g_strdup(answer.name);
			if (frontend != NULL) frontend->reset = TRUE;
		}
		
		if ((svc->node == NULL) || ((svc->node != NULL) && (!g_str_equal(svc->node, answer.nodename)))) {
			if (svc->node != NULL) g_free(svc->node);
			svc->node = g_strdup(answer.nodename);
			if (frontend != NULL) frontend->reset = TRUE;
		}
	}
	
	if (frontend != NULL) {
		frontend->snr = answer.snr;
		frontend->signal = answer.agc;
		frontend->ber = answer.ber;
		frontend->lock = answer.lock;
		frontend->carrier = answer.carrier;
		frontend->tuned = TRUE;
	}
	
	return TRUE;
}

gboolean udpclient_tune_service(struct _netgate *netgate)
{
	gchar *answerstr;
	//XML Parser
	GMarkupParser mp;
	GMarkupParseContext *mpc;
	GError *error = NULL;
	
	if ((answerstr = simple_data_request(netgate, REQUEST_SVCTUNE_GET, netgate->netsvcid)) == NULL) {
		g_clear_error(&netgate->error);
		g_set_error(&netgate->error, udpclient_error_quark, GSF_UDPCLIENT_ERROR, _("Failed to get answer from server"));
		return FALSE;
	}
	
	mp.start_element = udpclient_xml_get_element;
	mp.end_element = udpclient_xml_end_element;
	mp.text = udpclient_xml_get_value;
	mp.passthrough = NULL;
	mp.error = NULL; 
		
	mpc = g_markup_parse_context_new(&mp, 0, &answer, NULL);
	g_markup_parse_context_parse(mpc, answerstr, -1, &error);
	if (error != NULL) {
		g_warning("%s", error->message);
		g_error_free(error);
		g_markup_parse_context_free(mpc);
		answer.request = REQUEST_UNKNOWN;
	} else {
		g_markup_parse_context_free(mpc);
	}
	g_free(answerstr);
	
	if (answer.request == REQUEST_UNKNOWN) {
		g_clear_error(&netgate->error);
		g_set_error(&netgate->error, udpclient_error_quark, GSF_UDPCLIENT_ERROR, "%s", answer.error);
		return FALSE;
	}
	
	if (answer.result == 0) {
		g_clear_error(&netgate->error);
		g_set_error(&netgate->error, udpclient_error_quark, GSF_UDPCLIENT_ERROR, _("Tune operation not permitted in server"));
		return FALSE;
	}
	
	
	return TRUE;
}

void udpclient_xml_get_element(GMarkupParseContext *context, const gchar *element, const gchar **attr_names, const gchar **attr_values, gpointer data, GError **error)
{
	struct _answer *answer = data;
	gint i = 0;
	gint len;
	
	while ((attr_names[i] != NULL) && (attr_values[i] != NULL)) {
		if (strstr(attr_names[i], "ip") != NULL) {
			len = snprintf(answer->ip, sizeof(answer->ip), "%s", attr_values[i]);
			if (len > 0) answer->ip[len] = '\0'; else answer->ip[0] = '\0';
		} else if (strstr(attr_names[i], "id") != NULL) {
			answer->serviceid = strtoul(attr_values[i], (gchar **)NULL, 10);
		} else if (strstr(attr_names[i], "nodename") != NULL) {
			len = snprintf(answer->nodename, sizeof(answer->nodename), "%s", attr_values[i]);
			if (len > 0) answer->nodename[len] = '\0'; else answer->nodename[0] = '\0';
		}
		i++;
	}
	
	
	if (g_str_equal(element, "request")) udpclient_xml_parameter = XML_PARAM_REQUEST;
	else if (g_str_equal(element, "result")) udpclient_xml_parameter = XML_PARAM_RESULT;
	else if (g_str_equal(element, "datasize")) udpclient_xml_parameter = XML_PARAM_DATASIZE;
	else if (g_str_equal(element, "name")) udpclient_xml_parameter = XML_PARAM_NAME;
	else if (g_str_equal(element, "snr")) udpclient_xml_parameter = XML_PARAM_SNR;
	else if (g_str_equal(element, "agc")) udpclient_xml_parameter = XML_PARAM_AGC;
	else if (g_str_equal(element, "ber")) udpclient_xml_parameter = XML_PARAM_BER;
	else if (g_str_equal(element, "lock")) udpclient_xml_parameter = XML_PARAM_LOCK;
	else if (g_str_equal(element, "carrier")) udpclient_xml_parameter = XML_PARAM_CARRIER;
	else if (g_str_equal(element, "service")) udpclient_xml_parameter = XML_PARAM_SERVICE;
	else if (g_str_equal(element, "error")) udpclient_xml_parameter = XML_PARAM_ERROR;
	else udpclient_xml_parameter = XML_PARAM_NULL;
}

void udpclient_xml_get_value(GMarkupParseContext *context, const gchar *text, gsize size, gpointer data, GError **error)
{
	struct _answer *answer = data;
	gint len;
	
	if (udpclient_xml_parameter == XML_PARAM_NULL) return;
	
	switch (udpclient_xml_parameter) {
		case XML_PARAM_REQUEST:
			if (strstr(text, "SVCLIST_GET") != NULL) {
				answer->request = REQUEST_SVCLIST_GET;
			} else if (strstr(text, "SVCLIST_MPT") != NULL) {
				answer->request = REQUEST_SVCLIST_MPT;
			} else if (strstr(text, "SIGSTAT_GET") != NULL) {
				answer->request = REQUEST_SIGSTAT_GET;
			} else if (strstr(text, "SVCTUNE_GET") != NULL) {
				answer->request = REQUEST_SVCTUNE_GET;
			} else {
				answer->request = REQUEST_UNKNOWN;
			}
			break;
		case XML_PARAM_RESULT:
			answer->result = (gboolean)(strtoul(text, (gchar **)NULL, 10));
			break;
		case XML_PARAM_DATASIZE:
			answer->datasize = strtoul(text, (gchar **)NULL, 10);
			break;
		case XML_PARAM_NAME:
			len = snprintf(answer->name, sizeof(answer->name), "%s", text);
			if (len > 0) answer->name[len] = '\0'; else answer->name[0] = '\0';
			break;
		case XML_PARAM_SNR:
			answer->snr = strtoul(text, (gchar **)NULL, 10);
			break;
		case XML_PARAM_AGC:
			answer->agc = strtoul(text, (gchar **)NULL, 10);
			break;
		case XML_PARAM_BER:
			answer->ber = strtoul(text, (gchar **)NULL, 10);
			break;
		case XML_PARAM_LOCK:
			answer->lock = (gboolean)(strtoul(text, (gchar **)NULL, 10));
			break;
		case XML_PARAM_CARRIER:
			answer->carrier = (gboolean)(strtoul(text, (gchar **)NULL, 10));
			break;
		case XML_PARAM_SERVICE:
			len = snprintf(answer->servicename, sizeof(answer->servicename), "%s", text);
			if (len > 0) answer->servicename[len] = '\0'; else answer->servicename[0] = '\0';
			service_net_add(answer->serviceid, (gchar *)&answer->servicename);
			break;
		case XML_PARAM_ERROR:
			len = snprintf(answer->error, sizeof(answer->error), "%s", text);
			if (len > 0) answer->error[len] = '\0'; else answer->error[0] = '\0';
		default: break;
	}
	
}

void udpclient_xml_end_element(GMarkupParseContext *context, const gchar *element, gpointer data, GError **error)
{
	udpclient_xml_parameter = XML_PARAM_NULL;
	
	if (g_str_equal(element, "servicelist")) {
		if (udpclient_callback != NULL) {
			(udpclient_callback)(UDPCLIENT_STATUS_LIST_REFRESH, NULL);
		}
	}
}

gboolean udpclient_set_status_callback(udpclient_status_callback callback)
{
	udpclient_callback = callback;
}
