/*
 *      webif.c
 *      
 *      Copyright 2009 - 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <sys/mman.h>

#include "gsfcore.h"
//#include "mainwindow.h"
#include "dbstorage.h"
#include "resources.h"
#include "common/compression.h"
#include "common/base64.h"
#include "common/gost.h"


#define NETANS_SIGSTAT_XML_BUFFER  1024
#define NETANS_CONTROL_XML_BUFFER  1024

#define NETANS_SIGSTAT_GET_XML "<?xml version=\'1.0\' encoding='UTF-8'?>\n\n<netanswer ip=\"%s\" nodename=\"%s\">\n\t<name>%s</name>\n\t<snr>%u</snr>\n\t<agc>%u</agc>\n\t<ber>%u</ber>\n\t<lock>%u</lock>\n\t<carrier>%u</carrier>\n</netanswer>\n\n"
#define NETANS_CONTROL_GET_XML "<?xml version=\'1.0\' encoding='UTF-8'?>\n\n<control>\n\t<command result=\"%u\">%s</command>\n</control>\n\n"

#define HTTP_HDR_CONTENT_HTML      "Content-Type: text/html; charset=UTF-8"
#define HTTP_HDR_CONTENT_XML       "Content-Type: text/xml; charset=UTF-8"

#define HTTP_HDR_ENCODING_GZIP     "Content-Encoding: gzip"
#define HTTP_HDR_ENCODING_DEFLATE  "Content-Encoding: deflate"

#define HTTP_HEADER_200 "HTTP/1.1 200 OK\nDate: %s\nServer: GSF/1.0\nAccept-Ranges: bytes\nConnection: close\nContent-Length: %u%s"
#define HTTP_HEADER_401 "HTTP/1.1 401 Authorization Required\nDate: %s\nServer: GSF/1.0\nWWW-Authenticate: Basic realm=\"Need password to login\"\nAccept-Ranges: bytes\nConnection: close\nContent-Length: %u%s"
#define HTTP_HEADER_403 "HTTP/1.1 403 Forbidden\nDate: %s\nServer: GSF/1.0\nAccept-Ranges: bytes\nConnection: close\nContent-Length: %u%s"
#define HTTP_HEADER_404 "HTTP/1.1 404 Not Found\nDate: %s\nServer: GSF/1.0\nAccept-Ranges: bytes\nConnection: close\nContent-Length: %u%s"

enum {
	WEBIF_XML_SIGSTAT = 0,
	WEBIF_XML_SVCLIST,
	WEBIF_XML_CONTROL
};

enum {
	WEBIF_REQ_INDEX = 0,
	WEBIF_REQ_SIGSTAT,
	WEBIF_REQ_SVCLIST,
	WEBIF_REQ_CONTROL,
	WEBIF_REQ_FORBIDDEN,
	WEBIF_REQ_UNKNOWN
};

enum {
	WEBIF_FLAG_HTML = 0x00000000,
	WEBIF_FLAG_XML = 0x00000001,
	WEBIF_FLAG_GZIP = 0x00000002,
	WEBIF_FLAG_DEFLATE = 0x00000004,
	WEBIF_FLAG_HDR200 = 0x00000008,
	WEBIF_FLAG_HDR401 = 0x00000010,
	WEBIF_FLAG_HDR403 = 0x00000020,
	WEBIF_FLAG_HDR404 = 0x00000040
};

struct _reqparams {
	gint command;
	gulong id;
	gulong seq;
};

struct _template {
	gchar *mapping;
	ssize_t size;
};

struct _template *indextpl, *msg401tpl, *msg403tpl, *msg404tpl;

typedef struct {
	gchar *filename;
	struct _template **template;
} templates_table;

templates_table templates[] = {
	{RESOURCE_WEBIF_INDEX_TPL,  &indextpl},
	{RESOURCE_WEBIF_E401_TPL, &msg401tpl},
	{RESOURCE_WEBIF_E403_TPL, &msg403tpl},
	{RESOURCE_WEBIF_E404_TPL, &msg404tpl},
	{NULL,         NULL}
};

static char *webif_split_string(gchar *str, gchar *beginnig, gchar *end, gchar *buffer, ssize_t bufsize);
static struct _template *webif_cache_template(char *filename);
static gint webif_access_test(gchar *request, gchar *password);
static gint webif_compose_header(gchar *buffer, size_t bufsize, gint flags, guint length);
static gint webif_send_http_page(gint sockfd, struct _template *template, gint flags);
static gint webif_send_xml(gint sockfd, gint resource, gint flags, void *data);
static gint webif_parse_options(gchar *reqstr, struct _reqparams *reqparams);
static gint webif_async_recv(gint sockfd, gchar *buffer, gsize bufsize);
static gint webif_async_send(gint sockfd, gchar *buffer, gsize bufsize);

gint webif_start(struct _netgate *netgate, gboolean firsttime)
{
	gint listenfd;
	const gint sflag = 1;
	struct sockaddr_in address;
	gint i;
	gint flags;
	
	if (firsttime) {
		i = 0;
		while (templates[i].filename != NULL) {
			if ((*(struct _template **)(templates[i].template) = webif_cache_template(templates[i].filename)) == NULL) {
				*(struct _template **)(templates[i].template) = NULL;
				printf("[-] Failed to cache template \"%s\"\n", templates[i].filename);
				return -1;
			}
			i++;
		}
	}
	
	memset((void *)&address, 0, sizeof(address));
	
	address.sin_family = AF_INET;
	address.sin_addr.s_addr = htonl(INADDR_ANY);
	address.sin_port = htons(netgate->serverport);
	
	if ((listenfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		printf("[-] Failed to create listen socket for incoming TCP connections: %s\n", strerror(errno));
		return -1;
	}
		
	if (setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, &sflag, sizeof(sflag)) < 0) {
		printf("[-] Failed to set socket flags: %s\n", strerror(errno));
		close(listenfd);
		return -1;
	}
	
	if (bind(listenfd, (struct sockaddr *)&address, sizeof(struct sockaddr_in)) == -1) {
		printf("[-] Failed to bind listen socket for incoming TCP connections: %s\n", strerror(errno));
		close(listenfd);
		return -1;
	}
	
	if (listen(listenfd, 10) != 0) {
		perror("[-] Failed to turn listen mode on on socket");
		close(listenfd);
		return -1;
	}
	
	flags = fcntl(listenfd, F_GETFL, 0);
	fcntl(listenfd, F_SETFL, flags | O_NONBLOCK);
	
	printf("[+] Web interface started at port no. %s%i%s\n", "\033[92m", netgate->serverport, "\033[0m");
	
	return listenfd;
}

static char *webif_split_string(gchar *str, gchar *beginnig, gchar *end, gchar *buffer, ssize_t bufsize)
{
	gchar *first, *second;
	ssize_t linklen;
	
	if ((str == NULL) || (buffer == NULL) || (bufsize == 0)) return NULL;
	
	if ((beginnig != NULL) && ((first = strstr(str, beginnig)) != NULL)) {
		first += strlen(beginnig);
	} else return NULL;
	
	if ((second = strstr(first+1, end)) != NULL) {
		if ((linklen = second - first) > 0) {
			if (linklen > bufsize) linklen = bufsize;
			strncpy(buffer, first, linklen);
			buffer[linklen] = '\0';
			
			return buffer;
		}
	}
	
	return NULL;
}

static struct _template *webif_cache_template(char *filename)
{
	gint fd;
	struct stat statbuf;
	gchar *mapping;
	struct _template *template;
	
	if ((fd = open(filename, O_RDONLY)) == -1) {
		printf("Failed to open template file\n");
		return NULL;
	}
	
	if (fstat(fd, &statbuf) == -1) {
		printf("Failed to get template file info\n");
		close(fd);
		return NULL;
	}
	
	if (!S_ISREG(statbuf.st_mode)) {
		printf("Template file is not of suitable type\n");
		close(fd);
		return NULL;
	}
	
	if ((mapping = mmap(0, statbuf.st_size, PROT_READ, MAP_SHARED, fd, 0)) == MAP_FAILED) {
		printf("Failed to map template file into memory\n");
		close(fd);
		return NULL;
	}
	
	close(fd);
	
	template = (struct _template *)calloc(1, sizeof(struct _template));
	
	template->mapping = mapping;
	template->size = statbuf.st_size;
		
	return template;
}

static gint webif_access_test(gchar *request, gchar *password)
{
	gchar reqbuffer[4096], buffer[4096];
	gchar *upasswd;
	guchar uhash[GOST_HASH_SIZE];
	
	if (password == NULL) return 1;
	
	memset(reqbuffer, 0, sizeof(reqbuffer));
	
	if (webif_split_string(request, "Authorization: Basic ", "\n", (char *)&reqbuffer, sizeof(reqbuffer)) != NULL) {
		if (base64_decode((gchar *)&reqbuffer, (gchar *)&buffer, sizeof(buffer)) != NULL) {
			if ((upasswd = strchr((const gchar *)&buffer, ':')) != NULL) {
				upasswd = upasswd + 1;
				memset(uhash, 0, sizeof(uhash));
				gost_hash_new(upasswd, (guchar *)&uhash);
				if (strncmp(password, uhash, GOST_HASH_SIZE) == 0) {
					return 1;
				}
			}
		}
	}
	
	return 0;	
}

static gint webif_compose_header(gchar *buffer, size_t bufsize, gint flags, guint length)
{
	time_t timestamp;
	struct tm *tmp;
	gchar extbuffer[1024], tmbuffer[32];
	gchar *content, *encoding;
	gint extsize;
	gint noencoding;
	
	if ((!(flags & WEBIF_FLAG_HDR200)) && (!(flags & WEBIF_FLAG_HDR401)) && 
			(!(flags & WEBIF_FLAG_HDR403)) && (!(flags & WEBIF_FLAG_HDR404))) return -1;
		
	memset(buffer, 0, bufsize);
	memset(extbuffer, 0, sizeof(extbuffer));
	memset(tmbuffer, 0, sizeof(tmbuffer));
	
	timestamp = time(NULL);
	tmp = gmtime(&timestamp);
	snprintf((char *)&tmbuffer, sizeof(tmbuffer), "Sun, 1 Jan 1970 00:00:00 GMT");
	strftime((char *)&tmbuffer, sizeof(tmbuffer), "%a, %d %b %Y %T %Z", tmp);
	
	noencoding = 0;
	if (flags & WEBIF_FLAG_GZIP) {
		encoding = HTTP_HDR_ENCODING_GZIP;
	} else if (flags & WEBIF_FLAG_DEFLATE) {
		encoding = HTTP_HDR_ENCODING_DEFLATE;
	} else {
		noencoding = 1;
	}
	
	if (flags & WEBIF_FLAG_XML) {
		content = HTTP_HDR_CONTENT_XML;
	} else {
		content = HTTP_HDR_CONTENT_HTML;
	}
	
	if (noencoding) {
		extsize = snprintf((char *)&extbuffer, sizeof(extbuffer), "\n%s\r\n\r\n", content);
	} else {
		extsize = snprintf((char *)&extbuffer, sizeof(extbuffer), "\n%s\n%s\r\n\r\n", content, encoding);
	}
	
	if (flags & WEBIF_FLAG_HDR200) {
		extsize = snprintf(buffer, bufsize, HTTP_HEADER_200, tmbuffer, length, extbuffer);
	} else if (flags & WEBIF_FLAG_HDR401) {
		extsize = snprintf(buffer, bufsize, HTTP_HEADER_401, tmbuffer, length, extbuffer);
	} else if (flags & WEBIF_FLAG_HDR403) {
		extsize = snprintf(buffer, bufsize, HTTP_HEADER_403, tmbuffer, length, extbuffer);
	} else if (flags & WEBIF_FLAG_HDR404) {
		extsize = snprintf(buffer, bufsize, HTTP_HEADER_404, tmbuffer, length, extbuffer);
	}
	if (extsize < bufsize) buffer[extsize] = '\0';
	
	return extsize;
}

static gint webif_send_http_page(gint sockfd, struct _template *template, gint flags)
{
	gchar hdrbuffer[4096];
	gint hdrsize;
	gchar *cmpbuffer;
	gint cmpsize;
	gchar *buffer;
	gint buffersize;
	gint result;
	
	if ((template == NULL) || (flags == 0)) return -1;
	
	if ((cmpbuffer = calloc(1, template->size)) == NULL) return -1;
	strncpy(cmpbuffer, template->mapping, template->size);
	cmpsize = template->size;
	
	if (flags & WEBIF_FLAG_GZIP) {
		result = compression_gzip_http_compress(cmpbuffer, template->size);
		if (result > 0) {
			cmpsize = result;
		} else {
			flags = flags ^ WEBIF_FLAG_GZIP;
		}
	} else if (flags & WEBIF_FLAG_DEFLATE) {
		flags = flags ^ WEBIF_FLAG_DEFLATE;
	}
	
	if ((hdrsize = webif_compose_header((gchar *)&hdrbuffer, sizeof(hdrbuffer), flags, cmpsize)) == -1) {
		free(cmpbuffer);
		return -1;
	}
	
	buffersize = hdrsize + cmpsize;
	
	if ((buffer = (char *)calloc(1, buffersize)) == NULL) {
		free(cmpbuffer);
		return -1;
	}
	
	memcpy(buffer, hdrbuffer, hdrsize);
	memcpy(buffer+hdrsize, cmpbuffer, cmpsize); 
	
	result = webif_async_send(sockfd, buffer, buffersize);
	
	free(buffer);
	free(cmpbuffer);
	
	return result;
}

static gint webif_send_xml(gint sockfd, gint resource, gint flags, void *data)
{
	gchar hdrbuffer[4096];
	gint hdrsize;
	gchar *xmlbuffer;
	gint xmlsize;
	gchar *buffer;
	gint buffersize;
	gint result;
	struct _reqparams *reqparams;
	gchar *errorstr;
	gchar *svclist;
	guint cmdres;
	struct _sigstatus sigstatus;
	struct _netgate *netgate;
	
	
	if (flags == 0) return -1;
	
	if (resource == WEBIF_XML_SIGSTAT) {
		if ((xmlbuffer = calloc(1, NETANS_SIGSTAT_XML_BUFFER)) == NULL) return -1;
		gsfcore_get_signal_status(&sigstatus);
		netgate = (struct _netgate *)data;
		xmlsize = snprintf(xmlbuffer, NETANS_SIGSTAT_XML_BUFFER, NETANS_SIGSTAT_GET_XML, netgate->serveripstr, netgate->servernodename,
							sigstatus.svcname, sigstatus.snr, sigstatus.agc, sigstatus.ber, sigstatus.lock, sigstatus.carrier);
	} else if (resource == WEBIF_XML_SVCLIST) {
		if ((svclist = service_xml_list_get_string()) == NULL) return -1;	
		if ((xmlbuffer = strdup(svclist)) == NULL) return -1;
		xmlsize = strlen(xmlbuffer);
	} else if (resource == WEBIF_XML_CONTROL) {
		reqparams = (struct _reqparams *)data;
		cmdres = 0;
		switch (reqparams->command) {
			case 1: {
				if (reqparams->id != 0) {
					cmdres = (guint)gsfcore_tune_start(reqparams->id, SERVICE_TV_RADIO);
					errorstr = "Internal process";
				} else {
					cmdres = 0;
					errorstr = "Wrong service ID";
				}
				break;
			}
			
			default: {
				cmdres = 0;
				errorstr = "Wrong command ID";
				break;
			}
		}
		if ((xmlbuffer = calloc(1, NETANS_CONTROL_XML_BUFFER)) == NULL) return -1;
		xmlsize = snprintf(xmlbuffer, NETANS_CONTROL_XML_BUFFER, NETANS_CONTROL_GET_XML, cmdres, errorstr);
	}
	
	if (flags & WEBIF_FLAG_GZIP) {
		result = compression_gzip_http_compress(xmlbuffer, xmlsize);
		if (result > 0) {
			xmlsize = result;
		} else {
			flags = flags ^ WEBIF_FLAG_GZIP;
		}
	} else if (flags & WEBIF_FLAG_DEFLATE) {
		flags = flags ^ WEBIF_FLAG_DEFLATE;
	}
	
	if ((hdrsize = webif_compose_header((gchar *)&hdrbuffer, sizeof(hdrbuffer), flags, xmlsize)) == -1) {
		free(xmlbuffer);
		return -1;
	}
	
	buffersize = hdrsize + xmlsize;
	
	if ((buffer = (gchar *)calloc(1, buffersize)) == NULL) {
		free(xmlbuffer);
		return -1;
	}
	
	memcpy(buffer, hdrbuffer, hdrsize);
	memcpy(buffer+hdrsize, xmlbuffer, xmlsize);
	
	result = webif_async_send(sockfd, buffer, buffersize);
	
	free(buffer);
	free(xmlbuffer);
	
	return result;
}

static gint webif_parse_options(gchar *reqstr, struct _reqparams *reqparams)
{
	gchar *reqbuf;
	gchar *bsp, *bep;
	gint len;
	gchar *block;
	gchar *osp, *oep;
	
	if ((reqstr == NULL) || (reqparams == NULL)) return 0;
	
	bsp = strchr(reqstr, '?');
	bep = strstr(reqstr, " HTTP/");
	
	if ((bsp == NULL) || (bep == NULL) || (strchr(reqstr, '=') == NULL)) return 0;
	bsp++;
	len = bep - bsp;
	
	if ((reqbuf = calloc(1, len+1)) == NULL) return 0;
	strncpy(reqbuf, bsp, len);
	
	memset(reqparams, 0, sizeof(struct _reqparams));
	
	bep = strchr(reqbuf, '&');
	bsp = reqbuf;
	
	while (1) {
		if (bep != NULL) len = bep-bsp; else len = strlen(bsp);
		if ((block = calloc(1, len+1)) != NULL) {
			strncpy(block, bsp, len);
			if ((oep = strchr(block, '=')) != NULL) {
				osp = block;
				oep[0] = '\0';
				oep++;
				if (strcmp(osp, "cmd") == 0) {
					reqparams->command = atoi(oep);
				} else if (strcmp(osp, "id") == 0) {
					reqparams->id = strtoul(oep, (char **)NULL, 10);
				} else if (strcmp(osp, "seq") == 0) {
					reqparams->seq = strtoul(oep, (char **)NULL, 10);
				}
			}
			free(block);
		}
		if (bep == NULL) break;
		bsp = bep+1;
		bep = strchr(bsp, '&');
	}
	
	free(reqbuf);
	
	return 1;
}

static gint webif_async_recv(gint sockfd, gchar *buffer, gsize bufsize)
{
	struct timespec sleeptime = {.tv_sec = 0, .tv_nsec = 500 * 1000 * 1000 };	
	struct timeval tval;
	gint res, bytes;
	gulong starttime, endtime;
	
	bytes = 0;	
	memset(buffer, 0, bufsize);
	
	gettimeofday(&tval, NULL);
	starttime = tval.tv_sec*1000 + tval.tv_usec/1000; 
	
	while (bytes < bufsize) {
		res = recv(sockfd, &buffer[bytes], bufsize-bytes, 0);
		if (res > 0) {
			bytes += res;
			if ((buffer[bytes-2] == '\r') && (buffer[bytes-1] == '\n')) break;
		} else if (res == 0) {
			break;
		} else if ((res == -1) && (errno == EWOULDBLOCK)) {
			if ((buffer[bytes-2] == '\r') && (buffer[bytes-1] == '\n')) {
				break;
			} else {
				gettimeofday(&tval, NULL);
				endtime = tval.tv_sec*1000 + tval.tv_usec/1000;
				if ((endtime - starttime) <= 3000) {
					nanosleep(&sleeptime, NULL);
					continue;
				} else {
					bytes = 0;
					break;
				}
			}
		}
	}
	
	return bytes;
}

static gint webif_async_send(gint sockfd, gchar *buffer, gsize bufsize)
{
	struct timespec sleeptime = {.tv_sec = 0, .tv_nsec = 500 * 1000 * 1000 };	
	struct timeval tval;
	gint res, bytes;
	gulong starttime, endtime;
	
	bytes = 0;	
	
	gettimeofday(&tval, NULL);
	starttime = tval.tv_sec*1000 + tval.tv_usec/1000; 
	
	while (bytes < bufsize) {
		res = send(sockfd, &buffer[bytes], bufsize-bytes, 0);
		if (res > 0) {
			bytes += res;
			if (bytes == bufsize) break;
		} else if (res == 0) {
			break;
		} else if ((res == -1) && (errno == EWOULDBLOCK)) {
			if (bytes == bufsize) {
				break;
			} else {
				gettimeofday(&tval, NULL);
				endtime = tval.tv_sec*1000 + tval.tv_usec/1000;
				if ((endtime - starttime) <= 3000) {
					nanosleep(&sleeptime, NULL);
					continue;
				} else {
					break;
				}
			}
		}
	}
	
	return bytes;
}

gint webif_receive_message(struct _netgate *netgate)
{
	struct sockaddr_in clientaddr;
	socklen_t addrlen;
	gint clisockfd;
	gint flags;
	gchar message[16384], buffer[16384];
	gint bytes;
	gchar *fragment;
	gint reqid;
	gint reqflags;
	struct _reqparams reqparams;
	
	
	addrlen = sizeof(clientaddr);
	if ((clisockfd = accept(netgate->serverhttpsock, (struct sockaddr *)&clientaddr, &addrlen)) == -1) {
		printf("[-] Failed to accept connection\n");
		return 0;
	}
	
	flags = fcntl(clisockfd, F_GETFL, 0);
	fcntl(clisockfd, F_SETFL, flags | O_NONBLOCK);
	
	bytes = webif_async_recv(clisockfd, (gchar *)&message, sizeof(message));
	
	if (bytes > 0) {
		// message received
		if (strchr((gchar *)&message, '?') == NULL) {
			fragment = webif_split_string((gchar *)&message, "GET ", " HTTP/", (gchar *)&buffer, sizeof(buffer));
		} else {
			webif_parse_options((gchar *)&message, &reqparams);
			fragment = webif_split_string((gchar *)&message, "GET ", "?", (gchar *)&buffer, sizeof(buffer));
		}
		
		if ((strcmp(fragment, "/") == 0) || (strcmp(fragment, "/index.html") == 0)) {
			reqid = WEBIF_REQ_INDEX;
		} else if (strcmp(fragment, "/status.xml") == 0) {
			reqid = WEBIF_REQ_SIGSTAT;
		} else if (strcmp(fragment, "/services.xml") == 0) {
			if ((service_xml_list_get_length() > 0) && (netgate->serverexpsvclist)) {
				reqid = WEBIF_REQ_SVCLIST;
			} else {
				reqid = WEBIF_REQ_FORBIDDEN;
			}
		} else if (strcmp(fragment, "/control.xml") == 0) {
			if ((service_xml_list_get_length() > 0) && (netgate->serverexpsvclist)) {
				reqid = WEBIF_REQ_CONTROL;
			} else {
				reqid = WEBIF_REQ_FORBIDDEN;
			}
		} else {
			reqid = WEBIF_REQ_UNKNOWN;
		}
		
		reqflags = 0;
		fragment = webif_split_string((gchar *)&message, "Accept-Encoding: ", "\n", (gchar *)&buffer, sizeof(buffer));
		if (fragment != NULL) {
			if (strstr(fragment, "gzip") != NULL) {
				reqflags |= WEBIF_FLAG_GZIP;
			} else if (strstr(fragment, "deflate") != NULL) {
				reqflags |= WEBIF_FLAG_DEFLATE;
			}
		}
		
		if (reqid == WEBIF_REQ_UNKNOWN) {
			reqflags |= (WEBIF_FLAG_HTML | WEBIF_FLAG_HDR404);
			webif_send_http_page(clisockfd, msg404tpl, reqflags);
		} else {
			if (!webif_access_test((char *)&message, netgate->serverhash)) {
				reqflags |= (WEBIF_FLAG_HTML | WEBIF_FLAG_HDR401);
				webif_send_http_page(clisockfd, msg401tpl, reqflags);
			} else {
				switch (reqid) {
					case WEBIF_REQ_INDEX: {
						reqflags |= (WEBIF_FLAG_HTML | WEBIF_FLAG_HDR200);
						webif_send_http_page(clisockfd, indextpl, reqflags);
						break;
					}
					case WEBIF_REQ_SIGSTAT: {
						reqflags |= (WEBIF_FLAG_XML | WEBIF_FLAG_HDR200);
						webif_send_xml(clisockfd, WEBIF_XML_SIGSTAT, reqflags, netgate);
						break;
					}
					case WEBIF_REQ_SVCLIST: {
						reqflags |= (WEBIF_FLAG_XML | WEBIF_FLAG_HDR200);
						webif_send_xml(clisockfd, WEBIF_XML_SVCLIST, reqflags, NULL);
						break;
					}
					case WEBIF_REQ_CONTROL: {
						reqflags |= (WEBIF_FLAG_XML | WEBIF_FLAG_HDR200);
						webif_send_xml(clisockfd, WEBIF_XML_CONTROL, reqflags, (void *)&reqparams);
						break;
					}
					case WEBIF_REQ_FORBIDDEN: {
						reqflags |= (WEBIF_FLAG_HTML | WEBIF_FLAG_HDR403);
						webif_send_http_page(clisockfd, msg403tpl, reqflags);
						break;
					}
					default: break;
				}
			}
		}
		
	} else if (bytes < 0) {
		// receive error
		printf("[-] HTTP connection error\n");
	} else if (bytes == 0) {
		// receive socket closed
		printf("[-] HTTP connection closed\n");
	}
	
	shutdown(clisockfd, SHUT_RDWR);
	close(clisockfd);
}
